Package: remstats
Type: Package
Title: Computes Statistics for Relational Event History Data
Version: 4.0.0
Date: 2026-05-07
Authors@R: 
    c(
        person(
            given = "Giuseppe",
            family = "Arena",
            role = "aut",
            email = "g.arena@uva.nl",
            comment = c(ORCID = "0000-0001-5204-3326")),
        person(
            given = "Marlyne",
            family = "Meijerink-Bosman",
            role = "aut"),
        person(
            given = "Joris",
            family = "Mulder",
            role = c("aut", "cre"),
            email = "j.mulder3@tilburguniversity.edu"),
        person(
            given = "Diana",
            family = "Karimova",
            role = "ctb"),
        person(
            given = "Rumana",
            family = "Lakdawala",
            role = "ctb"),
        person(
            given = "Mahdi",
            family = "Shafiee Kamalabad",
            role = "ctb"),
        person(
            given = "Fabio",
            family = "Generoso Vieira",
            role = "ctb"),
        person(
            given = "Roger",
            family = "Leenders",
            role = "ctb",
            email = "r.t.a.j.leenders@tilburguniversity.edu")
    )
Maintainer: Joris Mulder <j.mulder3@tilburguniversity.edu>
Description: Computes a variety of statistics for relational event models (Meijerink et al., 2023, <doi:10.3758/s13428-022-01821-8>). Relational event models enable researchers to investigate exogenous and endogenous factors, and interactions, influencing the evolution of a time-ordered sequence of events. These models are categorized into tie-oriented models (Butts, C., 2008, <doi:10.1111/j.1467-9531.2008.00203.x>), where the probability of a dyad interacting next is modeled in a single step, and actor-oriented models (Stadtfeld, C., & Block, P., 2017, <doi:10.15195/v4.a14>), which first model the probability of a sender initiating an interaction and subsequently the probability of the sender's choice of receiver. The package is designed to compute a variety of statistics that summarize exogenous and endogenous influences on the event stream for both types of models.  
License: MIT + file LICENSE
URL: https://tilburgnetworkgroup.github.io/remstats/
BugReports: https://github.com/TilburgNetworkGroup/remstats/issues
LazyData: true
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.0.0)
Imports: Rcpp (>= 1.0.8.3), remify (>= 4.0.0), stats, graphics,
        grDevices
LinkingTo: Rcpp, RcppArmadillo, RcppProgress
Suggests: tinytest, knitr, rmarkdown
VignetteBuilder: knitr
LazyDataCompression: gzip
NeedsCompilation: yes
Packaged: 2026-05-07 14:04:19 UTC; jorismulder
Author: Giuseppe Arena [aut] (ORCID: <https://orcid.org/0000-0001-5204-3326>),
  Marlyne Meijerink-Bosman [aut],
  Joris Mulder [aut, cre],
  Diana Karimova [ctb],
  Rumana Lakdawala [ctb],
  Mahdi Shafiee Kamalabad [ctb],
  Fabio Generoso Vieira [ctb],
  Roger Leenders [ctb]
Repository: CRAN
Date/Publication: 2026-05-09 14:41:20 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-09 16:54:13 UTC; unix
Archs: remstats.so.dSYM

Package: remstats
Type: Package
Title: Computes Statistics for Relational Event History Data
Version: 4.0.0
Date: 2026-05-07
Authors@R: 
    c(
        person(
            given = "Giuseppe",
            family = "Arena",
            role = "aut",
            email = "g.arena@uva.nl",
            comment = c(ORCID = "0000-0001-5204-3326")),
        person(
            given = "Marlyne",
            family = "Meijerink-Bosman",
            role = "aut"),
        person(
            given = "Joris",
            family = "Mulder",
            role = c("aut", "cre"),
            email = "j.mulder3@tilburguniversity.edu"),
        person(
            given = "Diana",
            family = "Karimova",
            role = "ctb"),
        person(
            given = "Rumana",
            family = "Lakdawala",
            role = "ctb"),
        person(
            given = "Mahdi",
            family = "Shafiee Kamalabad",
            role = "ctb"),
        person(
            given = "Fabio",
            family = "Generoso Vieira",
            role = "ctb"),
        person(
            given = "Roger",
            family = "Leenders",
            role = "ctb",
            email = "r.t.a.j.leenders@tilburguniversity.edu")
    )
Maintainer: Joris Mulder <j.mulder3@tilburguniversity.edu>
Description: Computes a variety of statistics for relational event models (Meijerink et al., 2023, <doi:10.3758/s13428-022-01821-8>). Relational event models enable researchers to investigate exogenous and endogenous factors, and interactions, influencing the evolution of a time-ordered sequence of events. These models are categorized into tie-oriented models (Butts, C., 2008, <doi:10.1111/j.1467-9531.2008.00203.x>), where the probability of a dyad interacting next is modeled in a single step, and actor-oriented models (Stadtfeld, C., & Block, P., 2017, <doi:10.15195/v4.a14>), which first model the probability of a sender initiating an interaction and subsequently the probability of the sender's choice of receiver. The package is designed to compute a variety of statistics that summarize exogenous and endogenous influences on the event stream for both types of models.  
License: MIT + file LICENSE
URL: https://tilburgnetworkgroup.github.io/remstats/
BugReports: https://github.com/TilburgNetworkGroup/remstats/issues
LazyData: true
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.0.0)
Imports: Rcpp (>= 1.0.8.3), remify (>= 4.0.0), stats, graphics,
        grDevices
LinkingTo: Rcpp, RcppArmadillo, RcppProgress
Suggests: tinytest, knitr, rmarkdown
VignetteBuilder: knitr
LazyDataCompression: gzip
NeedsCompilation: yes
Packaged: 2026-05-07 14:04:19 UTC; jorismulder
Author: Giuseppe Arena [aut] (ORCID: <https://orcid.org/0000-0001-5204-3326>),
  Marlyne Meijerink-Bosman [aut],
  Joris Mulder [aut, cre],
  Diana Karimova [ctb],
  Rumana Lakdawala [ctb],
  Mahdi Shafiee Kamalabad [ctb],
  Fabio Generoso Vieira [ctb],
  Roger Leenders [ctb]
Repository: CRAN
Date/Publication: 2026-05-09 14:41:20 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-09 18:35:29 UTC; unix
Archs: remstats.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(boxplot,aomstats)
S3method(boxplot,tomstats)
S3method(plot,aomstats)
S3method(plot,tomstats)
S3method(print,remstats)
S3method(print,tomstats_sampled)
S3method(stack_stats,aomstats)
S3method(stack_stats,tomstats)
S3method(stack_stats,tomstats_sampled)
S3method(summary,remstats)
export(FEtype)
export(actor_effects)
export(aomstats)
export(average)
export(bind_remstats)
export(degreeDiff)
export(degreeMax)
export(degreeMin)
export(difference)
export(dyad)
export(event)
export(indegreeReceiver)
export(indegreeSender)
export(inertia)
export(isp)
export(itp)
export(maximum)
export(minimum)
export(osp)
export(otp)
export(outdegreeReceiver)
export(outdegreeSender)
export(psABA)
export(psABAB)
export(psABAY)
export(psABB)
export(psABBA)
export(psABBY)
export(psABX)
export(psABXA)
export(psABXB)
export(psABXY)
export(receive)
export(recencyContinue)
export(recencyReceiveReceiver)
export(recencyReceiveSender)
export(recencySendReceiver)
export(recencySendSender)
export(reciprocity)
export(remstats)
export(rrankReceive)
export(rrankSend)
export(same)
export(send)
export(sp)
export(spUnique)
export(stack_stats)
export(tie)
export(tie_effects)
export(tomstats)
export(totaldegreeDyad)
export(totaldegreeReceiver)
export(totaldegreeSender)
export(userStat)
importFrom(Rcpp,sourceCpp)
importFrom(grDevices,rainbow)
importFrom(graphics,axis)
importFrom(graphics,boxplot)
importFrom(graphics,bxp)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(remify,remify)
importFrom(stats,IQR)
importFrom(stats,quantile)
importFrom(stats,setNames)
useDynLib(remstats, .registration = TRUE)

REMSTATS - a Remote Statistics System (version 1.0.13a)

A system of programs to: 
	- gather data from remote machines,
	- store and maintain the data for long periods,
	- produce graphs and web-pages tieing them together
	- monitor the data for anomalous behaviours and issue alerts

Remstats' major difference from both orca and cricket is that I wanted lots
of graphs on the screen.  I'm not monitoring over a dialup, though it's not too
bad over 64K ISDN; I wanted to see lots of graphs together so I could correlate
things happening on one graph with other graphs.  It's already been useful
to me.  Please let me know if you find it useful.

MRTG and NOCOL provided the inspiration.  Rrdtool provided the guts.
Perl provided the platform.  Various other people provided bug-reports and 
suggestions and I wrote some glue to make it hang together.

Read RELEASENOTES for upgrade instructions, and new features. 
Read INSTALL for installations instructions.
Read ChangeLog for more detailed descriptions of what has changed.
Read BUGS for known problems that haven't been fixed yet.
Read FAQ for Frequently Asked Questions with their answers.
Read TODO for my list of things that would be nice to have in remstats.

Source available at http://remstats.sourceforge.net/release//index.html

Wed May 28 11:54:43 EDT 2003 Thomas Erskine <terskine@users.sourceforge.net>

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

# Actor info
info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

# Tie model
reh <- remify(edgelist, model = "tie")
effects <- ~ send(variable = "x1"):inertia()
stats <- remstats(reh, tie_effects = effects, attr_actors = info)

expect_stdout(print(stats))
expect_stdout(summary(stats))

# TODO(actor-model): actor model methods test pending aomstats implementation
# reh <- remify(edgelist, model = "actor")
# sender_effects <- ~ outdegreeSender()
# receiver_effects <- ~ receive(variable = "x1"):inertia()
# stats <- remstats(reh, sender_effects = sender_effects,
#     receiver_effects = receiver_effects, attr_actors = info)
# expect_stdout(print(stats))
# expect_stdout(summary(stats))
# stats <- remstats(reh, sender_effects = sender_effects, attr_actors = info)
# expect_stdout(print(stats))
# expect_stdout(summary(stats))
# stats <- remstats(reh, receiver_effects = receiver_effects, attr_actors = info)
# expect_stdout(print(stats))
# expect_stdout(summary(stats))

# Load data
data(history)

# plots for 'tomstats' objects

# Prepare data
reh <- remify(edgelist = history[,1:3], model = "tie")
# Compute effects
stats <- remstats(reh, tie_effects = ~ inertia())

# boxplot.tomstats

# Plot the 'inertia' distribution for 20 timepoints
expect_silent(boxplot(stats, effect = "inertia"))
expect_silent(boxplot(stats, effect = 1))
# Plot the 'inertia' distribution for 20 dyads
expect_silent(boxplot(stats, effect = "inertia", by = "dyads"))
# Plot the 'inertia' distribution for dyads 2:5
expect_silent(boxplot(stats, effect = "inertia", by = "dyads", subset = 2:5))

# plot.tomstats

# Plot the 'inertia' trajectories for 5 dyads
expect_silent(plot(stats, effect = "inertia"))
expect_silent(plot(stats, effect = 1))
# Plot the 'inertia' trajectory for a specific dyad
expect_silent(plot(stats, effect = "inertia", subset = 60))

# TODO(actor-model): # plots for 'aomstats' objects
# TODO(actor-model): 
# TODO(actor-model): # Prepare data
# TODO(actor-model): reh <- remify::remify(edgelist = history[,1:3], model = "actor")
# TODO(actor-model): # Compute effects
# TODO(actor-model): stats <- remstats(reh, sender_effects = ~ outdegreeSender(), receiver_effects = ~ outdegreeReceiver())
# TODO(actor-model): 
# TODO(actor-model): # boxplot.aomstats
# TODO(actor-model): 
# TODO(actor-model): # Plot the 'outdegreeSender' distribution for 20 timepoints
# TODO(actor-model): expect_silent(boxplot(stats, effect = "outdegreeSender", model = "sender"))
# TODO(actor-model): expect_silent(boxplot(stats, effect = 1, model = "sender"))
# TODO(actor-model): expect_silent(boxplot(x = stats, effect = 1, model = "receiver"))
# TODO(actor-model): # Plot the 'inertia' distribution for all 10 actors
# TODO(actor-model): expect_silent(boxplot(stats, effect = "outdegreeSender", model = "sender", by = "actors"))
# TODO(actor-model): 
# TODO(actor-model): # plot.aomstats
# TODO(actor-model): 
# TODO(actor-model): # Plot the 'outdegreeSender' trajectories 5 actors
# TODO(actor-model): expect_silent(plot(stats, effect = "outdegreeSender"))
# TODO(actor-model): expect_silent(plot(stats, effect = 1))
# TODO(actor-model): # Plot the 'outdegreeSender' trajectory for a specific actor
# TODO(actor-model): expect_silent(plot(stats, effect = "outdegreeSender", subset = 10))
# TODO(actor-model): 

#tomstats without case control sampling
tomstats0 <- function(effects, reh, attr_actors = NULL, attr_dyads = NULL,
                     memory = c("full", "window", "decay", "interval"),
                     memory_value = NA, start = 2, stop = Inf,
                     display_progress = FALSE) {
  # Check if the deprecated argument "attributes" is used

	method <- "pt"
	
  # Check if the deprecated "id" column is used in attr_actors
  if (!is.null(attr_actors)) {
    if (("id" %in% colnames(attr_actors)) & !("name" %in% colnames(attr_actors))) {
      warning("use 'name' instead of 'id' in 'attr_actors'")
      colnames(attr_actors)[which(colnames(attr_actors) == "id")] <- "name"
    }
  }

  if (!("remify" %in% class(reh))) stop("Expected a reh object of class remify")
  reh <- normalize_reh(reh)

  # Prepare all required inputs
  inputs <- prepare_tomstats(
    effects = effects, reh = reh,
    attr_actors = attr_actors, attr_dyads = attr_dyads,
    memory = memory, memory_value = memory_value,
    start = start, stop = stop, method = method
  )

  form <- inputs$form
  effects <- inputs$effects
  effectNames <- inputs$effectNames
  edgelist <- inputs$edgelist
  weights <- inputs$weights
  actors <- inputs$actor
  types <- inputs$types
  riskset <- inputs$riskset
  risksetMatrix <- inputs$risksetMatrix
  memory <- inputs$memory
  memory_value <- inputs$memory_value
  scaling <- inputs$scaling
  consider_type     <- inputs$consider_type
  consider_type_cpp <- inputs$consider_type_cpp
  covar <- inputs$covar
  interactions <- inputs$interactions
  start <- inputs$start
  stop <- inputs$stop
  method <- inputs$method
  with_type_riskset <- inputs$with_type_riskset
  needs_typed_riskset <- inputs$needs_typed_riskset
  C <- inputs$C

  # Inertia-needing effect names
  needs_inertia <- any(grepl("degree", effectNames)) |
    any(effectNames %in% c("inertia", "reciprocity", "otp", "itp", "osp", "isp", "sp"))

  # Compute the inertia building block (aggregated, all types)
  if (needs_inertia) {
  	inertia <- calculate_inertia(edgelist, weights, risksetMatrix, memory,
  															 memory_value, start, stop, display_progress, method
  	)
  } else {
  	inertia <- matrix()
  }

  # Compute statistics (unfiltered pass — correct for "ignore" effects)
  statistics <- compute_stats_tie(effectNames, edgelist, riskset,
    risksetMatrix, inertia, covar, interactions, memory, memory_value, scaling,
    consider_type_cpp, start, stop, reh$meta$directed, display_progress, method)

  # Add variable names to the statistics dimnames
  statistics <- add_variable_names(
    statistics, effectNames, effects,
    interactions
  )

  # For "separate"/"interact" effects with ext=TRUE and C>1: run C additional
  # filtered passes (one per past-event-type) to get per-type statistics.
  # Each pass filters the edgelist and inertia building block to type c only,
  # then computes only the "separate"/"interact" effects.
  stats_by_type <- NULL
  # Multi-pass needed for "separate" or "interact" with C>1, regardless of ext
  needs_multipass <- C > 1 &&
    any(consider_type[seq_along(effects)] %in% c(1L, 2L))

  if (needs_multipass) {
    # Indices of effects needing multi-pass (separate or interact)
    mp_idx <- which(consider_type[seq_along(effects)] %in% c(1L, 2L))
    mp_effectNames   <- effectNames[mp_idx]
    mp_scaling       <- scaling[mp_idx]
    mp_consider_type <- consider_type[mp_idx]  # 1L=separate, 2L=interact
    # consider_type_cpp for multi-pass effects: always FALSE (we filter edgelist)
    mp_ct_cpp <- rep(FALSE, length(mp_idx))

    stats_by_type <- vector("list", C)
    for (ci in seq_len(C)) {
      type_id_c <- types$typeID[ci]  # 0-based
      # Zero weights for non-type-c events — keeps full edgelist structure
      # intact (time indices must match) but only counts type-c events
      mask_c <- edgelist[, 4] == type_id_c
      wt_c <- weights * mask_c

      # Inertia building block for type c only (via zeroed weights)
      if (needs_inertia && length(mp_effectNames) > 0) {
        inertia_c <- calculate_inertia(edgelist, wt_c, risksetMatrix, memory,
          memory_value, start, stop, display_progress, method)
      } else {
        inertia_c <- matrix()
      }

      # Compute statistics using type-c inertia building block
      stats_c <- compute_stats_tie(
        mp_effectNames, edgelist, riskset,
        risksetMatrix, inertia_c, covar, interactions, memory, memory_value,
        mp_scaling, mp_ct_cpp, start, stop, reh$meta$directed,
        display_progress, method)

      # Add names (set directly to avoid length mismatch with interactions)
      dimnames(stats_c)[[3]] <- mp_effectNames

      # Pshift effects: overwrite with correct type-c values.
      # The masked-weights approach above does not affect pshifts because
      # compute_stats_tie ignores weights for pshift computation. Instead,
      # use R-level computation with correct method="pt" semantics (previous
      # time point events only).
      # ext=FALSE: always type-filter (separate dyads have no type column).
      # ext=TRUE + interact: type-filter rows (past-type x dyad-type semantics).
      # ext=TRUE + separate: leave as-is — base C++ stat is correct; all type
      #   slices share the same dyad pattern (defensible: "did ABAB occur,
      #   regardless of which dyad type you are?").
      ps_in_mp <- which(mp_effectNames %in% PSHIFT_NAMES)
      if (length(ps_in_mp) > 0) {
        event_type_ids_0 <- edgelist[, 4]
        M_out_ps <- stop - start + 1L
        for (ei in ps_in_mp) {
          pshift_num  <- PSHIFT_TYPE_NUM[mp_effectNames[ei]]
          is_interact <- mp_consider_type[ei] == 2L
          if (!is.null(inputs$prepR_untyped)) {
            # ext=FALSE: always type-filter
            rs_a1 <- inputs$prepR_untyped[, 1]
            rs_a2 <- inputs$prepR_untyped[, 2]
            correct_vals <- compute_pshift_type_c_tie(
              pshift_num, edgelist, event_type_ids_0, type_id_c,
              rs_a1, rs_a2, M_out_ps, start, stop
            )
            type_ids_riskset <- inputs$riskset[, 3]
            type_rows_ci <- which(type_ids_riskset == type_id_c)
            key_untyped <- paste(rs_a1, rs_a2, sep = "|")
            key_typed_c <- paste(inputs$riskset[type_rows_ci, 1],
                                 inputs$riskset[type_rows_ci, 2], sep = "|")
            dst <- match(key_typed_c, key_untyped)
            valid <- !is.na(dst)
            stats_c[, type_rows_ci[valid], ei] <- correct_vals[, dst[valid]]
          } else if (is_interact) {
            # ext=TRUE + interact: type-filter rows
            rs_a1 <- inputs$riskset[, 1]
            rs_a2 <- inputs$riskset[, 2]
            correct_vals <- compute_pshift_type_c_tie(
              pshift_num, edgelist, event_type_ids_0, type_id_c,
              rs_a1, rs_a2, M_out_ps, start, stop
            )
            stats_c[, , ei] <- correct_vals
          }
          # ext=TRUE + separate: leave stats_c as-is (base C++ stat)
        }
      }

      stats_by_type[[ci]] <- stats_c
    }
  }

  # Split consider_type stats into type-specific slices when needed
  if (needs_typed_riskset && C > 1) {
    statistics <- split_type_slices(
      statistics        = statistics,
      effects           = effects,
      consider_type     = consider_type,
      types             = types,
      prepR_typed       = inputs$riskset,
      with_type_riskset = with_type_riskset,
      prepR_untyped     = inputs$prepR_untyped,
      stats_by_type     = stats_by_type
    )
  }

  # Modify riskset output
  if (!with_type_riskset && needs_typed_riskset) {
    riskset <- build_untyped_riskset(reh, actors, types)
  } else {
    riskset <- modify_riskset(riskset, reh, actors, types)
  }

  # Format output
  class(statistics) <- c("tomstats", "remstats")
  attr(statistics, "model") <- "tie"
  attr(statistics, "formula") <- form
  attr(statistics, "riskset") <- riskset
  attr(statistics, "subset") <- data.frame(start = start + 1, stop = stop + 1)
  attr(statistics, "method") <- method
  # if (get_adjmat) {
  #   attr(statistics, "adjmat") <- adjmat
  # }
  rs <- attr(statistics, "riskset")  # data.frame with sender/receiver/id
  
  sender   <- as.character(rs$sender)
  receiver <- as.character(rs$receiver)
  
  keys <- paste0(sender, "|", receiver)
  
  idx <- seq_along(keys)
  if (!is.null(rs$id) && all(rs$id == idx)) idx <- rs$id
  
  dyad_keys <- stats::setNames(idx, keys)
  
  attr(statistics, "dyad_keys") <- dyad_keys
  

  # Output
  statistics
}



#' tomstats
#'
#' Computes statistics for modeling relational event history data
#' with the tie-oriented relational event model.
#'
#' @param effects an object of class \code{"\link[stats]{formula}"} (or one
#' that can be coerced to that class): a symbolic description of the effects in
#' the model for which statistics are computed, see 'Details' for the available
#' effects and their corresponding statistics
#' @param sampling Logical. If \code{TRUE}, statistics are computed using
#'   case–control (dyad) sampling rather than the full risk set. Default \code{FALSE}.
#' @param samp_num Integer. Number of dyads to include per event when
#'   \code{sampling = TRUE}. Must be smaller than or equal to the size of the
#'   active risk set. Ignored when \code{sampling = FALSE}.
#' @param seed Optional integer. Random seed used for dyad sampling. Setting
#'   this ensures reproducible sampling across calls. If \code{NULL}, the
#'   current RNG state is used.
#' @inheritParams remstats
#'
#' @section Effects:
#' The statistics to be computed are defined symbolically and should be
#' supplied to the \code{effects} argument in the form \code{~ effects}. The
#' terms are separated by + operators. For example:
#' \code{effects = ~ inertia() + otp()}. Interactions between two effects
#' can be included with * operators. For example:
#' \code{effects = ~ inertia()*otp()}. A list of available effects can be
#' obtained with \code{\link{tie_effects}()}.
#'
#' The majority of the statistics can be scaled in some way, see
#' the documentation of the \code{scaling} argument in the separate effect
#' functions for more information on this.
#'
#' The majority of the statistics can account for the event type
#' included as a dependent variable, see the documentation of the
#' \code{consider_type} argument in the separate effect functions for more
#' information on this.
#'
#' Note that events in the relational event history can be directed or
#' undirected. Some statistics are only defined for either directed or
#' undirected events (see the documentation of the statistics). Note that
#' undirected events are only available for the tie-oriented model.
#'
#' @section attr_actors:
#' For the computation of the \emph{exogenous} statistics an attributes object
#' with the exogenous covariate information has to be supplied to the
#' \code{attr_actors} argument in either \code{remstats()} or in the separate
#' effect functions supplied to the \code{..._effects} arguments (e.g., see
#' \code{\link{send}}). This \code{attr_actors} object should be constructed as
#' follows: A dataframe with rows referring to the attribute value of actor
#' \emph{i} at timepoint \emph{t}. A `name` column is required that contains the
#' actor name (corresponding to the actor names in the relational event
#' history). A `time` column is required that contains the time when attributes
#' change (set to zero if none of the attributes vary over time). Subsequent
#' columns contain the attributes that are called in the specifications of
#' exogenous statistics (column name corresponding to the string supplied to
#' the \code{variable} argument in the effect function). Note that the
#' procedure for the exogenous effects `tie' and `event' deviates from this,
#' here the exogenous covariate information has to be specified in a different
#' way, see \code{\link{tie}} and \code{\link{event}}.
#'
#' @section attr_dyads:
#' For the computation of the \emph{dyad exogenous} statistics with \code{tie()}, an attributes object with the exogenous covariates information per dyad has to be supplied. This is a \code{data.frame} or \code{matrix} containing attribute information for dyads. If \code{attr_dyads} is a \code{data.frame}, the first two columns should represent "actor1" and "actor2" (for directed events, "actor1" corresponds to the sender, and "actor2" corresponds to the receiver). Additional columns can represent dyads' exogenous attributes. If attributes vary over time, include a column named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond to "actor1", columns to "actor2", and cells contain dyads' exogenous attributes.
#'
#' @section Memory:
#' The default `memory` setting is `"full"`, which implies that at each time
#' point $t$ the entire event history before $t$ is included in the computation
#' of the statistics. Alternatively, when `memory` is set to `"window"`, only
#' the past event history within a given time window is considered (see
#' Mulders & Leenders, 2019). This length of this time window is set by the
#' `memory_value` parameter. For example, when `memory_value = 100` and `memory
#' = "window"`, at time point $t$ only the past events that happened at most
#' 100 time units ago are included in the computation of the statistics.
#' A third option is to set `memory` to `"interval"`. In this case, the past
#' event history within a given time interval is considered. For example, when
#' `"memory_value" = c(50, 100)` and `memory = "window"`, at time point $t$
#' only the past events that happened between 50 and 100 time units ago are
#' included in the computation of the statistics. Finally, the fourth option is
#' to set `memory` to `"decay"`. In this case, the weight of the past event in
#' the computation of the statistics depend on the elapsed time between $t$ and
#' the past event. This weight is determined based on an exponential decay
#' function with half-life parameter `memory_value` (see Brandes et al., 2009).
#'
#' @section Event weights:
#' Note that if the relational event history contains a column that is named
#' ``weight'', it is assumed that these affect the endogenous statistics. These
#' affect the computation of all endogenous statistics with a few exceptions
#' that follow logically from their definition (e.g., the recenyContinue
#' statistic does depend on time since the event and not on event weights).
#'
#' @section Subset the event history using 'start' and 'stop':
#' It is possible to compute statistics for a segment of the relational event 
#' sequence, based on the entire event history. This is done by specifying the 
#' 'start' and 'stop' values as the indices for the first and last event times 
#' for which statistics are needed. For instance, setting 'start = 5' and 'stop 
#' = 5' calculates statistics for the 5th event in the relational event 
#' sequence, considering events 1-4 in the history. Note that in cases of 
#' simultaneous events with the 'method' set to 'pt' (per timepoint), 'start' 
#' and 'stop' should correspond to the indices of the first and last 
#' \emph{unique} event timepoints for which statistics are needed. For example, 
#' if 'start = 5' and 'stop = 5', statistics are computed for the 5th unique 
#' timepoint in the relational event sequence, considering all events occurring 
#' at unique timepoints 1-4.
#'
#' @section Adjacency matrix:
#' Optionally, a previously computed adjacency matrix can be supplied. Note
#' that the endogenous statistics will be computed based on this adjacency
#' matrix. Hence, supplying a previously computed adjacency matrix can reduce
#' computation time but the user should be absolutely sure the adjacency matrix
#' is accurate.
#'
#' @return An object of class 'tomstats'. Array with the computed statistics,
#' where rows refer to time points, columns refer to potential relational event
#' (i.e., potential edges) in the risk set and slices refer to statistics.
#' Statistics are computed once per unique time point (per-timepoint "pt"
#' method), so the number of rows in the output equals \code{reh$M} (the number 
#' of  unique time points), which may be less than the total number of observed 
#' events when simultaneous events are present. The 'tomstats' object has the 
#' following attributes:
#'   \describe{
#'     \item{\code{model}}{Type of model that is estimated.}
#'     \item{\code{formula}}{Model formula, obtained from the formula inputted to 'tie_effects'.}
#'     \item{\code{riskset}}{The risk set used to construct the statistics.}
#'   }
#' @export
tomstats <- function(
		effects,
		reh,
		attr_actors = NULL,
		attr_dyads = NULL,
		memory = c("full", "window", "decay", "interval"),
		memory_value = NA,
		start = 2,
		stop = Inf,
		display_progress = FALSE,
		# new
		sampling = FALSE, # sampling = TRUE
		samp_num = NULL, # samp_num = 20
		seed = NULL
) {
	#remove not needed arguments
	
	method <- "pt"
	
	memory <- match.arg(memory)
	
	if (!isTRUE(sampling)) {
		# original behavior
		return(tomstats0(
			effects = effects,
			reh = reh,
			attr_actors = attr_actors,
			attr_dyads = attr_dyads,
			memory = memory,
			memory_value = memory_value,
			start = start,
			stop = stop,
			display_progress = display_progress
		))
		
	}else{
	
	# ---- sampling mode checks ----
	reh <- normalize_reh(reh)  # class already checked above in non-sampling dispatch
	if (is.null(samp_num) || length(samp_num) != 1L || is.na(samp_num)) {
		stop("sampling=TRUE requires a single non-missing integer 'samp_num'.")
	}
	samp_num <- as.integer(samp_num)
	if (samp_num <= 0L) stop("'samp_num' must be > 0.")
	
	if (!is.null(seed)) set.seed(as.integer(seed))
	
	# ---- prepare inputs exactly as current tomstats() does ----
	inputs <- prepare_tomstats(
		effects = effects,
		reh = reh,
		attr_actors = attr_actors,
		attr_dyads = attr_dyads,
		memory = memory,
		memory_value = memory_value,
		start = start,
		stop = stop,
		method = method
	)
	
	form <- inputs$form
	effectNames <- inputs$effectNames
	edgelist <- inputs$edgelist
	weights <- inputs$weights
	actors <- inputs$actor
	types <- inputs$types
	riskset <- inputs$riskset                 # base dyad list (size D_base)
	risksetMatrix <- inputs$risksetMatrix
	memory <- inputs$memory
	memory_value <- inputs$memory_value
	scaling <- inputs$scaling
	consider_type     <- inputs$consider_type
	consider_type_cpp <- inputs$consider_type_cpp
	covar <- inputs$covar
	interactions <- inputs$interactions
	start0 <- inputs$start                    # 0-based for C++
	stop0 <- inputs$stop                      # 0-based for C++
	method <- inputs$method
	with_type_riskset <- inputs$with_type_riskset
	needs_typed_riskset <- inputs$needs_typed_riskset
	C <- inputs$C
	
	# ---- construct row index -> event indices, and cases per row ----
	ed <- reh$edgelist
	if (is.null(ed) || !all(c("time","actor1","actor2") %in% names(ed))) {
		stop("reh$edgelist must contain columns time, actor1, actor2.")
	}
	
	# determine row timepoints/events used by compute_stats_tie (C++ uses start0/stop0 on time_points)
	if (method == "pt") {
		time_points_all <- sort(unique(ed$time))
		# start0/stop0 are 0-based indices into time_points_all
		row_times <- time_points_all[(start0 + 1L):(stop0 + 1L)]
		events_by_row <- lapply(row_times, function(tt) which(ed$time == tt))
	} else { # pe
		# C++ uses time_points = edgelist.col(0) and then subvec(start0, stop0)
		# That corresponds to event rows (start0+1):(stop0+1) in the edgelist passed into prepare_tomstats.
		ev_idx <- (start0 + 1L):(stop0 + 1L)
		events_by_row <- as.list(ev_idx)
	}
	
	M <- length(events_by_row)

	# For ext=FALSE with typed stats (needs_typed_riskset=TRUE, with_type_riskset=FALSE):
	# sample from the UNTYPED riskset so sample_map indexes ts_full columns directly.
	# C++ still receives the typed riskset; sample_map is expanded to typed positions
	# before the C++ call and collapsed back afterwards.
	# All other cases: sample from the base (possibly typed) riskset as before.
	sample_untyped <- needs_typed_riskset && !with_type_riskset
	sample_riskset <- if (sample_untyped) inputs$prepR_untyped else riskset

	D_base <- nrow(sample_riskset)
	if (samp_num > D_base) stop("samp_num cannot exceed base riskset size.")

	directed <- isTRUE(reh$meta$directed)

	make_key_untyped <- function(a1, a2) {
		if (!directed) { lo <- pmin(a1,a2); hi <- pmax(a1,a2) } else { lo <- a1; hi <- a2 }
		paste(lo, hi, sep="|")
	}
	make_key_typed <- function(a1, a2, ty) {
		if (!directed) { lo <- pmin(a1,a2); hi <- pmax(a1,a2) } else { lo <- a1; hi <- a2 }
		paste(lo, hi, ty, sep="|")
	}

	if (sample_untyped) {
		# Untyped sampling: build key directly from 0-based integer cols of sample_riskset.
		# Avoid modify_riskset — name lookup is fragile when prepR_untyped and the typed
		# riskset have different dyad subsets (e.g. active riskset where not all actor pairs
		# appear in both types). Use row index as 1-based id into sample_riskset.
		rs_key      <- make_key_untyped(sample_riskset[, 1], sample_riskset[, 2])
		key_to_base <- setNames(seq_len(nrow(sample_riskset)), rs_key)
	} else {
		# Typed sampling: build key directly from 0-based integer cols of riskset.
		# riskset col 4 is 0-based dyadID; convert to 1-based for key_to_base.
		rs_key      <- make_key_typed(riskset[, 1], riskset[, 2], riskset[, 3])
		key_to_base <- setNames(riskset[, 4] + 1L, rs_key)
	}

	# typed riskset ids should be 0..D_typed-1
	stopifnot(max(riskset[,4]) == nrow(riskset) - 1L)

	# ---- sample_map, case_pos, pi, log_pi ----
	# sample_map is 0-based, indexes sample_riskset (untyped when sample_untyped=TRUE)
	sample_map <- matrix(NA_integer_, nrow = M, ncol = samp_num)
	pi         <- matrix(NA_real_,    nrow = M, ncol = samp_num)
	log_pi     <- matrix(NA_real_,    nrow = M, ncol = samp_num)
	case_pos   <- case_mult <- vector("list", M)

	for (m in seq_len(M)) {
		ev_idx <- events_by_row[[m]]
		if (sample_untyped) {
			# Use 0-based actor IDs from the prepared edgelist matrix (cols 2,3)
			# to match the 0-based integer keys built from sample_riskset.
			keys <- make_key_untyped(edgelist[ev_idx, 2], edgelist[ev_idx, 3])
		} else if (C > 1L && ncol(edgelist) >= 4L) {
			keys <- make_key_typed(edgelist[ev_idx, 2], edgelist[ev_idx, 3], edgelist[ev_idx, 4])
		} else {
			keys <- make_key_untyped(edgelist[ev_idx, 2], edgelist[ev_idx, 3])
		}
		cases_all <- as.integer(unname(key_to_base[keys]) - 1L)
		tab <- table(cases_all)
		cases <- as.integer(names(tab))          # unique case dyads (0-based)
		case_mult[[m]] <- as.integer(tab)        # multiplicities
		m_t <- length(cases)

		if (samp_num < m_t) {
			idx <- sample.int(m_t, size = samp_num, replace = FALSE)
			cases <- cases[idx]
			case_mult[[m]] <- case_mult[[m]][idx]
			m_t <- length(cases)
		}

		c_m <- samp_num - m_t
		if (c_m > 0L) {
			pool     <- setdiff(0:(D_base - 1L), cases)
			controls <- sample(pool, size = c_m, replace = FALSE)
			S_m      <- c(cases, controls)
			pi_ctrl  <- c_m / (D_base - m_t)
			pi_m     <- c(rep(1, m_t), rep(pi_ctrl, c_m))
		} else {
			S_m  <- cases
			pi_m <- rep(1, m_t)
		}

		sample_map[m, ] <- as.integer(S_m)
		case_pos[[m]]   <- seq_len(m_t)
		pi[m, ]         <- pi_m
		log_pi[m, ]     <- log(pi_m)
	}
	storage.mode(sample_map) <- "integer"

	# ---- build sample_map_cpp for C++ (ext=FALSE only) ----
	# C++ needs 0-based typed riskset row indices.
	#
	# For sample_untyped=TRUE: run C separate C++ calls, one per type.
	# Each call uses sample_map_cpp_c [M x samp_num] pointing to the typed
	# riskset row for type c for each sampled untyped dyad (or the first
	# available row as fallback when that type doesn't exist for a dyad).
	# Results are assembled into named type slices in R without any zeroing:
	# the value for type ci comes from call ci, so it's always correct.
	# This avoids expanding sample_map to [M x samp_num*C] and handles both
	# Scenario A (1 type per dyad) and Scenario B (C types per dyad) correctly.
	#
	# For sample_untyped=FALSE: single C++ call as before.
	if (sample_untyped) {
		key_typed_all <- make_key_untyped(riskset[, 1], riskset[, 2])

		# For each untyped dyad and each type, find the typed riskset row (1-based).
		# If a dyad doesn't have a row for that type, fall back to any existing row
		# so C++ never receives NA. The value will be overridden to 0 in assembly.
		type_ids_0based <- types$typeID  # 0-based, length C

		# typed_row_map[[d]][ci] = 1-based riskset row for untyped dyad d, type ci
		# type_exists[[d]][ci]   = TRUE if that type actually exists for dyad d
		typed_row_map <- vector("list", D_base)
		type_exists   <- vector("list", D_base)
		for (d in seq_len(D_base)) {
			rows   <- which(key_typed_all == rs_key[d])
			exists <- sapply(type_ids_0based, function(tid) any(riskset[rows, 3] == tid))
			rows_c <- sapply(type_ids_0based, function(tid) {
				r <- rows[riskset[rows, 3] == tid]
				if (length(r) == 0L) rows[1L] else r[1L]  # fallback to any row
			})
			typed_row_map[[d]] <- rows_c
			type_exists[[d]]   <- exists
		}

		# Run one C++ call per type, each with [M x samp_num] sample_map
		# For "separate"/"interact": weight-mask to type-c events per pass
		needs_mp_samp <- any(consider_type[seq_along(inputs$effects)] %in% c(1L, 2L))
		mp_idx_samp  <- which(consider_type[seq_along(inputs$effects)] %in% c(1L, 2L))
		mp_eff_samp  <- if (needs_mp_samp) effectNames[mp_idx_samp] else effectNames
		mp_scal_samp <- if (needs_mp_samp) scaling[mp_idx_samp] else scaling
		mp_ct_samp   <- rep(FALSE, length(mp_eff_samp))

		stats_by_type <- vector("list", C)
		for (ci in seq_len(C)) {
			smap_c <- matrix(NA_integer_, nrow = M, ncol = samp_num)
			for (m in seq_len(M)) {
				for (s in seq_len(samp_num)) {
					d <- sample_map[m, s] + 1L          # 1-based untyped index
					smap_c[m, s] <- typed_row_map[[d]][ci] - 1L  # 0-based for C++
				}
			}
			storage.mode(smap_c) <- "integer"
			type_id_c <- types$typeID[ci]
			wt_c   <- if (needs_mp_samp) weights * (edgelist[, 4] == type_id_c) else weights
			eff_c  <- if (needs_mp_samp) mp_eff_samp  else effectNames
			scal_c <- if (needs_mp_samp) mp_scal_samp else scaling
			ct_c   <- if (needs_mp_samp) mp_ct_samp   else consider_type_cpp
			stats_by_type[[ci]] <- compute_stats_tie_sampled(
				effects = eff_c,
				edgelist = edgelist,
				weights = wt_c,
				riskset = riskset,
				risksetMatrix = risksetMatrix,
				covariates = covar,
				interactions = interactions,
				memory = memory,
				memory_value = memory_value,
				scaling = scal_c,
				consider_type = ct_c,
				start = start0,
				stop = stop0,
				directed = reh$meta$directed,
				display_progress = display_progress,
				method = method,
				sample_map = smap_c
			)

			# Pshift override: C++ ignores weights so weight masking has no effect.
			# Use R-level compute_pshift_type_c_tie indexed through sample_map.
			if (needs_mp_samp && !is.null(inputs$prepR_untyped)) {
				ps_in_mp_samp <- which(mp_eff_samp %in% PSHIFT_NAMES)
				if (length(ps_in_mp_samp) > 0) {
					rs_a1    <- inputs$prepR_untyped[, 1]
					rs_a2    <- inputs$prepR_untyped[, 2]
					M_out_ps <- stop0 - start0 + 1L
					sbt      <- stats_by_type[[ci]]
					idx1     <- sample_map + 1L   # [M x samp_num], 1-based into D_untyped
					for (ei in ps_in_mp_samp) {
						pshift_num   <- PSHIFT_TYPE_NUM[mp_eff_samp[ei]]
						correct_vals <- compute_pshift_type_c_tie(
							pshift_num, edgelist, edgelist[, 4], type_id_c,
							rs_a1, rs_a2, M_out_ps, start0, stop0
						)
						sbt[, , ei] <- matrix(
							correct_vals[cbind(rep(seq_len(M), samp_num), as.vector(idx1))],
							nrow = M, ncol = samp_num
						)
					}
					stats_by_type[[ci]] <- sbt
				}
			}
		}

		# Apply variable names to all per-type results.
		# When needs_mp_samp: stats_by_type[[ci]] has only mp_eff_samp effects.
		# Run one additional unmasked pass (all effects, full weights) then merge:
		# "ignore" effects come from the unmasked pass, "separate"/"interact" from
		# the type-c masked passes.
		if (needs_mp_samp) {
			# Unmasked pass using type-1 sample_map (smap for ci=1)
			smap_unmasked <- matrix(NA_integer_, nrow = M, ncol = samp_num)
			for (m in seq_len(M)) {
				for (s in seq_len(samp_num)) {
					d <- sample_map[m, s] + 1L
					smap_unmasked[m, s] <- typed_row_map[[d]][1L] - 1L
				}
			}
			storage.mode(smap_unmasked) <- "integer"
			full_stats <- compute_stats_tie_sampled(
				effects = effectNames,
				edgelist = edgelist,
				weights = weights,
				riskset = riskset,
				risksetMatrix = risksetMatrix,
				covariates = covar,
				interactions = interactions,
				memory = memory,
				memory_value = memory_value,
				scaling = scaling,
				consider_type = consider_type_cpp,
				start = start0,
				stop = stop0,
				directed = reh$meta$directed,
				display_progress = display_progress,
				method = method,
				sample_map = smap_unmasked
			)
			# Merge: for each type-c pass, replace mp_eff_samp slices with
			# the type-c filtered values; keep all other slices from full_stats
			for (ci in seq_len(C)) {
				sbt <- full_stats
				mp_arr <- stats_by_type[[ci]]
				for (ei in seq_along(mp_eff_samp)) {
					en <- mp_eff_samp[ei]
					p_full <- which(effectNames == en)
					if (length(p_full) > 0 && ei <= dim(mp_arr)[3]) {
						sbt[, , p_full] <- mp_arr[, , ei]
					}
				}
				stats_by_type[[ci]] <- add_variable_names(
					sbt, effectNames, inputs$effects, interactions)
			}
		} else {
			stats_by_type <- lapply(stats_by_type, function(s)
				add_variable_names(s, effectNames, inputs$effects, interactions))
		}
		statistics <- stats_by_type[[1]]

	} else {
		sample_map_cpp <- sample_map

		# ---- compute sampled statistics in C++ ----
		statistics <- compute_stats_tie_sampled(
			effects = effectNames,
			edgelist = edgelist,
			weights = weights,
			riskset = riskset,
			risksetMatrix = risksetMatrix,
			covariates = covar,
			interactions = interactions,
			memory = memory,
			memory_value = memory_value,
			scaling = scaling,
			consider_type = consider_type_cpp,
			start = start0,
			stop = stop0,
			directed = reh$meta$directed,
			display_progress = display_progress,
			method = method,
			sample_map = sample_map_cpp
		)
		statistics <- add_variable_names(statistics, effectNames, inputs$effects, interactions)

		# Multi-pass for "separate"/"interact" with ext=TRUE
		if (C > 1 && any(consider_type[seq_along(inputs$effects)] %in% c(1L, 2L))) {
			mp_idx_ext       <- which(consider_type[seq_along(inputs$effects)] %in% c(1L, 2L))
			mp_eff_ext       <- effectNames[mp_idx_ext]
			mp_scal_ext      <- scaling[mp_idx_ext]
			mp_ct_ext        <- rep(FALSE, length(mp_eff_ext))
			mp_consider_type_ext <- consider_type[mp_idx_ext]  # 1L=separate, 2L=interact
			stats_by_type <- vector("list", C)
			for (ci in seq_len(C)) {
				type_id_c <- types$typeID[ci]
				wt_c <- weights * (edgelist[, 4] == type_id_c)
				stats_c <- compute_stats_tie_sampled(
					effects = mp_eff_ext,
					edgelist = edgelist,
					weights = wt_c,
					riskset = riskset,
					risksetMatrix = risksetMatrix,
					covariates = covar,
					interactions = interactions,
					memory = memory,
					memory_value = memory_value,
					scaling = mp_scal_ext,
					consider_type = mp_ct_ext,
					start = start0,
					stop = stop0,
					directed = reh$meta$directed,
					display_progress = display_progress,
					method = method,
					sample_map = sample_map_cpp
				)
				dimnames(stats_c)[[3]] <- mp_eff_ext
				# Pshift override for ext=TRUE + interact only.
				# ext=TRUE + separate: base C++ stat is correct (all type slices
				# share the same dyad pattern — defensible for separate semantics).
				ps_in_ext <- which(mp_eff_ext %in% PSHIFT_NAMES &
				                   mp_consider_type_ext == 2L)
				if (length(ps_in_ext) > 0) {
					rs_a1    <- riskset[, 1]
					rs_a2    <- riskset[, 2]
					M_out_ps <- stop0 - start0 + 1L
					S        <- ncol(sample_map_cpp)
					idx1     <- sample_map_cpp + 1L
					for (ei in ps_in_ext) {
						pshift_num   <- PSHIFT_TYPE_NUM[mp_eff_ext[ei]]
						correct_vals <- compute_pshift_type_c_tie(
							pshift_num, edgelist, edgelist[, 4], type_id_c,
							rs_a1, rs_a2, M_out_ps, start0, stop0
						)
						stats_c[, , ei] <- matrix(
							correct_vals[cbind(rep(seq_len(M_out_ps), S), as.vector(idx1))],
							nrow = M_out_ps, ncol = S
						)
					}
				}
				stats_by_type[[ci]] <- stats_c
			}
		}
	}

	# ---- post-processing: type splitting ----
	#
	# For ext=FALSE (sample_untyped=TRUE): assemble type-split slices from
	# stats_by_type. For consider_type=TRUE effects, slice ci gets its value
	# from stats_by_type[[ci]], zeroed where type ci doesn't exist for the dyad.
	# For consider_type=FALSE and non-typed effects, all C calls return the same
	# aggregated value so we just use the first call.
	#
	# For ext=TRUE (with_type_riskset=TRUE): split_type_slices as before.
	if (needs_typed_riskset && C > 1) {
		if (sample_untyped) {
			statistics <- assemble_type_slices_sampled(
				stats_by_type = stats_by_type,
				effects       = inputs$effects,
				consider_type = consider_type,
				types         = types,
				sample_map    = sample_map,       # 0-based untyped indices
				type_exists   = type_exists
			)
		} else if (with_type_riskset) {
			statistics <- split_type_slices_sampled(
				statistics    = statistics,
				effects       = inputs$effects,
				consider_type = consider_type,
				types         = types,
				riskset       = riskset,
				sample_map    = sample_map + 1L,  # 1-based row index into riskset
				stats_by_type = if (exists("stats_by_type")) stats_by_type else NULL
			)
		}
	}

	# base riskset output: untyped when ext=FALSE, typed otherwise
	if (!with_type_riskset && needs_typed_riskset) {
		riskset_out <- build_untyped_riskset(reh, actors, types)
	} else {
		riskset_out <- modify_riskset(riskset, reh, actors, types)
	}
	
	class(statistics) <- c("tomstats_sampled", "remstats")
	attr(statistics, "model")           <- "tie"
	attr(statistics, "formula")         <- form
	attr(statistics, "riskset")         <- riskset_out
	attr(statistics, "sample_map")      <- sample_map + 1
	attr(statistics, "samp_num")        <- samp_num
	attr(statistics, "events_mult")     <- case_mult
	attr(statistics, "subset")          <- data.frame(start = start0 + 1L, stop = stop0 + 1L)
	attr(statistics, "method")          <- method
	attr(statistics, "dyad_keys")       <- key_to_base
	attr(statistics, "case_pos")        <- case_pos
	attr(statistics, "samp_prob")       <- pi
	attr(statistics, "log_samp_prob")   <- log_pi
	attr(statistics, "seed")            <- seed
	attr(statistics, "sampling_scheme") <- "uniform_wor_static_riskset"
	
	return(statistics)
	}
}

#' Combine two or more remstats objects
#'
#' Function to bind any number of remstats objects into one while duplicated
#' statistics in the combined object are removed based on their name.
#'
#' @param ... Any number of \code{\link{remstats}} objects. All the
#' \code{remstats} objects must have matching dimensions, except for the third
#' dimension. Note that duplicated statistics in the combined object are 
#' removed based on their name.
#'
#' @return \code{statistics } array with the combined statistics, where rows
#' refer to time points, columns refer to potential relational event (i.e.,
#' potential edges) in the risk set and slices refer to statistics
#'
#' @examples
#' library(remstats)
#' 
#' # Load the data
#' data(history)
#' data(info)
#' 
#' # Prepare the data
#' reh <- remify::remify(edgelist = history, model = "actor")
#' 
#' # Obtain two different statistics objects
#' effects1 <- ~ inertia():receive("extraversion") + otp()
#' stats1 <- remstats(receiver_effects = effects1, reh = reh, attr_actors = info)
#' effects2 <- ~ reciprocity()
#' stats2 <- remstats(receiver_effects = effects2, reh = reh, attr_actors = info)
#' 
#' # Bind the two statistics objects
#' statsC <- bind_remstats(stats1, stats2)
#'
#' @export
bind_remstats <- function(...) {
    arg.list <- list(...)

    # Check class
    stopifnot(
        "All objects should be of class remstats" =
            all(sapply(arg.list, function(x) any(class(x) == "remstats")))
    )

    firstclass <- class(arg.list[[1]])[1]
    if (!all(sapply(arg.list, function(x) class(x)[1] == firstclass))) {
        stop(paste0("All objects should be of class ", firstclass))
    }

    if (firstclass == "tomstats") {
        # Check risk set
        firstrs <- attr(arg.list[[1]], "riskset")
        stopifnot(
            "Risk sets of the remstats objects should be equal" =
                all(sapply(arg.list, function(x) attr(x, "riskset") == firstrs))
        )

        # Check names
        keep.names <- dimnames(arg.list[[1]])[[3]]
        keep.ids <- vector(mode = "list", length = length(arg.list))
        keep.ids[[1]] <- 1:(dim(arg.list[[1]])[3])
        for(i in 2:length(arg.list)) {
            current.names <- dimnames(arg.list[[i]])[[3]]
            keep.ids[[i]] <- which(!(current.names %in% keep.names))
            keep.names <- union(keep.names, current.names)
        }

        # Deal with empty keep.ids
        ids.length <- sapply(keep.ids, length)
        if (any(ids.length == 0)) {
            remove.arrays <- which(ids.length == 0)
            arg.list <- arg.list[-remove.arrays]
            keep.ids <- keep.ids[-remove.arrays]      
        }

        # Combine statistics
        statistics <- combine_stats(arg.list, keep.ids)

        # Names of the statistics
        # stat_names <- lapply(arg.list, function(x) dimnames(x)[[3]])
        # stat_names <- do.call(c, stat_names)
        # stat_names <- stat_names[!duplicated(stat_names)]
        dimnames(statistics) <- list(NULL, NULL, keep.names)

        # Duplicated stats warning
        n_stats <- sum(sapply(arg.list, function(x) dim(x)[3]))
        if (dim(statistics)[[3]] < n_stats) {
            warning("Duplicate statistics detected. Removing the duplicate slice.")
        }

        # Combine formula
        # form.list <- sapply(arg.list, function(x) attr(x, "form"))
        # processed_formulas <- sapply(form.list, function(formula) {
        #     formula_string <- as.character(formula)
        #     formula_string <- gsub("~", "", formula_string[-1])
        #     formula_string
        # })
        # combined_formula <- paste(processed_formulas, collapse = " + ")
        # final_formula <- paste("~", combined_formula)
        # final_formula <- stats::as.formula(final_formula)
        final_formula <- "Combined remstats object"

        # Format output
        class(statistics) <- class(arg.list[[1]])
        attr(statistics, "model") <- attr(arg.list[[1]], "model")
        attr(statistics, "formula") <- final_formula
        attr(statistics, "riskset") <- attr(arg.list[[1]], "riskset")
        attr(statistics, "subset") <- attr(arg.list[[1]], "subset")
        attr(statistics, "method") <- attr(arg.list[[1]], "method")
        attr(statistics, "adjmat") <- attr(arg.list[[1]], "adjmat")

        # Output
        return(statistics)
    }

    if (firstclass == "aomstats") {
        # Check actors set
        firstactors <- attr(arg.list[[1]], "actors")
        stopifnot(
            "Actors sets of the remstats objects should be equal" =
                all(sapply(arg.list, function(x) {
                    attr(x, "actors") ==
                        firstactors
                }))
        )

        statistics <- list()
        statistics$sender_stats <- NULL
        statistics$receiver_stats <- NULL

        # Combine rate statistics
        rate.list <- lapply(arg.list, function(x) x$sender_stats)
        # Filter empty list elements
        rate.list <- Filter(Negate(is.null), rate.list)
        if (length(rate.list) > 1) {
            # Check names
            keep.names <- dimnames(rate.list[[1]])[[3]]
            keep.ids <- vector(mode = "list", length = length(rate.list))
            keep.ids[[1]] <- 1:(dim(rate.list[[1]])[3])
            for(i in 2:length(rate.list)) {
                current.names <- dimnames(rate.list[[i]])[[3]]
                keep.ids[[i]] <- which(!(current.names %in% keep.names))
                keep.names <- union(keep.names, current.names)
            }

            # Deal with empty keep.ids
            ids.length <- sapply(keep.ids, length)
            if (any(ids.length == 0)) {
                remove.arrays <- which(ids.length == 0)
                rate.list <- rate.list[-remove.arrays]
                keep.ids <- keep.ids[-remove.arrays]      
            }

            # Combine statistics
            rate_statistics <- combine_stats(rate.list, keep.ids)

            # Names of the statistics
            # stat_names <- lapply(rate.list, function(x) dimnames(x)[[3]])
            # stat_names <- do.call(c, stat_names)
            # dimnames(rate_statistics) <- list(NULL, NULL, stat_names)
            dimnames(rate_statistics) <- list(NULL, NULL, keep.names)

            # Duplicated stats warning
            n_stats <- sum(sapply(arg.list, function(x) dim(x$sender_stats)[3]))
            if (dim(rate_statistics)[[3]] < n_stats) {
                warning("Duplicate statistics detected in the sender_stats. Removing the duplicate slice.")
            }

            # Save
            statistics$sender_stats <- rate_statistics

            # Combine formula
            # form.list <- sapply(arg.list, function(x) attr(x, "formula")$rate)
            # processed_formulas <- sapply(form.list, function(formula) {
            #     formula_string <- as.character(formula)
            #     formula_string <- gsub("~", "", formula_string[-1])
            #     formula_string
            # })
            # combined_formula <- paste(processed_formulas, collapse = " + ")
            # final_formula <- paste("~", combined_formula)
            # final_formula <- stats::as.formula(final_formula)
            final_formula <- "Combined remstats object"
            attr(statistics, "formula")$rate <- final_formula

        } else {

            if (length(rate.list) == 1) {
                statistics$sender_stats <- rate.list[[1]]
            }
            attr(statistics, "formula")$rate <- attr(statistics, "formula")$rate

        }

        # Combine choice statistics
        choice.list <- lapply(arg.list, function(x) x$receiver_stats)

        # Filter empty list elements
        choice.list <- Filter(Negate(is.null), choice.list)

        if (length(choice.list) > 1) {
            # Check names
            keep.names <- dimnames(choice.list[[1]])[[3]]
            keep.ids <- vector(mode = "list", length = length(choice.list))
            keep.ids[[1]] <- 1:(dim(choice.list[[1]])[3])
            for(i in 2:length(choice.list)) {
                current.names <- dimnames(choice.list[[i]])[[3]]
                keep.ids[[i]] <- which(!(current.names %in% keep.names))
                keep.names <- union(keep.names, current.names)
            }

            # Deal with empty keep.ids
            ids.length <- sapply(keep.ids, length)
            if (any(ids.length == 0)) {
                remove.arrays <- which(ids.length == 0)
                choice.list <- choice.list[-remove.arrays]
                keep.ids <- keep.ids[-remove.arrays]      
            }

            # Combine statistics
            choice_statistics <- combine_stats(choice.list, keep.ids)

            # Names of the statistics
            # stat_names <- lapply(choice.list, function(x) dimnames(x)[[3]])
            # stat_names <- do.call(c, stat_names)
            # dimnames(choice_statistics) <- list(NULL, NULL, stat_names)
            dimnames(choice_statistics) <- list(NULL, NULL, keep.names)

            # Duplicated stats warning
            n_stats <- sum(sapply(arg.list, function(x) dim(x$receiver_stats)[3]))
            if (dim(choice_statistics)[[3]] < n_stats) {
                warning("Duplicate statistics detected in the receiver_stats. Removing the duplicate slice.")
            }

            # Save
            statistics$receiver_stats <- choice_statistics

            # Combine formula
            # form.list <- sapply(arg.list, function(x) attr(x, "formula")$choice)
            # processed_formulas <- sapply(form.list, function(formula) {
            #     formula_string <- as.character(formula)
            #     formula_string <- gsub("~", "", formula_string[-1])
            #     formula_string
            # })
            # combined_formula <- paste(processed_formulas, collapse = " + ")
            # final_formula <- paste("~", combined_formula)
            # final_formula <- stats::as.formula(final_formula)
            final_formula <- "Combined remstats object"
            attr(statistics, "formula")$choice <- final_formula
        } else {
            if (length(choice.list) == 1) {
                statistics$receiver_stats <- choice.list[[1]]
            }
            attr(statistics, "formula")$choice <-
                attr(statistics, "formula")$choice
        }

        # Format output
        class(statistics) <- class(arg.list[[1]])
        attr(statistics, "model") <- attr(arg.list[[1]], "model")
        attr(statistics, "actors") <- attr(arg.list[[1]], "actors")
        attr(statistics, "subset") <- attr(arg.list[[1]], "subset")
        attr(statistics, "method") <- attr(arg.list[[1]], "method")

        # Output
        return(statistics)
    }
}

#' Printing Relational Event Network Statistics
#'
#' Print a \code{\link{remstats}} object in a user-friendly format.
#'
#' @param x object of class \code{\link{remstats}}.
#' @param ... further arguments passed to or from other methods.
#' 
#' @return The function prints formatted information about the remstats object to the console, presenting details about the relational event network statistics in a user-friendly format.
#'
#' @examples
#' rehObject <- remify::remify(edgelist = history, model = "tie")
#' remstatsObject <- remstats::remstats(reh = rehObject, tie_effects = ~ remstats::inertia())
#' print(remstatsObject)
#'
#' rehObject <- remify::remify(edgelist = history, model = "actor")
#' remstatsObject <- remstats::remstats(reh = rehObject, receiver_effects = ~ inertia())
#' print(remstatsObject)
#'
#' @method print remstats
#' @export
print.remstats <- function(x, ...) {
  if (!any(class(x) == "remstats")) {
    stop("Expected object of class 'remstats'")
  }

  title <- "Relational Event Network Statistics"
  model <- ifelse(any(class(x) == "tomstats"), "tie-oriented", "actor-oriented")
  model.title <- paste("> Model:", model, sep = " ")
  method <- attributes(x)$method
  method.fancy <- ifelse(method == "pt", "time point", "event")
  method.title <- paste("> Computation method: per", method.fancy)

  if (model == "tie-oriented") {
    row.title <- ifelse(method == "pt", "time points x", "events x")
    dim.stats <- dim(x)
    dim.long <- paste("> Dimensions:",
      dim.stats[1], row.title,
      dim.stats[2], "dyads x",
      dim.stats[3], "statistics",
      sep = " "
    )
    stats.title <- paste("> Statistics:")
    stats.names <- dimnames(x)[[3]]
    stats.names2 <- lapply(1:length(stats.names), function(i) {
      paste0("\t >> ", i, ": ", stats.names[i])
    })
    stats.names3 <- paste(stats.names2, collapse = "\n")
    cat(paste(title, model.title, method.title, dim.long, stats.title, stats.names3,
      sep = "\n"
    ))
  } else if (model == "actor-oriented") {
    if (is.null(x$sender_stats)) {
      sender.title <- "> Sender model: empty"
    } else {
      sender.title <- "> Sender model:"
      row.title <- ifelse(method == "pt", "time points x", "events x")
      dim.sender.stats <- dim(x$sender_stats)
      dim.sender.long <- paste("\t >> Dimensions:",
        dim.sender.stats[1], row.title,
        dim.sender.stats[2], "actors x",
        dim.sender.stats[3], "statistics",
        sep = " "
      )
      stats.sender.title <- paste("\t >> Statistics:")
      stats.sender.names <- dimnames(x$sender_stats)[[3]]
      stats.sender.names2 <- lapply(1:length(stats.sender.names), function(i) {
        paste0("\t \t >>> ", i, ": ", stats.sender.names[i])
      })
      stats.sender.names3 <- paste(stats.sender.names2, collapse = "\n")
    }

    if (is.null(x$receiver_stats)) {
      receiver.title <- "> Receiver model: empty"
    } else {
      receiver.title <- "> Receiver model:"
      dim.receiver.stats <- dim(x$receiver_stats)
      dim.receiver.long <- paste("\t >> Dimensions:",
        dim.receiver.stats[1], "events x",
        dim.receiver.stats[2], "actors x",
        dim.receiver.stats[3], "statistics",
        sep = " "
      )
      stats.receiver.title <- paste("\t >> Statistics:")
      stats.receiver.names <- dimnames(x$receiver_stats)[[3]]
      stats.receiver.names2 <- lapply(
        1:length(stats.receiver.names),
        function(i) {
          paste0("\t \t >>> ", i, ": ", stats.receiver.names[i])
        }
      )
      stats.receiver.names3 <- paste(stats.receiver.names2, collapse = "\n")
    }

    if (is.null(x$sender_stats)) {
      cat(paste(title, model.title, method.title,
        sender.title, receiver.title, dim.receiver.long, stats.receiver.title,
        stats.receiver.names3,
        sep = "\n"
      ))
    } else if (is.null(x$receiver_stats)) {
      cat(paste(title, model.title, method.title,
        sender.title, dim.sender.long, stats.sender.title, stats.sender.names3,
        sep = "\n"
      ))
    } else {
      cat(paste(title, model.title, method.title,
        sender.title, dim.sender.long, stats.sender.title, stats.sender.names3,
        receiver.title, dim.receiver.long, stats.receiver.title,
        stats.receiver.names3,
        sep = "\n"
      ))
    }
  }
}


#' @method print tomstats_sampled
#' @export
print.tomstats_sampled <- function(x, ...) {
	if (!inherits(x, "remstats")) stop("Expected object of class 'remstats'")
	
	title <- "Relational Event Network Statistics"
	method <- attr(x, "method")
	method.fancy <- ifelse(method == "pt", "time point", "event")
	
	# sampled tie-oriented
	model.title  <- "> Model: tie-oriented (sampled)"
	method.title <- paste("> Computation method: per", method.fancy)
	
	dim.stats <- dim(x)
	row.title <- ifelse(method == "pt", "time points x", "events x")
	dim.long <- paste("> Dimensions:",
										dim.stats[1], row.title,
										dim.stats[2], "sampled dyads x",
										dim.stats[3], "statistics")
	
	# sampling metadata (optional)
	samp_num <- attr(x, "samp_num")
	samp.scheme <- attr(x, "sampling_scheme")
	samp.line <- NULL
	if (!is.null(samp_num) || !is.null(samp.scheme)) {
		samp.line <- paste0("> Sampling: ",
												if (!is.null(samp.scheme)) samp.scheme else "TRUE",
												if (!is.null(samp_num)) paste0(" (S=", samp_num, ")") else "")
	}
	
	stats.title <- "> Statistics:"
	stats.names <- dimnames(x)[[3]]
	if (is.null(stats.names)) stats.names <- paste0("stat", seq_len(dim.stats[3]))
	stats.names3 <- paste(vapply(seq_along(stats.names),
															 function(i) paste0("\t >> ", i, ": ", stats.names[i]),
															 character(1)),
												collapse = "\n")
	
	cat(paste(c(title, model.title, method.title, dim.long, samp.line, stats.title, stats.names3),
						collapse = "\n"))
	invisible(x)
}


#' Relational Event Network Statistics Summaries
#'
#' Produce summaries of each statistic from a \code{\link{remstats}} object.
#'
#' @param object object of class \code{\link{remstats}}.
#' @param ... additional arguments affecting the summary produced.
#' 
#' @return The summaries provide information for each statistic included in the 
#' remstats object, offering insights into the distribution and characteristics 
#' of the data.
#'
#' @examples
#' rehObject <- remify::remify(edgelist = history, model = "tie")
#' remstatsObject <- remstats::remstats(reh = rehObject, tie_effects = ~ remstats::inertia())
#' summary(remstatsObject)
#'
#' rehObject <- remify::remify(edgelist = history, model = "actor")
#' remstatsObject <- remstats::remstats(reh = rehObject, receiver_effects = ~ inertia())
#' summary(remstatsObject)
#'
#' @method summary remstats
#' @export
summary.remstats <- function(object, ...) {
  if (!any(class(object) == "remstats")) {
    stop("Expected object of class 'remstats'")
  }

  model <- ifelse(inherits(object, "tomstats") || inherits(object, "tomstats_sampled"),
  								"tie-oriented", "actor-oriented")

  if (model == "tie-oriented") {
    out <- apply(object, 3, function(y) {
      summary(as.vector(y))
    })
  }

  if (model == "actor-oriented") {
    out <- lapply(object, function(y) {
      if (!is.null(y)) {
        apply(y, 3, function(z) {
          summary(as.vector(z))
        })
      } else {
        NULL
      }
    })
  }

  out
}

#' @importFrom remify remify

# Split consider_type=TRUE statistics into C type-specific slices for the
# sampled (CCS) path. Works on [M x S x P] arrays where S = samp_num and
# columns are typed riskset row indices (via sample_map_cpp).
#
# For each effect p where consider_type[p]=TRUE:
#   - produces C slices [M x S], where slice c keeps value at (m,s) only if
#     the sampled dyad's typeID == types$typeID[c], else 0.
# For consider_type=FALSE effects: kept as-is (C++ already aggregated).
# For non-typed effects: kept as-is.
#
# [param] statistics  array [M x S x P] from compute_stats_tie_sampled
# [param] effects     parsed effects list, length P
# [param] consider_type  logical vector length P
# [param] types       data.frame with columns typeName, typeID (0-based)
# [param] riskset     matrix [D x 4], col 3 = 0-based typeID per typed riskset row
# [param] sample_map  integer matrix [M x S], 1-based riskset row indices
#
# [return] array [M x S x P_out] with updated dimnames[[3]]
split_type_slices_sampled <- function(statistics, effects, consider_type,
																			types, riskset, sample_map,
																			stats_by_type = NULL) {
	M          <- dim(statistics)[1]
	S          <- dim(statistics)[2]
	C          <- nrow(types)
	stat_names <- dimnames(statistics)[[3]]

	# [M x S] matrix of typeIDs for each sampled dyad
	type_mat <- matrix(riskset[sample_map, 3L], nrow = M, ncol = S)

	new_slices <- list()

	for (p in seq_len(dim(statistics)[3])) {
		has_consider_type <- p <= length(effects) &&
			"consider_type" %in% names(effects[[p]])

		if (has_consider_type && consider_type[p] > 0L) {
			base_name <- stat_names[p]

			if (!is.null(stats_by_type)) {
				if (consider_type[p] == 2L) {
					# "interact": C^2 slices, zero non-matching dyad type positions
					for (ci in seq_len(C)) {
						past_type_name <- types$typeName[ci]
						sbt   <- stats_by_type[[ci]]
						p_sbt <- which(dimnames(sbt)[[3]] == base_name)
						if (length(p_sbt) == 0) next
						for (ci2 in seq_len(C)) {
							dyad_type_name <- types$typeName[ci2]
							dyad_type_id   <- types$typeID[ci2]
							slice_name <- paste0(base_name, ".", past_type_name,
							                     ".", dyad_type_name)
							slice2d <- sbt[, , p_sbt]
							slice2d[type_mat != dyad_type_id] <- 0
							new_mat <- array(slice2d, dim = c(M, S, 1))
							dimnames(new_mat)[[3]] <- slice_name
							new_slices[[length(new_slices) + 1]] <- new_mat
						}
					}
				} else {
					# "separate": C slices
					for (ci in seq_len(C)) {
						type_name  <- types$typeName[ci]
						slice_name <- paste0(base_name, ".", type_name)
						sbt   <- stats_by_type[[ci]]
						p_sbt <- which(dimnames(sbt)[[3]] == base_name)
						if (length(p_sbt) == 0) next
						slice2d <- sbt[, , p_sbt]
						new_mat <- array(slice2d, dim = c(M, S, 1))
						dimnames(new_mat)[[3]] <- slice_name
						new_slices[[length(new_slices) + 1]] <- new_mat
					}
				}
			} else if (consider_type[p] == 2L) {
				# "interact" without multi-pass: C^2 slices
				for (ci in seq_len(C)) {
					past_type_name <- types$typeName[ci]
					for (ci2 in seq_len(C)) {
						dyad_type_name <- types$typeName[ci2]
						dyad_type_id   <- types$typeID[ci2]
						slice_name <- paste0(base_name, ".", past_type_name,
						                     ".", dyad_type_name)
						slice2d <- statistics[, , p]
						slice2d[type_mat != dyad_type_id] <- 0
						new_mat <- array(slice2d, dim = c(M, S, 1))
						dimnames(new_mat)[[3]] <- slice_name
						new_slices[[length(new_slices) + 1]] <- new_mat
					}
				}
			} else {
				# "separate" without multi-pass: C slices
				for (ci in seq_len(C)) {
					type_name  <- types$typeName[ci]
					slice_name <- paste0(base_name, ".", type_name)
					slice2d <- statistics[, , p]
					new_mat <- array(slice2d, dim = c(M, S, 1))
					dimnames(new_mat)[[3]] <- slice_name
					new_slices[[length(new_slices) + 1]] <- new_mat
				}
			}
		} else {
			sl <- statistics[, , p, drop = FALSE]
			dimnames(sl)[[3]] <- stat_names[p]
			new_slices[[length(new_slices) + 1]] <- sl
		}
	}
	
	P_out  <- length(new_slices)
	snames <- sapply(new_slices, function(s) dimnames(s)[[3]])
	array(
		unlist(lapply(new_slices, function(s) s[, , 1])),
		dim      = c(M, S, P_out),
		dimnames = list(NULL, NULL, snames)
	)
}

# Assemble type-split statistics from C per-type C++ calls (ext=FALSE path).
#
# Called after running compute_stats_tie_sampled once per type (C calls total).
# stats_by_type[[ci]] is [M x samp_num x P] from the ci-th type's C++ call.
#
# For consider_type=TRUE effects:
#   - produces C named slices (e.g. "inertia.social", "inertia.work")
#   - slice ci gets its value from stats_by_type[[ci]]
#   - zeroed where type ci doesn't exist for that sampled dyad (type_exists)
# For consider_type=FALSE and non-typed effects:
#   - all C calls return the same aggregated value, use the first call
#
# [param] stats_by_type  list length C of arrays [M x samp_num x P]
# [param] effects        parsed effects list length P
# [param] consider_type  logical vector length P
# [param] types          data.frame with typeName, typeID (0-based), C rows
# [param] sample_map     integer matrix [M x samp_num], 0-based untyped indices
# [param] type_exists    list length D_base: logical vector length C
#
# [return] array [M x samp_num x P_out] with type-split dimnames[[3]]
assemble_type_slices_sampled <- function(stats_by_type, effects, consider_type,
                                          types, sample_map, type_exists) {
	M          <- dim(stats_by_type[[1]])[1]
	samp_num   <- dim(stats_by_type[[1]])[2]
	C          <- nrow(types)
	stat_names <- dimnames(stats_by_type[[1]])[[3]]

	new_slices <- list()

	for (p in seq_len(dim(stats_by_type[[1]])[3])) {
		has_consider_type <- p <= length(effects) &&
			"consider_type" %in% names(effects[[p]])

		if (has_consider_type && consider_type[p] > 0L) {
			# For each type ci, take the value from stats_by_type[[ci]].
			# Zero positions where type ci doesn't exist for the sampled dyad.
			base_name <- stat_names[p]
			for (ci in seq_len(C)) {
				type_name  <- types$typeName[ci]
				slice_name <- paste0(base_name, ".", type_name)

				slice2d <- stats_by_type[[ci]][, , p]   # [M x samp_num]
				# Zero positions where this type doesn't exist for the sampled dyad
				for (s in seq_len(samp_num)) {
					d_untyped  <- sample_map[, s] + 1L   # 1-based, length M
					# type_exists[[d]][ci] tells us if type ci exists for dyad d
					# sample_map is the same for all rows m, but d varies per row
					# Actually sample_map[m,s] varies per m too
					exists_vec <- sapply(seq_len(M), function(m)
						type_exists[[sample_map[m, s] + 1L]][ci])
					slice2d[!exists_vec, s] <- 0
				}
				new_mat <- array(slice2d, dim = c(M, samp_num, 1))
				dimnames(new_mat)[[3]] <- slice_name
				new_slices[[length(new_slices) + 1]] <- new_mat
			}
		} else {
			# consider_type=FALSE or non-typed: all C calls return same value, use first.
			sl <- stats_by_type[[1]][, , p, drop = FALSE]
			dimnames(sl)[[3]] <- stat_names[p]
			new_slices[[length(new_slices) + 1]] <- sl
		}
	}

	P_out  <- length(new_slices)
	snames <- sapply(new_slices, function(s) dimnames(s)[[3]])
	array(
		unlist(lapply(new_slices, function(s) s[, , 1])),
		dim      = c(M, samp_num, P_out),
		dimnames = list(NULL, NULL, snames)
	)
}

# Dual-compatibility normalizer for remify objects.
#
# Patches a reh object so that reh$meta and reh$ids are always populated,
# regardless of whether the object was produced by the old remify::remify()
# (which stores everything in attr()) or the new remify2() (which stores
# everything in named list elements). All downstream helpers can then safely
# use reh$meta$X and reh$ids$X without inline ternaries everywhere.
normalize_reh <- function(reh) {
  if (!is.null(reh$meta)) return(reh)  # already new-style, nothing to do

  # Build reh$meta from attributes
  reh$meta <- list(
    model      = attr(reh, "model"),
    directed   = attr(reh, "directed"),
    ordinal    = attr(reh, "ordinal"),
    weighted   = attr(reh, "weighted"),
    with_type  = isTRUE(attr(reh, "with_type")),
    riskset    = attr(reh, "riskset"),
    origin     = attr(reh, "origin"),
    dictionary = attr(reh, "dictionary")
  )

  # Build reh$ids from attributes / existing slots
  if (is.null(reh$ids)) {
    reh$ids <- list(
      dyad        = attr(reh, "dyad"),
      actor1      = attr(reh, "actor1"),
      actor2      = attr(reh, "actor2"),
      dyad_active = attr(reh, "dyad_active")
    )
    # Old remify stores these as columns on the edgelist_id matrix
    if (is.null(reh$ids$dyad) && !is.null(reh$edgelist_id)) {
      reh$ids$dyad   <- reh$edgelist_id[, "dyadID"]
      reh$ids$actor1 <- reh$edgelist_id[, "actor1ID"]
      reh$ids$actor2 <- reh$edgelist_id[, "actor2ID"]
    }
  }

  reh
}

# Prepare tomstats inputs
#
# This function prepares all the necessary information prior to using the cpp
# to compute statistics in tomstats. It processes the 'reh' object, matches
# memory settings, converts start and stop indices, prepares main effects,
# fixed effects, interaction effects, covariate information, and scaling
# information.
#
# [param] effects A formula object specifying the effects to be included in
# the analysis.
# [param] reh The reh object containing the event history data.
# [param] attr_actors (Optional) A data frame containing attribute data for the
# actors.
# [param] memory A character vector specifying the memory type, which can be
# one of "full", "window", "decay", or "interval".
# [param] memory_value (Optional) The value associated with the memory type.
# [param] start The starting index of the event history data to consider.
# [param] stop The ending index of the event history data to consider.
#
# [return] A list containing all the necessary information for input to the
# 'compute_stats_tie' function.
#
# [examples]
# inputs <- prepare_tomstats(
#   effects, reh, attr_actors, memory, memory_value,
#   start, stop
# )
prepare_tomstats <- function(
    effects, reh, attr_actors = NULL,
    attr_dyads = NULL, memory = c("full", "window", "decay", "interval"),
    memory_value = NA, start = 1, stop = Inf, method = c("pt", "pe")) {
  # Check if reh is of class remify
  if (!("remify" %in% class(reh))) {
    stop("Expected a reh object of class remify")
  }
  reh <- normalize_reh(reh)

  # Check if the reh object is prepared with the correct model argument
  if (reh$meta$model != "tie") {
    stop("The reh object should be prepared with the model argument set to `tie' if tie_effects are computed")
  }

  # Extract relevant elements from the prepared remify::remify object
  dyads <- unlist(reh$ids$dyad)
  actors <- reh$meta$dictionary$actors
  types <- reh$meta$dictionary$types

  # Convert actor IDs to cpp indexing
  actors$actorID <- actors$actorID - 1

  # Origin
  if (reh$meta$ordinal) {
    if (length(reh$meta$origin)==0) {
      reh$meta$origin <- reh$edgelist[1,1] - 1
    }
  }

  # Prepare the edgelist for cpp processing
  edgelist <- reh$edgelist[,1:3]
  edgelist$time <- cumsum(as.numeric(diff(c(reh$meta$origin, edgelist$time))))
  edgelist$actor1 <- unlist(reh$ids$actor1) - 1
  edgelist$actor2 <- unlist(reh$ids$actor2) - 1
  if (!reh$meta$directed) {
    edgelist[, c(2, 3)] <- t(apply(edgelist, 1, function(x) sort(c(x[2], x[3]))))
  }

  # Deal with event types
  if (is.null(types)) {
    # Prepare a default types object
    types <- data.frame(typeName = 0, typeID = 0)
  } else {
    # Convert type IDs to cpp indexing
    types$typeID <- types$typeID - 1
    edgelist$type <- reh$edgelist$type
    edgelist$type <- types$typeID[match(edgelist$type, types$typeName)]
  }

  # Prepare the event weights
  if (reh$meta$weighted) {
    weights <- as.numeric(reh$edgelist$weight)
  } else {
    weights <- rep(1, nrow(edgelist))
  }

  # Determine whether a typed riskset is needed.
  # A typed riskset is needed if:
  #   (a) the remify object was built with extend_riskset_by_type=TRUE, OR
  #   (b) C>1 and any effect requests consider_type=TRUE (prescan formula)
  with_type_riskset <- isTRUE(reh$meta$with_type_riskset)
  C <- nrow(types)

  needs_typed_riskset <- with_type_riskset || (C > 1 && {
    # Prescan: parse formula just enough to detect consider_type flags
    .pre_effects <- tryCatch(
      parse_formula(effects, "rem", reh$meta$ordinal),
      error = function(e) list()
    )
    any(sapply(.pre_effects, function(x) {
      !("consider_type" %in% names(x)) || {
        ct <- if ("consider_type" %in% names(x)) x$consider_type else "ignore"
        !(identical(ct, "ignore") || isFALSE(ct) || identical(ct, 0L) || identical(ct, 0))
      }
    }))
  })

  typeID_for_riskset <- if (needs_typed_riskset) types$typeID else 0L

  # Prepare the full risk set
  prepR <- get_riskset(
    actorID = actors$actorID,
    typeID = typeID_for_riskset,
    directed = reh$meta$directed
  )

  # Reduce risk set to "active" dyads only
  if (reh$meta$riskset == "active") {
    act_id <- setNames(actors$actorID, actors$actorName)
    
    # For active risksets: dyad_map_active is reliable.
    dyad_map <- reh$index$dyad_map_active %||% reh$index$dyad_map
    a1_ids   <- act_id[as.character(dyad_map$actor1)]
    a2_ids   <- act_id[as.character(dyad_map$actor2)]
    if (needs_typed_riskset) {
    	type_id      <- setNames(types$typeID, types$typeName)
    	t_ids        <- type_id[as.character(dyad_map$type)]
    	prepR_key    <- paste(prepR[, 1], prepR[, 2], prepR[, 3], sep = "|")
    	dyad_map_key <- paste(a1_ids, a2_ids, t_ids, sep = "|")
    } else {
    	# Untyped prepR: deduplicate actor pairs preserving riskset_idx order
    	prepR_key    <- paste(prepR[, 1], prepR[, 2], sep = "|")
    	pairs        <- paste(a1_ids, a2_ids, sep = "|")
    	dyad_map_key <- pairs[!duplicated(pairs)]
    }

    row_idx <- match(dyad_map_key, prepR_key)
    row_idx <- row_idx[!is.na(row_idx)]
    prepR <- prepR[row_idx, ]
    prepR[, 4] <- seq(0, nrow(prepR) - 1, 1)
  }

  # Get the risksetMatrix
  risksetMatrix <- convert_to_risksetMatrix(prepR, nrow(actors), nrow(types))

  # Match memory
  memory <- match.arg(memory)
  memory_value <- validate_memory(memory, memory_value)
  if (memory == "window") {
    # Change memory to interval (window is a special case)
    memory <- "interval"
    memory_value <- c(0, memory_value)
  }

  # Method for managing simultaneous events
  method <- match.arg(method)

  # Convert R start and stop indices to C++ (indexing starts at 0) and
  # check them
  subset <- prepare_subset(start, stop, edgelist, method, model = "tie")
  start <- subset$start
  stop <- subset$stop

  # Prepare main effects
  check_formula(effects)
  form <- effects
  effects <- parse_formula(form, "rem", reh$meta$ordinal)
  effectNames <- sapply(effects, function(x) x$effect)

  # Check correct specification effects
  if (!reh$meta$directed) {
    if (!all(effectNames %in% tie_effects(directed = FALSE))) {
      stop("Attempting to request effects that are not (yet) defined for undirected events")
    }
  }

  if (reh$meta$directed) {
    if (!all(effectNames %in% tie_effects(directed = TRUE))) {
      stop("Attempting to request effects that are not (yet) defined for directed events")
    }
  }

  # Prepare fixed effects
  if (any(effectNames == "FEtype")) {
    C <- nrow(types)
    if (C < 2) {
      stop("FEtype() is not defined when the number of event types is smaller than 2.")
    }
    FEeffects <- lapply(2:C, function(c) {
      x <- list()
      x$effect <- "FEtype"
      x$scaling <- 1
      x$typeName <- types$typeName[c]
      x$typeID <- types$typeID[c]
      x
    })

    pos <- which(effectNames == "FEtype")
    effects <- append(effects[-pos], FEeffects, pos - 1)
    effectNames <- append(effectNames[-pos], rep("FEtype", C - 1), pos - 1)
  }

  # Prepare interaction effects
  effects_int <- parse_int(form, "rem", effects, reh$meta$ordinal)
  effectNames <- append(effectNames, rep("interact", length(effects_int)), length(effectNames))
  interactions <- list()
  interactions[which(effectNames == "interact")] <- effects_int

  # Prepare covariate information
  covar <- process_covariate(effects, attr_actors, attr_dyads, actors, edgelist, reh, prepR)

  # Prepare scaling info (vector length p)
  scaling <- sapply(effects, function(x) {
    ifelse("scaling" %in% names(x), x$scaling, "none")
  })
  scaling <- append(scaling, rep("none", length(effects_int)), length(scaling))

  # Prepare consider_type info
  consider_type <- sapply(effects, function(x) {
    ct <- if ("consider_type" %in% names(x)) x$consider_type else "ignore"
    # Normalize: "ignore"/FALSE/0 -> 0L, "separate"/TRUE/1 -> 1L, "interact"/2 -> 2L
    if (identical(ct, "ignore") || isFALSE(ct) || identical(ct, 0L) || identical(ct, 0)) 0L
    else if (identical(ct, "separate") || isTRUE(ct) || identical(ct, 1L) || identical(ct, 1)) 1L
    else if (identical(ct, "interact") || identical(ct, 2L) || identical(ct, 2)) 2L
    else stop("consider_type must be \"ignore\", \"separate\", or \"interact\"")
  })
  consider_type <- append(
  	consider_type,
  	rep(if (nrow(types) > 1) 1L else 0L, length(effects_int)),
  	length(consider_type)
  )
  # C++ consider_type logic:
  # "ignore"   (0): FALSE — aggregate all types in C++
  # "separate" (1): FALSE — aggregate all types in C++, R splits without zeroing
  # "interact" (2): TRUE  — compute own-type per dyad in C++, R zeroes non-matching
  # Coerce "interact" to "separate" when ext=FALSE: the C^2 dyad-type x past-type
  # interaction requires a typed riskset (ext=TRUE). With ext=FALSE there is no
  # dyad-type dimension to interact with, so "interact" is meaningless.
  if (!with_type_riskset && any(consider_type == 2L)) {
    warning('"interact" requires extend_riskset_by_type = TRUE; coercing to "separate".')
    consider_type[consider_type == 2L] <- 1L
  }

  consider_type_cpp <- consider_type == 2L

  # Check consider_type info: warn if "separate"/"interact" requested with only one type
  if (nrow(types) == 1 & any(consider_type > 0L)) {
    warning("'consider_type' = \"separate\" or \"interact\" has no effect with only one event type: treating as \"ignore\".")
  }

  # Check correct scaling inertia statistic
  if (!reh$meta$directed) {
    if (any(sapply(effects, function(x) x$effect == "inertia"))) {
      idx <- which(sapply(effects, function(x) x$effect == "inertia"))
      if (any(scaling[idx] == "prop")) {
        stop("Proportional scaling of the inertia effect is not defined for undirected events")
      }
    }
  }

  # Output
  list(
    form = form,
    effects = effects,
    effectNames = effectNames,
    edgelist = as.matrix(edgelist),
    weights = weights,
    actors = actors,
    types = types,
    riskset = prepR,
    risksetMatrix = risksetMatrix,
    memory = memory,
    memory_value = memory_value,
    scaling = scaling,
    consider_type     = consider_type,      # integer 0/1/2
    consider_type_cpp = consider_type_cpp,  # logical for C++
    covar = covar,
    interactions = interactions,
    start = start,
    stop = stop,
    method = method,
    with_type_riskset = with_type_riskset,
    needs_typed_riskset = needs_typed_riskset,
    C = C,
    prepR_untyped = if (needs_typed_riskset && !with_type_riskset)
      get_prepR_untyped(reh, actors) else NULL
  )
}

# Check if Formula Effects are Specified as Functions
#
# This function checks if the effects in a given formula are specified as
# functions, as expected. It throws an error if any effect is not specified as
# a function.
#
# [param] formula A formula object specifying the effects.
#
# [examples]
# formula1 <- ~ inertia()
# formula2 <- ~ inertia
# check_formula(formula1)  # No error
# check_formula(formula2)  # Throws an error
check_formula <- function(formula) {
  formula_str <- as.character(formula)
  effect_names <- gsub("~ ", "", formula_str) # Extract effect names
  effect_names <- setdiff(effect_names, "~") # Remove "~"
  effect_names <- strsplit(effect_names, "\\+\\s*")[[1]] # Split effect names by "+"
  effect_names <- gsub("\\s+", "", effect_names) # Remove spaces from effect names


  # Check if effects are specified as functions
  incorrect_effects <- effect_names[!grepl("\\(.*\\)", effect_names)]
  incorrect_effects <- incorrect_effects[!grepl("1", incorrect_effects)]

  if (length(incorrect_effects) > 0) {
    stop(paste(
      "The following effects are not specified as functions:",
      paste(incorrect_effects, collapse = ", ")
    ))
  }
}

# Validate memory settings
#
# This function validates the memory settings for a specific memory type. It
# checks the validity of the supplied 'memory_value' based on the selected
# 'memory' type.
#
# [param] memory The memory type, which can be one of "full", "window",
# "decay", or "interval".
# [param] memory_value The value associated with the memory type.
#
# [return] The validated memory value based on the memory type.
#
# [examples]
# validated_value <- validate_memory("window", 10)
validate_memory <- function(memory, memory_value) {
  if (memory == "full") {
    memory_value <- Inf
  }
  if (memory == "window") {
    stopifnot(
      "A 'memory_value' of numeric type should be supplied when memory is 'window'" =
        length(memory_value) == 1 && is.numeric(memory_value)
    )
  }
  if (memory == "decay") {
    stopifnot(
      "A 'memory_value' of numeric type should be supplied when memory is 'decay'" =
        length(memory_value) == 1 && is.numeric(memory_value)
    )
    stopifnot(
      "Invalid 'memory_value': must be a positive non-zero value." =
        memory_value[1] > 0
    )
  }
  if (memory == "interval") {
    stopifnot(
      "Two 'memory_value' values of numeric type should be supplied when memory is 'interval'" =
        length(memory_value) == 2 && is.numeric(memory_value)
    )
    stopifnot(
      "The first 'memory_value' value should be lower than the second value" =
        memory_value[1] < memory_value[2]
    )
  }

  return(memory_value)
}

# Get all tie effects
#
# This function returns a vector of all available tie effects that can be used
# in tie-oriented models.
#
# [return] A character vector containing all tie effects.
#
# [examples]
# tie_effects <- all_tie_effects()
all_tie_effects <- function() {
  c(
    "baseline",
    "FEtype",

    # Exogenous stats
    "send",
    "receive",
    "tie",
    "same",
    "difference",
    "average",
    "minimum",
    "maximum",
    "event",

    # Endogenous stats
    "inertia",
    "reciprocity",
    "indegreeSender",
    "indegreeReceiver",
    "outdegreeSender",
    "outdegreeReceiver",
    "totaldegreeSender",
    "totaldegreeReceiver",
    "totaldegreeDyad",
    "degreeMin",
    "degreeMax",
    "degreeDiff",
    "otp",
    "itp",
    "osp",
    "isp",
    "sp",
    "psABBA",
    "psABBY",
    "psABXA",
    "psABXB",
    "psABXY",
    "psABAY",
    "psABAB",
    "rrankSend",
    "rrankReceive",
    "recencyContinue",
    "recencySendSender",
    "recencySendReceiver",
    "recencyReceiveSender",
    "recencyReceiveReceiver",
    "userStat",
    "interact"
  )
}

# Add variable names to statistics
#
# This function adds variable names to the statistics object based on the
# specified effects and interactions. It modifies the dimnames of the
# statistics object to include the variable names for specific effects.
#
# [param] statistics The statistics object to which variable names will be
# added.
# [param] all_effects A vector containing all possible effects.
# [param] effectsN A numeric vector specifying the indices of the effects in
# 'all_effects' that are present in the statistics.
# [param] effects A list containing information about the effects, including
# the effect name and associated variable.
# [param] interactions A list containing information about the interactions,
# including the indices of the interacting effects.
#
# [return] The modified statistics object with variable names added.
#
# [examples]
# statistics <- add_variable_names(
#   statistics, all_effects, effectsN, effects,
#   interactions
# )
add_variable_names <- function(statistics, effectNames, effects, interactions) {
  # Helper function to add variable name to effect
  add_variable_name <- function(effect, variable) {
    if (!is.null(variable)) {
      paste0(effect, "_", variable)
    } else {
      effect
    }
  }

  # Dimnames statistics
  dimnames(statistics) <- list(NULL, NULL, effectNames)

  # Add variable name to exogenous statistics
  exogenous_indices <- which(sapply(effects, function(x) {
    "variable" %in% names(x)
  }))

  dimnames(statistics)[[3]][exogenous_indices] <- sapply(
    effects[exogenous_indices],
    function(x) add_variable_name(x$effect, x$variable)
  )

  # Add counter to event name
  event_effects <- grepl("event", dimnames(statistics)[[3]])
  if (sum(event_effects) > 1) {
    dimnames(statistics)[[3]][event_effects] <- paste0("event", 1:sum(event_effects))
  }

  # Add the type to FEtype name
  FEtype_effects <- grep("FEtype", dimnames(statistics)[[3]])
  if (sum(FEtype_effects) > 1) {
    dimnames(statistics)[[3]][FEtype_effects] <- paste0("FEtype_", unlist(
      sapply(effects, function(x) {
        x$typeName
      })
    ))
  }

  # Add .unique
  unique_effects <- sapply(effects, function(x) {
    ifelse("scaling" %in% names(x), isTRUE(grepl("unique", x$scaling)), FALSE)
  })
  if (any(unique_effects)) {
    dimnames(statistics)[[3]][unique_effects] <- paste0(dimnames(statistics)[[3]][unique_effects], ".unique")
  }

  # "ignore": no suffix (clean output, e.g. "inertia" not "inertia.TypeAgg")
  type_effects <- sapply(effects, function(x) {
    ct <- if ("consider_type" %in% names(x)) x$consider_type else "ignore"
    identical(ct, "ignore") || isFALSE(ct) || identical(ct, 0L) || identical(ct, 0)
  })

  # Add variable name to interaction statistics
  interaction_index <- which(effectNames == "interact")
  dimnames(statistics)[[3]][interaction_index] <- sapply(
    interactions[interaction_index],
    function(x) {
      paste0(
        dimnames(statistics)[[3]][as.numeric(x[1])],
        ":",
        dimnames(statistics)[[3]][as.numeric(x[2])]
      )
    }
  )

  statistics # Return the modified statistics object
}

variable_check <- function(effect, dat) {
  if (effect == "send" | effect == "receive") {
    if (is.logical(dat$x)) {
      # Transform logical to binary
      dat$x <- as.numeric(dat$x)
    } else if (is.factor(dat$x)) {
      x_levels <- levels(dat$x[, , dat$variable])

      if (length(x_levels) != 2) {
        stop(paste0("The variable in '", effect, "' must have two factor levels or must be numeric."))
      } else {
        # Convert to binary variable
        dat$x <- ifelse(dat$x == x_levels[2], 1, 0)
      }
    } else if(!is.numeric(dat$x)) {
      stop(paste0("The variable in '", effect, "' must be numeric."))
    }
  }

  if (effect == "same") {
    # Force x to be numeric
    dat$x <- as.numeric(as.factor(dat$x))
  }

  if (effect == "average") {
    # x has to be numeric
    if (!is.numeric(dat$x)) {
      stop("The variable in 'average' must be numeric.")
    }
  }

  if (effect == "minimum") {
    # x has to be numeric
    if (!is.numeric(dat$x)) {
      stop("The variable in 'minimum' must be numeric.")
    }
  }

  if (effect == "maximum") {
    # x has to be numeric
    if (!is.numeric(dat$x)) {
      stop("The variable in 'maximum' must be numeric.")
    }
  }

  if (effect == "difference") {
    # x has to be numeric
    if (!is.numeric(dat$x)) {
      stop("The variable in 'difference' must be numeric.")
    }
  }

  return(dat)
}

# Prepare exogenous effect
#
# This function prepares an exogenous effect by matching the scaling, checking
# the presence of the variable in the attr_actors object, and collecting the
# necessary information in a data frame.
#
# [param] effect The effect name.
# [param] variable The variable name.
# [param] attr_actors A data frame containing attribute data for the actors.
# [param] scaling The scaling option for the effect, which can be "as.is" or
# "std".
#
# [return] A list containing the prepared exogenous effect information.
#
# [examples]
# exo_effect <- prep_exo(effect, variable, attr_actors, scaling)
prep_exo <- function(effect, variable, attr_actors, scaling) {
  # Check variable
  if (!is.character(variable)) {
    stop("The 'variable' argument should be a string.")
  }

  # Prepare effect
  if (is.null(attr_actors)) {
    list(
      effect = effect,
      variable = variable,
      x = NULL,
      scaling = scaling
    )
  } else {
    # Check if the variable name is in the attr_actors object
    if (!(variable %in% colnames(attr_actors))) {
      stop(paste0("Variable '", variable, "' not in attr_actors object for the '", effect, "' effect."))
    }

    # Check if the time variable is available
    if (!("time" %in% colnames(attr_actors))) {
      if (anyDuplicated(attr_actors$name) > 0) {
        warning("Duplicated actor names detected in 'attr_actors'. Did you mean to add a 'time' variable?")
      } else {
        attr_actors$time <- 0
      }
    }
    if (anyNA(attr_actors$time)) {
      stop("Missing 'time' values detected in 'attr_actors'.")
    }

    # Collect the information in a data.frame
    dat <- data.frame(
      name = attr_actors$name,
      time = attr_actors$time,
      x = attr_actors[, variable]
    )

    # Check the data class of the variable
    dat <- variable_check(effect, dat)

    # Warning for missing values
    if (anyNA(dat)) {
      warning("Missing values in the attr_actors object for the '", effect, "' effect can cause unexpected behavior.")
    }

    # Set the third column name equal to the variable name
    colnames(dat)[3] <- variable

    # Output
    list(
      effect = effect,
      variable = variable,
      x = dat,
      scaling = scaling
    )
  }
}

# Parse formula
#
# This function parses a formula and extracts the effects specified in the
# formula.
#
# [param] formula The formula specifying the effects.
# [param] type The type of model being used, either "rem" (for the tie-oriented
# model) or "rateEffects" or "choiceEffects".
# [param] ordinal Logical indicating whether the event data is ordinal.
#
# [return] A list containing the parsed effects.
#
# [examples]
# parsed_effects <- parse_formula(formula, type, ordinal)
parse_formula <- function(formula, type, ordinal = FALSE) {
  ft <- stats::terms(formula)

  var <- attr(ft, "variables")
  var <- as.list(var)[-1]

  effects <- lapply(var, function(y) {
    eval(y)
  })

  if (type == "rem" & !ordinal & attr(ft, "intercept") == 1) {
    effects <- append(
      effects,
      list(list(effect = "baseline")), 0
    )
  }
  if (type == "rateEffects" & !ordinal & attr(ft, "intercept") == 1) {
    effects <- append(
      effects,
      list(list(effect = "baseline")), 0
    )
  }

  attributes(effects)$model <- type
  effects
}

# Parse interaction terms
#
# This function parses a formula and extracts the interaction terms specified
# in the formula.
#
# [param] formula The formula specifying the interaction terms.
# [param] type The type of model being used, either "rem" or "rateEffects".
# [param] effects The list of effects parsed from the formula.
# [param] ordinal Logical indicating whether the event data is ordinal.
#
# [return] A list containing the parsed interaction terms.
#
# [examples]
# parsed_interactions <- parse_int(formula, type, effects, ordinal)
parse_int <- function(formula, type, effects, ordinal = FALSE) {
  ft <- stats::terms(formula)
  ft.order <- attr(ft, "order")
  ft.factor <- attr(ft, "factors")
  interactions <- which(ft.order > 1)
  interactions <- lapply(interactions, function(i) {
    if (type == "rem" & !ordinal & attr(ft, "intercept") == 1) {
      which(ft.factor[, i] > 0) + 1
    } else if (type == "rateEffects" & attr(ft, "intercept") == 1) {
      which(ft.factor[, i] > 0) + 1
    } else {
      which(ft.factor[, i] > 0)
    }
  })

  interactions <- if (any(sapply(effects, function(x) x$effect == "FEtype")) &
    any(ft.order > 1)) {
    pos <- which(sapply(effects, function(x) x$effect == "FEtype"))
    unlist(lapply(1:length(interactions), function(i) {
      if (any(interactions[[i]] %in% pos)) {
        change <- which(!interactions[[i]] == min(pos))
        x <- interactions[[i]]
        x[change] <- x[change] + length(pos) - 1
        FE <- which(interactions[[i]] == min(pos))
        lapply(pos, function(p) {
          x[FE] <- p
          x
        })
      } else {
        list(interactions[[i]])
      }
    }), recursive = F)
  } else {
    interactions
  }

  attributes(interactions)$model <- type
  interactions
}

# Modify Riskset
#
# This function modifies the riskset object by mapping actor and type IDs to their corresponding names in the event history data.
#
# [param] riskset The riskset object to be modified.
# [param] reh The reh object representing the event history data.
# [param] actors The actors information.
# [param] types The types information.
#
# [return] The modified riskset object.
#
# Split consider_type=TRUE statistics into C type-specific slices.
#
# Called after compute_stats_tie / compute_stats_tie_sampled when C > 1 and
# a typed riskset was used internally (needs_typed_riskset = TRUE).
#
# For each statistic p where consider_type[p] = TRUE:
#   - ext=TRUE  (with_type_riskset): keeps D_typed columns, zeros out
#               the non-matching type columns -> C slices of [M x D_typed]
#   - ext=FALSE (!with_type_riskset): extracts the D_dyad columns for that
#               type block only -> C slices of [M x D_dyad]
# For consider_type=FALSE statistics:
#   - ext=FALSE: strips from D_typed to first D_dyad columns (all blocks hold
#               the same aggregated value after transform_inertia)
#
# [param] statistics  array [M x D x P] returned by compute_stats_tie*
# [param] effects     parsed effects list (length P, before interactions)
# [param] consider_type  logical vector length P
# [param] types       data.frame with columns typeName, typeID (0-based)
# [param] prepR_typed matrix [D_typed x 4] — the typed riskset passed to C++
# [param] with_type_riskset  logical: was ext=TRUE?
#
# [return] array [M x D_out x P_out] with updated dimnames[[3]]
#
split_type_slices <- function(statistics, effects, consider_type,
															types, prepR_typed, with_type_riskset,
															prepR_untyped = NULL,
															stats_by_type = NULL) {
	C        <- nrow(types)
	D_typed  <- dim(statistics)[2]
	type_ids <- prepR_typed[, 3]

	if (!with_type_riskset && !is.null(prepR_untyped)) {
		D_dyad      <- nrow(prepR_untyped)
		key_untyped <- paste(prepR_untyped[, 1], prepR_untyped[, 2], sep = "|")
		key_typed   <- paste(prepR_typed[, 1],   prepR_typed[, 2],   sep = "|")
	} else {
		D_dyad <- D_typed %/% C
	}

	stat_names <- dimnames(statistics)[[3]]
	new_slices <- list()

	for (p in seq_len(dim(statistics)[3])) {
		has_consider_type <- p <= length(effects) &&
			"consider_type" %in% names(effects[[p]])

		if (has_consider_type && consider_type[p] > 0L) {
			base_name <- stat_names[p]

			if (!is.null(stats_by_type)) {

				if (consider_type[p] == 2L) {
					# "interact": C^2 slices — past-type x dyad-type
					for (ci in seq_len(C)) {
						past_type_name <- types$typeName[ci]
						sbt   <- stats_by_type[[ci]]
						p_sbt <- which(dimnames(sbt)[[3]] == base_name)
						if (length(p_sbt) == 0) next
						for (ci2 in seq_len(C)) {
							dyad_type_name <- types$typeName[ci2]
							dyad_type_id   <- types$typeID[ci2]
							slice_name <- paste0(base_name, ".", past_type_name, ".", dyad_type_name)
							new_mat <- sbt[, , p_sbt, drop = FALSE]
							new_mat[, which(type_ids != dyad_type_id), ] <- 0
							dimnames(new_mat)[[3]] <- slice_name
							new_slices[[length(new_slices) + 1]] <- new_mat
						}
					}
				} else {
					# "separate": C slices — past-type only
					for (ci in seq_len(C)) {
						type_id   <- types$typeID[ci]
						type_name <- types$typeName[ci]
						slice_name <- paste0(base_name, ".", type_name)
						sbt   <- stats_by_type[[ci]]
						p_sbt <- which(dimnames(sbt)[[3]] == base_name)
						if (length(p_sbt) == 0) next

						if (with_type_riskset) {
							# ext=TRUE: use per-type stats directly
							new_mat <- sbt[, , p_sbt, drop = FALSE]
						} else {
							# ext=FALSE: map type-c rows of typed riskset to D_dyad output
							type_rows <- which(type_ids == type_id)
							new_mat   <- array(0, dim = c(dim(sbt)[1], D_dyad, 1))
							if (!is.null(prepR_untyped)) {
								dst   <- match(key_typed[type_rows], key_untyped)
								valid <- !is.na(dst)
								new_mat[, dst[valid], 1] <- sbt[, type_rows[valid], p_sbt]
							} else {
								new_mat[, seq_along(type_rows), 1] <- sbt[, type_rows, p_sbt]
							}
						}
						dimnames(new_mat)[[3]] <- slice_name
						new_slices[[length(new_slices) + 1]] <- new_mat
					}
				}
				next
			}

			if (with_type_riskset && consider_type[p] == 2L) {
				# "interact" + ext=TRUE without multi-pass
				for (ci in seq_len(C)) {
					past_type_name <- types$typeName[ci]
					for (ci2 in seq_len(C)) {
						dyad_type_name <- types$typeName[ci2]
						dyad_type_id   <- types$typeID[ci2]
						slice_name <- paste0(base_name, ".", past_type_name, ".", dyad_type_name)
						new_mat <- statistics[, , p, drop = FALSE]
						new_mat[, which(type_ids != dyad_type_id), ] <- 0
						dimnames(new_mat)[[3]] <- slice_name
						new_slices[[length(new_slices) + 1]] <- new_mat
					}
				}
				next
			}

			# "separate" (or "interact" + ext=FALSE) without multi-pass
			for (ci in seq_len(C)) {
				type_id   <- types$typeID[ci]
				type_name <- types$typeName[ci]
				slice_name <- paste0(base_name, ".", type_name)

				if (with_type_riskset) {
					new_mat <- statistics[, , p, drop = FALSE]
				} else {
					type_rows <- which(type_ids == type_id)
					new_mat   <- array(0, dim = c(dim(statistics)[1], D_dyad, 1))
					if (!is.null(prepR_untyped)) {
						dst   <- match(key_typed[type_rows], key_untyped)
						valid <- !is.na(dst)
						new_mat[, dst[valid], 1] <- statistics[, type_rows[valid], p]
					} else {
						new_mat[, seq_along(type_rows), 1] <- statistics[, type_rows, p]
					}
				}
				dimnames(new_mat)[[3]] <- slice_name
				new_slices[[length(new_slices) + 1]] <- new_mat
			}

		} else {
			sl <- statistics[, , p, drop = FALSE]
			if (!with_type_riskset && dim(sl)[2] == D_typed) {
				if (!is.null(prepR_untyped)) {
					dst   <- match(key_untyped, key_typed)
					valid <- !is.na(dst)
					new_sl <- array(0, dim = c(dim(sl)[1], D_dyad, 1))
					new_sl[, which(valid), 1] <- sl[, dst[valid], 1]
					sl <- new_sl
				} else {
					sl <- sl[, seq_len(D_dyad), , drop = FALSE]
				}
			}
			dimnames(sl)[[3]] <- stat_names[p]
			new_slices[[length(new_slices) + 1]] <- sl
		}
	}

	M      <- dim(new_slices[[1]])[1]
	D_out  <- dim(new_slices[[1]])[2]
	P_out  <- length(new_slices)
	snames <- sapply(new_slices, function(s) dimnames(s)[[3]])
	array(
		unlist(lapply(new_slices, function(s) s[, , 1])),
		dim      = c(M, D_out, P_out),
		dimnames = list(NULL, NULL, snames)
	)
}
	
# Compute the raw untyped prepR matrix for the ext=FALSE case.
# Returns the integer matrix [D_dyad x 4] with 0-based actor/type/dyad IDs.
#
get_prepR_untyped <- function(reh, actors) {
  prepR_untyped <- get_riskset(
    actorID  = actors$actorID,
    typeID   = 0L,
    directed = reh$meta$directed
  )
  if (reh$meta$riskset == "active") {
    # Use dyad_map_active with actor-pair matching (same logic as prepare_tomstats).
    # dyad_map_active contains the correct active untyped dyads in riskset_idx order.
    dyad_map <- reh$index$dyad_map_active %||% reh$index$dyad_map
    act_id   <- setNames(actors$actorID, actors$actorName)
    a1_ids   <- act_id[as.character(dyad_map$actor1)]
    a2_ids   <- act_id[as.character(dyad_map$actor2)]
    pairs    <- paste(a1_ids, a2_ids, sep = "|")
    uniq_pairs <- pairs[!duplicated(pairs)]
    prepR_key  <- paste(prepR_untyped[, 1], prepR_untyped[, 2], sep = "|")
    row_idx    <- match(uniq_pairs, prepR_key)
    row_idx    <- row_idx[!is.na(row_idx)]
    prepR_untyped <- prepR_untyped[row_idx, ]
    prepR_untyped[, 4] <- seq(0, nrow(prepR_untyped) - 1, 1)
  }
  prepR_untyped
}

# Build the user-facing riskset data.frame for the ext=FALSE case, where the
# typed riskset was used internally but the output riskset should be untyped.
#
build_untyped_riskset <- function(reh, actors, types) {
  modify_riskset(get_prepR_untyped(reh, actors), reh, actors, types[1L, , drop = FALSE])
}

# [examples]
# modified_riskset <- modify_riskset(riskset, reh, actors, types)
modify_riskset <- function(riskset, reh, actors, types) {
  riskset <- as.data.frame(riskset)

  if (reh$meta$directed) {
    colnames(riskset) <- c("sender", "receiver", "type", "id")
    riskset$sender <- actors$actorName[match(riskset$sender, actors$actorID)]
    riskset$receiver <- actors$actorName[match(riskset$receiver, actors$actorID)]
    riskset$type <- types$typeName[match(riskset$type, types$typeID)]
  } else {
    colnames(riskset) <- c("actor1", "actor2", "type", "id")
    riskset$actor1 <- actors$actorName[match(riskset$actor1, actors$actorID)]
    riskset$actor2 <- actors$actorName[match(riskset$actor2, actors$actorID)]
    riskset$type <- types$typeName[match(riskset$type, types$typeID)]
  }

  riskset$id <- riskset$id + 1

  if (length(unique(riskset$type)) == 1) {
    riskset <- riskset[, -3]
  }

  riskset # Return the modified riskset object
}

# Process Covariate
#
# This function processes the covariates for different exogenous effects in
# the tie-oriented model.
#
# [param] effects A list of effects containing information about the covariates.
# [param] attr_actors The attr_actors object containing the covariate data.
# [param] actors The actors information.
# [param] edgelist The edgelist representing the network.
# [param] reh The reh object representing the event history data.
# [param] prepR The prepR object representing the risk set.
#
# [return] A list of processed covariates for each effect.
#
# [examples]
# covariates <- process_covariate(effects, attr_actors, actors, edgelist, reh, prepR)
process_covariate <- function(
    effects,
    attr_actors,
    attr_dyads,
    actors,
    edgelist,
    reh,
    prepR) {
  lapply(effects, function(x) {
    effect <- x$effect

    if (effect %in% c(
      "send",
      "receive",
      "same",
      "difference",
      "average",
      "minimum",
      "maximum"
    )) {
      if (is.null(x$x)) { # The variable has to be extracted from 'attr_actors'
        variable <- x$variable

        # Check if the variable name is in the 'attr_actors' object
        if (!(variable %in% colnames(attr_actors))) {
          stop(paste0("Variable '", variable, "' not in attr_actors object for the '", effect, "' effect."))
        }

        # Check if the 'time' variable is available
        if (!("time" %in% colnames(attr_actors))) {
          if (anyDuplicated(attr_actors$name) > 0) {
            warning("Duplicated actor names detected in 'attr_actors'. Did you mean to add a 'time' variable?")
          } else {
            attr_actors$time <- 0
          }
        }

        if (anyNA(attr_actors$time)) {
          stop("Missing 'time' values detected in 'attr_actors'.")
        }

        # Collect the information in a data.frame
        dat <- data.frame(
          name = attr_actors$name,
          time = attr_actors$time,
          x = attr_actors[, variable],
          stringsAsFactors = FALSE
        )

        # Check the data class of the variable
        dat <- variable_check(effect, dat)

        # Warning for missing values
        if (anyNA(dat)) {
          warning(paste0("Missing values in the attr_actors object for the '", effect, "' effect can cause unexpected behavior."))
        }
      } else { # The variable was supplied to the function in effects.R
        dat <- x$x
      }

      # Check the actors in the obtained 'dat' object
      dat$name <- as.character(dat$name)
      if (!all(actors[, 1] %in% dat$name)) {
        stop("Missing actors in the attr_actors object.")
      }

      # Convert 'actorName' to 'actorID'
      dat$name <- actors[match(dat$name, actors[, 1]), 2]

      # Set the third column name equal to the 'variable'
      colnames(dat)[3] <- x$variable

      # Order by actor ID
      dat <- dat[order(as.numeric(dat$name)), ]

      # Remove any redundant actors
      if (any(is.na(dat$name))) {
        warning(paste0("The 'attr_actors' object contains actors that are not in the risk set. These are not included in the computation of the statistics."))
        dat <- dat[!is.na(dat$name), ]
      }

      # Convert to matrix and output
      as.matrix(dat)
    } else if (effect %in% c("tie", "event", "FEtype")) {
      if (effect == "tie") {
        # Return: matrix (actor1, actor2, time, value)
        parse_tie(x, reh, attr_dyads)
      } else if (effect == "event") {
        if (NROW(x$x) != nrow(edgelist) & NROW(x$x) != NROW(unique(edgelist[, 1]))) {
          stop("Length of vector 'x' in event() does not match number of events in edgelist")
        }
        if (!is.numeric(x$x)) {
          x$x <- (as.numeric(factor(x$x)) - 1)
        }
        as.matrix(x$x)
      } else if (effect == "FEtype") {
        as.matrix(x$typeID)
      } 
    } else if (effect == "userStat") {
      if (NROW(x$x) != nrow(edgelist) & NROW(x$x) != NROW(unique(edgelist[, 1]))) {
        stop("Number of rows of matrix 'x' in userStat() does not match number of timepoints or number of events in edgelist")
      }

      if (NCOL(x$x) != nrow(prepR)) {
        stop("Number of columns of matrix 'x' in userStat() does not match number of dyads in risk set")
      }

      as.matrix(x$x)
    } else {
      matrix()
    }
  })
}

validate_aomstats_arguments <- function(attr_actors, reh) {
  reh <- normalize_reh(reh)
  # Check if the deprecated "id" column is used in attr_actors
  if (!is.null(attr_actors) & "id" %in% colnames(attr_actors) & !("name" %in% colnames(attr_actors))) {
    colnames(attr_actors) <- ifelse(colnames(attr_actors) == "id", "name", colnames(attr_actors))
    warning("Use 'name' instead of 'id' in 'attr_actors'")
  }

  # Check the reh class
  if (!inherits(reh, "remify")) {
    stop("Invalid argument: 'reh' must be a remify object.")
  }

  model <- reh$meta$model
  # Check the reh model
  if (is.null(model) || model != "actor") {
    stop("Invalid argument: 'reh' object should be prepared with the model argument set to 'actor' if effects for the actor-oriented model are computed.")
  }

  return(attr_actors)
}

# Compute aomstats with per-effect consider_type handling.
#
# For effects with consider_type="ignore": single C++ call, single slice.
# For effects with consider_type="separate": C filtered C++ calls per effect,
#   LOCF expansion to full M rows, C named slices per effect.
# Results are assembled in original effect order.
#
# [param] compute_fn     function(effect_names, edgelist, weights, start_0, stop_0)
#                        returns array [M_in x N x P_in] — raw C++ output
# [param] add_names_fn   function(st) adds dimnames to C++ output
# [param] edgelist       full edgelist matrix [M x 3]
# [param] weights        numeric vector length M
# [param] event_type_ids integer vector length M, 1-based typeID per event
# [param] types_df       data.frame: typeName, typeID (1-based)
# [param] effects_names  character vector length P of base effect names
# [param] consider_type  character vector length P: "ignore" or "separate"
# [param] with_type      logical: are there event types?
# [param] C              number of types
# [param] subset_start   0-based start index
# [param] subset_stop    0-based stop index
#

# Pshift effect names and their C++ type numbers (matching get_pshift in aomstats.cpp)
PSHIFT_NAMES <- c("psABAB", "psABBA", "psABXB", "psABXA", "psABAY", "psABBY", "psABXY")
PSHIFT_TYPE_NUM <- c(psABAB=1L, psABBA=2L, psABXB=3L, psABXA=4L,
                     psABAY=5L, psABBY=6L, psABXY=7L)

# Compute pshift.type_c for the ACTOR model (aomstats).
# For each output row m, finds the last type-c event before m and applies the
# pshift condition based on the current sender and the prev type-c sender/receiver.
# edgelist: [M x 3] matrix, cols = (time, actor1_0based, actor2_0based)
# event_type_ids: 1-based typeIDs, length M (same as nrow(edgelist))
# type_id: 1-based typeID to filter on
# Returns [M_out x N] matrix.
compute_pshift_type_c_receiver <- function(pshift_num, edgelist, event_type_ids,
																					 type_id, N, subset_start, subset_stop) {
	M_out     <- subset_stop - subset_start + 1L
	full_rows <- (subset_start + 1L):(subset_stop + 1L)  # 1-based rows in edgelist
	result    <- matrix(0, nrow = M_out, ncol = N)
	
	out_times    <- edgelist[full_rows, 1]
	unique_times <- sort(unique(out_times))
	
	for (tp_idx in seq_along(unique_times)) {
		curr_t <- unique_times[tp_idx]
		
		# Find previous time point
		if (tp_idx == 1L) {
			before_all <- which(edgelist[, 1] < curr_t)
			if (length(before_all) == 0L) next  # no previous events, all 0
			prev_t <- max(edgelist[before_all, 1])
		} else {
			prev_t <- unique_times[tp_idx - 1L]
		}
		
		####
		if (tp_idx == 1L) {
			before_all <- which(edgelist[, 1] < curr_t)
			cat("tp_idx=1: length(before_all)=", length(before_all), "\n")
			if (length(before_all) == 0L) next
			prev_t <- max(edgelist[before_all, 1])
			cat("prev_t=", prev_t, "\n")
			prev_ci_idx <- which(edgelist[, 1] == prev_t & event_type_ids == type_id)
			cat("type_id=", type_id, "length(prev_ci_idx)=", length(prev_ci_idx), "\n")
			cat("types at prev_t:", event_type_ids[which(edgelist[,1] == prev_t)], "\n")
		}
		####
		
		# Previous type-ci events at prev_t
		prev_ci_idx <- which(edgelist[, 1] == prev_t & event_type_ids == type_id)
		if (length(prev_ci_idx) == 0L) next
		
		# Output rows at curr_t
		m_rows <- which(out_times == curr_t)
		
		for (m in m_rows) {
			curr_s <- edgelist[full_rows[m], 2] + 1L  # 1-based
			for (prev_e in prev_ci_idx) {
				prev_s <- edgelist[prev_e, 2] + 1L
				prev_r <- edgelist[prev_e, 3] + 1L
				if (pshift_num == 1L) {        # AB-AB
					if (curr_s == prev_s) result[m, prev_r] <- 1
				} else if (pshift_num == 2L) { # AB-BA
					if (curr_s == prev_r) result[m, prev_s] <- 1
				} else if (pshift_num == 3L) { # AB-XB
					if (curr_s != prev_s && curr_s != prev_r) result[m, prev_r] <- 1
				} else if (pshift_num == 4L) { # AB-XA
					if (curr_s != prev_s && curr_s != prev_r) result[m, prev_s] <- 1
				} else if (pshift_num == 5L) { # AB-AY
					if (curr_s == prev_s) {
						result[m, ] <- 1; result[m, prev_s] <- 0; result[m, prev_r] <- 0
					}
				} else if (pshift_num == 6L) { # AB-BY
					if (curr_s == prev_r) {
						result[m, ] <- 1; result[m, prev_s] <- 0; result[m, prev_r] <- 0
					}
				} else if (pshift_num == 7L) { # AB-XY
					if (curr_s != prev_s && curr_s != prev_r) {
						result[m, ] <- 1; result[m, prev_s] <- 0; result[m, prev_r] <- 0
					}
				}
			}
		}
	}
	result
}

# Compute pshift.type_c for the TIE model (tomstats), untyped riskset.
# For each output row m, finds the last type-c event and applies the pshift
# condition for all D dyads (actor1_0based, actor2_0based).
# edgelist: [M x 4] matrix, cols = (time, actor1_0based, actor2_0based, type_0based)
# event_type_ids: 0-based typeIDs, length M
# type_id: 0-based typeID to filter on
# rs_a1, rs_a2: 0-based actor IDs for each of D dyads
# Returns [M_out x D] matrix.
compute_pshift_type_c_tie <- function(pshift_num, edgelist, event_type_ids,
																			type_id, rs_a1, rs_a2,
																			M_out, start, stop) {
	D         <- length(rs_a1)
	full_rows <- (start + 1L):(stop + 1L)   # 1-based
	result    <- matrix(0, nrow = M_out, ncol = D)
	
	out_times    <- edgelist[full_rows, 1]
	unique_times <- sort(unique(out_times))
	
	for (tp_idx in seq_along(unique_times)) {
		curr_t <- unique_times[tp_idx]
		
		# Previous time point events of type ci
		if (tp_idx == 1L) {
			before_all <- which(edgelist[, 1] < curr_t)
			if (length(before_all) == 0L) next
			prev_t <- max(edgelist[before_all, 1])
		} else {
			prev_t <- unique_times[tp_idx - 1L]
		}
		
		prev_ci_idx <- which(edgelist[, 1] == prev_t & event_type_ids == type_id)
		if (length(prev_ci_idx) == 0L) next
		
		# Output rows at curr_t
		m_rows <- which(out_times == curr_t)
		
		for (prev_e in prev_ci_idx) {
			prev_s <- edgelist[prev_e, 2]
			prev_r <- edgelist[prev_e, 3]
			
			for (m in m_rows) {
				curr_s <- edgelist[full_rows[m], 2]   # 0-based sender of this output event
				
				if (pshift_num == 1L) {        # AB-AB
					if (curr_s == prev_s)
						result[m, rs_a1 == prev_s & rs_a2 == prev_r] <- 1
				} else if (pshift_num == 2L) { # AB-BA
					if (curr_s == prev_r)
						result[m, rs_a1 == prev_r & rs_a2 == prev_s] <- 1
				} else if (pshift_num == 3L) { # AB-XB
					if (curr_s != prev_s && curr_s != prev_r)
						result[m, rs_a1 != prev_s & rs_a1 != prev_r & rs_a2 == prev_r] <- 1
				} else if (pshift_num == 4L) { # AB-XA
					if (curr_s != prev_s && curr_s != prev_r)
						result[m, rs_a1 != prev_s & rs_a1 != prev_r & rs_a2 == prev_s] <- 1
				} else if (pshift_num == 5L) { # AB-AY
					if (curr_s == prev_s)
						result[m, rs_a1 == prev_s & rs_a2 != prev_r & rs_a2 != prev_s] <- 1
				} else if (pshift_num == 6L) { # AB-BY
					if (curr_s == prev_r)
						result[m, rs_a1 == prev_r & rs_a2 != prev_s & rs_a2 != prev_r] <- 1
				} else if (pshift_num == 7L) { # AB-XY
					if (curr_s != prev_s && curr_s != prev_r)
						result[m, rs_a1 != prev_s & rs_a1 != prev_r &
									 	rs_a2 != prev_s & rs_a2 != prev_r] <- 1
				}
			}
		}
	}
	result
}

# [return] array [M_out x N x P_out] with correct dimnames[[3]]
compute_aomstats_with_type <- function(compute_fn, add_names_fn,
																			 edgelist, weights, event_type_ids,
																			 types_df, effects_names, consider_type,
																			 with_type, C, subset_start, subset_stop, N) {
	M     <- nrow(edgelist)
	M_out <- subset_stop - subset_start + 1L
	out_rows <- (subset_start + 1L):(subset_stop + 1L)
	
	ignore_idx   <- which(consider_type == "ignore" | !with_type | C <= 1L)
	separate_idx <- which(consider_type == "separate" & with_type & C > 1L)
	
	result_slices <- vector("list", length(effects_names))
	
	# ── ignore effects: single C++ call ───────────────────────────────────────
	if (length(ignore_idx) > 0L) {
		ign_names <- effects_names[ignore_idx]
		st_ign    <- compute_fn(ign_names, edgelist, weights,
														subset_start, subset_stop)
		dimnames(st_ign)[[3]] <- ign_names
		for (k in seq_along(ignore_idx)) {
			p  <- ignore_idx[k]
			sl <- st_ign[, , k, drop = FALSE]
			dimnames(sl)[[3]] <- ign_names[k]
			result_slices[[p]] <- sl
		}
	}
	
	# ── separate effects ──────────────────────────────────────────────────────
	if (length(separate_idx) > 0L) {
		for (p in separate_idx) {
			eff_name    <- effects_names[p]
			type_slices <- vector("list", C)
			
			if (eff_name %in% PSHIFT_NAMES) {
				# Pshift: use R-level per-type computation with correct
				# method="pt" semantics (previous time point events only).
				pshift_num <- PSHIFT_TYPE_NUM[eff_name]
				for (ci in seq_len(C)) {
					type_id   <- types_df$typeID[ci]
					type_name <- types_df$typeName[ci]
					st_c <- compute_pshift_type_c_receiver(
						pshift_num, edgelist, event_type_ids, type_id,
						N, subset_start, subset_stop
					)
					st_c <- array(st_c, dim = c(M_out, N, 1L))
					dimnames(st_c)[[3]] <- paste0(eff_name, ".", type_name)
					type_slices[[ci]] <- st_c
				}
			} else {
				# Weighted-sum effects: masked-weights C++ call
				for (ci in seq_len(C)) {
					type_id   <- types_df$typeID[ci]
					type_name <- types_df$typeName[ci]
					weights_c <- weights * (event_type_ids == type_id)
					st_c <- compute_fn(eff_name, edgelist, weights_c,
														 subset_start, subset_stop)
					st_c <- array(as.numeric(st_c), dim = c(M_out, N, 1L))
					dimnames(st_c)[[3]] <- paste0(eff_name, ".", type_name)
					type_slices[[ci]] <- st_c
				}
			}
			
			result_slices[[p]] <- type_slices
		}
	}
	
	# ── Assemble in original effect order ─────────────────────────────────────
	final_slices <- list()
	for (p in seq_along(result_slices)) {
		sl <- result_slices[[p]]
		if (is.array(sl)) {
			final_slices[[length(final_slices) + 1L]] <- sl
		} else if (is.list(sl)) {
			for (ci in seq_len(C)) {
				final_slices[[length(final_slices) + 1L]] <- sl[[ci]]
			}
		}
	}
	
	P_out  <- length(final_slices)
	snames <- sapply(final_slices, function(s) dimnames(s)[[3]])
	N_out  <- dim(final_slices[[1]])[2]
	array(
		unlist(lapply(final_slices, function(s) s[, , 1L])),
		dim      = c(M_out, N_out, P_out),
		dimnames = list(NULL, NULL, snames)
	)
}

# Compute type-specific aomstats and expand to full sequence with LOCF.
#
# For each type c:
#   1. Filter edgelist to type-c events only -> [M_c x 3]
#   2. Call compute_stats_sender/receiver on filtered edgelist -> [M_c x N x P]
#   3. Expand back to [M x N x P] using last-observation-carried-forward:
#      between consecutive type-c events, stats are constant (no type-c
#      event occurred). Before the first type-c event, stats are zero.
#
# Each base effect p produces C named slices: "effectname.typename"
#
# [param] compute_fn     function(edgelist, weights, subset_start, subset_stop)
#                        returns [M_c x N x P] array with named dimnames[[3]]
# [param] edgelist       full edgelist matrix [M x 3], 0-based actor IDs
# [param] weights        numeric vector length M
# [param] event_type_ids integer vector length M, 1-based typeID per event
# [param] types_df       data.frame: typeName, typeID (1-based), C rows
# [param] effect_names   character vector length P of base effect names
# [param] M_out          number of output rows (after start/stop subsetting)
# [param] start_0        0-based start index passed to C++
# [param] stop_0         0-based stop index passed to C++
#
# [return] array [M_out x N x P*C] with type-split dimnames[[3]]
compute_aom_type_separate <- function(compute_fn, edgelist, weights,
                                       event_type_ids, types_df,
                                       effect_names, M_out, start_0, stop_0) {
  M <- nrow(edgelist)
  C <- nrow(types_df)

  # Output indices (1-based) in the full sequence
  out_rows <- (start_0 + 1L):(stop_0 + 1L)  # 1-based positions in edgelist

  new_slices <- list()

  for (ci in seq_len(C)) {
    type_id <- types_df$typeID[ci]   # 1-based
    type_name <- types_df$typeName[ci]

    # Filter to type-c events in the FULL sequence (for history)
    keep_full <- event_type_ids == type_id   # logical length M
    el_c      <- edgelist[keep_full, , drop=FALSE]
    wts_c     <- weights[keep_full]
    M_c       <- nrow(el_c)

    if (M_c == 0L) {
      # No events of this type — all slices are zero
      for (p in seq_along(effect_names)) {
        sl <- array(0, dim=c(M_out, ncol(edgelist), 1L))
        # N comes from first successful call; defer — fill below
        new_slices[[length(new_slices) + 1L]] <- list(
          type=type_name, effect=effect_names[p], data=NULL)
      }
      next
    }

    # C++ start/stop within the filtered sequence:
    # We want output rows corresponding to out_rows intersected with type-c
    # positions. But we need ALL M_c rows to carry forward correctly,
    # so call C++ on the full filtered sequence (start=1, stop=M_c).
    st_c <- compute_fn(el_c, wts_c, 0L, M_c - 1L)  # [M_c x N x P]
    N    <- dim(st_c)[2]
    P    <- dim(st_c)[3]

    # Positions in original sequence where type-c events occur (1-based)
    type_c_pos <- which(keep_full)  # length M_c

    # Expand st_c to full sequence [M x N x P] using LOCF
    # Before first type-c event: zero (array init)
    st_full <- array(0, dim=c(M, N, P))
#    print(c(M, N, P))
    for (k in seq_len(M_c)) {
#    	print(st_c[k,,])
    	from <- type_c_pos[k]
    	to   <- if (k < M_c) type_c_pos[k + 1L] - 1L else M
    	print(c(from,to))
    	for (row in from:to) {
    		st_full[row, , ] <- st_c[k, , ]
    	}
    }

    # Subset to output rows
    st_out <- st_full[out_rows, , , drop=FALSE]  # [M_out x N x P]

    for (p in seq_len(P)) {
      slice_name <- paste0(dimnames(st_c)[[3]][p], ".", type_name)
      sl <- st_out[, , p, drop=FALSE]
      dimnames(sl)[[3]] <- slice_name
      new_slices[[length(new_slices) + 1L]] <- sl
    }
  }

  # Remove any NULL placeholders and get N from first real slice
  valid <- Filter(Negate(is.null), new_slices)
  # Handle zero-type case
  valid <- lapply(new_slices, function(x) {
    if (is.list(x) && is.null(x$data)) {
      array(0, dim=c(M_out, dim(valid[[which(sapply(new_slices, is.array))]])[2], 1))
    } else x
  })
  valid <- Filter(is.array, new_slices)

  P_out  <- length(valid)
  snames <- sapply(valid, function(s) dimnames(s)[[3]])
  N_out  <- dim(valid[[1]])[2]
  array(
    unlist(lapply(valid, function(s) s[, , 1])),
    dim      = c(M_out, N_out, P_out),
    dimnames = list(NULL, NULL, snames)
  )
}

prepare_aomstats_edgelist <- function(reh) {
  reh <- normalize_reh(reh)
  # Extract the edgelist
  edgelist <- reh$edgelist

  # Transform to cpp indexing
  edgelist[, 2] <- unlist(reh$ids$actor1) - 1
  edgelist[, 3] <- unlist(reh$ids$actor2) - 1

  # Origin
  if (reh$meta$ordinal) {
  	if (length(reh$meta$origin)==0) {
  		reh$meta$origin <- edgelist[1,1] - 1
  	}
  }

  # Transform to numeric
  edgelist[, 1] <- cumsum(
    as.numeric(
      diff(
        c(reh$meta$origin, edgelist[, 1])
      )
    )
  )

  # Transform to matrix
  edgelist <- as.matrix(edgelist[,1:3])

  return(edgelist)
}

prepare_aomstats_actors <- function(reh) {
  reh <- normalize_reh(reh)
  # Extract the actors
  actors <- reh$meta$dictionary$actors

  # Transform to cpp indexing
  actors$actorID <- actors$actorID - 1

  return(actors)
}

prepare_aomstats_event_weights <- function(reh) {
  reh <- normalize_reh(reh)
  if (!(reh$meta$weighted)) {
    weights <- rep(1, nrow(reh$edgelist))
  } else {
    weights <- reh$edgelist$weight
  }
  return(weights)
}

prepare_subset <- function(start, stop, edgelist, method, model) {
  if (start < 1) {
    stop("The 'start' value should be set to 1 or a larger number.")
  }
  start <- start - 1
  if (stop == Inf) {
    if (method == "pe") {
      stop <- nrow(edgelist)
    } else if (method == "pt") {
      stop <- NROW(unique(edgelist[, 1]))
    }
  }
  stop <- stop - 1
  if (stop < start) {
    stop("The 'stop' value cannot be smaller than the 'start' value.")
  }
  if (method == "pe") {
    if (stop > nrow(edgelist)) {
      stop("The 'stop' value cannot be larger than the number of events in 'reh'.")
    }
  }
  if (method == "pt") {
    if (stop > NROW(unique(edgelist[, 1]))) {
      stop("The 'stop' value cannot be larger than the number of unique time points in 'reh'.")
    }
  }
  if (model == "receiver" & method == "pt") {
    # For the receiver model, we need all timepoints not only the unique one
    unique_times <- unique(edgelist[, 1])
    start_time <- unique_times[start + 1]
    stop_time <- unique_times[stop + 1]
    all_times <- edgelist[, 1]
    start <- min(which(all_times == start_time)) - 1
    stop <- max(which(all_times == stop_time)) - 1
  }
  list(start = start, stop = stop)
}

prepare_sender_effects <- function(sender_formula, ordinal = FALSE) {
  check_formula(sender_formula)
  sender_effects <- parse_formula(sender_formula, "rateEffects", ordinal)
  sender_effects_names <- sapply(sender_effects, function(x) x$effect)

  # Check correct specification effects
  if (!all(sender_effects_names %in% actor_effects(step = "sender"))) {
    stop(paste("Invalid effect: Attempting to request effects that are not defined for the sender activity model"))
  }

  # Prepare interaction sender_effects
  sender_effects_int <- parse_int(
    sender_formula, "rateEffects",
    sender_effects
  )
  sender_effects_names <- append(sender_effects_names, rep("interact", length(sender_effects_int)), length(sender_effects_names))
  sender_interactions <- list()
  sender_interactions[which(sender_effects_names == "interact")] <- sender_effects_int

  # Output
  list(
    sender_effects = sender_effects,
    sender_effects_names = sender_effects_names,
    sender_interactions = sender_interactions
  )
}

prepare_aomstats_scaling <- function(effects, interactions) {
  scaling <- sapply(effects, function(x) {
    ifelse("scaling" %in% names(x), x$scaling, "none")
  })
  scaling <- append(scaling, rep("none", sum(!(sapply(interactions, is.null)))), length(scaling))
  scaling
}

prepare_receiver_effects <- function(receiver_formula) {
  check_formula(receiver_formula)
  receiver_effects <- parse_formula(receiver_formula, "choiceEffects")
  receiver_effects_names <- sapply(receiver_effects, function(x) x$effect)

  # Check correct specification effects
  if (!all(receiver_effects_names %in% actor_effects(step = "receiver"))) {
    stop(paste("Invalid effect: Attempting to request effects that are not defined for the receiver choice model"))
  }

  # Prepare interaction receiver_effects
  receiver_effects_int <- parse_int(
    receiver_formula, "choiceEffects",
    receiver_effects
  )
  receiver_effects_names <- append(receiver_effects_names, rep("interact", length(receiver_effects_int)), length(receiver_effects_names))
  receiver_interactions <- list()
  receiver_interactions[which(receiver_effects_names == "interact")] <- receiver_effects_int

  # Output
  list(
    receiver_effects = receiver_effects,
    receiver_effects_names = receiver_effects_names,
    receiver_interactions = receiver_interactions
  )
}

# Actor-oriented model: covariates ----------------------------------------
validate_attr_actors <- function(attr_actors, x) {

  # Check if the variable name is in the attr_actors object
  if (!(x$variable %in% colnames(attr_actors))) {
    stop(paste0("Variable '", x$variable, "' not in attr_actors object for the '", x$effect, "' effect."))
  }

  # Check if the 'time' variable is available
  if (!("time" %in% colnames(attr_actors))) {
    if (anyDuplicated(attr_actors$name) > 0) {
      warning("Duplicated actor names detected in 'attr_actors'. Did you mean to add a 'time' variable?")
    } else {
      attr_actors$time <- 0
    }
  }

  if (anyNA(attr_actors$time)) {
    stop("Missing 'time' values detected in 'attr_actors'.")
  }

  return(attr_actors)
}

prepare_covariate_data <- function(attr_actors, x, actors) {
  # Collect the information in a data.frame
  dat <- data.frame(
    name = attr_actors$name,
    time = attr_actors$time,
    x = attr_actors[, x$variable]
  )

  # Check the data class of the variable
  dat <- variable_check(x$effect, dat)

  # Convert 'actorName' to 'actorID'
  dat$name <- actors[match(dat$name, actors[, 1]), 2]

  # Set the third colum name equal to the 'variable'
  colnames(dat)[3] <- x$variable

  dat
}

validate_covariate_data <- function(dat, x, actors) {
  # Warning for missing values
  if (anyNA(dat)) {
    warning(paste0("Missing values in the attr_actors object for the '", x$effect, "' effect can cause unexpected behavior."))
  }
  # Check if all actors are in the attr_actors
  if (!all(actors[, 2] %in% dat$name)) {
    stop("Missing actors in the attr_actors object.")
  }
  # Check for actors in the attr_actors object that are not in the
  # risk set
  if (any(is.na(dat$name))) {
    warning(paste0("attr_actors contain actors that are not in the risk set. These are not included in the computation of the statistics."))
    dat <- dat[!is.na(dat$name), ]
  }
  as.matrix(dat)
}

validate_user_stat <- function(x, edgelist, actors) {
  if (NCOL(x) != nrow(actors)) {
    stop("Number of columns of matrix 'x' in userStat() does not match the number of actors")
  }
  as.matrix(x)
}

validate_covariates <- function(
    x, attr_actors, attr_dyads, actors, edgelist,
    reh) {
  if (x$effect %in% c("send", "receive", "same", "difference", "average")) {
    if (is.null(x$x)) {
      attr_actors <- validate_attr_actors(attr_actors, x)
      dat <- prepare_covariate_data(attr_actors, x, actors)
      validate_covariate_data(dat, x, actors)
    } else {
      dat <- x$x
      dat$name <- actors[match(dat$name, actors[, 1]), 2]
      validate_covariate_data(dat, x, actors)
    }
  } else if (x$effect == "tie") {
    parse_tie(x, reh, attr_dyads)
  } else if (x$effect == "userStat") {
    validate_user_stat(x$x, edgelist, actors)
  } else {
    matrix()
  }
}

prepare_sender_covariates <- function(sender_effects, attr_actors, actors, edgelist, reh) {
  sender_covar <- lapply(sender_effects, function(x) {
    validate_covariates(x, attr_actors, NULL, actors, edgelist, reh)
  })
  sender_covar
}

prepare_receiver_covariates <- function(
    receiver_effects, attr_actors, attr_dyads, actors,
    edgelist, reh) {
  receiver_covar <- lapply(receiver_effects, function(x) {
    validate_covariates(x, attr_actors, attr_dyads, actors, edgelist, reh)
  })

  return(receiver_covar)
}

# tie/dyad effect ---------------------------------------------------------
prep_tie <- function(variable, attr_dyads, scaling) {
  if (is.null(attr_dyads)) {
    list(
      effect = "tie",
      variable = variable,
      x = NULL,
      scaling = scaling
    )
  } else {
    # Wide or long format?
    # - First check
    wide <- ifelse(nrow(attr_dyads) == ncol(attr_dyads), TRUE, FALSE)
    # - Second check
    wide <- ifelse(wide & !(variable %in% colnames(attr_dyads)), TRUE, FALSE)

    # Prep long format
    if (!wide) {
      # Check for presence of time-column
      if ("time" %in% colnames(attr_dyads)) {
        time <- attr_dyads[, "time"]
      } else {
        # Check for multiple actor appearances
        if (any(duplicated(attr_dyads[, c(1, 2)]))) {
          warning("Duplicated dyads in 'attr_dyads': Did you mean to include a 'time' column?")
        }
        time <- rep(0, nrow(attr_dyads))
      }

      # Check for presence of variable
      if (!(variable %in% colnames(attr_dyads))) {
        stop(paste("Invalid variable: Cannot find", variable, "in attr_dyads"))
      }

      # Check for missing values
      if (anyNA(time)) {
        warning("Missing time values in 'attr_dyads'")
      }

      if (anyNA(attr_dyads[, 1]) | anyNA(attr_dyads[, 2])) {
        warning("Missing actors in 'attr_dyads'")
      }

      if (anyNA(attr_dyads[, variable])) {
        warning("Missing variable values in 'attr_dyads'")
      }

      # Collect data
      dat <- data.frame(
        actor1 = attr_dyads[, 1],
        actor2 = attr_dyads[, 2],
        time = time,
        x = attr_dyads[, variable]
      )

      colnames(dat)[4] <- variable
    }

    if (wide) {
      dat <- attr_dyads
    }

    # Output
    list(
      effect = "tie",
      variable = variable,
      x = dat,
      scaling = scaling
    )
  }
}

tie_convert_wide_to_long <- function(x, variable, reh) {
  actors <- reh$meta$dictionary$actors

  # Check actors in x
  if (is.null(rownames(x)) | is.null(colnames(x))) {
    if (nrow(x) != nrow(actors) | ncol(x) != nrow(actors)) {
      stop("Invalid attr_dyads object: Number of rows and columns should match number of actors in the network")
    }
    rownames(x) <- colnames(x) <- actors$actorName
  }

  # Error message in the case of missing actors
  if ((!all(actors$actorName %in% rownames(x))) |
    (!all(actors$actorName %in% colnames(x)))) {
    warning("Missing actors in 'attr_dyads'")
  }

  # Recode
  rownames(x) <- actors$actorID[match(rownames(x), actors$actorName)] - 1
  colnames(x) <- actors$actorID[match(colnames(x), actors$actorName)] - 1

  # Reorder
  x <- x[order(as.numeric(rownames(x))), ]
  x <- x[, order(as.numeric(colnames(x)))]

  # Undirected events
  if (!reh$meta$directed) {
    if (!isSymmetric(x)) {
      if (all(is.na(x[upper.tri(x)]))) {
        x[upper.tri(x)] <- t(x)[upper.tri(x)]
      } else if (all(is.na(x[lower.tri(x)]))) {
        x[lower.tri(x)] <- t(x)[lower.tri(x)]
      } else {
        warning("Undirected events: attr_dyads matrix is expected to be symmetrical")
      }
    }
  }

  # Error message in the case of missing values
  if (anyNA(x[lower.tri(x)]) | anyNA(x[upper.tri(x)])) {
    warning("Missing variable values in 'attr_dyads'")
  }

  # Get row and column names
  row_ids <- rownames(x)
  col_ids <- colnames(x)

  # Initialize empty vectors to store data
  row_data <- character(0)
  col_data <- character(0)
  value_data <- numeric(0)

  # Iterate through the matrix
  for (i in 1:nrow(x)) {
    for (j in 1:ncol(x)) {
      row_data <- c(row_data, row_ids[i])
      col_data <- c(col_data, col_ids[j])
      value_data <- c(value_data, x[i, j])
    }
  }

  # Create a data frame
  result_df <- data.frame(
    actor1 = as.integer(row_data),
    actor2 = as.integer(col_data),
    time = 0,
    x = value_data
  )
  colnames(result_df)[4] <- variable
  result_df
  as.matrix(result_df)
}

parse_tie <- function(prepped_effect, reh, attr_dyads) {
  x <- prepped_effect$x

  if (is.null(x)) {
    x <- prep_tie(prepped_effect$variable, attr_dyads, prepped_effect$scaling)$x
  }
  # Check the format
  wide <- ifelse("actor1" %in% colnames(x), FALSE, TRUE)
  if (wide) {
    tie_convert_wide_to_long(x, prepped_effect$variable, reh)
    # return: converted matrix
  } else {
    actors <- reh$meta$dictionary$actors
    x$actor1 <- actors$actorID[match(x$actor1, actors$actorName)] - 1
    x$actor2 <- actors$actorID[match(x$actor2, actors$actorName)] - 1
    as.matrix(x) # return
  }
}

#' Stack remstats for model fitting
#'
#' Stack a \code{tomstats} object into long format suitable for fitting with
#' standard R model functions. Each row corresponds to one dyad in the risk
#' set at one event time point.
#'
#' For interval timing (\code{ordinal = FALSE}), the stacked data can be
#' fitted with a Poisson GLM using \code{log_interevent} as an offset:
#' \code{glm(obs ~ -1 + . - event - dyad - log_interevent,
#'           offset = log_interevent, family = poisson, data = stacked$remstats_stack)}
#'
#' For ordinal timing (\code{ordinal = TRUE}), use conditional logistic
#' regression stratified by event:
#' \code{survival::clogit(obs ~ -1 + . - event - dyad +
#'                        strata(event), data = stacked$remstats_stack)}
#'
#' @param stats A \code{tomstats} object (output of \code{remstats()} or
#'   \code{tomstats()}).
#' @param reh A \code{remify} object (output of \code{remify::remify2()}).
#'
#' @return A list with elements:
#' \describe{
#'   \item{remstats_stack}{Data frame in long format with columns: \code{event}
#'     (event index), \code{dyad} (dyad index 1..D), all statistic columns,
#'     \code{obs} (1 = observed event, 0 = non-event), and
#'     \code{log_interevent} (log inter-event time; only for interval timing).}
#'   \item{subset}{Integer vector of length 2: first and last event index.}
#'   \item{D}{Number of dyads in the risk set.}
#'   \item{E}{Number of events (time points).}
#'   \item{ordinal}{Logical: whether the ordinal likelihood applies.}
#' }
#'
#' @export
stack_stats <- function(stats, reh) {
  UseMethod("stack_stats")
}

#' @export
#' @method stack_stats tomstats
stack_stats.tomstats <- function(stats, reh) {

  if (!inherits(reh, "remify")) stop("'reh' must be a remify object.")

  # ── Dimensions ──────────────────────────────────────────────────────────────
  D  <- dim(stats)[2]   # number of dyads
  E  <- dim(stats)[1]   # number of time points in stats
  subset_idx <- as.integer(unlist(attr(stats, "subset")))  # [start, stop]

  # ── Ordinal flag ─────────────────────────────────────────────────────────────
  ordinal <- if (!is.null(reh$meta)) isTRUE(reh$meta$ordinal) else
             isTRUE(attr(reh, "ordinal"))

  # ── Riskset type ─────────────────────────────────────────────────────────────
  riskset <- if (!is.null(reh$meta))
    (reh$meta$riskset_source %||% reh$meta$riskset)
  else
    attr(reh, "riskset")

  # ── Stack statistics: [E*D x P] ──────────────────────────────────────────────
  stat_glm <- as.data.frame(
    do.call(rbind, lapply(seq_len(E), function(e) cbind(e, stats[e, , ])))
  )
  colnames(stat_glm)[1] <- "event"

  # ── Offset: log inter-event time (interval timing only) ──────────────────────
  if (!ordinal) {
    iet <- reh$intereventTime[subset_idx[1]:subset_idx[2]]
    stat_glm$log_interevent <- rep(log(iet), each = D)
  }

  # ── Response: observed events per dyad per time point ────────────────────────
  # dyad_active / dyad contain 1-based column indices into the stats array
  if (!is.null(reh$ids)) {
    # New remify2 format
    if (riskset %in% c("active", "manual")) {
      # dyad_active is [N x 1] matrix — flatten to vector, then subset
      dyad_vec <- as.vector(reh$ids$dyad_active)
    } else {
      dyad_vec <- as.vector(reh$ids$dyad)
    }
  } else {
    # Old remify format
    if (riskset %in% c("active", "manual")) {
      dyad_vec <- as.vector(attr(reh, "dyadIDactive") %||% attr(reh, "dyadID"))
    } else {
      dyad_vec <- as.vector(attr(reh, "dyadID"))
    }
  }

  stat_glm$obs <- unlist(lapply(seq_len(E), function(e) {
    ev_idx <- subset_idx[1] + e - 1L
    obs_dyad <- dyad_vec[ev_idx]
    tabulate(obs_dyad, nbins = D)
  }))

  # ── Dyad index ───────────────────────────────────────────────────────────────
  stat_glm$dyad <- rep(seq_len(D), E)

  list(
    remstats_stack = stat_glm,
    subset  = subset_idx,
    D       = D,
    E       = E,
    ordinal = ordinal
  )
}

# NULL-coalescing operator (defined locally if not already available)
if (!exists("%||%")) {
  `%||%` <- function(a, b) if (!is.null(a)) a else b
}

#' @export
#' @method stack_stats tomstats_sampled
stack_stats.tomstats_sampled <- function(stats, reh) {

  if (!inherits(reh, "remify")) stop("'reh' must be a remify object.")

  # ── Dimensions ──────────────────────────────────────────────────────────────
  E  <- dim(stats)[1]   # number of events
  S  <- dim(stats)[2]   # sampled set size per event
  subset_idx <- as.integer(unlist(attr(stats, "subset")))

  # ── Sampling attributes ──────────────────────────────────────────────────────
  sample_map <- attr(stats, "sample_map")  # [E x S] 1-based dyad indices
  samp_prob  <- attr(stats, "samp_prob")   # [E x S] inclusion probabilities
  case_pos   <- attr(stats, "case_pos")    # list of length E, 1-based

  if (is.null(sample_map) || is.null(samp_prob) || is.null(case_pos)) {
    stop("'stats' is missing required sampling attributes.")
  }

  # ── Ordinal flag ─────────────────────────────────────────────────────────────
  ordinal <- if (!is.null(reh$meta)) isTRUE(reh$meta$ordinal) else
             isTRUE(attr(reh, "ordinal"))

  # ── Stack statistics: [E*S x P] ──────────────────────────────────────────────
  stat_glm <- as.data.frame(
    do.call(rbind, lapply(seq_len(E), function(e) cbind(e, stats[e, , ])))
  )
  colnames(stat_glm)[1] <- "event"

  # ── Offset: log inter-event time (interval only) ─────────────────────────────
  if (!ordinal) {
    iet <- reh$intereventTime[subset_idx[1]:subset_idx[2]]
    stat_glm$log_interevent <- rep(log(iet), each = S)
  }

  # ── Response: 1 for case, 0 for control ──────────────────────────────────────
  stat_glm$obs <- unlist(lapply(seq_len(E), function(e) {
    is_case <- integer(S)
    cp <- case_pos[[e]]  # 1-based case positions
    if (length(cp) > 0) is_case[cp] <- 1L
    is_case
  }))

  # ── Importance weights: 1 for cases, 1/pi_s for controls ─────────────────────
  stat_glm$weight <- unlist(lapply(seq_len(E), function(e) {
    w <- 1.0 / samp_prob[e, ]       # start with 1/pi_s for all
    cp <- case_pos[[e]]              # 1-based case positions
    if (length(cp) > 0) w[cp] <- 1.0  # cases get weight 1
    w
  }))

  # ── Sampled dyad index (1-based within active riskset) ───────────────────────
  stat_glm$dyad <- as.vector(t(sample_map))  # [E*S]: dyad ID for each row

  list(
    remstats_stack = stat_glm,
    subset  = subset_idx,
    S       = S,
    E       = E,
    ordinal = ordinal
  )
}


#' @export
#' @method stack_stats aomstats
stack_stats.aomstats <- function(stats, reh) {
	
	if (!inherits(reh, "remify")) stop("'reh' must be a remify object.")
	
	# ── Dimensions ───────────────────────────────────────────────────────────────
	subset_idx <- as.integer(unlist(attr(stats, "subset")))  # [start, stop]
	E  <- subset_idx[2] - subset_idx[1] + 1L
	N  <- reh$N
	
	# ── Ordinal flag ─────────────────────────────────────────────────────────────
	ordinal <- isTRUE(reh$meta$ordinal)
	
	# ── Observed sender/receiver per event ───────────────────────────────────────
	# ids$actor1 and ids$actor2 are 1-based
	obs_sender   <- reh$ids$actor1[ subset_idx[1]:subset_idx[2] ]
	obs_receiver <- reh$ids$actor2[ subset_idx[1]:subset_idx[2] ]
	
	# ── Inter-event times (interval only) ────────────────────────────────────────
	if (!ordinal) {
		iet <- reh$intereventTime[ subset_idx[1]:subset_idx[2] ]
		log_iet <- log(iet)
	}
	
	# ── Receiver riskset (list indexed by 1-based sender ID) ─────────────────────
	rec_riskset <- reh$receiver_riskset  # list of length N, each N-1 1-based IDs
	
	# ── SENDER STACK ─────────────────────────────────────────────────────────────
	sender_stack <- if (!is.null(stats$sender_stats)) {
		
		ss <- stats$sender_stats  # [E x N x Ks]
		Ks <- dim(ss)[3]
		stat_names <- dimnames(ss)[[3]]
		
		df <- as.data.frame(
			do.call(rbind, lapply(seq_len(E), function(e) cbind(e, ss[e, , ])))
		)
		colnames(df) <- c("event", stat_names)
		
		if (!ordinal) {
			df$log_interevent <- rep(log_iet, each = N)
		}
		
		df$obs   <- unlist(lapply(seq_len(E), function(e) {
			tabulate(obs_sender[e], nbins = N)
		}))
		df$actor <- rep(seq_len(N), E)
		
		df
		
	} else NULL
	
	# ── RECEIVER STACK ───────────────────────────────────────────────────────────
	receiver_stack <- if (!is.null(stats$receiver_stats)) {
		
		rs <- stats$receiver_stats  # [E x N x Kr]
		Kr <- dim(rs)[3]
		stat_names <- dimnames(rs)[[3]]
		
		do.call(rbind, lapply(seq_len(E), function(e) {
			s_id  <- obs_sender[e]
			r_ids <- rec_riskset[[ s_id ]]
			obs_r <- obs_receiver[e]
			
			mat  <- matrix(rs[e, r_ids, ], nrow = length(r_ids), ncol = Kr)
			df_e <- as.data.frame(mat)
			colnames(df_e) <- stat_names
			
			df_e$obs   <- as.integer(r_ids == obs_r)
			df_e$actor <- r_ids
			df_e$event <- e
			df_e[, c("event", stat_names, "obs", "actor")]
		}))
		
	} else NULL
	
	list(
		sender_stack   = sender_stack,
		receiver_stack = receiver_stack,
		subset         = subset_idx,
		N              = N,
		E              = E,
		ordinal        = ordinal
	)
}

#' remstats
#'
#' Computes statistics for modeling relational events with a tie-oriented or 
#' actor-oriented approach.
#'
#' @section Effects:
#' The statistics to be computed are defined symbolically and should be
#' supplied to the \code{tie_effects} (for the tie-oriented model), or
#' \code{sender_effects} and/or \code{receiver_effects} (for the actor-oriented
#' model) argument in the form \code{~ effects}. The terms are separated by +
#' operators. For example: \code{effects = ~ inertia() + otp()}. Interactions
#' between two effects can be included with * or : operators. For example:
#' \code{effects = ~ inertia():otp()}. A list of available effects
#' can be obtained with \code{\link{tie_effects}()} and
#' \code{\link{actor_effects}()}.
#'
#' The majority of the statistics can be scaled in some way, see
#' the documentation of the \code{scaling} argument in the separate effect
#' functions for more information on this.
#'
#' The majority of the statistics can account for the event type
#' included as a dependent variable, see the documentation of the
#' \code{consider_type} argument in the separate effect functions for more
#' information on this. Note that this option is only available for the
#' tie-oriented model.
#'
#' Note that events in the relational event history can be directed or
#' undirected. Some statistics are only defined for either directed or
#' undirected events (see the documentation of the statistics). Note that
#' undirected events are only available for the tie-oriented model.
#'
#' @section attr_actors:
#' For the computation of the \emph{exogenous} statistics an attributes 
#' object with the exogenous covariate information has to be supplied to the
#' \code{attr_actors} argument in either \code{remstats()} or in the separate
#' effect functions supplied to the \code{..._effects} arguments (e.g., see
#' \code{\link{send}}). This \code{attr_actors} object should be constructed as
#' follows: A dataframe with rows referring to the attribute value of actor
#' \emph{i} at timepoint \emph{t}. A `name` column is required that contains the
#' actor name (corresponding to the actor names in the relational event
#' history). A `time` column is required that contains the time when attributes
#' change (set to zero if none of the attributes vary over time). Subsequent
#' columns contain the attributes that are called in the specifications of
#' exogenous statistics (column name corresponding to the string supplied to
#' the \code{variable} argument in the effect function). Note that the
#' procedure for the exogenous effects `tie' and `event' deviates from this,
#' here the exogenous covariate information has to be specified in a different
#' way, see \code{\link{tie}} and \code{\link{event}}.
#' 
#' @section attr_dyads:  
#' For the computation of the \emph{dyad exogenous} statistics with 
#' \code{tie()}, an attributes object with the exogenous covariates information 
#' per dyad has to be supplied. This is a \code{data.frame} or \code{matrix} 
#' containing attribute information for dyads. If \code{attr_dyads} is a 
#' \code{data.frame}, the first two columns should represent "actor1" and 
#' "actor2" (for directed events, "actor1" corresponds to the sender, and 
#' "actor2" corresponds to the receiver). Additional columns can represent 
#' dyads' exogenous attributes. If attributes vary over time, include a column 
#' named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond 
#' to "actor1", columns to "actor2", and cells contain dyads' exogenous 
#' attributes.
#'
#' @section Memory:
#' The default `memory` setting is `"full"`, which implies that at each time
#' point $t$ the entire event history before $t$ is included in the computation
#' of the statistics. Alternatively, when `memory` is set to `"window"`, only
#' the past event history within a given time window is considered (see
#' Mulders & Leenders, 2019). This length of this time window is set by the
#' `memory_value` parameter. For example, when `memory_value = 100` and `memory
#' = "window"`, at time point $t$ only the past events that happened at most
#' 100 time units ago are included in the computation of the statistics.
#' A third option is to set `memory` to `"interval"`. In this case, the past
#' event history within a given time interval is considered. For example, when
#' `"memory_value" = c(50, 100)` and `memory = "window"`, at time point $t$
#' only the past events that happened between 50 and 100 time units ago are
#' included in the computation of the statistics. Finally, the fourth option is
#' to set `memory` to `"decay"`. In this case, the weight of the past event in
#' the computation of the statistics depend on the elapsed time between $t$ and
#' the past event. This weight is determined based on an exponential decay
#' function with half-life parameter `memory_value` (see Brandes et al., 2009).
#'
#' @section Event weights:
#' Note that if the relational event history contains a column that is named
#' ``weight'', it is assumed that these affect the endogenous statistics. These
#' affect the computation of all endogenous statistics with a few exceptions
#' that follow logically from their definition (e.g., the recenyContinue
#' statistic does depend on time since the event and not on event weights).
#'
#' @section Subset the event history using 'start' and 'stop':
#' It is possible to compute statistics for a segment of the relational event 
#' sequence, based on the entire event history. This is done by specifying the 
#' 'start' and 'stop' values as the indices for the first and last event times 
#' for which statistics are needed. For instance, setting 'start = 5' and 'stop 
#' = 5' calculates statistics for the 5th event in the relational event 
#' sequence, considering events 1-4 in the history. Note that in cases of 
#' simultaneous events with the 'method' set to 'pt' (per timepoint), 'start' 
#' and 'stop' should correspond to the indices of the first and last 
#' \emph{unique} event timepoints for which statistics are needed. For example, 
#' if 'start = 5' and 'stop = 5', statistics are computed for the 5th unique 
#' timepoint in the relational event sequence, considering all events occurring 
#' at unique timepoints 1-4.
#'
#' @section Adjacency matrix:
#' Optionally, a previously computed adjacency matrix can be supplied. Note
#' that the endogenous statistics will be computed based on this adjacency
#' matrix. Hence, supplying a previously computed adjacency matrix can reduce
#' computation time but the user should be absolutely sure the adjacency matrix
#' is accurate.
#'
#' @param tie_effects an object of class \code{"\link[stats]{formula}"} (or one
#' that can be coerced to that class): a symbolic description of the effects in
#' the tie-oriented model for which statistics are computed, see 'Details' for
#' the available effects and their corresponding statistics
#' @param sender_effects an object of class \code{"\link[stats]{formula}"} (or
#' one that can be coerced to that class): a symbolic description of the
#' effects in the sender activity rate step of the actor-oriented model for
#' which statistics are computed, see `Details'
#' @param receiver_effects an object of class \code{"\link[stats]{formula}"}
#' (or one that can be coerced to that class): a symbolic description of the
#' effects in the receiver choice step of model for which statistics are
#' computed, see `Details'
#' @param reh an object of class \code{"\link[remify]{remify}"} characterizing 
#' the relational event history.
#' @param attr_actors optionally, an object of class
#' \code{"\link[base]{data.frame}"} that contains exogenous attributes for 
#' actors (see Details).
#' @param attr_dyads optionally, an object of class \code{data.frame} or 
#' \code{matrix} containing attribute information for dyads (see Details).
#' @param memory The memory to be used. See `Details'.
#' @param memory_value Numeric value indicating the memory parameter. Default 
#' is \code{NA}, which is only valid for \code{memory = "full"} (no memory
#' parameter required). See `Details'.
#' @param start an optional integer value, specifying the index of the first
#' time or event in the relational event history for which statistics must be 
#' computed (see 'Details'). Default is \code{2}: the first event has no history and is used only to initialize statistics, not to fit the model.
#' @param stop an optional integer value, specifying the index of the last
#' time or event in the relational event history for which statistics must be 
#' computed (see 'Details')
#' @param sampling Logical. If \code{TRUE}, statistics are computed using
#'   case–control (dyad) sampling rather than the full risk set. Default \code{FALSE}.
#'   Only supported for a \code{tie} model.
#' @param samp_num Integer. Number of dyads to include per event when
#'   \code{sampling = TRUE}. Must be smaller than or equal to the size of the
#'   active risk set. Ignored when \code{sampling = FALSE}.
#'   Only supported for a \code{tie} model.
#' @param display_progress should a progress bar for the computation of the
#' endogenous statistics be shown (TRUE) or not (FALSE)?
#' @param seed Optional integer. Random seed used for dyad sampling. Setting
#'   this ensures reproducible sampling across calls. If \code{NULL}, the
#'   current RNG state is used.
#'
#' @return An object of class 'remstats'. In case of the 
#' tie-oriented model, an array with the computed statistics, where rows refer 
#' to time points, columns refer to potential relational event (i.e., potential 
#' edges) in the risk set and slices refer to statistics. In case of the 
#' actor-oriented model, list with in the first element the statistics for the 
#' sender activity rate step and in the second element the statistics for the 
#' receiver choice step, where rows refer to time points, columns refer to 
#' potential senders or receivers, respectively. Statistics are computed once
#' per unique time point (per-timepoint "pt" method), so the number of rows in
#' the output equals \code{reh$M} (the number of  unique time points), which may
#' be less than the total number of observed events when simultaneous events
#' are present. The 'remstats' object has the following attributes: 
#'   \describe{
#'     \item{\code{model}}{Type of model that is estimated, obtained from the 
#'      remify object inputted to 'reh'.}
#'     \item{\code{formula}}{Model formula, obtained from the formula inputted 
#'      to 'tie_effects', 'sender_effects' and/or 'receiver_effects', depending 
#'      on the model.}
#'     \item{\code{riskset}}{For the tie-oriented model, the risk set used to 
#'      construct the statistics.}
#'     \item{\code{actors}}{For the actor-oriented model, the set of actors 
#'      used to construct the statistics, obtained from the remify object 
#'      inputted to 'reh'.}
#'   }
#'
#' @examples
#' library(remstats)
#'
#' # Tie-oriented model
#' eff <- ~ inertia():send("extraversion") + otp()
#' reh_tie <- remify::remify(edgelist = history, model = "tie")
#' remstats(reh = reh_tie, tie_effects = eff, attr_actors = info)
#'
#' # Actor-oriented model
#' seff <- ~ send("extraversion")
#' reff <- ~ receive("agreeableness") + inertia() + otp()
#' reh_actor <- remify::remify(edgelist = history, model = "actor")
#' remstats(
#'     reh = reh_actor, sender_effects = seff, receiver_effects = reff,
#'     attr_actors = info
#' )
#'
#' @references Butts, C. T. (2008). A relational event framework for social
#' action. Sociological Methodology, 38(1), 155–200.
#' \doi{10.1111/j.1467-9531.2008.00203.x},
#' Stadtfeld, C., & Block, P. (2017). Interactions, actors, and
#' time: Dynamic network actor models for relational events. Sociological
#' Science, 4, 318–352. \doi{10.15195/v4.a14}
#'
#' @export
remstats <- function(
    reh, 
    tie_effects = NULL, 
    sender_effects = NULL,
    receiver_effects = NULL, 
    attr_actors = NULL, 
    attr_dyads = NULL, 
    memory = c("full", "window", "decay", "interval"),
    memory_value = NA, 
    start = 2, 
    stop = Inf,
    display_progress = FALSE,
    sampling = FALSE,
    samp_num = NULL,
    seed = NULL
) {

    # Check if the deprecated "id" column is used in attr_actors
    if (!is.null(attr_actors)) {
        if (("id" %in% colnames(attr_actors)) & !("name" %in% colnames(attr_actors))) {
        warning("use 'name' instead of 'id' in 'attr_actors'")
        colnames(attr_actors)[which(colnames(attr_actors) == "id")] <- "name"
        }
    }

    # Check the reh
    if (!("remify" %in% class(reh))) {
        stop("Expected a 'reh' object of class remify")
    }

    if (reh$meta$model == "tie") {
    	out <- tomstats(
    		effects = tie_effects, 
    		reh = reh,
    		attr_actors = attr_actors, 
    		attr_dyads = attr_dyads, 
    		memory = memory, 
    		memory_value = memory_value, 
    		start = start, 
    		stop = stop, 
    		display_progress = display_progress,
    		sampling = sampling,
    		samp_num = samp_num,
    		seed = seed
    	)
    }

    if (reh$meta$model == "actor") {
        out <- aomstats(
            reh = reh, 
            sender_effects = sender_effects,
            receiver_effects = receiver_effects,
            attr_actors = attr_actors, 
            attr_dyads = attr_dyads, 
            memory = memory, 
            memory_value = memory_value, 
            start = start, 
            stop = stop, 
            display_progress = display_progress
        )
    }

    out
}

#' baseline
#'
#' Specifies an intercept for the tie-oriented model or the sender activity
#' rate step of the actor-oriented model in the same manner as in
#' \code{\link[stats]{lm}} (see Details).
#'
#' @details
#' A baseline effect is automatically specified for the tie-oriented model and
#' the sender activity rate step of the actor-oriented model when the
#' \code{ordinal} argument in \code{\link{remstats}}, \code{\link{tomstats}},
#' \code{\link{aomstats}} is set to FALSE (default) and automatically removed
#' when this argument is set to TRUE. Alternatively, a baseline effect can be
#' explicitly specified by adding '1' to the equation or explicitly removed by
#' adding '-1' to the equation.
#'
#' The baseline effect refers to the baseline tendency to interact. In the
#' tie-oriented model, the log-inverse of the estimated parameter translates to
#' the average number of observed events per time unit per dyad. In the
#' actor-oriented model, the log-inverse of the estimated parameter translates
#' to the average number of observed events per time unit per actor. The
#' statistic is equal to one for all dyads resp. actors in the riskset at all
#' timepoints.
#'
#' @name baseline
#' @aliases intercept
#'
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(reh = reh_tie, tie_effects = ~1)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = ~1)
NULL

#' send
#'
#' Specifies the statistic for a "send" effect in the tie-oriented model or the
#' actor activity rate step of the actor-oriented model. A "send" effect refers
#' to an exogenous actor attribute that affects actor \emph{i}'s rate of
#' sending events.
#'
#' @param variable string with the name of the column in the
#' \code{attr_actors} object for which the statistic has to be computed.
#' @param attr_actors optionally, an object of class
#' \code{\link[base]{data.frame}} that contains the attribute, see
#' 'Details.'
#' @param scaling the method for scaling the statistic. Default is to not scale
#' the statistic. Alternatively, standardization of the
#' statistic per time point can be requested with "std".
#' @param attr_data Deprecated argument. Please use 'attr_actors' instead.
#'
#' @details
#' The statistic at timepoint \emph{t} is equal to the value of the exogenous
#' attribute for actor \emph{i} at time \emph{t} for all dyads in the risk set
#' that have actor \emph{i} as sender. Note that a "send" effect is only
#' defined for directed relational events.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#' 
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#'
#' @examples
#' data(history)
#' data(info)
#' 
#' # Tie-oriented model
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ send("extraversion")
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' # Actor-oriented model
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = effects, attr_actors = info)
#'
#' @export
send <- function(variable, attr_actors = NULL, scaling = c("none", "std"), 
  attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  # Prep
  prep_exo("send", variable, attr_actors, scaling)
}

#' receive
#'
#' Specifies the statistic for a "receive" effect in the tie-oriented model or
#' the receiver choice step of the actor-oriented model. A "receive" effect
#' refers to an exogenous actor attribute that affects actor \emph{i}'s rate of
#' receiving events.
#'
#' @details
#' The statistic at timepoint \emph{t} is equal to the value of the exogenous
#' attribute for actor \emph{i} at time \emph{t} for all dyads in the riskset
#' that have actor \emph{i} as receiver. Note that a "receive" effect is only
#' defined for directed relational events.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#'
#' @inheritParams send
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' data(history)
#' data(info)
#' 
#' # Tie-oriented model
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ receive("extraversion")
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' # Actor-oriented model
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects, attr_actors = info)
#'
#' @export
receive <- function(variable, attr_actors = NULL, scaling = c("none", "std"), attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  # Prep
  prep_exo("receive", variable, attr_actors, scaling)
}

#' tie
#'
#' Specifies the statistic for a "tie" (or, "dyad") effect.
#'
#' @details
#' The "tie" effect or "dyad" effect refers to an exogenous dyad attribute that influences dyad \emph{(i,j)}'s interaction rate (in tie-oriented models) or the probability of actor \emph{j} being chosen as a receiver for the event sent by the active sender \emph{i} (in actor-oriented models). The statistic represents the value of the exogenous attribute for dyad \emph{(i,j)} in the \code{attr_dyads} data.
#'
#' @param variable A string specifying the attribute to compute the statistic. If \code{attr_dyads} is a \code{data.frame}, this refers to the column name in \code{attr_actors}. If \code{attr_dyads} is a \code{matrix}, this corresponds to the name of the exogenous attribute, used to label the statistic in the resulting \code{remstats} object.
#' @param attr_dyads A \code{data.frame} or \code{matrix} containing attribute information for dyads. If \code{attr_dyads} is a \code{data.frame}, the first two columns should represent "actor1" and "actor2" (for directed events, "actor1" corresponds to the sender, and "actor2" corresponds to the receiver). Additional columns can represent dyads' exogenous attributes. If attributes vary over time, include a column named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond to "actor1", columns to "actor2", and cells contain dyads' exogenous attributes.
#' @param scaling The method for scaling the statistic. The default is no scaling. Alternatively, standardization of the statistic per time point can be requested with "std".
#' @param x Deprecated argument. Please use 'attr_dyads' instead.
#' @param variableName Deprecated argument. Please use 'variable' instead.
#'
#' @aliases dyad
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' data(history)
#' data(both_male_long)
#' effect <- ~ tie(variable = "both_male", attr_dyads = both_male_long)
#' reh <- remify::remify(history, model = "tie")
#' remstats(reh = reh, tie_effects = effect)
#'
#' data(both_male_wide)
#' effect <- ~ tie(variable = "both_male", attr_dyads = both_male_wide)
#' reh <- remify::remify(history, model = "tie")
#' remstats(reh = reh, tie_effects = effect)
#'
#' @export
tie <- function(variable, attr_dyads = NULL, scaling = c("none", "std"), x, variableName) {
  # Deal with old function set-up
  if (!missing(x)) {
    warning("The 'x' argument in 'tie()' is deprecated. Please use 'attr_dyads' instead.")
    attr_dyads <- x
  }
  if (!missing(variableName)) {
    warning("The 'variableName' argument in 'tie()' is deprecated. Please use 'variable' instead.")
    variable <- variableName
  }
  if(!is.character(variable)) {
    stop("The 'variable' argument should be of type character.")
  }
  if(!is.null(attr_dyads) & !is.matrix(attr_dyads) & !is.data.frame(attr_dyads)) {
    stop("The 'attr_dyads' argument should be of type matrix or data.frame.")
  }

  # Match scaling
  if ("as.is" %in% scaling) {
    warning("The 'scaling' value 'as.is' is deprecated. Use 'none' instead.")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  # Prep
  prep_tie(variable, attr_dyads, scaling)
}

#' @export
dyad <- tie

#' same
#'
#' Specifies the statistic for a "same" effect in the tie-oriented model or the
#' receiver choice step of the actor-oriented model. A "same" effect refers to
#' an exogenous actor attribute that affects dyad \emph{(i,j)}'s rate of
#' interacting (tie-oriented model) or actor \emph{j}'s probability of being
#' chosen as a receiver for the event send by the active sender \emph{i} at
#' time \emph{t} (actor-oriented model) based on whether actors \emph{i} and
#' \emph{j} have the same value (or not) on this attribute.
#'
#' @details
#' The statistic at timepoint \emph{t} is equal to one for dyads \emph{(i,j)}
#' that have the same value on the attribute at timepoint \emph{t}
#' (tie-oriented model) or one for receivers \emph{j} that have the same value
#' on the attribute as the active sender \emph{i} at timepoint \emph{t}
#' (actor-oriented model) and equal to 0 for dyads and receivers that do not
#' have the same value.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#'
#' @inheritParams send
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ same("age")
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects, attr_actors = info)
#'
#' @export
same <- function(variable, attr_actors = NULL, attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }
  # Prep
  prep_exo("same", variable, attr_actors, "none")
}

#' difference
#'
#' Specifies the statistic for a "difference" effect in the tie-oriented model
#' or the receiver choice step of the actor-oriented model. A difference effect
#' refers to an exogenous actor attribute that affects dyad \emph{(i,j)}'s rate
#' of interacting (tie-oriented model) or actor \emph{j}'s probability of being
#' chosen as a receiver for the event send by the active sender \emph{i} at
#' time \emph{t} (actor-oriented model) based on the difference between the
#' values of actors \emph{i} and \emph{j} on this attribute.
#'
#' @details
#' The statistic at timepoint \emph{t} is equal to the (absolute) difference
#' between the values of actor \emph{i} and \emph{j} on the attribute at
#' timepoint \emph{t}.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#'
#' @inheritParams send
#' @param absolute Logical value indicating whether the difference values
#' should be converted to the absolute difference (default is TRUE).
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' # Example for tie-oriented model
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ difference("extraversion", absolute = TRUE)
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' # Example for actor-oriented model
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects, attr_actors = info)
#'
#' @export
difference <- function(variable,
                       attr_actors = NULL,
                       scaling = c("none", "std"),
                       absolute = TRUE,
                       attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }

  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  scaling <- ifelse(absolute, paste0(scaling, "_abs"), scaling)
  # Prep
  prep_exo("difference", variable, attr_actors, scaling)
}

#' average
#'
#' Specifies the statistic for an "average" effect in the tie-oriented model or
#' the receiver choice step of the actor-oriented model. An "average" effect
#' refers to an exogenous actor attribute that affects dyad \emph{(i,j)}'s rate
#' of interacting (tie-oriented model) or actor \emph{j}'s probability of being
#' chosen as a receiver for the event send by the active sender \emph{i} at
#' time \emph{t} (actor-oriented model) based on the average of the values of
#' actors \emph{i} and \emph{j} on this attribute.
#'
#' @details
#' The statistic at timepoint \emph{t} for dyad \emph{(i,j)} is equal to the
#' average of the values of actor \emph{i} and \emph{j} on the attribute at
#' timepoint \emph{t}.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#'
#' @inheritParams send
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ average("extraversion")
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects, attr_actors = info)
#'
#' @export
average <- function(variable, attr_actors = NULL, scaling = c("none", "std"), attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }

  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  # Prep
  prep_exo(effect = "average", variable = variable, attr_actors = attr_actors, scaling = scaling)
}

#' minimum
#'
#' Specifies the statistic for a "minimum" effect in the tie-oriented model. A
#' "minimum" effect refers to an exogenous actor attribute that affects dyad
#' \emph{(i,j)}'s rate of interacting based on the minimum of the values of
#' actors \emph{i} and \emph{j} on this attribute.
#'
#' @details
#' The statistic at timepoint \emph{t} for dyad \emph{(i,j)} is equal to the
#' minimum of the values of actor \emph{i} and \emph{j} on the attribute at
#' timepoint \emph{t}.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @inheritParams send
#'
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ minimum("extraversion")
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' @export
minimum <- function(variable, attr_actors = NULL, scaling = c("none", "std"), attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }

  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  # Prep
  prep_exo("minimum", variable, attr_actors, scaling)
}

#' maximum
#'
#' Specifies the statistic for a "maximum" effect in the tie-oriented model. A
#' "maximum" effect refers to an exogenous actor attribute that affects dyad
#' \emph{(i,j)}'s rate of interacting based on the maximum of the values of
#' actors \emph{i} and \emph{j} on this attribute.
#'
#' @details
#' The statistic at timepoint \emph{t} for dyad \emph{(i,j)} is equal to the
#' maximum of the values of actor \emph{i} and \emph{j} on the attribute at
#' timepoint \emph{t}.
#'
#' Construct the `attr_actors` object as a data frame where each row represents 
#' the attribute value of actor \emph{i} at timepoint \emph{t}:
#' \itemize{
#'   \item{name:} The actors' name.
#'   \item{time:} The time when the attribute values change.
#'   \item{variable:} The third column contains the attribute used in the
#'   specification of the "difference" effect. The column name should correspond
#'   to the string supplied to the \code{variable} argument in the `difference()`
#'   function.
#' }
#' 
#' Note that it is possible to omit the `attr_actors` object in the call of
#' \code{difference()} and, instead, supply it in the call of \code{remstats()}
#' for multiple exogenous effects.
#'
#' @inheritParams send
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic. 
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ maximum("extraversion")
#' remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)
#'
#' @export
maximum <- function(variable, attr_actors = NULL, scaling = c("none", "std"), attr_data) {
  # Deal with old function set-up
  if (!missing(attr_data)) {
    warning("The 'attr_data' argument in 'send()' is deprecated. Please use 'attr_actors' instead.")
    attr_actors <- attr_data
  }

  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }
  # Prep
  prep_exo("maximum", variable, attr_actors, scaling)
}

#' event
#'
#' Specifies the statistic for an "event" effect in the tie-oriented model. An
#' "event" effect refers to an exogenous event attribute that affects the
#' waiting time between events.
#'
#' @details
#' The statistic at timepoint \emph{t} is for all dyads in the risk set equal
#' to the attribute of the event at timepoint \emph{t}.
#'
#' @param x vector with the event attribute
#' @param variableName optionally, a string indicating the variable name, used
#' for the dimnames of the output statistics object
#'
#' @seealso \code{\link{FEtype}}
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' \donttest{
#'   reh_tie <- remify::remify(history, model = "tie")
#'   data(history, package = "remstats")
#'   history$work <- ifelse(history$setting == "work", 1, 0)
#'   effects <- ~ event(x = history$work, variableName = "setting_is_work")
#'   remstats(reh = reh_tie, tie_effects = effects)
#' }
#'
#' @export
event <- function(x, variableName = NULL) {
  # Missing values
  if (anyNA(x)) {
    stop("Vector 'x' in event() contains missing values.")
  }

  # Output
  list(
    effect = "event",
    x = x,
    variable = variableName
  )
}

#' FEtype
#'
#' Specifies the statistic for fixed effects for event types in the
#' tie-oriented model.
#'
#' @details
#' Fixed effects for event types capture the variation in event rate across
#' different event types (e.g., see Butts, 2008). The specification of FEtype
#' results in the creation of C-1 statistics, were C is the number of different
#' event types in the riskset. Let one of the event types, e.g. \emph{c = 1},
#' represent the reference category. Than, for every event type
#' \emph{c = 2, ..., C}, a statistic is created that at timepoint \emph{t} for
#' dyad \emph{(i,j,c)} is equal to 1 if \emph{c} is equal to the respective
#' event type and equal to 0 otherwise (i.e., dummy variables are created).
#' Note that specifying fixed effects for event types is only available when
#' event types are modeled in the dependent variable.
#'
#' @seealso \code{\link{event}}
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' history$type <- history$setting
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ FEtype()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
FEtype <- function() {

  # Add effect
  call_args <- list()
  call_args$effect <- "FEtype"

  return(call_args)
}

#' inertia
#'
#' Specifies the statistic for an inertia effect in the tie-oriented model or
#' the receiver choice step of the actor-oriented model.
#'
#' @details
#' An inertia effect refers to the tendency for dyads to repeatedly interact
#' with each other (tie-oriented model) or for actors to repeatedly choose the
#' same actor as receiver of their events (actor-oriented model). The statistic
#' at timepoint \emph{t} for dyad \emph{(i,j)} resp. receiver \emph{j} is
#' equal to the number of \emph{(i,j)} events before timepoint \emph{t}.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' inertia count by the outdegree of the sender ("prop"), the statistic refers
#' to the fraction of messages send by actor i that were send to actor j. If
#' actor i hasn't send any messages yet it can be assumed that every actor is
#' equally likely to receive a message from i and the statistic is set equal to
#' 1/(n-1), where n refers to the number of actors. The resulting statistic is
#' similar to the "FrPSndSnd" statistic in the R package 'relevent', or the
#' persistence statistic in Section 2.2.2 of Butts (2008). Note that this
#' scaling method is only defined for directed events.
#'
#' @param scaling the method for scaling the inertia statistic. Default is to
#' not scale the statistic (scaling = "none"). Alternatively, the
#' statistics can be scaled by specifying 'prop', in which raw counts are
#' divided by the outdegree of the sender at time t (see 'details') or
#' standardization of the raw counts per time point can be requested with 'std'.
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ inertia()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
inertia <- function(scaling = c("none", "prop", "std"),
                    consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "inertia"

  return(call_args)
}

#' reciprocity
#'
#' Specifies the statistic for a reciprocity effect in the tie-oriented model
#' or the receiver choice step of the actor-oriented model.
#'
#' @details
#' A reciprocity effect refers to the tendency for actors to reciprocate past
#' interactions. The statistic at timepoint \emph{t} for dyad \emph{(i,j)}
#' (tie-oriented model) or receiver \emph{j} (actor-oriented model) is equal to
#' the number of \emph{(j,i)} events before timepoint \emph{t}. Note that a
#' reciprocity effect is only defined for directed events.
#'
#' Optionally, a scaling method can be set with \code{scaling}.  By scaling the
#' reciprocity count by the indegree of the sender, the statistic refers to the
#' fraction of messages received by actor i that were received from actor j. If
#' actor i hasn't received any messages yet it can be assumed that actor i is
#' equally likely to receive a message from every actor and the statistic is
#' set equal to 1/(n-1), where n refers to the number of actors. The resulting
#' statistic is similar to the "FrRecSnd" statistic in the R package 'relevent'.
#'
#' @param scaling the method for scaling the reciprocity statistic. Default is
#' to not scale the statistic but keep the raw 'counts'. Alternatively, the
#' statistics can be scaled by 'prop', in which raw counts are
#' divided by the indegree of the sender at time t (see 'details') or
#' standardization of the raw counts per time point can be requested with 'std'.
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ reciprocity()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
reciprocity <- function(scaling = c("none", "prop", "std"),
                        consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "reciprocity"

  return(call_args)
}

#' indegreeSender
#'
#' Specifies the statistic for an `indegreeSender` effect in the tie-oriented
#' model or the sender activity rate step of the actor-oriented model.
#'
#' @details
#' An indegree of the sender effect refers to the tendency for actors to send
#' events if they have received more past events. The statistic at timepoint
#' \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or sender \emph{i}
#' (actor-oriented model) is equal to the number of events received by
#' actor \emph{i} before timepoint \emph{t}. Note that the 'indegreeSender'
#' effect is only defined for directed events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events, the statistic refers to the
#' fraction of past events that were received by actor i. At the first time
#' point, when no events did previously occur, it is assumed that every actor
#' is equally likely to send a message and the statistic is set equal to 1/n,
#' where n refers to the number of actors.
#'
#' @param scaling the method for scaling the degree statistic. Default is
#' to not scale the statistic (scaling = "none"). Alternatively,
#' scaling of the raw degree counts by the number of past events at time t can
#' be requested with 'prop' or standardization of the raw degree counts per
#' time point can be requested with 'std'.
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @aliases degree indegree
#' @seealso \code{\link{indegreeReceiver}}, \code{\link{outdegreeSender}},
#' \code{\link{outdegreeReceiver}}, \code{\link{totaldegreeSender}}, or
#' \code{\link{totaldegreeReceiver}} for other types of degree effects.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ indegreeSender()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = effects)
#'
#' @export
indegreeSender <- function(scaling = c("none", "prop", "std"),
                           consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "indegreeSender"

  return(call_args)
}

#' indegreeReceiver
#'
#' Specifies the statistic for an `indegreeReceiver` effect in the tie-oriented
#' model or the receiver choice step of the actor-oriented model.
#'
#' @details
#' An indegree of the receiver effect refers to the tendency for actors to
#' receive events if they have received more past events. The statistic at
#' timepoint \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or
#' receiver \emph{j} (actor-oriented model) is equal to the number of events
#' received by actor \emph{j} before timepoint \emph{t}. Note that the
#' 'indegreeReceiver' effect is only defined for directed events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events, the statistic refers to the
#' fraction of past events that were received by actor j. At the first time
#' point, when no events did previously occur, it is assumed that every actor
#' is equally likely to receive a message and the statistic is set equal to 1/
#' n, where n refers to the number of actors.
#'
#' @inheritParams indegreeSender
#'
#' @seealso \code{\link{indegreeSender}}, \code{\link{outdegreeSender}},
#' \code{\link{outdegreeReceiver}}, \code{\link{totaldegreeSender}}, or
#' \code{\link{totaldegreeReceiver}} for other types of degree effects.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ indegreeReceiver()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
indegreeReceiver <- function(scaling = c("none", "prop", "std"),
                             consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "indegreeReceiver"

  return(call_args)
}

#' outdegreeSender
#'
#' Specifies the statistic for an `outdegreeSender` effect in the tie-oriented
#' model or the sender activity rate step of the actor-oriented model.
#'
#' @details
#' An outdegree of the sender effect refers to the tendency for actors to send
#' events if they have send more past events. The statistic at timepoint
#' \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or sender \emph{i}
#' (actor-oriented model) is equal to the number of events send by
#' actor \emph{i} before timepoint \emph{t}. Note that the 'outdegreeSender'
#' effect is only defined for directed events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events, the statistic refers to the
#' fraction of past events that were send by actor i. At the first time
#' point, when no events did previously occur, it is assumed that every actor
#' is equally likely to send a message and the statistic is set equal to 1/n,
#' where n refers to the number of actors.
#'
#' @inheritParams indegreeSender
#'
#' @aliases outdegree
#' @seealso \code{\link{indegreeSender}}, \code{\link{indegreeReceiver}},
#' \code{\link{outdegreeReceiver}}, \code{\link{totaldegreeSender}}, or
#' \code{\link{totaldegreeReceiver}} for other types of degree effects.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ outdegreeSender()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = effects)
#'
#' @export
outdegreeSender <- function(scaling = c("none", "prop", "std"),
                            consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "outdegreeSender"

  return(call_args)
}

#' outdegreeReceiver
#'
#' Specifies the statistic for an `outdegreeReceiver` effect in the
#' tie-oriented model or the receiver choice step of the actor-oriented model.
#'
#' @details
#' An outdegree of the receiver effect refers to the tendency for actors to
#' receive events if they have send more past events. The statistic at
#' timepoint \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or
#' receiver \emph{j} (actor-oriented model) is equal to the number of events
#' send by actor \emph{j} before timepoint \emph{t}. Note that the
#' 'outdegreeReceiver' effect is only defined for directed events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events, the statistic refers to the
#' fraction of past events that were send by actor j. At the first time
#' point, when no events did previously occur, it is assumed that every actor
#' is equally likely to receive a message and the statistic is set equal to
#' 1/n, where n refers to the number of actors.
#'
#' @inheritParams indegreeSender
#'
#' @seealso \code{\link{indegreeSender}}, \code{\link{indegreeReceiver}},
#' \code{\link{outdegreeSender}}, \code{\link{totaldegreeSender}}, or
#' \code{\link{totaldegreeReceiver}} for other types of degree effects.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ outdegreeReceiver()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
outdegreeReceiver <- function(scaling = c("none", "prop", "std"),
                              consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "outdegreeReceiver"

  return(call_args)
}

#' totaldegreeSender
#'
#' Specifies the statistic for an `totaldegreeSender` effect in the
#' tie-oriented model or the sender activity rate step of the actor-oriented
#' model.
#'
#' @details
#' A total degree of the sender effect refers to the tendency for actors to
#' send events if they have send and received more past events. The statistic
#' at timepoint \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or sender
#' \emph{i} (actor-oriented model) is equal to the number of events
#' send and received by actor \emph{i} before timepoint \emph{t}. Note that the
#' 'totaldegreeSender' effect is only defined for directed events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events times two, the statistic
#' refers to the fraction of past events times two that involved actor i. At
#' the first time point, when no events did previously occur, it is assumed
#' that every actor is equally likely to send a message and the statistic is
#' set equal to 1/n, where n refers to the number of actors.
#'
#' @param scaling the method for scaling the degree statistic. Default is
#' to not scale the statistic (scaling = "none"). Alternatively,
#' scaling of the raw degree counts by two times the number of past events at
#' time t can be requested with 'prop' or standardization of the raw degree
#' counts per time point can be requested with 'std'.
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @aliases totaldegree
#' @seealso \code{\link{indegreeSender}}, \code{\link{indegreeReceiver}},
#' \code{\link{outdegreeSender}}, \code{\link{outdegreeReceiver}}, or
#' \code{\link{totaldegreeReceiver}} for other types of degree effects.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' effects <- ~ totaldegreeSender()
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = effects)
#'
#' @export
totaldegreeSender <- function(scaling = c("none", "prop", "std"),
                              consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "totaldegreeSender"

  return(call_args)
}

#' totaldegreeReceiver
#'
#' Specifies the statistic for an `totaldegreeReceiver` effect in the
#' tie-oriented model or the receiver choice step of the actor-oriented model.
#'
#' @details
#' A total degree of the receiver effect refers to the tendency for actors to
#' receive events if they have send and received more past events. The
#' statistic at timepoint \emph{t} for dyad \emph{(i,j)} (tie-oriented model)
#' or receiver \emph{j} (actor-oriented model) is equal to the number
#' of events send and received by actor \emph{j} before timepoint \emph{t}.
#' Note that the 'totaldegreeReceiver' effect is only defined for directed
#' events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events times two, the statistic
#' refers to the fraction of past events times two that involved actor j. At
#' the first time point, when no events did previously occur, it is assumed
#' that every actor is equally likely to receive a message and the statistic is
#' set equal to 1/n, where n refers to the number of actors.
#'
#' @inheritParams totaldegreeSender
#'
#' @seealso \code{\link{indegreeSender}}, \code{\link{indegreeReceiver}},
#' \code{\link{outdegreeSender}}, \code{\link{outdegreeReceiver}}, or
#' \code{\link{totaldegreeSender}} for other types of degree effects.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ totaldegreeReceiver()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
totaldegreeReceiver <- function(scaling = c("none", "prop", "std"),
                                consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "totaldegreeReceiver"

  return(call_args)
}

#' totaldegreeDyad
#'
#' Specifies the statistic for a 'totaldegreeDyad' effect.
#'
#' @inheritParams totaldegreeSender
#'
#' @details
#' The 'totaldegreeDyad' effect refers to the tendency of pairs of actors
#' (dyads) to increase their interaction rate as the total degree (number of
#' interactions) of both actors in the pair goes up. To calculate this effect
#' for a specific pair (i,j) at a given timepoint (t), we sum the degrees of
#' the two actors in the dyad (i,j).
#'
#' Additionally, there is an optional scaling method, which can be chosen using
#' the 'scaling' method. When the 'prop' scaling method is applied, the degree
#' count is divided by two times the total number of past events. This scaling
#' converts the statistic into a fraction, representing the proportion of past
#' events in which at least one actor in the dyad was involved. For the first
#' timepoint, where no events have previously occurred, it is assumed that each
#' actor is equally likely to be involved in an event. In this case, the
#' statistic is set to 1 divided by the total number of actors (N).
#'
#' The totaldegreeDyad effect is defined for the tie-oriented model and is
#' applicable to both directed and undirected events.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ totaldegreeDyad()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
totaldegreeDyad <- function(scaling = c("none", "prop", "std"),
                            consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "totaldegreeDyad"

  return(call_args)
}

#' degreeDiff
#'
#' Specifies the statistic for a `degreeDiff` effect in the tie-oriented
#' model.
#'
#' @param scaling the method for scaling the degree statistic. Default is to
#' not scale the statistic (scaling = "none"). Alternatively, standardization
#' of the degree difference per time point can be requested with `std`.
#' @inheritParams totaldegreeSender
#'
#' @details
#' A degreeDiff effect refers to the tendency for dyads to increase their
#' interaction rate if the absolute difference in degree for the two actors in
#' the pair increases. The statistic at timepoint \emph{t} for dyad
#' \emph{(i,j)} is equal to the difference between the following two values:
#' the number of events before timepoint \emph{t} that involved actor \emph{i}
#' and actor \emph{j}, respectively. The degreeDiff effect is only defined for
#' undirected events.
#'
#' @seealso \code{\link{degreeMin}}, \code{\link{degreeMax}} or
#' \code{\link{totaldegreeDyad}} for other types of degree effects for
#' undirected events.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ degreeDiff()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
degreeDiff <- function(scaling = c("none", "std"), consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "degreeDiff"

  return(call_args)
}

#' degreeMin
#'
#' Specifies the statistic for an `degreeMin` effect in the tie-oriented
#' model with undirected events.
#'
#' @details
#' An degreeMin effect refers to the tendency for dyads to increase their
#' interaction rate if the total degree of the least active actor in the pair
#' increases. The statistic at timepoint \emph{t} for dyad \emph{(i,j)} is
#' equal to the minimum of the following two values: the number of events
#' before timepoint \emph{t} that involved actor \emph{i} and actor \emph{j},
#' respectively. Note that the degreeMin effect is only defined for undirected
#' events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events, the statistic refers to the
#' fraction of past events that the least active actor was involved in. At the
#' first time point, when no events did previously occur, it is assumed that
#' every actor is equally likely to be involved in an event and the statistic
#' is set equal to 1/n, where n refers to the number of actors.
#'
#' @inheritParams totaldegreeSender
#'
#' @seealso \code{\link{degreeDiff}}, \code{\link{degreeMax}} or
#' \code{\link{totaldegreeDyad}} for other types of degree effects for
#' undirected events.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ degreeMin()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
degreeMin <- function(scaling = c("none", "prop", "std"),
                      consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "degreeMin"

  return(call_args)
}

#' degreeMax
#'
#' Specifies the statistic for an `degreeMax` effect in the tie-oriented
#' model with undirected events.
#'
#' @details
#' An degreeMax effect refers to the tendency for dyads to increase their
#' interaction rate if the total degree of the most active actor in the pair
#' increases. The statistic at timepoint \emph{t} for dyad \emph{(i,j)} is
#' equal to the maximum of the following two values: the number of events
#' before timepoint \emph{t} that involved actor \emph{i} and actor \emph{j},
#' respectively. Note that the degreeMax effect is only defined for undirected
#' events.
#'
#' Optionally, a scaling method can be set with \code{scaling}. By scaling the
#' degree count by the total number of past events, the statistic refers to the
#' fraction of past events that the most active actor was involved in. At the
#' first time point, when no events did previously occur, it is assumed that
#' every actor is equally likely to be involved in an event and the statistic
#' is set equal to 1/n, where n refers to the number of actors.
#'
#' @inheritParams totaldegreeSender
#'
#' @seealso \code{\link{degreeDiff}}, \code{\link{degreeMin}} or
#' \code{\link{totaldegreeDyad}} for other types of degree effects for
#' undirected events.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ degreeMax()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
degreeMax <- function(scaling = c("none", "prop", "std"),
                      consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

 # Collect info
  call_args <- list()
  call_args$scaling <- scaling
  call_args$consider_type <- consider_type
  call_args$effect <- "degreeMax"

  return(call_args)
}

#' otp
#'
#' Specifies the statistic for an outgoing two-path effect.
#'
#' @details
#' The outgoing two-path effect describes the propensity of dyads to interact
#' based on the number of past outgoing two-paths between them. By default, the
#' statistic at timepoint t for the dyad (i,j) is computed as the sum of the
#' minimum occurrences of past (i,h) and (h,j) events across all actors h.
#'
#' When the unique parameter is set to TRUE, a different approach is taken.
#' In this case, the statistic counts the number of actors h that contribute to
#' the creation of a new, distinct two-path between actors i and j.
#'
#' Additionally, it is possible to specify a scaling method using the scaling
#' parameter.
#'
#' Please note that the outgoing two-path effect, 'otp', is exclusively defined
#' for directed events.
#'
#' @param unique A logical value indicating whether to sum the minimum of
#' events with third actors (FALSE, default) or the number of third actors that
#' create a new, unique two-path (TRUE). See details for more information.
#' @param scaling The method for scaling the triad statistic. The default value
#' is "none", which means the statistic is not scaled. Alternatively, you can
#' set it to "std" to request standardization of the raw counts per time point.
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @aliases triad
#' @seealso \code{\link{itp}}, \code{\link{osp}}, or \code{\link{isp}} for
#' other types of triadic effects for directed relational events and
#' \code{\link{sp}} for triadic effects for undirected relational events.
#'
#' @references Butts, C. (2008). A relational event framework for social
#' action. Sociological Methodology.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ otp()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
otp <- function(unique = FALSE, scaling = c("none", "std"),
                consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- ifelse(unique, paste0(scaling, "_unique"), scaling)
  call_args$consider_type <- consider_type
  call_args$effect <- "otp"

  return(call_args)
}

#' itp
#'
#' Specifies the statistic for an incoming two-path effect.
#'
#' @details
#' The incoming two-path effect describes the propensity of dyads to interact
#' based on the number of past incoming two-paths between them. By default, the
#' statistic at timepoint t for the dyad (i,j) is computed as the sum of the
#' minimum occurrences of past (j,h) and (h,i) events across all actors h.
#'
#' When the unique parameter is set to TRUE, a different approach is taken.
#' In this case, the statistic counts the number of actors h that contribute to
#' the creation of a new, distinct two-path between actors i and j.
#'
#' Additionally, it is possible to specify a scaling method using the scaling
#' parameter.
#'
#' Please note that the incoming two-path effect, 'itp', is exclusively defined
#' for directed events.
#'
#' @inheritParams otp
#'
#' @seealso \code{\link{otp}}, \code{\link{osp}}, or \code{\link{isp}} for
#' other types of triadic effects for directed relational events and
#' \code{\link{sp}} for triadic effects for undirected relational events.
#'
#' @references Butts, C. (2008). A relational event framework for social
#' action. Sociological Methodology.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ itp()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
itp <- function(unique = FALSE, scaling = c("none", "std"),
                consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- ifelse(unique, paste0(scaling, "_unique"), scaling)
  call_args$consider_type <- consider_type
  call_args$effect <- "itp"

  return(call_args)
}

#' osp
#'
#' Specifies the statistic for an outgoing shared partners effect.
#'
#' @details
#' The outgoing shared partners effect describes the propensity of dyads to
#' interact based on the number of past outgoing shared partners between them.
#' By default, the statistic at timepoint t for the dyad (i,j) is computed as
#' the sum of the minimum occurrences of past (i,h) and (j,h) events across all
#' actors h.
#'
#' When the unique parameter is set to TRUE, a different approach is taken.
#' In this case, the statistic counts the number of actors h that contribute to
#' the creation of a new, distinct shared partner between actors i and j.
#'
#' Additionally, it is possible to specify a scaling method using the scaling
#' parameter.
#'
#' Please note that the outgoing shared partners effect, 'osp', is exclusively
#' defined for directed events.
#'
#' @param unique A logical value indicating whether to sum the minimum of
#' events with third actors (FALSE, default) or the number of third actors that
#' create a new, unique shared partner (TRUE). See details for more information.
#' @param scaling the method for scaling the triad statistic. Default is to not
#' scale the statistic but keep the raw 'counts'. Alternatively,
#' standardization of the raw counts per time point can be requested
#' with 'std'.
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @seealso \code{\link{otp}}, \code{\link{itp}}, or \code{\link{isp}} for
#' other types of triadic effects for directed relational events and
#' \code{\link{sp}} for triadic effects for undirected relational events.
#'
#' @references Butts, C. (2008). A relational event framework for social
#' action. Sociological Methodology.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ osp()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
osp <- function(unique = FALSE, scaling = c("none", "std"), consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- ifelse(unique, paste0(scaling, "_unique"), scaling)
  call_args$consider_type <- consider_type
  call_args$effect <- "osp"

  return(call_args)
}

#' isp
#'
#' Specifies the statistic for an incoming shared partners effect.
#'
#' @details
#' The incoming shared partners effect describes the propensity of dyads to
#' interact based on the number of past incoming shared partners between them.
#' By default, the statistic at timepoint t for the dyad (i,j) is computed as
#' the sum of the minimum occurrences of past (h,i) and (h,j) events across all
#' actors h.
#'
#' When the unique parameter is set to TRUE, a different approach is taken.
#' In this case, the statistic counts the number of actors h that contribute to
#' the creation of a new, distinct shared partner between actors i and j.
#'
#' Additionally, it is possible to specify a scaling method using the scaling
#' parameter.
#'
#' Please note that the incoming shared partners effect, 'isp', is exclusively
#' defined for directed events.
#'
#' @inheritParams osp
#'
#' @seealso \code{\link{otp}}, \code{\link{itp}}, or \code{\link{osp}} for
#' other types of triadic effects for directed relational events and
#' \code{\link{sp}} for triadic effects for undirected relational events.
#'
#' @references Butts, C. (2008). A relational event framework for social
#' action. Sociological Methodology.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ isp()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, receiver_effects = effects)
#'
#' @export
isp <- function(
    unique = FALSE, scaling = c("none", "std"),
    consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- ifelse(unique, paste0(scaling, "_unique"), scaling)
  call_args$consider_type <- consider_type
  call_args$effect <- "isp"

  return(call_args)
}

#' sp
#'
#' Specifies the statistic for a shared partners effect for undirected events.
#'
#' @details
#' The shared partners effect describes the propensity of dyads to interact
#' based on the number of past shared partners between them. By default, the
#' statistic at timepoint t for the undirected dyad (i,j) is computed as
#' the sum of the minimum occurrences of past undirected (i,h) and undirected
#' (j,h) events across all actors h.
#'
#' When the unique parameter is set to TRUE, a different approach is taken.
#' In this case, the statistic counts the number of actors h that contribute to
#' the creation of a new, distinct shared partner between actors i and j.
#'
#' Additionally, it is possible to specify a scaling method using the scaling
#' parameter.
#'
#' Please note that the shared partners effect, 'sp', is exclusively defined
#' for undirected events.
#'
#' @inheritParams osp
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @seealso \code{\link{otp}}, \code{\link{itp}}, \code{\link{osp}}, or
#' \code{\link{isp}} for triadic effects for directed relational events.
#'
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ sp()
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' @export
sp <- function(
    unique = FALSE, scaling = c("none", "std"),
    consider_type = "ignore") {
  # Match scaling
  if ("as.is" %in% scaling) {
    warning("use 'scaling' is 'none' instead of 'as.is'")
    scaling <- "none"
  } else {
    scaling <- match.arg(scaling)
  }

  # Collect info
  call_args <- list()
  call_args$scaling <- ifelse(unique, paste0(scaling, "_unique"), scaling)
  call_args$consider_type <- consider_type
  call_args$effect <- "sp"

  return(call_args)
}

#' spUnique
#'
#' Deprecated. Use \code{\link{sp}}.
#'
#' @returns Warning.
#' 
#' @export
spUnique <- function() {
  .Deprecated("sp")
}

#' psABBA
#'
#' Specifies the statistic for a participation shift AB-BA.
#'
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @details
#' The AB-BA pshift effect refers to one of Gibson's (2003) dyadic participation shifts. The AB-BA pshift refers to the tendency for immediate reciprocation (the next sender is the previous receiver and the next receiver is the previous sender). For each timepoint t, the psABBA statistic is equal to one for the dyad that will create the participation shift if it would occur in the edgelist at time t and equal to zero for the dyads that will not create this participation shift. If consider_type is set to TRUE, the type of the AB event and the type of the BA event have to be equal. If it is set to FALSE, the participation shift is set to one for every BA event, regardless of the event type. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed. Note that the AB-BA pshift is only defined for directed events.
#'
#' @aliases pshift
#' @seealso \code{\link{psABBY}}, \code{\link{psABXA}}, \code{\link{psABXB}},
#' \code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation
#' shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ psABBA()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
psABBA <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABBA"
  return(call_args)
}

#' psABB
#'
#' Specifies the statistic for a participation shift AB-B in the sender step of the actor-oriented model.
#'
#' @details
#' The AB-B participation shift refers to the tendency for immediate reciprocation (the next sender is the previous receiver). For each timepoint t, the psABBA statistic is equal to one for the actor (i.e, the previous event receiver) that will create the participation shift if it would occur as sender in the edgelist at time t and equal to zero for the actors that will not create this participation shift. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed.
#'
#' @seealso \code{\link{psABA}} or \code{\link{psABX}} for exploring alternative participation shifts in the sender step of the actor-oriented model.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = ~ psABB())
#'
#' @export
psABB <- function() {
  call_args <- list()
  call_args$effect <- "psABB"
  return(call_args)
}

#' psABBY
#'
#' Specifies the statistic for a participation shift AB-BY.
#'
#' @inheritParams psABBA
#'
#' @details
#' The AB-BY participation shift refers to one of Gibson's (2003) dyadic participation shifts. The AB-BY pshift refers to a tendency for turn receiving (here, the next sender is the previous receiver and the next receiver is not in the current previous). For each timepoint t, the psABBY statistic is equal to one for the dyads that will create the participation shift if they would occur in the edgelist at time t and equal to zero for the dyads that will not create this participation shift. If consider_type is set to TRUE, the type of the AB event and the type of the BY events have to be equal. If it is set to FALSE, the participation shift is set to one for every BY event, regardless of the event type. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed. Note that the AB-BY pshift is only defined for directed events.
#'
#' @seealso \code{\link{psABBA}}, \code{\link{psABXA}}, \code{\link{psABXB}},
#' \code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation
#' shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ psABBY()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
psABBY <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABBY"
  return(call_args)
}

#' psABXA
#'
#' Specifies the statistic for a participation shift AB-XA.
#'
#' @inheritParams psABBA
#'
#' @details
#' The AB-XA participation shift refers to one of Gibson's (2003) dyadic participation shifts. The AB-XA pshift refers to a tendency for turn usurping (here, the next sender is not in the previous event and the next receiver is the previous sender). For each timepoint t, the psABXA statistic is equal to one for the dyads that will create the participation shift if they would occur in the edgelist at time t and equal to zero for the dyads that will not create this participation shift. If consider_type is set to TRUE, the type of the AB event and the type of the XA events have to be equal. If it is set to FALSE, the participation shift is set to one for every XA event, regardless of the event type. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the pshift is observed. Note that the AB-XA pshift is only defined for directed events.
#'
#' @seealso \code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXB}}, \code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ psABXA()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
psABXA <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABXA"
  return(call_args)
}

#' psABX
#'
#' Specifies the statistic for a participation shift AB-X in the sender step of
#' the actor-oriented model.
#'
#' @details
#' The AB-X participation shift refers to a tendency for turn usurping (here,
#' the next sender is not in the previous event). For each timepoint t, the
#' psABX statistic is equal to one for the actors that will create the
#' participation shift if they would occur as the sender in the edgelist at
#' time t and equal to zero for the actors that will not create this
#' participation shift. If multiple events in the edgelist occur at the same
#' time point, the order of these events determines whether the p-shift is
#' observed.
#'
#' @seealso \code{\link{psABA}} or \code{\link{psABB}} for exploring
#' alternative participation shifts in the sender step of the actor-oriented
#' model.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(reh = reh_actor, sender_effects = ~ psABX())
#'
#' @export
psABX <- function() {
  call_args <- list()
  call_args$effect <- "psABX"
  return(call_args)
}

#' psABXB
#'
#' Specifies the statistic for a participation shift AB-XB.
#'
#' @inheritParams psABBA
#'
#' @details
#' The AB-XB participation shift refers to one of Gibson's (2003) dyadic participation shifts. The AB-XB pshift refers to a tendency for turn usurping (here, the next sender is not in the previous event and the next receiver is the previous receiver). For each timepoint t, the psABXB statistic is equal to one for the dyads that will create the participation shift if they would occur in the edgelist at time t and equal to zero for the dyads that will not create this participation shift. If consider_type is set to TRUE, the type of the AB event and the type of the XB events have to be equal. If it is set to FALSE, the participation shift is set to one for every XB event, regardless of the event type. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed. Note that the AB-XB pshift is only defined for directed events.
#'
#' @seealso \code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXA}},
#' \code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation
#' shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ psABXB()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
psABXB <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABXB"
  return(call_args)
}

#' psABXY
#'
#' Specifies the statistic for a participation shift AB-XY.
#'
#' @inheritParams psABBA
#'
#' @details
#' The AB-XY participation shift refers to one of Gibson's (2003) dyadic participation shifts. The AB-XY pshift refers to a tendency for turn usurping (here, the next sender and the next receiver are not in the previous event). For each timepoint t, the psABXY statistic is equal to one for the dyads that will create the participation shift if they would occur in the edgelist at time t and equal to zero for the dyads that will not create this participation shift. If consider_type is set to TRUE, the type of the AB event and the type of the XY events have to be equal. If it is set to FALSE, the participation shift is set to one for every XY event, regardless of the event type. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed. Note that the AB-XY pshift is only defined for directed events.
#'
#' @seealso \code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXA}},
#' \code{\link{psABXB}} or \code{\link{psABAY}} for other dyadic participation
#' shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ psABXY()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' @export
psABXY <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABXY"
  return(call_args)
}

#' psABAY
#'
#' Specifies the statistic for a participation shift AB-AY.
#'
#' @inheritParams psABBA
#'
#' @details
#' One of Gibson's (2003) dyadic participation shifts. The AB-AY participation
#' shift refers to a tendency for \emph{turn continuing}. For directed events,
#' the sender (A) in the current event is the same as the sender in the
#' previous event (A), and the receiver (Y) is different from the previous
#' receiver (B). In undirected events, one of the current actors (A) matches
#' one of the actors in the previous events (A or B), while the other actor (Y)
#' is different.
#'
#' To identify these shifts, a statistic 'psABAY' is calculated for each pair
#' of actors at a given timepoint (t). If the pair's interaction follows the
#' AB-AY pattern, the statistic is set equal to one; otherwise, it's set to
#' zero.
#'
#' Additionally, the types of the AB and AY events can be taken into account. If
#' 'consider_type' is 'TRUE', the type of the AB event and the type of the AY
#' event must match for the shift to occur. If 'consider_type' is 'FALSE', the
#' shift happens for every AY event, regardless of the event type.
#'
#' @seealso \code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXA}},
#' \code{\link{psABXB}}, \code{\link{psABXY}} or \code{\link{psABAB}} for other
#' dyadic participation shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh <- remify::remify(history, model = "tie")
#' effects <- ~ psABAY()
#' remstats(reh = reh, tie_effects = effects)
#'
#' @export
psABAY <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABAY"
  return(call_args)
}

#' psABAB
#'
#' Specifies the statistic for a pshift AB-AB effect.
#'
#' @inheritParams psABBA
#'
#' @details
#'
#' Refers to the tendency for the same dyads to keep interacting. For directed
#' events, the next sender and receiver are equal to the previous sender and
#' receiver. For undirected events, the next actor pair is equal to the current
#' actor pair. For each timepoint t, the psABAB statistic is equal to one for
#' the dyads that will create the participation shift if they would occur in
#' the edgelist at time t and equal to zero for the dyads that will not create
#' this participation shift. If consider_type is set to TRUE, the type of the
#' two subsequent AB events have to be equal. If it is set to FALSE, the
#' participation shift is set to one for every AB event, regardless of the
#' event type. If multiple events in the edgelist occur at the same time point,
#' the order of these events determines whether the p-shift is observed.
#'
#' @seealso \code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXA}},
#' \code{\link{psABXB}}, \code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation shifts.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
#' effects <- ~ psABAB()
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' @export
psABAB <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "psABAB"
  return(call_args)
}

#' psABA
#'
#' Specifies the statistic for a participation shift AB-A in the sender step of
#' the actor-oriented model.
#'
#' @details
#'
#' Refers to the tendency for the same actor to keep initiating events: The
#' next sender is equal to the previous sender. For each timepoint t, the psABA
#' statistic is equal to one for the actor that will create the participation
#' shift if they would occur in the edgelist as the sender at time t and equal
#' to zero for the actors that will not create this participation shift. If
#' multiple events in the edgelist occur at the same time point, the order of
#' these events determines whether the p-shift is observed.
#'
#' @seealso \code{\link{psABB}} or \code{\link{psABX}} for exploring
#' alternative participation shifts in the sender step of the actor-oriented
#' model.
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(sender_effects = ~ psABA(), reh = reh_actor)
#'
#' @export
psABA <- function() {
  call_args <- list()
  call_args$effect <- "psABA"
  return(call_args)
}


#' rrankSend
#'
#' Specifies the statistic for a recency rank send effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the
#' \code{receiver_effects} argument of \code{\link{aomstats}}.
#'
#' The rrankSend effect refers to a rank-based recency effect, as described in
#' section 2.2.5 of Butts (2008). For each timepoint t, for directed dyad (i,j)
#' the statistic is equal to the inverse of the rank of receiver j among the
#' actors to which sender i has most recently send past events. Note that the
#' 'rrankSend' effect is only defined for directed events.
#'
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @aliases recencyRank rrank
#' @seealso \code{\link{rrankReceive}}, \code{\link{recencySendSender}},
#' \code{\link{recencySendReceiver}}, \code{\link{recencyReceiveSender}},
#' \code{\link{recencyReceiveReceiver}} and \code{\link{recencyContinue}} for
#' other type of recency effects
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ rrankSend()
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(receiver_effects = effects, reh = reh_actor)
#'
#' @export
rrankSend <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "rrankSend"
  return(call_args)
}

#' rrankReceive
#'
#' Specifies the statistic for a recency rank receive effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the
#' \code{receiver_effects} argument of \code{\link{aomstats}}.
#'
#' The rrankSend effect refers to a rank-based recency effect, as described in
#' section 2.2.5 of Butts (2008). For each timepoint t, for directed dyad (i,j)
#' the statistic is equal to the inverse of the rank of receiver j among the
#' actors from which sender i has most recently received past events. Note that
#' the 'rrankReceive' effect is only defined for directed events.
#'
#' @inheritParams rrankSend
#'
#' @seealso \code{\link{rrankSend}}, \code{\link{recencySendSender}},
#' \code{\link{recencySendReceiver}}, \code{\link{recencyReceiveSender}},
#' \code{\link{recencyReceiveReceiver}} and \code{\link{recencyContinue}} for
#' other type of recency effects
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#'
#' reh_tie <- remify::remify(history, model = "tie")
#' effects <- ~ rrankReceive()
#' remstats(reh = reh_tie, tie_effects = effects)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(receiver_effects = effects, reh = reh_actor)
#'
#' @export
rrankReceive <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "rrankReceive"
  return(call_args)
}


#' recencySendSender
#'
#' Specifies the statistic for a recency send of sender effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the \code{sender_effects}
#' argument of \code{\link{aomstats}}.
#'
#' The recencySendSender effect refers to a recency statistic similar to what
#' is described in Vu et al. (2017) and Mulder and Leenders (2019). For each
#' timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the time
#' that has past since sender i was last active as sender + 1). Note that the
#' 'recencySendSender' effect is only defined for directed events.
#'
#' @param consider_type character. Controls how event types are handled:
#'   \code{"ignore"} (default): aggregate over all event types (one statistic);
#'   \code{"separate"}: compute C type-specific statistics, where the type-c
#'   statistic for a dyad reflects past type-c events on that actor pair
#'   ;
#'   \code{"interact"}: compute C^2 statistics capturing past-event-type x
#'   dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
#'   Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
#'   (-> \code{"separate"}) for backward compatibility.
#' @seealso \code{\link{rrankSend}}, \code{\link{rrankReceive}},
#' \code{\link{recencySendReceiver}}, \code{\link{recencyReceiveSender}},
#' \code{\link{recencyReceiveReceiver}} and \code{\link{recencyContinue}} for
#' other type of recency effects
#'
#' @examples
#' effects <- ~ recencySendSender()
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(sender_effects = effects, reh = reh_actor)
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @export
recencySendSender <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "recencySendSender"
  return(call_args)
}

#' recencySendReceiver
#'
#' Specifies the statistic for a recency send of receiver effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the
#' \code{receiver_effects} argument of \code{\link{aomstats}}.
#'
#' The recencySendReceiver effect refers to a recency statistic similar to what
#' is described in Vu et al. (2017) and Mulder and Leenders (2019). For each
#' timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the time
#' that has past since receiver j was last active as sender + 1). Note that the
#' 'recencySendReceiver' effect is only defined for directed events.
#'
#' @inheritParams recencySendSender
#'
#' @seealso \code{\link{rrankSend}}, \code{\link{rrankReceive}},
#' \code{\link{recencySendSender}}, \code{\link{recencyReceiveSender}},
#' \code{\link{recencyReceiveReceiver}} and \code{\link{recencyContinue}} for
#' other type of recency effects
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' effects <- ~ recencySendReceiver()
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(receiver_effects = effects, reh = reh_actor)
#'
#' @export
recencySendReceiver <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "recencySendReceiver"
  return(call_args)
}

#' recencyReceiveSender
#'
#' Specifies the statistic for a recency receive of sender effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the
#' \code{sender_effects} argument of \code{\link{aomstats}}.
#'
#' The recencyReceiveSender effect refers to a recency statistic similar to
#' what is described in Vu et al. (2017) and Mulder and Leenders (2019). For
#' each timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the
#' time that has past since sender i was last active as receiver + 1). Note
#' that the 'recencyReceiveSender' effect is only defined for directed events.
#'
#' @inheritParams recencySendSender
#'
#' @seealso \code{\link{rrankSend}}, \code{\link{rrankReceive}},
#' \code{\link{recencySendSender}}, \code{\link{recencySendReceiver}},
#' \code{\link{recencyReceiveReceiver}} and \code{\link{recencyContinue}} for
#' other type of recency effects
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' effects <- ~ recencyReceiveSender()
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(sender_effects = effects, reh = reh_actor)
#'
#' @export
recencyReceiveSender <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "recencyReceiveSender"
  return(call_args)
}

#' recencyReceiveReceiver
#'
#' Specifies the statistic for a recency receive of receiver effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the
#' \code{receiver_effects} argument of \code{\link{aomstats}}.
#'
#' The recencyReceiveReceiver effect refers to a recency statistic similar to
#' what is described in Vu et al. (2017) and Mulder and Leenders (2019). For
#' each timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the
#' time that has past since receiver j was last active as receiver + 1). Note
#' that the 'recencyReceiveReceiver' effect is only defined for directed
#' events.
#'
#' @inheritParams recencySendSender
#'
#' @seealso \code{\link{rrankSend}}, \code{\link{rrankReceive}},
#' \code{\link{recencySendSender}}, \code{\link{recencyReceiveSender}},
#' \code{\link{recencyReceiveSender}} and \code{\link{recencyContinue}} for
#' other type of recency effects
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' effects <- ~ recencyReceiveReceiver()
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(receiver_effects = effects, reh = reh_actor)
#'
#' @export
recencyReceiveReceiver <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "recencyReceiveReceiver"
  return(call_args)
}

#' recencyContinue
#'
#' Specifies the statistic for a recency continue effect in the
#' \code{effects} argument of \code{\link{tomstats}} or the
#' \code{receiver_effects} argument of \code{\link{aomstats}}.
#'
#' The recencyContinue effect refers to a recency statistic similar to what is
#' described in Vu et al. (2017) and Mulder and Leenders (2019). For each
#' timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the time
#' that has past since the dyad was last active + 1).
#'
#' @inheritParams recencySendSender
#'
#' @aliases recency
#'
#' @seealso \code{\link{rrankSend}}, \code{\link{rrankReceive}},
#' \code{\link{recencySendSender}}, \code{\link{recencyReceiveSender}},
#' \code{\link{recencyReceiveSender}} and \code{\link{recencyReceiveReceiver}}
#' for other type of recency effects
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' effects <- ~ recencyContinue()
#' reh_tie <- remify::remify(history, model = "tie")
#' remstats(tie_effects = effects, reh = reh_tie)
#'
#' reh_actor <- remify::remify(history, model = "actor")
#' remstats(receiver_effects = effects, reh = reh_actor)
#'
#' @export
recencyContinue <- function(consider_type = "ignore") {
  call_args <- list()
  call_args$consider_type <- consider_type
  call_args$effect <- "recencyContinue"
  return(call_args)
}

#' userStat
#'
#' Allows the user to add its own pre-computed statistic to the statistics
#' object and, optionally, interact this statistic with other statistics in the
#' formula.
#'
#' @param x Matrix with number of rows equal to the number of events and number
#' of columns equal to the number of dyads in the network (tie-oriented model) 
#' or the number of actors in the network (actor-oriented model)
#' @param variableName Optionally, a string with the name of the statistic. 
#'
#' @returns List with all information required by `remstats::remstats()` to 
#' compute the statistic.
#' 
#' @examples
#' \donttest{
#'  reh <- remify::remify(history, model = "tie")
#'  actor101Events <- which(history$actor1 == "101" | history$actor2 == "101")
#'  actor101_stat <- t(sapply(seq_len(nrow(history)), function(i) {
#'    rep(i %in% actor101Events, reh$D)
#'  }))
#'  
#'  # Main effects only
#'  effects <- ~ userStat(x = actor101_stat, variableName = "actor101event")
#'  remstats(reh = reh, tie_effects = effects)
#'  
#'  # Model with interaction effects
#'  interaction_effects <- ~ inertia() *
#'    userStat(x = actor101_stat, variableName = "actor101event")
#'  remstats(reh = reh, tie_effects = interaction_effects)
#' }
#' @export
userStat <- function(x, variableName = NULL) {
  # Missing values
  if (anyNA(x)) {
    warning("Matrix 'x' in userStat() contains missing values.")
  }

  # Output
  list(
    effect = "userStat",
    x = x,
    variable = variableName,
    scaling = "none"
  )
}

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

# Actor info
info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

# Tie model
reh <- remify::remify(edgelist, model = "tie")
effects <- ~ send(variable = "x1"):inertia()
stats <- remstats(reh, tie_effects = effects, attr_actors = info)

interaction <- stats[, , "send_x1"] * stats[, , "inertia"]
expect_equal(stats[, , "send_x1:inertia"], interaction)

# # Actor model
# reh <- remify::remify(edgelist, model = "actor")
# sender_effects <- ~ send(variable = "x1"):outdegreeSender()
# receiver_effects <- ~ receive(variable = "x1"):inertia()
# stats <- remstats(reh, sender_effects = sender_effects, 
#     receiver_effects = receiver_effects, attr_actors = info)
# sender_stats <- stats$sender_stats
# receiver_stats <- stats$receiver_stats
# 
# sender_interaction <- sender_stats[, , "send_x1"] * 
#   sender_stats[, , "outdegreeSender"]
# expect_equal(sender_stats[, , "send_x1:outdegreeSender"], 
#   sender_interaction)
# 
# receiver_interaction <- receiver_stats[, , "receive_x1"] * 
#   receiver_stats[, , "inertia"]
# expect_equal(receiver_stats[, , "receive_x1:inertia"], receiver_interaction)

# Tests for typed events: consider_type x extend_riskset_by_type orthogonality
#
# Covers:
#   - riskset = "full",   directed = TRUE
#   - riskset = "active", directed = TRUE
#   - riskset = "manual", directed = TRUE
#   - riskset = "full",   directed = FALSE
#
# For each combination, tests all four cases:
#   ext=TRUE,  consider_type="interact" -> [M x D_typed x C^2] named stat.social.social etc
#   ext=TRUE,  consider_type="separate" -> [M x D_typed x C]   named stat.social, stat.work
#   ext=TRUE,  consider_type="ignore" -> [M x D_typed x 1]  named stat
#   ext=FALSE, consider_type="separate" -> [M x D_dyad x C] named stat.social, stat.work
#   ext=FALSE, consider_type="ignore" -> [M x D_dyad x 1]  named stat
#
# Note: type slices are in alphabetical order of the type label ("social" < "work").

data(history, package = "remstats")
edgelist <- history
colnames(edgelist)[colnames(edgelist) == "setting"] <- "type"

M <- nrow(edgelist)   # 115

# ---------------------------------------------------------------------------
# Helper: compute expected weighted inertia for a given dyad at event m.
# ---------------------------------------------------------------------------
expected_inertia <- function(edgelist, m, a1, a2, type = NULL, directed = TRUE) {
  prior <- edgelist[seq_len(m - 1), ]
  if (!directed) {
    swap <- prior$actor1 > prior$actor2
    tmp <- prior$actor1[swap]
    prior$actor1[swap] <- prior$actor2[swap]
    prior$actor2[swap] <- tmp
    lo <- min(a1, a2); hi <- max(a1, a2)
    a1 <- lo; a2 <- hi
  }
  prior <- prior[prior$actor1 == a1 & prior$actor2 == a2, ]
  if (!is.null(type)) prior <- prior[prior$type == type, ]
  sum(prior$weight)
}

# ---------------------------------------------------------------------------
# Value checker: tests inertia values at given events for a specific dyad.
# Slices are in alphabetical type order: "inertia.social" first, "inertia.work" second.
# ---------------------------------------------------------------------------
check_inertia_values <- function(stats_TT, stats_TF, stats_FT, stats_FF,
                                  rs_TT, rs_FT,
                                  edgelist, a1, a2,
                                  events, label, directed = TRUE) {
  s_col <- if ("sender" %in% names(rs_TT)) "sender" else "actor1"
  r_col <- if ("receiver" %in% names(rs_TT)) "receiver" else "actor2"
  if (!directed) { lo <- min(a1,a2); hi <- max(a1,a2); a1 <- lo; a2 <- hi }

  col_work   <- which(rs_TT[[s_col]] == a1 & rs_TT[[r_col]] == a2 & rs_TT$type == "work")
  col_social <- which(rs_TT[[s_col]] == a1 & rs_TT[[r_col]] == a2 & rs_TT$type == "social")
  col_dyad   <- which(rs_FT[[s_col]] == a1 & rs_FT[[r_col]] == a2)

  for (m in events) {
    exp_w <- expected_inertia(edgelist, m, a1, a2, type = "work",   directed = directed)
    exp_s <- expected_inertia(edgelist, m, a1, a2, type = "social", directed = directed)
    exp_a <- expected_inertia(edgelist, m, a1, a2, type = NULL,     directed = directed)

    if (length(col_work) > 0) {
    	expect_equal(as.numeric(stats_TT[m, col_work, "inertia.work"]), exp_w,
    							 info = paste(label, "TT.work m=", m))
    	expect_equal(as.numeric(stats_TF[m, col_work, "inertia"]), exp_a,
    							 info = paste(label, "TF.ignore m=", m))
    }
    if (length(col_social) > 0) {
    	expect_equal(as.numeric(stats_TT[m, col_social, "inertia.social"]), exp_s,
    							 info = paste(label, "TT.social m=", m))
    }
    if (length(col_dyad) > 0) {
      expect_equal(as.numeric(stats_FT[m, col_dyad, "inertia.work"]),    exp_w,
        info = paste(label, "FT.work m=",    m))
      expect_equal(as.numeric(stats_FT[m, col_dyad, "inertia.social"]),  exp_s,
        info = paste(label, "FT.social m=",  m))
      expect_equal(as.numeric(stats_FF[m, col_dyad, "inertia"]), exp_a,
        info = paste(label, "FF.ignore m=", m))
    }
  }
}

# ---------------------------------------------------------------------------
# Consistency: FT/FF values must match the corresponding columns of TT/TF.
#
# The FT type-slice checks are valid for all riskset types: ext=FALSE just
# restricts the riskset to D_dyad rows, but the per-type values are computed
# from the same typed inertia matrix as TT.
#
# The FF TypeAgg check (FF.TypeAgg == TF.TypeAgg) is only valid for FULL
# risksets. For active/manual risksets, TF's TypeAgg at a (actor1,actor2,type)
# row only accumulates events of that type (because the C++ risksetMatrix only
# has a valid slot for observed type), while FF accumulates all types for that
# pair. They can genuinely differ when a dyad was observed in one type but not
# the other. Pass check_TypeAgg=FALSE to skip this for active/manual risksets.
# ---------------------------------------------------------------------------
check_consistency <- function(stats_TT, stats_TF, stats_FT, stats_FF,
                               rs_TT, rs_FT, label,
                               check_TypeAgg = TRUE) {
  s_col <- if ("sender" %in% names(rs_TT)) "sender" else "actor1"
  r_col <- if ("receiver" %in% names(rs_TT)) "receiver" else "actor2"

  social_cols <- which(rs_TT$type == "social")
  work_cols   <- which(rs_TT$type == "work")

  key_FT   <- paste(rs_FT[[s_col]], rs_FT[[r_col]], sep = "|")
  key_soc  <- paste(rs_TT[[s_col]][social_cols], rs_TT[[r_col]][social_cols], sep = "|")
  key_work <- paste(rs_TT[[s_col]][work_cols],   rs_TT[[r_col]][work_cols],   sep = "|")

  m_soc  <- match(key_FT, key_soc)
  m_work <- match(key_FT, key_work)
  valid_s <- which(!is.na(m_soc))
  valid_w <- which(!is.na(m_work))

  expect_equal(
    stats_FT[, valid_s, "inertia.social"],
    stats_TT[, social_cols[m_soc[valid_s]], "inertia.social"],
    info = paste(label, "FT.social == TT[social_cols]")
  )
  expect_equal(
    stats_FT[, valid_w, "inertia.work"],
    stats_TT[, work_cols[m_work[valid_w]], "inertia.work"],
    info = paste(label, "FT.work == TT[work_cols]")
  )

  # FF inertia == TF inertia: only valid for full risksets.
  # For active/manual, TF TypeAgg at a type-row only sums events of that type
  # (missing-type slot = -1 in risksetMatrix → skipped in transform_inertia),
  # while FF always sums all types. They can legitimately differ.
  if (check_TypeAgg) {
    both <- intersect(key_FT[valid_s], key_FT[valid_w])
    valid_both_FT <- which(key_FT %in% both)
    if (length(valid_both_FT) > 0) {
      soc_idx <- social_cols[match(key_FT[valid_both_FT], key_soc)]
      valid_match <- which(!is.na(soc_idx))
      if (length(valid_match) > 0) {
        expect_equal(
          stats_FF[, valid_both_FT[valid_match], "inertia"],
          stats_TF[, soc_idx[valid_match], "inertia"],
          info = paste(label, "FF.inertia == TF.inertia (full riskset only)")
        )
      }
    }
  }
}

# ===========================================================================
# SECTION 1: riskset = "full", directed = TRUE
# ===========================================================================
N       <- 10L
D_dyad  <- N * (N - 1L)   # 90
D_typed <- D_dyad * 2L    # 180

reh_full_T <- remify(edgelist, model = "tie", riskset = "full",
                      extend_riskset_by_type = TRUE)
reh_full_F <- remify(edgelist, model = "tie", riskset = "full",
                      extend_riskset_by_type = FALSE)

expect_true(reh_full_T$meta$with_type_riskset)
expect_false(reh_full_F$meta$with_type_riskset)

stats_TT_full <- remstats(reh_full_T, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_TT_full_int <- remstats(reh_full_T, tie_effects = ~ inertia(consider_type = "interact"), start = 1)
stats_TF_full <- remstats(reh_full_T, tie_effects = ~ inertia(consider_type = FALSE), start = 1)
stats_FT_full <- remstats(reh_full_F, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_FF_full <- remstats(reh_full_F, tie_effects = ~ inertia(consider_type = FALSE), start = 1)

# Shapes: baseline + inertia.social + inertia.work = 3; baseline + inertia.TypeAgg = 2
expect_equal(dim(stats_TT_full), c(M, D_typed, 3L), info = "full TT shape")
expect_equal(dim(stats_TF_full), c(M, D_typed, 2L), info = "full TF shape")
expect_equal(dim(stats_FT_full), c(M, D_dyad,  3L), info = "full FT shape")
expect_equal(dim(stats_FF_full), c(M, D_dyad,  2L), info = "full FF shape")

# Dimnames: types in alphabetical order ("social" < "work")
expect_equal(dimnames(stats_TT_full)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_TF_full)[[3]], c("baseline", "inertia"))
expect_equal(dimnames(stats_FT_full)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_FF_full)[[3]], c("baseline", "inertia"))

# Riskset attributes
rs_TT_full <- attr(stats_TT_full, "riskset")
rs_FT_full <- attr(stats_FT_full, "riskset")
expect_true("type"  %in% colnames(rs_TT_full))
expect_false("type" %in% colnames(rs_FT_full))
expect_equal(nrow(rs_TT_full), D_typed)
expect_equal(nrow(rs_FT_full), D_dyad)

# Zero-pattern for TT (social cols have zero work-type inertia, vice versa)
social_cols_full <- which(rs_TT_full$type == "social")
work_cols_full   <- which(rs_TT_full$type == "work")
# "interact": social dyad cols should have zero work-inertia (.social.work slice)
expect_true(all(stats_TT_full_int[, social_cols_full, "inertia.social.work"] == 0))
expect_true(all(stats_TT_full_int[, work_cols_full, "inertia.work.social"] == 0))

# Consistency
check_consistency(stats_TT_full, stats_TF_full, stats_FT_full, stats_FF_full,
                  rs_TT_full, rs_FT_full, "full/directed")

# Spot values
check_inertia_values(stats_TT_full, stats_TF_full, stats_FT_full, stats_FF_full,
                     rs_TT_full, rs_FT_full, edgelist,
                     a1 = 105, a2 = 109, events = c(6L, 113L, 114L, 115L),
                     label = "full/dir (105->109)")
check_inertia_values(stats_TT_full, stats_TF_full, stats_FT_full, stats_FF_full,
                     rs_TT_full, rs_FT_full, edgelist,
                     a1 = 105, a2 = 104, events = c(113L, 114L, 115L),
                     label = "full/dir (105->104)")
check_inertia_values(stats_TT_full, stats_TF_full, stats_FT_full, stats_FF_full,
                     rs_TT_full, rs_FT_full, edgelist,
                     a1 = 107, a2 = 105, events = c(113L, 114L, 115L),
                     label = "full/dir (107->105)")

# ===========================================================================
# SECTION 2: riskset = "active", directed = TRUE
# ===========================================================================
reh_active_T <- remify(edgelist, model = "tie", riskset = "active",
                        extend_riskset_by_type = TRUE)
reh_active_F <- remify(edgelist, model = "tie", riskset = "active",
                        extend_riskset_by_type = FALSE)

stats_TT_act <- remstats(reh_active_T, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_TT_act_int <- remstats(reh_active_T, tie_effects = ~ inertia(consider_type = "interact"), start = 1)
stats_TF_act <- remstats(reh_active_T, tie_effects = ~ inertia(consider_type = FALSE), start = 1)
stats_FT_act <- remstats(reh_active_F, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_FF_act <- remstats(reh_active_F, tie_effects = ~ inertia(consider_type = FALSE), start = 1)

rs_TT_act <- attr(stats_TT_act, "riskset")
rs_FT_act <- attr(stats_FT_act, "riskset")
D_act_typed <- nrow(rs_TT_act)
D_act_dyad  <- nrow(rs_FT_act)

# Shapes
expect_equal(dim(stats_TT_act), c(M, D_act_typed, 3L), info = "active TT shape")
expect_equal(dim(stats_TF_act), c(M, D_act_typed, 2L), info = "active TF shape")
expect_equal(dim(stats_FT_act), c(M, D_act_dyad,  3L), info = "active FT shape")
expect_equal(dim(stats_FF_act), c(M, D_act_dyad,  2L), info = "active FF shape")

# Dimnames
expect_equal(dimnames(stats_TT_act)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_TF_act)[[3]], c("baseline", "inertia"))
expect_equal(dimnames(stats_FT_act)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_FF_act)[[3]], c("baseline", "inertia"))

# Riskset
expect_true("type"  %in% colnames(rs_TT_act))
expect_false("type" %in% colnames(rs_FT_act))
# Note: for active riskset D_typed != necessarily 2*D_dyad (work/social active sets differ)

# Zero-pattern
social_cols_act <- which(rs_TT_act$type == "social")
work_cols_act   <- which(rs_TT_act$type == "work")
# "interact": social dyad cols should have zero work-inertia (.social.work slice)
expect_true(all(stats_TT_act_int[, social_cols_act, "inertia.social.work"] == 0))
expect_true(all(stats_TT_act_int[, work_cols_act, "inertia.work.social"] == 0))

# Consistency (only dyads in both type sets for TypeAgg comparison)
check_consistency(stats_TT_act, stats_TF_act, stats_FT_act, stats_FF_act,
                  rs_TT_act, rs_FT_act, "active/directed", check_TypeAgg = FALSE)

# Spot values
check_inertia_values(stats_TT_act, stats_TF_act, stats_FT_act, stats_FF_act,
                     rs_TT_act, rs_FT_act, edgelist,
                     a1 = 105, a2 = 109, events = c(6L, 113L, 114L, 115L),
                     label = "active/dir (105->109)")
check_inertia_values(stats_TT_act, stats_TF_act, stats_FT_act, stats_FF_act,
                     rs_TT_act, rs_FT_act, edgelist,
                     a1 = 107, a2 = 105, events = c(113L, 114L, 115L),
                     label = "active/dir (107->105)")

# ===========================================================================
# SECTION 3: riskset = "manual", directed = TRUE
# ===========================================================================
manual_rs <- data.frame(
  actor1 = c(105, 105, 113, 109, 101),
  actor2 = c(109, 113, 107, 105, 115)
)

reh_manual_T <- suppressWarnings(
  remify(edgelist, model = "tie", riskset = "manual",
          manual.riskset = manual_rs, extend_riskset_by_type = TRUE)
)
reh_manual_F <- suppressWarnings(
  remify(edgelist, model = "tie", riskset = "manual",
          manual.riskset = manual_rs, extend_riskset_by_type = FALSE)
)

stats_TT_man <- remstats(reh_manual_T, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_TT_man_int <- remstats(reh_manual_T, tie_effects = ~ inertia(consider_type = "interact"), start = 1)
stats_TF_man <- remstats(reh_manual_T, tie_effects = ~ inertia(consider_type = FALSE), start = 1)
stats_FT_man <- remstats(reh_manual_F, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_FF_man <- remstats(reh_manual_F, tie_effects = ~ inertia(consider_type = FALSE), start = 1)

rs_TT_man <- attr(stats_TT_man, "riskset")
rs_FT_man <- attr(stats_FT_man, "riskset")
D_man_typed <- nrow(rs_TT_man)
D_man_dyad  <- nrow(rs_FT_man)

# Shapes
expect_equal(dim(stats_TT_man), c(M, D_man_typed, 3L), info = "manual TT shape")
expect_equal(dim(stats_TF_man), c(M, D_man_typed, 2L), info = "manual TF shape")
expect_equal(dim(stats_FT_man), c(M, D_man_dyad,  3L), info = "manual FT shape")
expect_equal(dim(stats_FF_man), c(M, D_man_dyad,  2L), info = "manual FF shape")

# Dimnames
expect_equal(dimnames(stats_TT_man)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_TF_man)[[3]], c("baseline", "inertia"))
expect_equal(dimnames(stats_FT_man)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_FF_man)[[3]], c("baseline", "inertia"))

# Riskset
expect_true("type"  %in% colnames(rs_TT_man))
expect_false("type" %in% colnames(rs_FT_man))

# Zero-pattern
social_cols_man <- which(rs_TT_man$type == "social")
work_cols_man   <- which(rs_TT_man$type == "work")
# "interact": social dyad cols should have zero work-inertia (.social.work slice)
expect_true(all(stats_TT_man_int[, social_cols_man, "inertia.social.work"] == 0))
expect_true(all(stats_TT_man_int[, work_cols_man, "inertia.work.social"] == 0))

# Consistency
check_consistency(stats_TT_man, stats_TF_man, stats_FT_man, stats_FF_man,
                  rs_TT_man, rs_FT_man, "manual/directed", check_TypeAgg = FALSE)

# Spot values: (105->109) is in the manual riskset
check_inertia_values(stats_TT_man, stats_TF_man, stats_FT_man, stats_FF_man,
                     rs_TT_man, rs_FT_man, edgelist,
                     a1 = 105, a2 = 109, events = c(6L, 113L, 114L, 115L),
                     label = "manual/dir (105->109)")

# ===========================================================================
# SECTION 4: riskset = "full", directed = FALSE
# ===========================================================================
reh_undir_T <- remify(edgelist, model = "tie", riskset = "full",
                        directed = FALSE, extend_riskset_by_type = TRUE)
reh_undir_F <- remify(edgelist, model = "tie", riskset = "full",
                        directed = FALSE, extend_riskset_by_type = FALSE)

stats_TT_undir <- remstats(reh_undir_T, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_TT_undir_int <- remstats(reh_undir_T, tie_effects = ~ inertia(consider_type = "interact"), start = 1)
stats_TF_undir <- remstats(reh_undir_T, tie_effects = ~ inertia(consider_type = FALSE), start = 1)
stats_FT_undir <- remstats(reh_undir_F, tie_effects = ~ inertia(consider_type = "separate"), start = 1)
stats_FF_undir <- remstats(reh_undir_F, tie_effects = ~ inertia(consider_type = FALSE), start = 1)

rs_TT_undir <- attr(stats_TT_undir, "riskset")
rs_FT_undir <- attr(stats_FT_undir, "riskset")
# Derive D from actual riskset (avoids assumption about self-loops / actor pairing)
D_undir_typed <- nrow(rs_TT_undir)
D_undir_dyad  <- nrow(rs_FT_undir)
expect_equal(D_undir_typed, 2L * D_undir_dyad, info = "undir: D_typed == 2 * D_dyad")

# Shapes
expect_equal(dim(stats_TT_undir), c(M, D_undir_typed, 3L), info = "undir TT shape")
expect_equal(dim(stats_TF_undir), c(M, D_undir_typed, 2L), info = "undir TF shape")
expect_equal(dim(stats_FT_undir), c(M, D_undir_dyad,  3L), info = "undir FT shape")
expect_equal(dim(stats_FF_undir), c(M, D_undir_dyad,  2L), info = "undir FF shape")

# Dimnames
expect_equal(dimnames(stats_TT_undir)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_TF_undir)[[3]], c("baseline", "inertia"))
expect_equal(dimnames(stats_FT_undir)[[3]], c("baseline", "inertia.social", "inertia.work"))
expect_equal(dimnames(stats_FF_undir)[[3]], c("baseline", "inertia"))

# Riskset
expect_true("type"  %in% colnames(rs_TT_undir))
expect_false("type" %in% colnames(rs_FT_undir))

# Zero-pattern
social_cols_undir <- which(rs_TT_undir$type == "social")
work_cols_undir   <- which(rs_TT_undir$type == "work")
# "interact": social dyad cols should have zero work-inertia (.social.work slice)
expect_true(all(stats_TT_undir_int[, social_cols_undir, "inertia.social.work"] == 0))
expect_true(all(stats_TT_undir_int[, work_cols_undir, "inertia.work.social"] == 0))

# Consistency
check_consistency(stats_TT_undir, stats_TF_undir, stats_FT_undir, stats_FF_undir,
                  rs_TT_undir, rs_FT_undir, "full/undirected")

# Spot values
check_inertia_values(stats_TT_undir, stats_TF_undir, stats_FT_undir, stats_FF_undir,
                     rs_TT_undir, rs_FT_undir, edgelist,
                     a1 = 105, a2 = 109, events = c(6L, 113L, 114L, 115L),
                     label = "full/undir (105-109)", directed = FALSE)
check_inertia_values(stats_TT_undir, stats_TF_undir, stats_FT_undir, stats_FF_undir,
                     rs_TT_undir, rs_FT_undir, edgelist,
                     a1 = 107, a2 = 105, events = c(113L, 114L, 115L),
                     label = "full/undir (105-107)", directed = FALSE)

# ===========================================================================
# SECTION 5: mixed consider_type in one formula, ext=FALSE
# ===========================================================================
stats_mix <- remstats(reh_full_F,
  tie_effects = ~ inertia(consider_type = "interact") +
                  outdegreeSender(consider_type = FALSE), start = 1)

# baseline + inertia.social + inertia.work + outdegreeSender.TypeAgg = 4 slices
expect_equal(dim(stats_mix), c(M, D_dyad, 4L), info = "mixed: shape")
expect_equal(dimnames(stats_mix)[[3]],
  c("baseline", "inertia.social", "inertia.work", "outdegreeSender"),
  info = "mixed: dimnames")

# inertia slices must match standalone FT
expect_equal(stats_mix[, , "inertia.social"], stats_FT_full[, , "inertia.social"],
  info = "mixed: inertia.social matches standalone FT")
expect_equal(stats_mix[, , "inertia.work"],   stats_FT_full[, , "inertia.work"],
  info = "mixed: inertia.work matches standalone FT")

# Condition 2: Undirected events, tie-oriented model with active risk set

# Small edgelist
edgelist <- data.frame(
  time = 1:10,
  actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
  actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
)

# Statistics
reh <- remify::remify(edgelist, model = "tie", directed = FALSE, 
  riskset = "active")
effects <- ~ degreeDiff() + degreeMin() + degreeMax() + totaldegreeDyad() +
  inertia() + sp() + sp(unique = TRUE) + psABAB() + psABAY()
stats <- remstats(reh, tie_effects = effects, start = 1)
riskset <- attr(stats, "riskset")

# baseline
expect_equal(stats[, , "baseline"], matrix(1, nrow = nrow(edgelist), ncol = nrow(riskset)))

# degreeMin
degreeMin <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 1, 0, 0, 0, 0),
  c(1, 1, 0, 1, 0, 0),
  c(1, 2, 0, 1, 0, 0),
  c(2, 3, 0, 2, 0, 0),
  c(3, 3, 0, 3, 0, 0),
  c(3, 3, 1, 3, 1, 1),
  c(4, 4, 1, 4, 1, 1),
  c(4, 4, 1, 5, 1, 1),
  c(4, 4, 2, 6, 2, 2)
)
expect_equal(stats[, , "degreeMin"], degreeMin)

# degreeMax
degreeMax <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(1, 1, 1, 1, 0, 1),
  c(2, 2, 2, 1, 1, 1),
  c(3, 3, 3, 2, 1, 2),
  c(3, 3, 3, 3, 2, 3),
  c(3, 4, 3, 4, 3, 4),
  c(3, 5, 3, 5, 3, 5),
  c(4, 5, 4, 5, 4, 5),
  c(5, 6, 4, 6, 5, 6),
  c(6, 6, 4, 6, 6, 6)
)
expect_equal(stats[, , "degreeMax"], degreeMax)

# degreeDiff
degreeDiff <- degreeMax - degreeMin
expect_equal(stats[, , "degreeDiff"], degreeDiff)

# totaldegreeDyad
totaldegreeDyad <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(1, 2, 1, 1, 0, 1),
  c(3, 3, 2, 2, 1, 1),
  c(4, 5, 3, 3, 1, 2),
  c(5, 6, 3, 5, 2, 3),
  c(6, 7, 3, 7, 3, 4),
  c(6, 8, 4, 8, 4, 6),
  c(8, 9, 5, 9, 5, 6),
  c(9, 10, 5, 11, 6, 7),
  c(10, 10, 6, 12, 8, 8)
)
expect_equal(stats[, , "totaldegreeDyad"], totaldegreeDyad)

# ccp
# duration <- c(3, 2, 1, 1, 3, 2, 1, 1, 1, 1)
# ccp_stats <- remstats(reh, tie_effects = ~ ccp(duration = duration))
# ccp <- rbind(
#   matrix(0, ncol = nrow(riskset)),
#   c(0, 0, 0, 0, 0, 0),
#   c(0, 0, 0, 1, 0, 0),
#   c(0, 0, 0, 1, 0, 0),
#   c(0, 0, 0, 0, 0, 0),
#   c(0, 0, 0, 0, 0, 0),
#   c(0, 0, 0, 0, 1, 0),
#   c(0, 1, 0, 0, 1, 0),
#   c(0, 0, 0, 0, 0, 0),
#   c(0, 0, 0, 0, 0, 0)
# )
# expect_equal(ccp_stats[, , "ccp"], ccp)

# inertia
inertia <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 1, 0, 0, 0, 0),
  c(1, 1, 0, 0, 0, 0),
  c(1, 2, 0, 0, 0, 0),
  c(1, 2, 0, 1, 0, 0),
  c(1, 2, 0, 2, 0, 0),
  c(1, 2, 0, 2, 0, 1),
  c(2, 2, 0, 2, 0, 1),
  c(2, 2, 0, 3, 0, 1),
  c(2, 2, 0, 3, 1, 1)
)
expect_equal(stats[, , "inertia"], inertia)

# sp
sp <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 0, 0, 0, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(1, 1, 0, 1, 0, 0),
  c(2, 1, 0, 1, 0, 0),
  c(2, 1, 1, 1, 1, 0),
  c(2, 2, 1, 2, 1, 0),
  c(2, 2, 1, 2, 1, 0),
  c(2, 2, 2, 3, 1, 1)
)
expect_equal(stats[, , "sp"], sp)

# sp.unique
sp.unique <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 0, 0, 0, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(1, 1, 0, 1, 0, 0),
  c(1, 1, 0, 1, 0, 0),
  c(1, 1, 1, 1, 1, 0),
  c(1, 1, 1, 1, 1, 0),
  c(1, 1, 1, 1, 1, 0),
  c(1, 1, 2, 2, 1, 1)
)
expect_equal(stats[, , "sp.unique"], sp.unique)

# psABAB
psABAB <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 1, 0, 0, 0, 0),
  c(1, 0, 0, 0, 0, 0),
  c(0, 1, 0, 0, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(0, 0, 0, 0, 0, 1),
  c(1, 0, 0, 0, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(0, 0, 0, 0, 1, 0)
)
expect_equal(stats[, , "psABAB"], psABAB)

# psABAY
psABAY <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(1, 0, 1, 1, 0, 1),
  c(0, 1, 1, 1, 1, 0),
  c(1, 0, 1, 1, 0, 1),
  c(1, 1, 0, 0, 1, 1),
  c(1, 1, 0, 0, 1, 1),
  c(0, 1, 1, 1, 1, 0),
  c(0, 1, 1, 1, 1, 0),
  c(1, 1, 0, 0, 1, 1),
  c(1, 0, 1, 1, 0, 1)
)
expect_equal(stats[, , "psABAY"], psABAY)

# test standardization
std_effects <- ~
  degreeMin(scaling = "std") + degreeMax(scaling = "std") +
  degreeDiff(scaling = "std") + totaldegreeDyad(scaling = "std") +
  inertia(scaling = "std") + sp(scaling = "std") + 
  sp(scaling = "std", unique = TRUE)
std_stats <- remstats(reh, tie_effects = std_effects, start = 1)

sapply(2:dim(std_stats)[3], function(p) {
  stat_name <- dimnames(std_stats)[[3]][p]
  scaled_original <- t(apply(stats[, , stat_name], 1, scale))
  scaled_original[which(apply(stats[, , stat_name], 1, sd) == 0), ] <-
    rep(0, ncol(stats))
  expect_equal(std_stats[, , stat_name], scaled_original)
})

# test proportional scaling
prop_effects <- ~ inertia(scaling = "prop")
expect_error(remstats(reh, tie_effects = prop_effects, start = 1),
  pattern = "not defined")

prop_effects <- ~ degreeMin(scaling = "prop") + degreeMax(scaling = "prop") + 
  totaldegreeDyad(scaling = "prop") 
prop_stats <- remstats(reh, tie_effects = prop_effects, start = 1)

sapply(2:3, function(p) {
  stat_name <- dimnames(prop_stats)[[3]][p]
  scaled_original <- stats[,,stat_name] / (1:nrow(stats)-1)
  scaled_original[1,] <- 1/4
  expect_equal(prop_stats[,,stat_name], scaled_original)
}) # degreeMin and degreeMax

# totaldegreeDyad
prop_totaldegreeDyad <- stats[,,"totaldegreeDyad"] / (2*(1:nrow(stats)-1))
prop_totaldegreeDyad[1,] <- prop_totaldegreeDyad[1,] <- 1/4
expect_equal(prop_stats[,,"totaldegreeDyad"], prop_totaldegreeDyad)

# Test method -------------------------------------------------------------
# Small change to the times in the edgelist
edgelist <- data.frame(
  time = c(1, 2, 3, 4, 5, 5, 5, 6, 7, 8),
  actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
  actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
)

reh <- remify::remify(edgelist, model = "tie", directed = FALSE, 
  riskset = "active")

# Selection of effects that have unique underlying cpp functions
effects <- ~ degreeMin() + sp() 

# Method = "pt"
pt_stats <- remstats(reh, tie_effects = effects, start = 1)
riskset <- attr(pt_stats, "riskset")

# degreeMin
degreeMin <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 1, 0, 0, 0, 0),
  c(1, 1, 0, 1, 0, 0),
  c(1, 2, 0, 1, 0, 0),
  c(2, 3, 0, 2, 0, 0),
  c(4, 4, 1, 4, 1, 1),
  c(4, 4, 1, 5, 1, 1),
  c(4, 4, 2, 6, 2, 2)
)
expect_equal(pt_stats[, , "degreeMin"], degreeMin)

# sp
sp <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 0, 0, 0, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(0, 0, 0, 1, 0, 0),
  c(1, 1, 0, 1, 0, 0),
  c(2, 2, 1, 2, 1, 0),
  c(2, 2, 1, 2, 1, 0),
  c(2, 2, 2, 3, 1, 1)
)
expect_equal(pt_stats[, , "sp"], sp)

# tests of whethed the case-control sampled versions of tomstats yields the same stats are the full analysis
# for the sampled dyads.
# here this is tested a dyadic ("tie") model with active riskset, directed events, time-sensitive (not ordinal) model
# Model: tie, manual riskset, time likelihood, undirected events, exponential memory decay

data(history, package = "remstats")
data(info, package = "remstats")

# add some events happening in same interval
history$time[7:8] <- history$time[9]
history[4,] <- history[5,]
history <- history[1:35,]

# take subset for test
start1 <- 3
stop1 <- 30

samp_num <- 5

check_sampled_equals_full <- function(effects,
																			samp_num = 10L,
																			seed = 1L,
																			tol = 1e-12,
																			attr_dyads = NULL) {

	reh <- remify::remify(edgelist = history, model = "tie",
												 riskset = "full", ordinal = TRUE, directed = FALSE)

	ts_samp <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1, memory = "decay", memory_value = 3
	)

	# reproducibility (same seed/args)
	ts_samp2 <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1, memory = "decay", memory_value = 3
	)
	expect_equal(ts_samp, ts_samp2, tol = tol)
	expect_equal(attr(ts_samp, "sample_map"), attr(ts_samp2, "sample_map"))

	ts_full <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads,
		sampling = FALSE, start = start1, stop = stop1, memory = "decay", memory_value = 3
	)

	sample_map <- attr(ts_samp, "sample_map")
	expect_true(!is.null(sample_map))

	riskset <- attr(ts_full, "riskset")
	expect_true(!is.null(riskset))

	# dyad_id <- as.integer(riskset[, ncol(riskset)])
	# dyad_id_key <- if (min(dyad_id, na.rm = TRUE) == 0L) dyad_id else (dyad_id - 1L)
	# col_index_by_dyad <- setNames(seq_along(dyad_id_key), dyad_id_key)

	M <- dim(ts_samp)[1]
	S <- dim(ts_samp)[2]
	#statnum <- 1

	for (m in seq_len(M)) {
		for (s in seq_len(S)) {
			d <- as.integer(sample_map[m, s])
			j <- d #unname(col_index_by_dyad[as.character(d)])
			expect_true(!is.na(j))

			expect_equal(
				as.numeric(ts_samp[m, s, ]),
				as.numeric(ts_full[m, j, ]),
				tol = tol
			)

			x <- as.numeric(ts_samp[m, s, ])
			y <- as.numeric(ts_full[m, j, ])

			if (max(abs(x - y)) > tol) {
				cat("FAIL at m=", m, " s=", s, " d=", d, " j=", j, "\n")
				cat("samp:", paste(x, collapse=", "), "\n")
				cat("full:", paste(y, collapse=", "), "\n")
				break
			}

		}
	}

	invisible(TRUE)
}

effects_exo <- ~
	same("sex", info) +
	difference("age", info) +
	average("extraversion", info) +
	minimum("age", info) +
	maximum("age", info)

# Optional: dyad covariates shipped with remstats (skip if not present)
dyad_cov_tests <- function() {
	if (exists("both_male_wide", where = asNamespace("remstats"), inherits = FALSE)) {
		data(both_male_wide, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_wide),
			samp_num = samp_num, seed = seed, tol = tol,
			attr_dyads = both_male_wide
		)
	}
	if (exists("both_male_long", where = asNamespace("remstats"), inherits = FALSE)) {
		data(both_male_long, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_long),
			samp_num = samp_num, seed = seed, tol = tol,
			attr_dyads = both_male_long
		)
	}
	invisible(TRUE)
}

tests <- list(
	inertia_recip = ~ inertia(),
	degrees       = ~ totaldegreeDyad() + degreeMin() + degreeMax() + degreeDiff(),
	triads        = ~ sp(),
	recency       = ~ recencyContinue(),
	pshifts       = ~ psABAY() + psABAB(),
	exo_stats     = effects_exo
)

for (nm in names(tests)) {
	if (is.function(tests[[nm]])) {
		tests[[nm]]()
	} else {
		check_sampled_equals_full(tests[[nm]])
	}
}

# --- userStat test for undirected + ordinal (remify2) ---
make_actor_event_userstat <- function(reh, actor_id, riskset_df) {
	# M = number of unique ordinal time points used by tomstats under method="pt"
	time_points <- sort(unique(reh$edgelist$time))
	M <- length(time_points)
	
	# D = number of dyads in the (base) riskset
	D <- nrow(riskset_df)
	
	involved_at_time <- vapply(time_points, function(tt) {
		any(reh$edgelist$time == tt &
					(reh$edgelist$actor1 == actor_id | reh$edgelist$actor2 == actor_id))
	}, logical(1))
	
	matrix(as.numeric(involved_at_time), nrow = M, ncol = D)
}
userstat_actor_test <- function(start1,stop1) {
	reh <- remify::remify(
		edgelist = history, model = "tie",
		riskset = "full", ordinal = TRUE, directed = FALSE
	)
	actor101Events <- which(history$actor1 == "101" | history$actor2 == "101")
	actor101_stat <- t(sapply(seq_len(nrow(history)), function(i) {
		rep(i %in% actor101Events, reh$D)
	}))#[start1:stop1,]
	effects_us <- ~ userStat(x = actor101_stat, variableName = "actor101event")
	
	check_sampled_equals_full(
		effects_us,
		samp_num = 5L, seed = 1L, tol = 1e-12
	)
}



# dim(ts_samp)
# dim(ts_full)
# m <- 31
# head(history)
# head(reh$edgelist)
# 
# 
# exp(-(3 - 2) * log(2) / 3) * 1.64
# exp(-(3 - 1) * log(2) / 3) * 1.33
# 
# ev_idx <- events_by_row[[5]]
# c(ed$actor1[ev_idx], ed$actor2[ev_idx])
# keys <- make_key(ed$actor1[ev_idx], ed$actor2[ev_idx])
# key_to_base[keys]   # should be 26

# Create example remstats objects
edgelist <- data.frame(
  time = 1:4,
  actor1 = sample(1:3, 4, replace = T),
  actor2 = sample(4:6, 4, replace = T)
)

# --- Check for the tie-oriented model
reh <- remify::remify(edgelist, model = "tie", actors = 1:6)

# Create example remstats objects
rs1 <- remstats(reh = reh, 
  tie_effects = ~ inertia() + reciprocity() + otp())
rs2 <- remstats(reh = reh, tie_effects = ~ outdegreeSender() + reciprocity())
rs3 <- remstats(reh = reh, tie_effects = ~ reciprocity())

# Combine remstats objects
expect_warning(combined_stats <- bind_remstats(rs1, rs2, rs3))

# Check dimensions
expect_equal(dim(combined_stats), c(3, 30, 5))

# Check model attribute
expect_equal(attr(combined_stats, "model"), "tie")

# Check formula attribute
expect_equal(attr(combined_stats, "formula"), "Combined remstats object")

# Check riskset attribute
expect_equal(attr(combined_stats, "riskset"), attr(rs1, "riskset"))

# Check adjmat attribute
expect_equal(attr(combined_stats, "adjmat"), attr(rs1, "adjmat"))

# Check statistics
if (at_home()) {
  expected_stats <- abind::abind(rs1, rs2[, , "outdegreeSender"], along = 3)
  attributes(combined_stats) <- NULL
  attributes(expected_stats) <- NULL
  expect_true(identical(expected_stats, combined_stats))
}

# Check different remstats objects
# TODO(actor-model): reh_actor <- remify::remify(edgelist, model = "actor", actors = 1:6)
# TODO(actor-model): rs3 <- remstats(reh = reh, tie_effects = ~ inertia())
# TODO(actor-model): rs4 <- remstats(reh = reh_actor, receiver_effects = ~ otp())
# TODO(actor-model): expect_error(bind_remstats(rs3, rs4), "All objects should be of class")

# TODO(actor-model): # --- Check for the actor-oriented model
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor", actors = 1:6)
# TODO(actor-model): 
# TODO(actor-model): # Create example remstats objects
# TODO(actor-model): rs1 <- remstats(reh = reh, 
# TODO(actor-model):   receiver_effects = ~ inertia() + reciprocity() + otp())
# TODO(actor-model): rs2 <- remstats(reh = reh, 
# TODO(actor-model):   receiver_effects = ~ reciprocity() + itp())
# TODO(actor-model): rs3 <- remstats(reh = reh, 
# TODO(actor-model):   receiver_effects = ~ indegreeReceiver())
# TODO(actor-model): 
# TODO(actor-model): # Combine remstats objects
# TODO(actor-model): expect_warning(combined_stats <- bind_remstats(rs1, rs2, rs3))
# TODO(actor-model): 
# TODO(actor-model): # Check statistics
# TODO(actor-model): if (at_home()) {
# TODO(actor-model):   expected_stats <- abind::abind(rs1$receiver_stats, 
# TODO(actor-model):     rs2$receiver_stats[,,"itp"], along = 3)
# TODO(actor-model):   expected_stats <- abind::abind(expected_stats, 
# TODO(actor-model):     rs3$receiver_stats[,,"indegreeReceiver"], along = 3)
# TODO(actor-model):   attributes(combined_stats$receiver_stats) <- NULL
# TODO(actor-model):   attributes(expected_stats) <- NULL
# TODO(actor-model):   expect_true(identical(expected_stats, combined_stats$receiver_stats))
# TODO(actor-model): }
# TODO(actor-model): 
# TODO(actor-model): # Create example remstats objects
# TODO(actor-model): rs4 <- remstats(reh = reh, 
# TODO(actor-model):   sender_effects = ~ outdegreeSender() + recencySendSender())
# TODO(actor-model): rs5 <- remstats(reh = reh, 
# TODO(actor-model):   sender_effects = ~ indegreeSender() + recencyReceiveSender())
# TODO(actor-model): rs6 <- remstats(reh = reh,
# TODO(actor-model):   sender_effects = ~ totaldegreeSender())
# TODO(actor-model): 
# TODO(actor-model): # Combine remstats objects
# TODO(actor-model): expect_warning(combined_stats <- bind_remstats(rs4, rs5, rs6))
# TODO(actor-model): 
# TODO(actor-model): # Check statistics
# TODO(actor-model): if (at_home()) {
# TODO(actor-model):   expected_stats <- abind::abind(rs4$sender_stats, 
# TODO(actor-model):     rs5$sender_stats[,,-1], along = 3)
# TODO(actor-model):   expected_stats <- abind::abind(expected_stats, rs6$sender_stats[,,-1], along = 3)
# TODO(actor-model):   attributes(combined_stats$sender_stats) <- NULL
# TODO(actor-model):   attributes(expected_stats) <- NULL
# TODO(actor-model):   expect_true(identical(expected_stats, combined_stats$sender_stats))
# TODO(actor-model): }
# TODO(actor-model): 
# TODO(actor-model): # Create example remstats objects
# TODO(actor-model): rs7 <- remstats(reh = reh, 
# TODO(actor-model):   receiver_effects = ~ inertia() + reciprocity() + otp(),
# TODO(actor-model):   sender_effects = ~ outdegreeSender() + recencySendSender())
# TODO(actor-model): rs8 <- remstats(reh = reh, 
# TODO(actor-model):   receiver_effects = ~ reciprocity() + itp(),
# TODO(actor-model):   sender_effects = ~ indegreeSender() + recencyReceiveSender())
# TODO(actor-model): rs9 <- remstats(reh = reh,
# TODO(actor-model):   receiver_effects = ~ indegreeReceiver(),
# TODO(actor-model):   sender_effects = ~ totaldegreeSender())
# TODO(actor-model): 
# TODO(actor-model): # Combine remstats objects
# TODO(actor-model): expect_warning(combined_stats <- bind_remstats(rs7, rs8, rs9))
# TODO(actor-model): 
# TODO(actor-model): # Check statistics
# TODO(actor-model): if (at_home()) {
# TODO(actor-model):   # Sender effects
# TODO(actor-model):   expected_stats <- abind::abind(rs7$sender_stats, 
# TODO(actor-model):     rs8$sender_stats[,,-1], along = 3)
# TODO(actor-model):   expected_stats <- abind::abind(expected_stats, 
# TODO(actor-model):     rs9$sender_stats[,,-1], along = 3)
# TODO(actor-model):   attributes(combined_stats$sender_stats) <- NULL
# TODO(actor-model):   attributes(expected_stats) <- NULL
# TODO(actor-model):   expect_true(identical(expected_stats, combined_stats$sender_stats))
# TODO(actor-model):   
# TODO(actor-model):   # Receiver effects
# TODO(actor-model):   expected_stats <- abind::abind(rs7$receiver_stats, 
# TODO(actor-model):     rs8$receiver_stats[,,"itp"], along = 3)
# TODO(actor-model):   expected_stats <- abind::abind(expected_stats, 
# TODO(actor-model):     rs9$receiver_stats[,,"indegreeReceiver"], along = 3)
# TODO(actor-model):   attributes(combined_stats$receiver_stats) <- NULL
# TODO(actor-model):   attributes(expected_stats) <- NULL
# TODO(actor-model):   expect_true(identical(expected_stats, combined_stats$receiver_stats))
# TODO(actor-model): }

# Condition 1: Directed events, tie-oriented model with active risk set

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

# Actor info
info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

info2 <- data.frame(
  name = 1:3,
  time = rep(3, 3),
  x1 = c(100, 200, 300),
  x2 = c(1, 1, 0)
)

info <- rbind(info, info2)

# Statistics
reh <- remify::remify(edgelist, model = "tie", riskset = "active")
effects <-	~ send(variable = "x1") + receive(variable = "x1") + 
  average(variable = "x1") + difference(variable = "x1") +
  maximum(variable = "x1") + minimum(variable = "x1") +
  same(variable = "x2") 

stats <- remstats(reh, tie_effects = effects, attr_actors = info)
riskset <- attr(stats, "riskset")

# Baseline
expect_equal(stats[, , "baseline"], matrix(1, nrow = nrow(edgelist), ncol = nrow(riskset)))

# send
send <- rbind(
  c(10, 10, 20, 20, 30),
  c(10, 10, 20, 20, 30),
  c(100, 100, 200, 200, 300),
  c(100, 100, 200, 200, 300),
  c(100, 100, 200, 200, 300)
)
expect_equal(stats[, , "send_x1"], send)

# receive
receive <- rbind(
  c(20, 30, 10, 30, 20),
  c(20, 30, 10, 30, 20),
  c(200, 300, 100, 300, 200),
  c(200, 300, 100, 300, 200),
  c(200, 300, 100, 300, 200)
)
expect_equal(stats[, , "receive_x1"], receive)

# average
average <- rbind(
  c(15, 20, 15, 25, 25),
  c(15, 20, 15, 25, 25),
  c(150, 200, 150, 250, 250),
  c(150, 200, 150, 250, 250),
  c(150, 200, 150, 250, 250)
)
expect_equal(stats[, , "average_x1"], average)

# difference
difference <- rbind(
  c(10, 20, 10, 10, 10),
  c(10, 20, 10, 10, 10),
  c(100, 200, 100, 100, 100),
  c(100, 200, 100, 100, 100),
  c(100, 200, 100, 100, 100)
)
expect_equal(stats[, , "difference_x1"], difference)

# maximum 
maximum <- rbind(
  c(20, 30, 20, 30, 30),
  c(20, 30, 20, 30, 30),
  c(200, 300, 200, 300, 300),
  c(200, 300, 200, 300, 300),
  c(200, 300, 200, 300, 300)
)
expect_equal(stats[, , "maximum_x1"], maximum)

# minimum
minimum <- rbind(
  c(10, 10, 10, 20, 20),
  c(10, 10, 10, 20, 20),
  c(100, 100, 100, 200, 200),
  c(100, 100, 100, 200, 200),
  c(100, 100, 100, 200, 200)
)
expect_equal(stats[, , "minimum_x1"], minimum)

# same
same <- rbind(
  c(0, 0, 0, 1, 1),
  c(0, 0, 0, 1, 1),
  c(1, 0, 1, 0, 0),
  c(1, 0, 1, 0, 0),
  c(1, 0, 1, 0, 0)
)
expect_equal(stats[, , "same_x2"], same)

# test difference absolute = FALSE
effects3 <- ~ difference(variable = "x1", absolute = FALSE)
stats3 <- remstats(reh, tie_effects = effects3, attr_actors = info)

difference <- rbind(
  c(-10, -20, 10, -10, 10),
  c(-10, -20, 10, -10, 10),
  c(-100, -200, 100, -100, 100),
  c(-100, -200, 100, -100, 100),
  c(-100, -200, 100, -100, 100)
)
expect_equal(stats3[, , "difference_x1"], difference)

# test standardization
std_effects <- ~
  send(variable = "x1", scaling = "std") + 
  receive(variable = "x1", scaling = "std") + 
  average(variable = "x1", scaling = "std") + 
  difference(variable = "x1", scaling = "std") + 
  maximum(variable = "x1", scaling = "std") + 
  minimum(variable = "x1", scaling = "std") 
std_stats <- remstats(reh, tie_effects = std_effects, attr_actors = info)

sapply(2:dim(std_stats)[3], function(p) {
  stat_name <- dimnames(std_stats)[[3]][p]
  scaled_original <- t(apply(stats[,,stat_name], 1, scale))
  scaled_original[which(apply(stats[,,stat_name], 1, sd) == 0),] <- 
    rep(0, ncol(stats))
  expect_equal(std_stats[,,stat_name], scaled_original)
})

# Test method -------------------------------------------------------------
# Small change to the times in the edgelist
edgelist <- data.frame(
  time = c(1, 2, 3, 3, 4),
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

reh <- remify::remify(edgelist, model = "tie", riskset = "active")

# Method = "pt"
# Selection of effects that have unique underlying cpp functions
effects <- ~ send(variable = "x1") +
  average(variable = "x1") 

pt_stats <- remstats(reh, tie_effects = effects, attr_actors = info, 
  method = "pt")
riskset <- attr(pt_stats, "riskset")

# send
send <- rbind(
  c(10, 10, 20, 20, 30),
  c(10, 10, 20, 20, 30),
  c(100, 100, 200, 200, 300),
  c(100, 100, 200, 200, 300)
)
expect_equal(pt_stats[, , "send_x1"], send)

# average
average <- rbind(
  c(15, 20, 15, 25, 25),
  c(15, 20, 15, 25, 25),
  c(150, 200, 150, 250, 250),
  c(150, 200, 150, 250, 250)
)
expect_equal(pt_stats[, , "average_x1"], average)

# Method = "pe"
# Selection of effects that have unique underlying cpp functions
effects <- ~ send(variable = "x1") +
  average(variable = "x1")  

pe_stats <- remstats(reh, tie_effects = effects, attr_actors = info, 
  method = "pe")
riskset <- attr(pt_stats, "riskset")

# send
send <- rbind(
  c(10, 10, 20, 20, 30),
  c(10, 10, 20, 20, 30),
  c(100, 100, 200, 200, 300),
  c(100, 100, 200, 200, 300),
  c(100, 100, 200, 200, 300)
)
expect_equal(pe_stats[, , "send_x1"], send)

# average
average <- rbind(
  c(15, 20, 15, 25, 25),
  c(15, 20, 15, 25, 25),
  c(150, 200, 150, 250, 250),
  c(150, 200, 150, 250, 250),
  c(150, 200, 150, 250, 250)
)
expect_equal(pe_stats[, , "average_x1"], average)

# # Condition 9: Actor-oriented model
# 
# # Small edgelist
# edgelist <- data.frame(
#   time = 1:5,
#   actor1 = c(1, 1, 2, 2, 3),
#   actor2 = c(2, 3, 1, 3, 2)
# )
# 
# # Actor info
# info <- data.frame(
#   name = 1:3,
#   time = rep(0, 3),
#   x1 = c(10, 20, 30),
#   x2 = c(0, 1, 1)
# )
# 
# info2 <- data.frame(
#   name = 1:3,
#   time = rep(3, 3),
#   x1 = c(100, 200, 300),
#   x2 = c(1, 1, 0)
# )
# 
# info <- rbind(info, info2)
# 
# # Statistics
# reh <- remify::remify(edgelist, model = "actor")
# sender_effects <- ~ send(variable = "x1") 
# receiver_effects <- ~ receive(variable = "x1") + 
#   average(variable = "x1") + difference(variable = "x1") + 
#   same(variable = "x2") 
# stats <- remstats(reh = reh,
#   sender_effects = sender_effects,
#   receiver_effects = receiver_effects, 
#   attr_actors = info
# )
# sender_stats <- stats$sender_stats
# receiver_stats <- stats$receiver_stats
# actors <- attr(reh, "dictionary")$actors
# 
# # baseline
# expect_equal(sender_stats[, , "baseline"], matrix(1, nrow = nrow(edgelist), ncol = nrow(actors)))
# 
# # send
# send <- rbind(
#   c(10, 20, 30),
#   c(10, 20, 30),
#   c(100, 200, 300),
#   c(100, 200, 300),
#   c(100, 200, 300)
# )
# expect_equal(sender_stats[, , "send_x1"], send)
# 
# # receive
# receive <- rbind(
#   c(10, 20, 30),
#   c(10, 20, 30),
#   c(100, 200, 300),
#   c(100, 200, 300),
#   c(100, 200, 300)
# )
# expect_equal(receiver_stats[, , "receive_x1"], receive)
# 
# # average
# average <- rbind(
#   c(10, 15, 20),
#   c(10, 15, 20),
#   c(150, 200, 250),
#   c(150, 200, 250),
#   c(200, 250, 300)
# )
# expect_equal(receiver_stats[, , "average_x1"], average)
# 
# # difference
# difference <- rbind(
#   c(0, 10, 20),
#   c(0, 10, 20),
#   c(100, 0, 100),
#   c(100, 0, 100),
#   c(200, 100, 0)
# )
# expect_equal(receiver_stats[, , "difference_x1"], difference)
# 
# # same 
# same <- rbind(
#   c(1, 0, 0),
#   c(1, 0, 0),
#   c(1, 1, 0),
#   c(1, 1, 0),
#   c(0, 0, 1)
# )
# expect_equal(receiver_stats[, , "same_x2"], same)
# 
# # test standardization
# reh <- remify::remify(edgelist, model = "actor")
# std_sender_effects <- ~ send(variable = "x1", scaling = "std")
# std_receiver_effects <- ~ receive(variable = "x1", scaling = "std") + 
#   average(variable = "x1", scaling = "std") + 
#   difference(variable = "x1", scaling = "std") 
# std_stats <- remstats(reh = reh,
#   sender_effects = std_sender_effects,
#   receiver_effects = std_receiver_effects, 
#   attr_actors = info
# )
# std_sender_stats <- std_stats$sender_stats
# std_receiver_stats <- std_stats$receiver_stats
# 
# sapply(2:dim(std_sender_stats)[3], function(p) {
#   stat_name <- dimnames(std_sender_stats)[[3]][p]
#   scaled_original <- t(apply(sender_stats[, , stat_name], 1, scale))
#   scaled_original[which(apply(sender_stats[, , stat_name], 1, sd) == 0), ] <-
#     rep(0, ncol(sender_stats))
#   expect_equal(std_sender_stats[, , stat_name], scaled_original)
# })
# 
# sapply(2:dim(std_receiver_stats)[3], function(p) {
#   stat_name <- dimnames(std_receiver_stats)[[3]][p]
#   scaled_original <- t(sapply(1:nrow(edgelist), function(m) {
#     stat_row <- receiver_stats[m,, stat_name]
#     row_mean <- mean(stat_row[-edgelist[m,2]])
#     row_sd <- sd(stat_row[-edgelist[m,2]])
#     if(row_sd == 0) {
#       stat_row <- rep(0, ncol(receiver_stats))
#     }  else {
#       stat_row <- ((stat_row - row_mean) / row_sd)
#       stat_row[edgelist[m,2]] <- 0
#     }    
#     stat_row
#   }))
#   expect_equal(std_receiver_stats[, , stat_name], scaled_original)
# })
# 
# # Test method -------------------------------------------------------------
# # Small change to the times in the edgelist
# edgelist <- data.frame(
#   time = c(1, 2, 3, 3, 4),
#   actor1 = c(1, 1, 2, 2, 3),
#   actor2 = c(2, 3, 1, 3, 2)
# )
# 
# reh <- remify::remify(edgelist, model = "actor", riskset = "active")
# 
# # Method = "pt"
# # Selection of effects that have unique underlying cpp functions
# sender_effects <- ~ send(variable = "x1") 
# receiver_effects <- ~ receive(variable = "x1") + 
#   average(variable = "x1") 
# pt_stats <- remstats(reh = reh,
#   sender_effects = sender_effects,
#   receiver_effects = receiver_effects, 
#   attr_actors = info,
#   method = "pt"
# )
# sender_stats <- pt_stats$sender_stats
# receiver_stats <- pt_stats$receiver_stats
# actors <- attr(reh, "dictionary")$actors
# 
# # send
# send <- rbind(
#   c(10, 20, 30),
#   c(10, 20, 30),
#   c(100, 200, 300),
#   c(100, 200, 300)
# )
# expect_equal(sender_stats[, , "send_x1"], send)
# 
# # receive
# receive <- rbind(
#   c(10, 20, 30),
#   c(10, 20, 30),
#   c(100, 200, 300),
#   c(100, 200, 300),
#   c(100, 200, 300)
# )
# expect_equal(receiver_stats[, , "receive_x1"], receive)
# 
# # average
# average <- rbind(
#   c(10, 15, 20),
#   c(10, 15, 20),
#   c(150, 200, 250),
#   c(150, 200, 250),
#   c(200, 250, 300)
# )
# expect_equal(receiver_stats[, , "average_x1"], average)
# 
# # Method = "pe"
# 
# # Selection of effects that have unique underlying cpp functions
# sender_effects <- ~ send(variable = "x1") 
# receiver_effects <- ~ receive(variable = "x1") + 
#   average(variable = "x1") 
# pe_stats <- remstats(reh = reh,
#   sender_effects = sender_effects,
#   receiver_effects = receiver_effects, 
#   attr_actors = info,
#   method = "pe"
# )
# sender_stats <- pe_stats$sender_stats
# receiver_stats <- pe_stats$receiver_stats
# actors <- attr(reh, "dictionary")$actors
# 
# # send
# send <- rbind(
#   c(10, 20, 30),
#   c(10, 20, 30),
#   c(100, 200, 300),
# 	c(100, 200, 300),
#   c(100, 200, 300)
# )
# expect_equal(sender_stats[, , "send_x1"], send)
# 
# # receive
# receive <- rbind(
#   c(10, 20, 30),
#   c(10, 20, 30),
#   c(100, 200, 300),
#   c(100, 200, 300),
#   c(100, 200, 300)
# )
# expect_equal(receiver_stats[, , "receive_x1"], receive)
# 
# # average
# average <- rbind(
#   c(10, 15, 20),
#   c(10, 15, 20),
#   c(150, 200, 250),
#   c(150, 200, 250),
#   c(200, 250, 300)
# )
# expect_equal(receiver_stats[, , "average_x1"], average)

# Spot-check: aomstats consider_type="separate" vs tomstats consider_type=TRUE
#
# Full event history (9915 events) so stats are rich, but start=9900 so only
# ~15 output rows are computed. All receivers checked at all output rows.
#
# Effects covered (all directed, consider_type="separate" vs TRUE):
#   - inertia, reciprocity
#   - degrees: indegreeSender, outdegreeSender, indegreeReceiver, outdegreeReceiver
#   - triads:  otp, itp, osp, isp
#   - pshifts: psABBA, psABBY, psABXA, psABXB, psABXY, psABAY, psABAB
#   - rrank:   rrankSend, rrankReceive
#   - recency: recencySendSender, recencySendReceiver,
#              recencyReceiveSender, recencyReceiveReceiver
#
# Memory types: full, decay, window, interval

library(remify)
library(remstats)
library(tinytest)

data(randomREH, package = "remify")
el <- randomREH$edgelist

reh_tie <- remify(el, actors = randomREH$actors, directed = TRUE,
                   origin = randomREH$origin, model = "tie",
                   extend_riskset_by_type = FALSE)
reh_actor <- remify(el, actors = randomREH$actors, directed = TRUE,
                     origin = randomREH$origin, model = "actor")

el_reh <- reh_tie$edgelist
N      <- length(randomREH$actors)
dict   <- reh_tie$meta$dictionary$actors[, 1]
dyad_col <- function(s, j) (s - 1L) * (N - 1L) + j - as.integer(j > s)

START <- 9900L
tol   <- 1e-7

run_check <- function(tie_formula, aom_formula, memory, memory_value, label) {
  cat("\n---", label, "---\n")

  ts_tie <- tomstats(
    tie_formula, reh = reh_tie, sampling = FALSE,
    memory = memory, memory_value = memory_value,
    start = START
  )

  ts_aom <- aomstats(
    reh = reh_actor, receiver_effects = aom_formula,
    memory = memory, memory_value = memory_value,
    start = START
  )

  slices <- dimnames(ts_tie)[[3]]
  slices <- slices[slices != "baseline"]
  M_out  <- dim(ts_aom$receiver_stats)[1]

  for (m in seq_len(M_out)) {
    event_row   <- START - 1L + m
    sender_name <- as.character(el_reh$actor1[event_row])
    sender_id   <- which(dict == sender_name)
    for (j in seq_len(N)) {
      if (j == sender_id) next
      d <- dyad_col(sender_id, j)
      for (sl in slices) {
        expect_equal(
          ts_aom$receiver_stats[m, j, sl],
          ts_tie[m, d, sl],
          tolerance = tol,
          info = sprintf("%s m=%d j=%d slice=%s", label, m, j, sl)
        )
      }
    }
  }
  cat("  done:", length(slices), "slices x", M_out, "events x", N - 1L,
      "receivers =", length(slices) * M_out * (N - 1L), "checks\n")
}

# ---------------------------------------------------------------------------
# Formula groups
# ---------------------------------------------------------------------------

f_base_tie <- ~ inertia(consider_type = TRUE) +
                 reciprocity(consider_type = TRUE)
f_base_aom <- ~ inertia(consider_type = "separate") +
                 reciprocity(consider_type = "separate")

f_deg_tie  <- ~ indegreeReceiver(consider_type = TRUE) +
                 outdegreeReceiver(consider_type = TRUE) +
                 totaldegreeReceiver(consider_type = TRUE)
f_deg_aom  <- ~ indegreeReceiver(consider_type = "separate") +
                 outdegreeReceiver(consider_type = "separate") +
                 totaldegreeReceiver(consider_type = "separate")

f_triad_tie <- ~ otp(consider_type = TRUE) + itp(consider_type = TRUE) +
                  osp(consider_type = TRUE) + isp(consider_type = TRUE)
f_triad_aom <- ~ otp(consider_type = "separate") + itp(consider_type = "separate") +
                  osp(consider_type = "separate") + isp(consider_type = "separate")

f_ps_tie <- ~ psABBA(consider_type = TRUE) + psABBY(consider_type = TRUE) +
               psABXA(consider_type = TRUE) + psABXB(consider_type = TRUE) +
               psABXY(consider_type = TRUE) + psABAY(consider_type = TRUE) +
               psABAB(consider_type = TRUE)
f_ps_aom <- ~ psABBA(consider_type = "separate") + psABBY(consider_type = "separate") +
               psABXA(consider_type = "separate") + psABXB(consider_type = "separate") +
               psABXY(consider_type = "separate") + psABAY(consider_type = "separate") +
               psABAB(consider_type = "separate")

f_rr_tie <- ~ rrankSend(consider_type = TRUE) + rrankReceive(consider_type = TRUE)
f_rr_aom <- ~ rrankSend(consider_type = "separate") + rrankReceive(consider_type = "separate")

f_rec_tie <- ~ recencySendReceiver(consider_type = TRUE) +
                recencyReceiveReceiver(consider_type = TRUE) +
                recencyContinue(consider_type = TRUE)
f_rec_aom <- ~ recencySendReceiver(consider_type = "separate") +
                recencyReceiveReceiver(consider_type = "separate") +
                recencyContinue(consider_type = "separate")

# ---------------------------------------------------------------------------
# Full memory
# ---------------------------------------------------------------------------
run_check(f_base_tie,  f_base_aom,  "full", NA, "full / base")
run_check(f_deg_tie,   f_deg_aom,   "full", NA, "full / degrees")
run_check(f_triad_tie, f_triad_aom, "full", NA, "full / triads")
run_check(f_ps_tie,    f_ps_aom,    "full", NA, "full / pshifts")
run_check(f_rr_tie,    f_rr_aom,    "full", NA, "full / rrank")
run_check(f_rec_tie,   f_rec_aom,   "full", NA, "full / recency")

# ---------------------------------------------------------------------------
# Decay memory — primary regression test for the bug fix                     
# ---------------------------------------------------------------------------
run_check(f_base_tie,  f_base_aom,  "decay", 100, "decay / base")
run_check(f_deg_tie,   f_deg_aom,   "decay", 100, "decay / degrees")
run_check(f_triad_tie, f_triad_aom, "decay", 100, "decay / triads")
run_check(f_ps_tie,    f_ps_aom,    "decay", 100, "decay / pshifts")
run_check(f_rr_tie,    f_rr_aom,    "decay", 100, "decay / rrank")
run_check(f_rec_tie,   f_rec_aom,   "decay", 100, "decay / recency")

# ---------------------------------------------------------------------------
# Window and interval                                                        
# ---------------------------------------------------------------------------
run_check(f_base_tie, f_base_aom, "window",   1000,         "window / base")
run_check(f_deg_tie,  f_deg_aom,  "window",   1000,         "window / degrees")
run_check(f_base_tie, f_base_aom, "interval", c(500, 2000), "interval / base")
run_check(f_deg_tie,  f_deg_aom,  "interval", c(500, 2000), "interval / degrees")

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

# Event weights
weights <- c(0.15, 0.25, 0.35, 0.45, 0.55)
edgelist$weight <- weights

# Test for tie-oriented model (weights are used in "adjmat", only need to test for one statistic)
reh <- remify(edgelist, model = "tie")
effects <- ~ inertia()
stats <- remstats(reh, tie_effects = effects, start = 1)
riskset <- attr(stats, "riskset")

inertia <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0.15, 0, 0, 0, 0, 0),
  c(0.15, 0.25, 0, 0, 0, 0),
  c(0.15, 0.25, 0.35, 0, 0, 0),
  c(0.15, 0.25, 0.35, 0.45, 0, 0)
)
expect_equal(stats[, , "inertia"], inertia)

# TODO(actor-model): actor-oriented model weights test pending aomstats implementation
# reh <- remify(edgelist, model = "actor")
# sender_effects <- ~
#   indegreeSender() + outdegreeSender() + totaldegreeSender()
# receiver_effects <- ~
#   indegreeReceiver() + outdegreeReceiver() + totaldegreeReceiver() +
#     inertia() + reciprocity() +
#     isp() + itp() + osp() + otp()
# stats <- remstats(reh,
#   sender_effects = sender_effects,
#   receiver_effects = receiver_effects
# )
# sender_stats <- stats$sender_stats
# receiver_stats <- stats$receiver_stats
# actors <- reh$meta$dictionary$actors
# ... (weight value checks for each statistic)

# Condition 2: Undirected events, tie-oriented model with active risk set

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 4, 2)
)

# Actor info
info <- data.frame(
  name = 1:4,
  time = rep(0, 4),
  x1 = c(10, 20, 30, 40),
  x2 = c(0, 1, 1, 0)
)

info2 <- data.frame(
  name = 1:4,
  time = rep(3, 4),
  x1 = c(100, 200, 300, 400),
  x2 = c(1, 1, 0, 0)
)

info <- rbind(info, info2)

# Statistics
reh <- remify::remify(edgelist, model = "tie", directed = FALSE, 
  riskset = "active")
effects <- ~ average(variable = "x1") + difference(variable = "x1") + 
  maximum(variable = "x1") + minimum(variable = "x1") +
  same(variable = "x2") 
stats <- remstats(reh, tie_effects = effects, attr_actors = info)
riskset <- attr(stats, "riskset")

# Baseline
expect_equal(stats[, , "baseline"], matrix(1, nrow = nrow(edgelist), ncol = nrow(riskset)))

# average
average <- rbind(
  c(15, 20, 25, 30),
  c(15, 20, 25, 30),
  c(150, 200, 250, 300),
  c(150, 200, 250, 300),
  c(150, 200, 250, 300)
)
expect_equal(stats[, , "average_x1"], average)

# difference
difference <- rbind(
  c(10, 20, 10, 20),
  c(10, 20, 10, 20),
  c(100, 200, 100, 200),
  c(100, 200, 100, 200),
  c(100, 200, 100, 200)
)
expect_equal(stats[, , "difference_x1"], difference)

# maximum 
maximum <- rbind(
  c(20, 30, 30, 40),
  c(20, 30, 30, 40),
  c(200, 300, 300, 400),
  c(200, 300, 300, 400),
  c(200, 300, 300, 400)
)
expect_equal(stats[, , "maximum_x1"], maximum)

# minimum
minimum <- rbind(
  c(10, 10, 20, 20),
  c(10, 10, 20, 20),
  c(100, 100, 200, 200),
  c(100, 100, 200, 200),
  c(100, 100, 200, 200)
)
expect_equal(stats[, , "minimum_x1"], minimum)

# same
same <- rbind(
  c(0, 0, 1, 0),
  c(0, 0, 1, 0),
  c(1, 0, 0, 0),
  c(1, 0, 0, 0),
  c(1, 0, 0, 0)
)
expect_equal(stats[, , "same_x2"], same)

# test standardization
std_effects <- ~
  average(variable = "x1", scaling = "std") + 
  difference(variable = "x1", scaling = "std") + 
  maximum(variable = "x1", scaling = "std") + 
  minimum(variable = "x1", scaling = "std") 
std_stats <- remstats(reh, tie_effects = std_effects, attr_actors = info)

sapply(2:dim(std_stats)[3], function(p) {
  stat_name <- dimnames(std_stats)[[3]][p]
  scaled_original <- t(apply(stats[,,stat_name], 1, scale))
  scaled_original[which(apply(stats[,,stat_name], 1, sd) == 0),] <- 
    rep(0, ncol(stats))
  expect_equal(std_stats[,,stat_name], scaled_original)
})

# tests of whethed the case-control sampled versions of tomstats yields the same stats are the full analysis
# for the sampled dyads.
# here this is tested a dyadic ("tie") model with active riskset, directed events, time-sensitive (not ordinal) model
# memroy decay
# with event types

library(tinytest)

data(history, package = "remstats")
data(info, package = "remstats")

# add some events happening in same interval
history$time[7:8] <- history$time[9]
history[4,] <- history[5,]
history <- history[1:33,]

# take subset for test
start1 <- 2
stop1 <- 20

check_sampled_equals_full <- function(effects,
																			samp_num = 5L,
																			seed = 1L,
																			tol = 1e-12,
																			attr_dyads = NULL) {
	
	reh <- remify::remify(edgelist = history, model = "tie", riskset = "manual", manual.riskset = history[1:15,2:3])
	
	ts_samp <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads, memory = "decay", memory_value = 1000,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1
	)
	
	# reproducibility (same seed/args)
	ts_samp2 <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads, memory = "decay", memory_value = 1000,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1
	)
	expect_equal(ts_samp, ts_samp2, tol = tol)
	expect_equal(attr(ts_samp, "sample_map"), attr(ts_samp2, "sample_map"))
	
	ts_full <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads, memory = "decay", memory_value = 1000,
		sampling = FALSE, start = start1, stop = stop1
	)
	
	sample_map <- attr(ts_samp, "sample_map")
	expect_true(!is.null(sample_map))
	
	riskset <- attr(ts_full, "riskset")
	expect_true(!is.null(riskset))
	
	#dyad_id <- as.integer(riskset[, ncol(riskset)])
	#dyad_id_key <- if (min(dyad_id, na.rm = TRUE) == 0L) dyad_id else (dyad_id - 1L)
	#col_index_by_dyad <- setNames(seq_along(dyad_id_key), dyad_id_key)
	
	M <- dim(ts_samp)[1]
	S <- dim(ts_samp)[2]
	
	for (m in seq_len(M)) {
		for (s in seq_len(S)) {
			d <- as.integer(sample_map[m, s])
			j <- d #unname(col_index_by_dyad[as.character(d)])
			expect_true(!is.na(j))
			
			expect_equal(
				as.numeric(ts_samp[m, s, ]),
				as.numeric(ts_full[m, j, ]),
				tol = tol
			)
			
			x <- as.numeric(ts_samp[m, s, ])
			y <- as.numeric(ts_full[m, j, ])
			
			if (max(abs(x - y)) > tol) {
				cat("FAIL at m=", m, " s=", s, " d=", d, " j=", j, "\n", as.character(effects))
				cat("samp:", paste(x, collapse=", "), "\n")
				cat("full:", paste(y, collapse=", "), "\n")
				break
			}
			
		}
	}
	
	invisible(TRUE)
}

tests <- list(
	inertia_recip = ~ inertia() + reciprocity(),
	inertia_recip_prop = ~ inertia(scaling="prop") + reciprocity(scaling="prop"),
	degrees       = ~ indegreeSender() + outdegreeSender() + indegreeReceiver() + outdegreeReceiver(),
	degrees_prop  = ~ indegreeSender(scaling="prop") + outdegreeSender(scaling="prop") + indegreeReceiver(scaling="prop") + outdegreeReceiver(scaling="prop"),
	triads        = ~ otp() + itp() + isp() + osp(),
	# scaling = 'std' (use ccs correction)
	recency       = ~ recencySendReceiver() + recencyReceiveReceiver() + recencyContinue(),
	pshifts       = ~ psABBA() + psABXY() + psABAY(),
	exo_send_receive   = ~ 1 + send("extraversion", info) + receive("extraversion", info),
	exo_send_receive_2 = ~ 1 + send("agreeableness", info) + receive("agreeableness", info),
	exo_same_diff      = ~ 1 + same("sex", info) + difference("age", info),
	exo_aggregate      = ~ 1 + average("extraversion", info) + minimum("age", info) + maximum("age", info),
	
	dyad_tie_wide = function() {
		data(both_male_wide, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_wide),
			attr_dyads = both_male_wide
		)
	},
	dyad_tie_long = function() {
		data(both_male_long, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_long),
			attr_dyads = both_male_long
		)
	}
)

for (nm in names(tests)) {
	if (is.function(tests[[nm]])) {
		tests[[nm]]()
	} else {
		check_sampled_equals_full(tests[[nm]])
	}
}

# m <- 1
# dim(ts_samp)
# dim(ts_full)
# head(history)
# head(reh$edgelist_id)
# riskset[sample_map[m,],]
# ts_samp[m,,]
# ts_full[m,sample_map[m,],]

# exp(- (345 - 238) * log(2) / 1000) * 1.33
# exp(- (317 - 238) * log(2) / 1000) * 1.33
# exp(- (317 - 238) * log(2) / 1000) * 1.33

# FAIL at m= 1  s= 5  d= 35  j= 35 
# samp: 1, 1.25912912954483, 0 
# full: 1, 1.33, 0 

# tests of whethed the case-control sampled versions of tomstats yields the same stats are the full analysis
# for the sampled dyads.
# here this is tested a dyadic ("tie") model with active riskset, directed events, time-sensitive (not ordinal) model
# memroy decay

data(history, package = "remstats")
data(info, package = "remstats")

# add some events happening in same interval
history$time[7:8] <- history$time[9]
history[4,] <- history[5,]
history <- history[1:25,]

# take subset for test
start1 <- 2
stop1 <- 22

check_sampled_equals_full <- function(effects,
																			samp_num = 5L,
																			seed = 1L,
																			tol = 1e-12,
																			attr_dyads = NULL) {
	
	reh <- remify::remify(edgelist = history, model = "tie", riskset = "active")
	
	ts_samp <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads, memory = "decay", memory_value = 1000,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1
	)
	
	# reproducibility (same seed/args)
	ts_samp2 <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads, memory = "decay", memory_value = 1000,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1
	)
	expect_equal(ts_samp, ts_samp2, tol = tol)
	expect_equal(attr(ts_samp, "sample_map"), attr(ts_samp2, "sample_map"))
	
	ts_full <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads, memory = "decay", memory_value = 1000,
		sampling = FALSE, start = start1, stop = stop1
	)
	
	sample_map <- attr(ts_samp, "sample_map")
	expect_true(!is.null(sample_map))
	
	riskset <- attr(ts_full, "riskset")
	expect_true(!is.null(riskset))
	
	#dyad_id <- as.integer(riskset[, ncol(riskset)])
	#dyad_id_key <- if (min(dyad_id, na.rm = TRUE) == 0L) dyad_id else (dyad_id - 1L)
	#col_index_by_dyad <- setNames(seq_along(dyad_id_key), dyad_id_key)
	
	M <- dim(ts_samp)[1]
	S <- dim(ts_samp)[2]
	
	for (m in seq_len(M)) {
		for (s in seq_len(S)) {
			d <- as.integer(sample_map[m, s])
			j <- d #unname(col_index_by_dyad[as.character(d)])
			expect_true(!is.na(j))
			
			expect_equal(
				as.numeric(ts_samp[m, s, ]),
				as.numeric(ts_full[m, j, ]),
				tol = tol
			)
			
			x <- as.numeric(ts_samp[m, s, ])
			y <- as.numeric(ts_full[m, j, ])
			
			if (max(abs(x - y)) > tol) {
				cat("FAIL at m=", m, " s=", s, " d=", d, " j=", j, "\n")
				cat("samp:", paste(x, collapse=", "), "\n")
				cat("full:", paste(y, collapse=", "), "\n")
				break
			}
			
		}
	}
	
	invisible(TRUE)
}

tests <- list(
	inertia_recip = ~ inertia() + reciprocity(),
	inertia_recip_prop = ~ inertia(scaling="prop") + reciprocity(scaling="prop"),
	degrees       = ~ indegreeSender() + outdegreeSender() + indegreeReceiver() + outdegreeReceiver(),
	degrees_prop  = ~ indegreeSender(scaling="prop") + outdegreeSender(scaling="prop") + indegreeReceiver(scaling="prop") + outdegreeReceiver(scaling="prop"),
	triads        = ~ otp() + itp() + isp() + osp(),
	# scaling = 'std' (use ccs correction)
	# add tests when events have types.
	recency       = ~ recencySendReceiver() + recencyReceiveReceiver() + recencyContinue(),
	pshifts       = ~ psABBA() + psABXY() + psABAY(),
	exo_send_receive   = ~ 1 + send("extraversion", info) + receive("extraversion", info),
	exo_send_receive_2 = ~ 1 + send("agreeableness", info) + receive("agreeableness", info),
	exo_same_diff      = ~ 1 + same("sex", info) + difference("age", info),
	exo_aggregate      = ~ 1 + average("extraversion", info) + minimum("age", info) + maximum("age", info),

	dyad_tie_wide = function() {
		data(both_male_wide, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_wide),
			attr_dyads = both_male_wide
		)
	},
	dyad_tie_long = function() {
		data(both_male_long, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_long),
			attr_dyads = both_male_long
		)
	}
)

for (nm in names(tests)) {
	if (is.function(tests[[nm]])) {
		tests[[nm]]()
	} else {
		check_sampled_equals_full(tests[[nm]])
	}
}

# m <- 1
# dim(ts_samp)
# dim(ts_full)
# head(history)
# head(reh$edgelist_id)
# ts_samp[m,,]
# ts_full[m,sample_map[m,],]
# 
# exp(- (345 - 238) * log(2) / 1000) * 1.33
# exp(- (317 - 238) * log(2) / 1000) * 1.33

# Small edgelist
edgelist <- data.frame(
  time = 1:10,
  actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
  actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
)

# Statistics
reh <- remify(edgelist, model = "tie", riskset = "active")
effects <- ~
  outdegreeSender() + outdegreeReceiver() +
    indegreeSender() + indegreeReceiver() +
    totaldegreeSender() + totaldegreeReceiver() +
    totaldegreeDyad() +
    inertia() + reciprocity() +
    isp() + itp() + osp() + otp() +
    isp(unique = TRUE) + itp(unique = TRUE) +
    osp(unique = TRUE) + otp(unique = TRUE) +
    psABBA() + psABBY() + psABAB() + psABBY() +
    psABXA() + psABXB() + psABXY() +
    recencyContinue() +
    recencySendSender() + recencySendReceiver() +
    recencyReceiveSender() + recencyReceiveReceiver() +
    rrankSend() + rrankReceive()
output <- capture.output(
  stats <- remstats(reh, tie_effects = effects, display_progress = TRUE)
)
expect_true(all(grepl("Calculating", output)))

# TODO(actor-model): actor-oriented model progress test pending aomstats implementation
# reh <- remify(edgelist, model = "actor")
# sender_effects <- ~
#   indegreeSender() + outdegreeSender() + totaldegreeSender() +
#     recencySendSender() + recencyReceiveSender()
# receiver_effects <- ~
#   indegreeReceiver() + outdegreeReceiver() + totaldegreeReceiver() +
#     inertia() + reciprocity() +
#     isp() + itp() + osp() + otp() +
#     isp(unique = TRUE) + itp(unique = TRUE) +
#     osp(unique = TRUE) + otp(unique = TRUE) +
#     recencyContinue() + recencySendReceiver() + recencyReceiveReceiver() +
#     rrankSend() + rrankReceive()
# output <- capture.output(
#   stats <- remstats(reh,
#     sender_effects = sender_effects,
#     receiver_effects = receiver_effects, display_progress = TRUE,
#   )
# )
# expect_true(all(grepl("Calculating", output)))

# tie-oriented model 
# --------------------------------------------------------------------
# Small edgelist
edgelist <- data.frame(
  time = 1:10,
  actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
  actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
)

reh <- remify::remify(edgelist, model = "tie", riskset = "active")
effects <- ~ inertia()

stats_window <- remstats(
  reh = reh, tie_effects = effects, memory = "window",
  memory_value = 5, start = 1
)
riskset <- attr(stats_window, "riskset")

inertia_window <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(1, 0, 0, 0, 0, 0, 0),
  c(1, 1, 0, 0, 0, 0, 0),
  c(2, 1, 0, 0, 0, 0, 0),
  c(2, 1, 1, 0, 0, 0, 0),
  c(2, 1, 1, 0, 1, 0, 0),
  c(1, 1, 1, 0, 1, 0, 1),
  c(1, 1, 1, 0, 1, 0, 1),
  c(0, 1, 2, 0, 1, 0, 1),
  c(0, 1, 1, 1, 1, 0, 1)
)
expect_equal(stats_window[, , "inertia"], inertia_window)

stats_interval <- remstats(
  reh = reh, tie_effects = effects,
  memory = "interval", memory_value = c(2, 5), start = 1
)

intertia_interval <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(0, 0, 0, 0, 0, 0, 0),
  c(0, 0, 0, 0, 0, 0, 0),
  c(1, 0, 0, 0, 0, 0, 0),
  c(1, 1, 0, 0, 0, 0, 0),
  c(2, 1, 0, 0, 0, 0, 0),
  c(1, 1, 1, 0, 0, 0, 0),
  c(1, 0, 1, 0, 1, 0, 0),
  c(0, 0, 1, 0, 1, 0, 1),
  c(0, 1, 0, 0, 1, 0, 1)
)
expect_equal(stats_interval[, , "inertia"], intertia_interval)

stats_decay <- remstats(
  reh = reh, tie_effects = effects,
  memory = "decay", memory_value = 5, start = 1
)

f <- function(time, time_event) {
  exp(-(time - time_event) * (log(2) / 5)) #* (log(2) / 5)
}

inertia_decay <- rbind(
  matrix(0, ncol = nrow(riskset)),
  c(f(1, 1), 0, 0, 0, 0, 0, 0),
  c(f(2, 1), f(2, 2), 0, 0, 0, 0, 0),
  c(f(3, 1) + f(3, 3), f(3, 2), 0, 0, 0, 0, 0),
  c(f(4, 1) + f(4, 3), f(4, 2), f(4, 4), 0, 0, 0, 0),
  c(f(5, 1) + f(5, 3), f(5, 2), f(5, 4), 0, f(5, 5), 0, 0),
  c(f(6, 1) + f(6, 3), f(6, 2), f(6, 4), 0, f(6, 5), 0, f(6, 6)),
  c(f(7, 1) + f(7, 3), f(7, 2) + f(7, 7), f(7, 4), 0, f(7, 5), 0, f(7, 6)),
  c(f(8, 1) + f(8, 3), f(8, 2) + f(8, 7), f(8, 4) + f(8, 8), 0, f(8, 5), 0, f(8, 6)),
  c(f(9, 1) + f(9, 3), f(9, 2) + f(9, 7), f(9, 4) + f(9, 8), f(9, 9), f(9, 5), 0, f(9, 6))
)

expect_equal(stats_decay[, , "inertia"], inertia_decay)

# TODO(actor-model): # actor-oriented model 
# TODO(actor-model): # --------------------------------------------------------------------
# TODO(actor-model): # Small edgelist
# TODO(actor-model): edgelist <- data.frame(
# TODO(actor-model):   time = 1:10,
# TODO(actor-model):   actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
# TODO(actor-model):   actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): sender_effects <- ~
# TODO(actor-model):   indegreeSender() + outdegreeSender() + totaldegreeSender() +
# TODO(actor-model):     recencySendSender() + recencyReceiveSender()
# TODO(actor-model): receiver_effects <- ~
# TODO(actor-model):   indegreeReceiver() + outdegreeReceiver() + totaldegreeReceiver() +
# TODO(actor-model):     inertia() + reciprocity() +
# TODO(actor-model):     isp() + itp() + osp() + otp() +
# TODO(actor-model):     isp(unique = TRUE) + itp(unique = TRUE) + 
# TODO(actor-model):     osp(unique = TRUE) + otp(unique = TRUE) +
# TODO(actor-model):     recencyContinue() + recencySendReceiver() + recencyReceiveReceiver() +
# TODO(actor-model):     rrankSend() + rrankReceive()
# TODO(actor-model): stats_window <- remstats(reh, sender_effects = sender_effects,
# TODO(actor-model):   receiver_effects = receiver_effects, memory = "window", memory_value = 5
# TODO(actor-model): )
# TODO(actor-model): sender_statsW <- stats_window$sender_stats
# TODO(actor-model): receiver_statsW <- stats_window$receiver_stats
# TODO(actor-model): actors <- attr(reh, "dictionary")$actors
# TODO(actor-model): 
# TODO(actor-model): # outdegreeSender
# TODO(actor-model): outdegreeSender <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(1, 0, 0, 0),
# TODO(actor-model):   c(1, 1, 0, 0),
# TODO(actor-model):   c(2, 1, 0, 0),
# TODO(actor-model):   c(2, 2, 0, 0),
# TODO(actor-model):   c(2, 2, 1, 0),
# TODO(actor-model):   c(1, 2, 1, 1),
# TODO(actor-model):   c(1, 2, 1, 1),
# TODO(actor-model):   c(0, 3, 1, 1),
# TODO(actor-model):   c(0, 3, 1, 1)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(sender_statsW[, , "outdegreeSender"], outdegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # indegreeSender
# TODO(actor-model): indegreeSender <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 2, 0),
# TODO(actor-model):   c(1, 0, 3, 0),
# TODO(actor-model):   c(1, 1, 3, 0),
# TODO(actor-model):   c(1, 1, 3, 0),
# TODO(actor-model):   c(1, 1, 3, 0),
# TODO(actor-model):   c(1, 1, 3, 0),
# TODO(actor-model):   c(1, 1, 2, 1)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(sender_statsW[, , "indegreeSender"], indegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # totaldegreeSender
# TODO(actor-model): totaldegreeSender <- indegreeSender + outdegreeSender
# TODO(actor-model): expect_equal(sender_statsW[, , "totaldegreeSender"], totaldegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # outdegreeReceiver
# TODO(actor-model): outdegreeReceiver <- outdegreeSender
# TODO(actor-model): expect_equal(receiver_statsW[, , "outdegreeReceiver"], outdegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # indegreeReceiver
# TODO(actor-model): indegreeReceiver <- indegreeSender
# TODO(actor-model): expect_equal(receiver_statsW[, , "indegreeReceiver"], indegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # totaldegreeReceiver
# TODO(actor-model): totaldegreeReceiver <- indegreeReceiver + outdegreeReceiver
# TODO(actor-model): expect_equal(receiver_statsW[, , "totaldegreeReceiver"], totaldegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # inertia
# TODO(actor-model): inertia <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(1, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 2, 0),
# TODO(actor-model):   c(0, 0, 1, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsW[, , "inertia"], inertia)
# TODO(actor-model): 
# TODO(actor-model): # reciprocity
# TODO(actor-model): reciprocity <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 1, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(2, 1, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 1, 0, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsW[, , "reciprocity"], reciprocity)
# TODO(actor-model): 
# TODO(actor-model): # itp
# TODO(actor-model): itp <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 1, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(1, 0, 0, 1),
# TODO(actor-model):   c(1, 0, 0, 1),
# TODO(actor-model):   c(0, 0, 0, 1),
# TODO(actor-model):   c(0, 0, 1, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsW[, , "itp"], itp)
# TODO(actor-model): 
# TODO(actor-model): # interval memory
# TODO(actor-model): stats_interval <- remstats(reh, sender_effects = sender_effects,
# TODO(actor-model):   receiver_effects = receiver_effects, memory = "interval", 
# TODO(actor-model):   memory_value = c(2, 5)
# TODO(actor-model): )
# TODO(actor-model): sender_statsI <- stats_interval$sender_stats
# TODO(actor-model): receiver_statsI <- stats_interval$receiver_stats
# TODO(actor-model): actors <- attr(reh, "dictionary")$actors
# TODO(actor-model): 
# TODO(actor-model): # outdegreeSender
# TODO(actor-model): outdegreeSender <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(1, 0, 0, 0),
# TODO(actor-model):   c(1, 1, 0, 0),
# TODO(actor-model):   c(2, 1, 0, 0),
# TODO(actor-model):   c(1, 2, 0, 0),
# TODO(actor-model):   c(1, 1, 1, 0),
# TODO(actor-model):   c(0, 1, 1, 1),
# TODO(actor-model):   c(0, 1, 1, 1)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(sender_statsI[, , "outdegreeSender"], outdegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # indegreeSender
# TODO(actor-model): indegreeSender <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 1, 0),
# TODO(actor-model):   c(1, 0, 2, 0),
# TODO(actor-model):   c(1, 0, 2, 0),
# TODO(actor-model):   c(0, 1, 2, 0),
# TODO(actor-model):   c(0, 1, 2, 0),
# TODO(actor-model):   c(1, 1, 1, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(sender_statsI[, , "indegreeSender"], indegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # totaldegreeSender
# TODO(actor-model): totaldegreeSender <- indegreeSender + outdegreeSender
# TODO(actor-model): expect_equal(sender_statsI[, , "totaldegreeSender"], totaldegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # outdegreeReceiver
# TODO(actor-model): outdegreeReceiver <- outdegreeSender
# TODO(actor-model): expect_equal(receiver_statsI[, , "outdegreeReceiver"], outdegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # indegreeReceiver
# TODO(actor-model): indegreeReceiver <- indegreeSender
# TODO(actor-model): expect_equal(receiver_statsI[, , "indegreeReceiver"], indegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # totaldegreeReceiver
# TODO(actor-model): totaldegreeReceiver <- indegreeReceiver + outdegreeReceiver
# TODO(actor-model): expect_equal(receiver_statsI[, , "totaldegreeReceiver"], totaldegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # inertia
# TODO(actor-model): inertia <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(1, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 1, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsI[, , "inertia"], inertia)
# TODO(actor-model): 
# TODO(actor-model): # reciprocity
# TODO(actor-model): reciprocity <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(1, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 1, 0),
# TODO(actor-model):   c(0, 0, 0, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsI[, , "reciprocity"], reciprocity)
# TODO(actor-model): 
# TODO(actor-model): # itp
# TODO(actor-model): itp <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 1, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(1, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 1),
# TODO(actor-model):   c(0, 0, 0, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsI[, , "itp"], itp)
# TODO(actor-model): 
# TODO(actor-model): # decay memory
# TODO(actor-model): stats_decay <- remstats(reh, sender_effects = sender_effects,
# TODO(actor-model):   receiver_effects = receiver_effects, memory = "decay", 
# TODO(actor-model):   memory_value = 5
# TODO(actor-model): )
# TODO(actor-model): sender_statsD <- stats_decay$sender_stats
# TODO(actor-model): receiver_statsD <- stats_decay$receiver_stats
# TODO(actor-model): actors <- attr(reh, "dictionary")$actors
# TODO(actor-model): 
# TODO(actor-model): # outdegreeSender
# TODO(actor-model): outdegreeSender <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(f(1, 1), 0, 0, 0),
# TODO(actor-model):   c(f(2, 1), f(2, 2), 0, 0),
# TODO(actor-model):   c(f(3, 1) + f(3, 3), f(3, 2), 0, 0),
# TODO(actor-model):   c(f(4, 1) + f(4, 3), f(4, 2) + f(4, 4), 0, 0),
# TODO(actor-model):   c(f(5, 1) + f(5, 3), f(5, 2) + f(5, 4), f(5, 5), 0),
# TODO(actor-model):   c(f(6, 1) + f(6, 3), f(6, 2) + f(6, 4), f(6, 5), f(6, 6)),
# TODO(actor-model):   c(f(7, 1) + f(7, 3), f(7, 2) + f(7, 4) + f(7, 7), f(7, 5), f(7, 6)),
# TODO(actor-model):   c(f(8, 1) + f(8, 3), f(8, 2) + f(8, 4) + f(8, 7) + f(8, 8), f(8, 5), f(8, 6)),
# TODO(actor-model):   c(f(9, 1) + f(9, 3), f(9, 2) + f(9, 4) + f(9, 7) + f(9, 8) + f(9, 9), f(9, 5), f(9, 6))
# TODO(actor-model): )
# TODO(actor-model): expect_equal(sender_statsD[, , "outdegreeSender"], outdegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # indegreeSender
# TODO(actor-model): indegreeSender <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, f(1, 1), 0),
# TODO(actor-model):   c(f(2, 2), 0, f(2, 1), 0),
# TODO(actor-model):   c(f(3, 2), 0, f(3, 1) + f(3, 3), 0),
# TODO(actor-model):   c(f(4, 2), 0, f(4, 1) + f(4, 3) + f(4, 4), 0),
# TODO(actor-model):   c(f(5, 2), f(5, 5), f(5, 1) + f(5, 3) + f(5, 4), 0),
# TODO(actor-model):   c(f(6, 2), f(6, 5), f(6, 1) + f(6, 3) + f(6, 4) + f(6, 6), 0),
# TODO(actor-model):   c(f(7, 2) + f(7, 7), f(7, 5), f(7, 1) + f(7, 3) + f(7, 4) + f(7, 6), 0),
# TODO(actor-model):   c(f(8, 2) + f(8, 7), f(8, 5), f(8, 1) + f(8, 3) + f(8, 4) + f(8, 6) + f(8, 8), 0),
# TODO(actor-model):   c(f(9, 2) + f(9, 7), f(9, 5), f(9, 1) + f(9, 3) + f(9, 4) + f(9, 6) + f(9, 8), f(9, 9))
# TODO(actor-model): )
# TODO(actor-model): expect_equal(sender_statsD[, , "indegreeSender"], indegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # totaldegreeSender
# TODO(actor-model): totaldegreeSender <- indegreeSender + outdegreeSender
# TODO(actor-model): expect_equal(sender_statsD[, , "totaldegreeSender"], totaldegreeSender)
# TODO(actor-model): 
# TODO(actor-model): # outdegreeReceiver
# TODO(actor-model): outdegreeReceiver <- outdegreeSender
# TODO(actor-model): expect_equal(receiver_statsD[, , "outdegreeReceiver"], outdegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # indegreeReceiver
# TODO(actor-model): indegreeReceiver <- indegreeSender
# TODO(actor-model): expect_equal(receiver_statsD[, , "indegreeReceiver"], indegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # totaldegreeReceiver
# TODO(actor-model): totaldegreeReceiver <- indegreeReceiver + outdegreeReceiver
# TODO(actor-model): expect_equal(receiver_statsD[, , "totaldegreeReceiver"], totaldegreeReceiver)
# TODO(actor-model): 
# TODO(actor-model): # inertia
# TODO(actor-model): inertia <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, f(2, 1), 0),
# TODO(actor-model):   c(f(3, 2), 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(f(6, 2), 0, f(6, 4), 0),
# TODO(actor-model):   c(f(7, 2) + f(7, 7), 0, f(7, 4), 0),
# TODO(actor-model):   c(f(8, 2) + f(8, 7), 0, f(8, 4) + f(8, 8), 0),
# TODO(actor-model):   c(0, 0, f(9, 6), 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsD[, , "inertia"], inertia)
# TODO(actor-model): 
# TODO(actor-model): # reciprocity
# TODO(actor-model): reciprocity <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, f(2, 2), 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(f(4, 1) + f(4, 3), f(4, 4), 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, f(6, 5), 0),
# TODO(actor-model):   c(0, 0, f(7, 5), 0),
# TODO(actor-model):   c(0, 0, f(8, 5), 0),
# TODO(actor-model):   c(0, f(9, 9), 0, 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsD[, , "reciprocity"], reciprocity)
# TODO(actor-model): 
# TODO(actor-model): # itp
# TODO(actor-model): itp <- rbind(
# TODO(actor-model):   matrix(0, ncol = nrow(actors)),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(0, f(4, 2), 0, 0),
# TODO(actor-model):   c(0, 0, 0, 0),
# TODO(actor-model):   c(f(6, 5), 0, 0, f(6, 5)),
# TODO(actor-model):   c(f(7, 5), 0, 0, f(7, 5)),
# TODO(actor-model):   c(f(8, 5), 0, 0, f(8, 5)),
# TODO(actor-model):   c(0, 0, f(9, 5), 0)
# TODO(actor-model): )
# TODO(actor-model): expect_equal(receiver_statsD[, , "itp"], itp)
# TODO(actor-model): 

# Tests for stack_stats.tomstats()
# Checks correct dimensions, structure, response, offset, and ordinal support.

library(tinytest)

data(history, package = "remstats")
data(info,    package = "remstats")
colnames(history)[colnames(history) == "setting"] <- "type"
history_sub <- history[1:40, ]

effects <- ~ inertia(consider_type = FALSE) +
               indegreeSender(consider_type = FALSE) +
               outdegreeSender(consider_type = FALSE)

# ---------------------------------------------------------------------------
# SECTION 1: Interval timing, active riskset
# ---------------------------------------------------------------------------
reh_int <- remify::remify(edgelist = history_sub, model = "tie",
                            riskset = "active", extend_riskset_by_type = TRUE)

ts_int <- remstats::tomstats(effects, reh = reh_int,
                               attr_actors = info,
                               memory = "decay", memory_value = 1000,
                               start = 2, stop = 30,
                               sampling = FALSE)

stacked_int <- remstats::stack_stats(ts_int, reh_int)

# Return structure
expect_true(is.list(stacked_int),
  info = "interval: returns a list")
expect_true(all(c("remstats_stack", "subset", "D", "E", "ordinal") %in%
                  names(stacked_int)),
  info = "interval: list has correct elements")

df_int <- stacked_int$remstats_stack
D_int   <- stacked_int$D
E_int   <- stacked_int$E

# Dimensions
expect_equal(nrow(df_int), D_int * E_int,
  info = "interval: nrow = D * E")
expect_equal(D_int, nrow(attr(ts_int, "riskset")),
  info = "interval: D matches riskset size")
expect_equal(E_int, dim(ts_int)[1],
  info = "interval: E matches number of time points")

# Column names
expect_true(all(c("event", "obs", "dyad", "log_interevent") %in%
                  colnames(df_int)),
  info = "interval: required columns present")
expect_true(all(c("baseline", "inertia", "indegreeSender", "outdegreeSender") %in%
                  colnames(df_int)),
  info = "interval: statistic columns present")

# ordinal flag
expect_false(stacked_int$ordinal,
  info = "interval: ordinal flag is FALSE")

# Response: total observed events should equal E (one event per time point)
expect_equal(sum(df_int$obs), E_int,
  info = "interval: total observed events equals E")

# Response is 0/1 integer (no simultaneous events in this data)
expect_true(all(df_int$obs %in% c(0L, 1L)),
  info = "interval: obs is binary")

# Dyad index cycles 1..D for each event
expect_equal(df_int$dyad, rep(seq_len(D_int), E_int),
  info = "interval: dyad index cycles correctly")

# Event index: each event block has the correct event number
expect_equal(df_int$event, rep(seq_len(E_int), each = D_int),
  info = "interval: event index correct")

# Offset: log_interevent matches reh interevent times
iet_expected <- log(reh_int$intereventTime[2:30])
expect_equal(df_int$log_interevent[seq(1, nrow(df_int), by = D_int)],
  iet_expected,
  tolerance = 1e-10,
  info = "interval: log_interevent values match reh intereventTime")

# Baseline column is all 1
expect_true(all(df_int$baseline == 1),
  info = "interval: baseline column is all 1")

# subset stored correctly
expect_equal(stacked_int$subset, c(2L, 30L),
  info = "interval: subset stored correctly")

# ---------------------------------------------------------------------------
# SECTION 2: Ordinal timing, active riskset
# ---------------------------------------------------------------------------
reh_ord <- remify::remify(edgelist = history_sub, model = "tie",
                            riskset = "active", ordinal = TRUE, extend_riskset_by_type = TRUE)

ts_ord <- remstats::tomstats(effects, reh = reh_ord,
                               attr_actors = info,
                               memory = "decay", memory_value = 1000,
                               start = 2, stop = 30,
                               sampling = FALSE)

stacked_ord <- remstats::stack_stats(ts_ord, reh_ord)
df_ord <- stacked_ord$remstats_stack

# ordinal flag
expect_true(stacked_ord$ordinal,
  info = "ordinal: ordinal flag is TRUE")

# No log_interevent column for ordinal
expect_false("log_interevent" %in% colnames(df_ord),
  info = "ordinal: no log_interevent column")

# Dimensions same as interval
expect_equal(nrow(df_ord), stacked_ord$D * stacked_ord$E,
  info = "ordinal: nrow = D * E")

# Response correct
expect_equal(sum(df_ord$obs), stacked_ord$E,
  info = "ordinal: total observed events equals E")

expect_true(all(df_ord$obs %in% c(0L, 1L)),
  info = "ordinal: obs is binary")

# ---------------------------------------------------------------------------
# SECTION 3: Interval timing, full riskset
# ---------------------------------------------------------------------------
reh_full <- remify::remify(edgelist = history_sub, model = "tie",
                             riskset = "full", extend_riskset_by_type = TRUE)

ts_full <- remstats::tomstats(effects, reh = reh_full,
                                attr_actors = info,
                                memory = "decay", memory_value = 1000,
                                start = 2, stop = 30,
                                sampling = FALSE)

stacked_full <- remstats::stack_stats(ts_full, reh_full)
df_full <- stacked_full$remstats_stack

expect_equal(stacked_full$D, dim(ts_full)[2],
  info = "full riskset: D matches stats dimension")
expect_equal(nrow(df_full), stacked_full$D * stacked_full$E,
  info = "full riskset: nrow = D * E")
expect_equal(sum(df_full$obs), stacked_full$E,
  info = "full riskset: total observed events equals E")
expect_false(stacked_full$ordinal,
  info = "full riskset: ordinal flag is FALSE")
expect_true("log_interevent" %in% colnames(df_full),
  info = "full riskset: log_interevent present")

# ---------------------------------------------------------------------------
# SECTION 4: start/stop subsetting respected
# ---------------------------------------------------------------------------
ts_sub <- remstats::tomstats(effects, reh = reh_int,
                               attr_actors = info,
                               memory = "decay", memory_value = 1000,
                               start = 5, stop = 20,
                               sampling = FALSE)

stacked_sub <- remstats::stack_stats(ts_sub, reh_int)

expect_equal(stacked_sub$subset, c(5L, 20L),
  info = "subset: stored correctly")
expect_equal(stacked_sub$E, dim(ts_sub)[1],
  info = "subset: E matches stats rows")
expect_equal(nrow(stacked_sub$remstats_stack),
  stacked_sub$D * stacked_sub$E,
  info = "subset: nrow correct for subset")
expect_equal(sum(stacked_sub$remstats_stack$obs), stacked_sub$E,
  info = "subset: total events correct")

# ---------------------------------------------------------------------------
# SECTION 5: Sampled tomstats — interval
# ---------------------------------------------------------------------------
samp_num <- 5L

ts_samp <- remstats::tomstats(effects, reh = reh_int,
                                attr_actors = info,
                                memory = "decay", memory_value = 1000,
                                start = 2, stop = 30,
                                sampling = TRUE, samp_num = samp_num, seed = 1L)

stacked_samp <- remstats::stack_stats(ts_samp, reh_int)
df_samp <- stacked_samp$remstats_stack

# Return structure
expect_true(is.list(stacked_samp),
  info = "sampled interval: returns list")
expect_true(all(c("remstats_stack", "subset", "S", "E", "ordinal") %in%
                  names(stacked_samp)),
  info = "sampled interval: list has correct elements")

# Dimensions: E*S rows
expect_equal(nrow(df_samp), stacked_samp$E * stacked_samp$S,
  info = "sampled interval: nrow = E * S")
expect_equal(stacked_samp$S, samp_num,
  info = "sampled interval: S matches samp_num")
expect_equal(stacked_samp$E, dim(ts_samp)[1],
  info = "sampled interval: E matches stats rows")

# Column names
expect_true(all(c("event", "obs", "dyad", "weight", "log_interevent") %in%
                  colnames(df_samp)),
  info = "sampled interval: required columns present")
expect_true(all(c("inertia", "indegreeSender", "outdegreeSender") %in%
                  colnames(df_samp)),
  info = "sampled interval: statistic columns present")

# ordinal flag
expect_false(stacked_samp$ordinal,
  info = "sampled interval: ordinal flag is FALSE")

# Response: exactly one case per event (no simultaneous events here)
expect_equal(sum(df_samp$obs), stacked_samp$E,
  info = "sampled interval: total cases equals E")
expect_true(all(df_samp$obs %in% c(0L, 1L)),
  info = "sampled interval: obs is binary")

# Weights: cases have weight 1, controls have weight 1/pi_s
case_rows <- df_samp$obs == 1L
expect_true(all(df_samp$weight[case_rows] == 1.0),
  info = "sampled interval: case weights are 1")
expect_true(all(df_samp$weight[!case_rows] > 0),
  info = "sampled interval: control weights are positive")

# With uniform sampling, control weights = (D-1)/(samp_num-1)
D_active <- nrow(attr(ts_samp, "riskset"))
expected_control_weight <- (D_active - 1) / (samp_num - 1)
expect_equal(
  mean(df_samp$weight[!case_rows]),
  expected_control_weight,
  tolerance = 1e-10,
  info = "sampled interval: control weights correct for uniform sampling"
)

# subset stored correctly
expect_equal(stacked_samp$subset, c(2L, 30L),
  info = "sampled interval: subset stored correctly")

# ---------------------------------------------------------------------------
# SECTION 6: Sampled tomstats — ordinal
# ---------------------------------------------------------------------------
ts_samp_ord <- remstats::tomstats(effects, reh = reh_ord,
                                    attr_actors = info,
                                    memory = "decay", memory_value = 1000,
                                    start = 2, stop = 30,
                                    sampling = TRUE, samp_num = samp_num, seed = 1L)

stacked_samp_ord <- remstats::stack_stats(ts_samp_ord, reh_ord)
df_samp_ord <- stacked_samp_ord$remstats_stack

# ordinal flag
expect_true(stacked_samp_ord$ordinal,
  info = "sampled ordinal: ordinal flag is TRUE")

# No log_interevent for ordinal
expect_false("log_interevent" %in% colnames(df_samp_ord),
  info = "sampled ordinal: no log_interevent column")

# Dimensions
expect_equal(nrow(df_samp_ord), stacked_samp_ord$E * stacked_samp_ord$S,
  info = "sampled ordinal: nrow = E * S")

# Response and weights
expect_equal(sum(df_samp_ord$obs), stacked_samp_ord$E,
  info = "sampled ordinal: total cases equals E")
expect_true(all(df_samp_ord$weight[df_samp_ord$obs == 1L] == 1.0),
  info = "sampled ordinal: case weights are 1")

# test-aomstats-typed.R
#
# Verifies aomstats() with typed events against the tomstats() tie model
# (extend_riskset_by_type = FALSE) as the ground truth, for multiple stats
# and memory types.
#
# Design:
#   For every valid (event m, receiver j) cell:
#     aomstats$receiver_stats[m, j, stat] ==
#       tomstats[m, dyad_col(sender_m, j), stat]
#   For sender stats analogously (any reference receiver j != i).
#
# The dyad_col formula (1-based, derived from remify2 internals):
#   d(s, j) = (s-1)*(N-1) + j - as.integer(j > s)
#
# Sections:
#   1  Dimensions and slice names
#   2  Invariants: separate sums to ignore; LOCF; zero before first type event
#   3  Full memory: receiver ignore  vs tie
#   4  Full memory: receiver separate vs tie (consider_type = TRUE)
#   5  Decay memory: receiver ignore  vs tie
#   6  Decay memory: receiver separate vs tie
#   7  Sender stats (outdegreeSender, indegreeSender): full and decay

library(tinytest)
library(remify)
library(remstats)

# ---------------------------------------------------------------------------
# Data
# ---------------------------------------------------------------------------
data(randomREH, package = "remify")
el <- randomREH$edgelist[1:100, ]

reh_actor <- remify(
  edgelist = el,
  actors   = randomREH$actors,
  directed = TRUE,
  origin   = randomREH$origin,
  model    = "actor"
)

# Tie model with untyped riskset (extend = FALSE) is the reference
reh_tie <- remify(
  edgelist               = el,
  actors                 = randomREH$actors,
  directed               = TRUE,
  origin                 = randomREH$origin,
  model                  = "tie",
  extend_riskset_by_type = FALSE
)

N          <- length(randomREH$actors)   # number of actors
M          <- nrow(el) - 1L              # aomstats rows  (events 2:100)
tol        <- 1e-10
types      <- sort(unique(el$type))      # competition, conflict, cooperation
actor1_ids <- reh_actor$ids$actor1       # 1-based sender ID for every event

# Directed dyad column (1-based) in the full untyped tie-model riskset.
# Equivalent to remify2's 0-based formula a1*(N-1)+a2-int(a2>a1)+1
# translated to 1-based actor IDs.
dyad_col <- function(s, j, N) (s - 1L) * (N - 1L) + j - as.integer(j > s)

# ---------------------------------------------------------------------------
# Helper: check aomstats receiver_stats[,,stat] == tomstats[,,stat]
#   for a random sample of events, all receivers per event.
#
# aom_arr : [M x N x P] from aomstats()$receiver_stats
# tie_arr : [M x D x P] from tomstats()
# stat    : slice name (must exist in both arrays)
# ---------------------------------------------------------------------------
check_aom_tie_receiver <- function(aom_arr, tie_arr, stat,
                                   n_events = 15L, seed = 1L) {
  set.seed(seed)
  M_loc  <- dim(aom_arr)[1]
  events <- sample(seq_len(M_loc), min(n_events, M_loc))
  for (m in events) {
    s <- actor1_ids[m + 1L]   # sender at aomstats output row m (1-based)
    for (j in seq_len(N)) {
      if (j == s) next
      d <- dyad_col(s, j, N)
      expect_equal(
        aom_arr[m, j, stat],
        tie_arr[m, d, stat],
        tolerance = tol,
        info = paste0("stat=", stat, " m=", m, " s=", s, " j=", j)
      )
    }
  }
}

# Helper: check aomstats sender_stats[,,stat] == tomstats[,,stat]
#   Sender stats are receiver-independent: for any reference receiver j != i,
#   tie_arr[m, dyad_col(i, j), stat] carries the same value.
#   We use j_ref = (i %% N) + 1  (cycles, never self-loops).
check_aom_tie_sender <- function(aom_sender, tie_arr, stat,
                                 n_events = 15L, seed = 2L) {
  set.seed(seed)
  M_loc  <- dim(aom_sender)[1]
  events <- sample(seq_len(M_loc), min(n_events, M_loc))
  for (m in events) {
    for (i in seq_len(N)) {
      j_ref <- (i %% N) + 1L
      d     <- dyad_col(i, j_ref, N)
      expect_equal(
        aom_sender[m, i, stat],
        tie_arr[m, d, stat],
        tolerance = tol,
        info = paste0("sender stat=", stat, " m=", m, " i=", i)
      )
    }
  }
}

# ===========================================================================
# SECTION 1: Dimensions and slice names
# ===========================================================================

rec_stats_ig <- c("inertia", "indegreeReceiver", "outdegreeReceiver", "reciprocity")

ts_ig <- aomstats(
  reh              = reh_actor,
  receiver_effects = ~ inertia() + indegreeReceiver() +
                       outdegreeReceiver() + reciprocity()
)
ts_sep <- aomstats(
  reh              = reh_actor,
  receiver_effects = ~ inertia(consider_type = "separate") +
                       indegreeReceiver(consider_type = "separate") +
                       outdegreeReceiver(consider_type = "separate") +
                       reciprocity(consider_type = "separate")
)

expect_equal(dim(ts_ig$receiver_stats),  c(M, N, 4L),
             info = "ignore: receiver dims [M x N x 4]")
expect_equal(dim(ts_sep$receiver_stats), c(M, N, 12L),
             info = "separate: receiver dims [M x N x 12]")

expect_equal(
  dimnames(ts_ig$receiver_stats)[[3]],
  rec_stats_ig,
  info = "ignore: correct receiver slice names"
)

expected_sep <- unlist(lapply(rec_stats_ig, function(s) paste0(s, ".", types)))
expect_true(
  all(expected_sep %in% dimnames(ts_sep$receiver_stats)[[3]]),
  info = "separate: all type slices present"
)

# Sender stats
ts_s_ig <- aomstats(
  reh            = reh_actor,
  sender_effects = ~ outdegreeSender() + indegreeSender()
)
ts_s_sep <- aomstats(
  reh            = reh_actor,
  sender_effects = ~ outdegreeSender(consider_type = "separate") +
                     indegreeSender(consider_type = "separate")
)

expect_equal(dim(ts_s_ig$sender_stats),  c(M, N, 3L),
             info = "sender ignore: dims [M x N x 2]")
expect_equal(dim(ts_s_sep$sender_stats), c(M, N, 7L),
             info = "sender separate: dims [M x N x 6]")

expect_equal(
  dimnames(ts_s_ig$sender_stats)[[3]],
  c("baseline","outdegreeSender", "indegreeSender"),
  info = "sender ignore: correct slice names"
)

# ===========================================================================
# SECTION 2: Invariants
# ===========================================================================

# 2.1  Separate slices sum to ignore — receiver stats
for (stat in rec_stats_ig) {
  sep_sum <- Reduce("+", lapply(types, function(tp)
    ts_sep$receiver_stats[,, paste0(stat, ".", tp)]
  ))
  expect_equal(sep_sum, ts_ig$receiver_stats[,, stat], tolerance = tol,
               info = paste("receiver separate sums to ignore:", stat))
}

# 2.2  Separate slices sum to ignore — sender stats
for (stat in c("outdegreeSender", "indegreeSender")) {
  sep_sum <- Reduce("+", lapply(types, function(tp)
    ts_s_sep$sender_stats[,, paste0(stat, ".", tp)]
  ))
  expect_equal(sep_sum, ts_s_ig$sender_stats[,, stat], tolerance = tol,
               info = paste("sender separate sums to ignore:", stat))
}

# 2.3  LOCF: between consecutive type-t events, separate stats are constant.
#   Representative check: inertia.competition.
comp_full <- which(el$type == "competition")
comp_rows <- comp_full - 1L
comp_rows <- comp_rows[comp_rows >= 1L & comp_rows <= M]
if (length(comp_rows) >= 2L) {
  from <- comp_rows[1]; to <- comp_rows[2] - 1L
  if (to > from) {
    val_from <- ts_sep$receiver_stats[from, , "inertia.competition"]
    for (k in seq(from + 1L, to)) {
      expect_equal(
        ts_sep$receiver_stats[k, , "inertia.competition"],
        val_from, tolerance = tol,
        info = paste("LOCF inertia.competition: row", k)
      )
    }
  }
}

# 2.4  Zero before first event of each type — check inertia per type
for (tp in types) {
  first_full <- min(which(el$type == tp))
  first_out  <- first_full - 1L
  if (first_out > 1L) {
    zero_vals <- ts_sep$receiver_stats[seq_len(first_out - 1L), ,
                                       paste0("inertia.", tp)]
    expect_equal(sum(abs(zero_vals)), 0, tolerance = tol,
                 info = paste("zero before first event: type =", tp))
  }
}

# ===========================================================================
# SECTION 3: Full memory — receiver ignore vs tomstats
# ===========================================================================

ts_tie_ig <- tomstats(
  ~ inertia() + indegreeReceiver() + outdegreeReceiver() + reciprocity(),
  reh = reh_tie, sampling = FALSE
)

for (stat in rec_stats_ig) {
  check_aom_tie_receiver(ts_ig$receiver_stats, ts_tie_ig, stat, n_events = 15L)
}

# ===========================================================================
# SECTION 4: Full memory — receiver separate vs tomstats (consider_type=TRUE)
# ===========================================================================

ts_tie_sep <- tomstats(
  ~ inertia(consider_type = TRUE) +
    indegreeReceiver(consider_type = TRUE) +
    outdegreeReceiver(consider_type = TRUE) +
    reciprocity(consider_type = TRUE),
  reh = reh_tie, sampling = FALSE
)

for (stat in rec_stats_ig) {
  for (tp in types) {
    check_aom_tie_receiver(
      ts_sep$receiver_stats, ts_tie_sep,
      stat  = paste0(stat, ".", tp),
      n_events = 15L, seed = 1L
    )
  }
}

# ===========================================================================
# SECTION 5: Decay memory — receiver ignore vs tomstats
# ===========================================================================

ts_ig_d <- aomstats(
  reh              = reh_actor,
  receiver_effects = ~ inertia() + indegreeReceiver() +
                       outdegreeReceiver() + reciprocity(),
  memory = "decay", memory_value = 100
)

ts_tie_ig_d <- tomstats(
  ~ inertia() + indegreeReceiver() + outdegreeReceiver() + reciprocity(),
  reh = reh_tie, sampling = FALSE,
  memory = "decay", memory_value = 100
)

for (stat in rec_stats_ig) {
  check_aom_tie_receiver(ts_ig_d$receiver_stats, ts_tie_ig_d, stat,
                         n_events = 15L, seed = 2L)
}

# ===========================================================================
# SECTION 6: Decay memory — receiver separate vs tomstats
# ===========================================================================

ts_sep_d <- aomstats(
  reh              = reh_actor,
  receiver_effects = ~ inertia(consider_type = "separate") +
                       indegreeReceiver(consider_type = "separate") +
                       outdegreeReceiver(consider_type = "separate") +
                       reciprocity(consider_type = "separate"),
  memory = "decay", memory_value = 100
)

ts_tie_sep_d <- tomstats(
  ~ inertia(consider_type = TRUE) +
    indegreeReceiver(consider_type = TRUE) +
    outdegreeReceiver(consider_type = TRUE) +
    reciprocity(consider_type = TRUE),
  reh = reh_tie, sampling = FALSE,
  memory = "decay", memory_value = 100
)

# 6.1  Separate sums to ignore under decay
for (stat in rec_stats_ig) {
  sep_d_sum <- Reduce("+", lapply(types, function(tp)
    ts_sep_d$receiver_stats[,, paste0(stat, ".", tp)]
  ))
  expect_equal(sep_d_sum, ts_ig_d$receiver_stats[,, stat], tolerance = tol,
               info = paste("decay separate sums to ignore:", stat))
}

# 6.2  Value-level match against tie model
for (stat in rec_stats_ig) {
  for (tp in types) {
    check_aom_tie_receiver(
      ts_sep_d$receiver_stats, ts_tie_sep_d,
      stat     = paste0(stat, ".", tp),
      n_events = 15L, seed = 3L
    )
  }
}

# ===========================================================================
# SECTION 7: Sender stats — full and decay memory
# ===========================================================================

# --- 7.1  Full memory -------------------------------------------------------
ts_tie_s_ig <- tomstats(
  ~ outdegreeSender() + indegreeSender(),
  reh = reh_tie, sampling = FALSE
)

for (stat in c("outdegreeSender", "indegreeSender")) {
  check_aom_tie_sender(ts_s_ig$sender_stats, ts_tie_s_ig, stat, n_events = 15L)
}

# Separate sums to ignore already checked in Section 2.2; spot-check values
for (stat in c("outdegreeSender", "indegreeSender")) {
  ts_tie_s_tp <- tomstats(
    as.formula(paste0("~ ", stat, "(consider_type = TRUE)")),
    reh = reh_tie, sampling = FALSE
  )
  for (tp in types) {
    sl <- paste0(stat, ".", tp)
    if (sl %in% dimnames(ts_s_sep$sender_stats)[[3]] &&
        sl %in% dimnames(ts_tie_s_tp)[[3]]) {
      check_aom_tie_sender(ts_s_sep$sender_stats, ts_tie_s_tp, sl,
                           n_events = 10L, seed = 4L)
    }
  }
}

# --- 7.2  Decay memory ------------------------------------------------------
ts_s_ig_d <- aomstats(
  reh            = reh_actor,
  sender_effects = ~ outdegreeSender() + indegreeSender(),
  memory = "decay", memory_value = 100
)
ts_tie_s_ig_d <- tomstats(
  ~ outdegreeSender() + indegreeSender(),
  reh = reh_tie, sampling = FALSE,
  memory = "decay", memory_value = 100
)

for (stat in c("outdegreeSender", "indegreeSender")) {
  check_aom_tie_sender(ts_s_ig_d$sender_stats, ts_tie_s_ig_d, stat,
                       n_events = 15L, seed = 5L)
}

# test-pshift-types.R
# Tests correct type-splitting of pshift statistics for typed events.
# Verifies that pshift slices are only non-zero at rows where the
# triggering event at the previous time point matches the slice type.

library(tinytest)
library(remify)
library(remstats)

# Small directed edgelist, 2 types
el <- data.frame(
	time   = 1:5,
	actor1 = c(1, 2, 1, 3, 2),
	actor2 = c(2, 3, 3, 2, 1),
	type   = c("social", "work", "social", "social", "work")
)

reh_FALSE <- remify(el, model = "tie", directed = TRUE,
										 extend_riskset_by_type = FALSE)
reh_TRUE  <- remify(el, model = "tie", directed = TRUE,
										 extend_riskset_by_type = TRUE)

eff_sep <- ~ psABBA(consider_type = "separate") +
	psABAB(consider_type = "separate") +
	psABAY(consider_type = "separate")
eff_int <- ~ psABBA(consider_type = "interact") +
	psABAB(consider_type = "interact") +
	psABAY(consider_type = "interact")
eff_ign <- ~ psABBA(consider_type = "ignore") +
	psABAB(consider_type = "ignore") +
	psABAY(consider_type = "ignore")

ts_FALSE     <- tomstats(eff_sep, reh = reh_FALSE, sampling = FALSE, start = 2)
ts_sep       <- tomstats(eff_sep, reh = reh_TRUE,  sampling = FALSE, start = 2)
ts_int       <- tomstats(eff_int, reh = reh_TRUE,  sampling = FALSE, start = 2)
ts_ign_FALSE <- tomstats(eff_ign, reh = reh_FALSE, sampling = FALSE, start = 2)
ts_ign_TRUE  <- tomstats(eff_ign, reh = reh_TRUE,  sampling = FALSE, start = 2)

# ---------------------------------------------------------------------------
# Row 4: previous event = (3,2,social) — only social slices can be non-zero
# ---------------------------------------------------------------------------

# ext=FALSE "separate": all .work slices must be 0
expect_true(all(ts_FALSE[4, , grep("\\.work", dimnames(ts_FALSE)[[3]])] == 0),
						info = "ext=FALSE row4: all .work slices = 0")

# correct dyads non-zero in .social
expect_equal(as.numeric(ts_FALSE[4, 4, "psABBA.social"]), 1,
						 info = "ext=FALSE row4: psABBA.social dyad (2,3) = 1")
expect_equal(as.numeric(ts_FALSE[4, 6, "psABAB.social"]), 0,   # curr_sender=2 != prev_sender=3
						 info = "ext=FALSE row4: psABAB.social dyad (3,2) = 0")
expect_equal(as.numeric(ts_FALSE[4, 5, "psABAY.social"]), 0,   # curr_sender=2 != prev_sender=3
						 info = "ext=FALSE row4: psABAY.social dyad (3,1) = 0")

# ext=TRUE "separate": pshift pattern is same across all type slices
# (dyad pattern reflects the ABAB regardless of dyad type)
expect_true(all(ts_sep[4, , grep("psABBA\\.social", dimnames(ts_sep)[[3]])] ==
									ts_sep[4, , grep("psABBA\\.work",   dimnames(ts_sep)[[3]])]),
						info = "ext=TRUE separate row4: social and work slices identical")

# ext=TRUE "interact": all work.* slices must be 0 (past was not work)
expect_true(all(ts_int[4, , grep("work\\.", dimnames(ts_int)[[3]])] == 0),
						info = "ext=TRUE interact row4: all work.* slices = 0")

# ---------------------------------------------------------------------------
# Row 1 (time=3): previous time point = time=2, TWO simultaneous events:
# (1,2,social) and (2,3,work) — both .social AND .work can be non-zero
# ---------------------------------------------------------------------------
el2 <- data.frame(
	time   = c(2, 2, 3:5),
	actor1 = c(1, 2, 1, 3, 2),
	actor2 = c(2, 3, 3, 2, 1),
	type   = c("social", "work", "social", "social", "work")
)
reh2 <- remify(el2, model = "tie", directed = TRUE,
								extend_riskset_by_type = FALSE)
ts2  <- tomstats(eff_sep, reh = reh2, sampling = FALSE, start = 2)

# Row 2 corresponds to time=3, previous time point = time=2
# (1,2,social): psABAB.social → dyad (1,2) = 1 (sender 1 == curr sender 1)
# (2,3,work):   psABAB.work  → dyad (2,3) = 0 (sender 2 != curr sender 1)
expect_equal(as.numeric(ts2[2, 1, "psABAB.social"]), 1,
						 info = "simultaneous row2: psABAB.social dyad (1,2) = 1")
expect_equal(as.numeric(ts2[2, 1, "psABAB.work"]), 0,
						 info = "simultaneous row2: psABAB.work dyad (1,2) = 0")
expect_equal(as.numeric(ts2[2, 4, "psABAB.social"]), 0,
						 info = "simultaneous row2: psABAB.social dyad (2,3) = 0")
expect_equal(as.numeric(ts2[2, 4, "psABAB.work"]), 0,
						 info = "simultaneous row2: psABAB.work dyad (2,3) = 0")

# ---------------------------------------------------------------------------
# consider_type = "ignore": single slice, no type splitting, unaffected by fix
# ---------------------------------------------------------------------------

# Only 3 slices (no type splitting)
expect_equal(dimnames(ts_ign_FALSE)[[3]],
						 c("baseline", "psABBA", "psABAB", "psABAY"),
						 info = "ignore ext=FALSE: no type-split slices")
expect_equal(dimnames(ts_ign_TRUE)[[3]],
						 c("baseline", "psABBA", "psABAB", "psABAY"),
						 info = "ignore ext=TRUE: no type-split slices")

# Row 4 (previous = 3,2,social): correct dyads non-zero
# ext=FALSE (6 dyads)
expect_equal(as.numeric(ts_ign_FALSE[4, 4, "psABBA"]), 1,
						 info = "ignore ext=FALSE row4: psABBA dyad (2,3) = 1")
expect_equal(as.numeric(ts_ign_FALSE[4, 6, "psABAB"]), 1,
						 info = "ignore ext=FALSE row4: psABAB dyad (3,2) = 1")
expect_equal(as.numeric(ts_ign_FALSE[4, 5, "psABAY"]), 1,
						 info = "ignore ext=FALSE row4: psABAY dyad (3,1) = 1")

# ext=TRUE (12 dyads): both type rows should be 1 since ignore aggregates
expect_equal(as.numeric(ts_ign_TRUE[4, 4,  "psABBA"]), 1,
						 info = "ignore ext=TRUE row4: psABBA dyad (2,3,social) = 1")
expect_equal(as.numeric(ts_ign_TRUE[4, 10, "psABBA"]), 1,
						 info = "ignore ext=TRUE row4: psABBA dyad (2,3,work) = 1")
expect_equal(as.numeric(ts_ign_TRUE[4, 6,  "psABAB"]), 1,
						 info = "ignore ext=TRUE row4: psABAB dyad (3,2,social) = 1")
expect_equal(as.numeric(ts_ign_TRUE[4, 12, "psABAB"]), 1,
						 info = "ignore ext=TRUE row4: psABAB dyad (3,2,work) = 1")

# test-aomstats-vs-tie.R
#
# Direct comparison of aomstats() and tomstats() at a single event.
#
# Strategy:
#   - Fix on output row m = 99 (i.e. edgelist event 100, since aomstats
#     starts at event 2 so output row m corresponds to edgelist row m+1).
#   - Inspect the sender at that event, then pick 5 specific receivers.
#   - Look up the corresponding dyad column in tomstats via the explicit
#     formula: d(s, j) = (s-1)*(N-1) + j - as.integer(j > s)   [1-based]
#   - Compare values directly — no loops, no sampling, no automation.

library(tinytest)
library(remify)
library(remstats)

data(randomREH, package = "remify")
el <- randomREH$edgelist[1:100, ]
N  <- length(randomREH$actors)
tol <- 1e-10

reh_actor <- remify(
  edgelist = el, actors = randomREH$actors,
  directed = TRUE, origin = randomREH$origin, model = "actor"
)
reh_tie <- remify(
  edgelist = el, actors = randomREH$actors,
  directed = TRUE, origin = randomREH$origin, model = "tie",
  extend_riskset_by_type = FALSE
)

# aomstats output row 99 = edgelist row 100
m <- 99L
sender_id <- reh_actor$ids$actor1[m + 1L]   # 1-based sender at event 100
cat("event 100: sender_id =", sender_id, "\n")

# Pick 5 receivers that are not the sender
receivers <- setdiff(1:8, sender_id)[1:5]
cat("checking receivers:", receivers, "\n")

dyad_col <- function(s, j) (s - 1L) * (N - 1L) + j - as.integer(j > s)

# ---------------------------------------------------------------------------
# BLOCK A: full memory, inertia + indegreeReceiver + outdegreeReceiver + reciprocity
# ---------------------------------------------------------------------------
effects <- ~ inertia() + indegreeReceiver() + outdegreeReceiver() + reciprocity()
effects_sep <- ~ inertia(consider_type = "separate") +
                 indegreeReceiver(consider_type = "separate") +
                 outdegreeReceiver(consider_type = "separate") +
                 reciprocity(consider_type = "separate")

ts_aom  <- aomstats(reh = reh_actor, receiver_effects = effects)
ts_tie  <- tomstats(effects, reh = reh_tie, sampling = FALSE)
ts_sep  <- aomstats(reh = reh_actor, receiver_effects = effects_sep)
ts_tie_sep <- tomstats(
  ~ inertia(consider_type = TRUE) + indegreeReceiver(consider_type = TRUE) +
    outdegreeReceiver(consider_type = TRUE) + reciprocity(consider_type = TRUE),
  reh = reh_tie, sampling = FALSE
)

types <- sort(unique(el$type))

for (j in receivers) {
  d <- dyad_col(sender_id, j)
  cat(sprintf("  j=%d  d=%d\n", j, d))

  # ignore vs tie
  expect_equal(ts_aom$receiver_stats[m, j, "inertia"],
               ts_tie[m, d, "inertia"], tolerance = tol,
               info = sprintf("full/ignore inertia          j=%d", j))
  expect_equal(ts_aom$receiver_stats[m, j, "indegreeReceiver"],
               ts_tie[m, d, "indegreeReceiver"], tolerance = tol,
               info = sprintf("full/ignore indegreeReceiver  j=%d", j))
  expect_equal(ts_aom$receiver_stats[m, j, "outdegreeReceiver"],
               ts_tie[m, d, "outdegreeReceiver"], tolerance = tol,
               info = sprintf("full/ignore outdegreeReceiver j=%d", j))
  expect_equal(ts_aom$receiver_stats[m, j, "reciprocity"],
               ts_tie[m, d, "reciprocity"], tolerance = tol,
               info = sprintf("full/ignore reciprocity       j=%d", j))

  # separate vs tie (per type)
  for (tp in types) {
    sl <- paste0("inertia.", tp)
    expect_equal(ts_sep$receiver_stats[m, j, sl],
                 ts_tie_sep[m, d, sl], tolerance = tol,
                 info = sprintf("full/separate %s j=%d", sl, j))
  }
}

# ---------------------------------------------------------------------------
# BLOCK B: decay memory, same stats
# ---------------------------------------------------------------------------
effects_d     <- ~ inertia() + indegreeReceiver() + outdegreeReceiver() + reciprocity()
effects_sep_d <- ~ inertia(consider_type = "separate") +
                   indegreeReceiver(consider_type = "separate") +
                   outdegreeReceiver(consider_type = "separate") +
                   reciprocity(consider_type = "separate")
effects_tie_sep_d <- ~ inertia(consider_type = TRUE) +
                       indegreeReceiver(consider_type = TRUE) +
                       outdegreeReceiver(consider_type = TRUE) +
                       reciprocity(consider_type = TRUE)

ts_aom_d     <- aomstats(reh = reh_actor, receiver_effects = effects_d,
                          memory = "decay", memory_value = 100)
ts_tie_d     <- tomstats(effects_d, reh = reh_tie, sampling = FALSE,
                           memory = "decay", memory_value = 100)
ts_sep_d     <- aomstats(reh = reh_actor, receiver_effects = effects_sep_d,
                          memory = "decay", memory_value = 100)
ts_tie_sep_d <- tomstats(effects_tie_sep_d, reh = reh_tie, sampling = FALSE,
                           memory = "decay", memory_value = 100)

for (j in receivers) {
  d <- dyad_col(sender_id, j)

  expect_equal(ts_aom_d$receiver_stats[m, j, "inertia"],
               ts_tie_d[m, d, "inertia"], tolerance = tol,
               info = sprintf("decay/ignore inertia          j=%d", j))
  expect_equal(ts_aom_d$receiver_stats[m, j, "indegreeReceiver"],
               ts_tie_d[m, d, "indegreeReceiver"], tolerance = tol,
               info = sprintf("decay/ignore indegreeReceiver  j=%d", j))
  expect_equal(ts_aom_d$receiver_stats[m, j, "outdegreeReceiver"],
               ts_tie_d[m, d, "outdegreeReceiver"], tolerance = tol,
               info = sprintf("decay/ignore outdegreeReceiver j=%d", j))
  expect_equal(ts_aom_d$receiver_stats[m, j, "reciprocity"],
               ts_tie_d[m, d, "reciprocity"], tolerance = tol,
               info = sprintf("decay/ignore reciprocity       j=%d", j))

  for (tp in types) {
    sl <- paste0("inertia.", tp)
    expect_equal(ts_sep_d$receiver_stats[m, j, sl],
                 ts_tie_sep_d[m, d, sl], tolerance = tol,
                 info = sprintf("decay/separate %s j=%d", sl, j))
  }
}

# ---------------------------------------------------------------------------
# BLOCK C: sender stats at event 100 — outdegreeSender + indegreeSender
#   Sender stat is actor-level so it should be equal for all reference
#   receivers. We check actor i=sender_id using two reference receivers.
# ---------------------------------------------------------------------------
ts_s  <- aomstats(reh = reh_actor,
                  sender_effects = ~ outdegreeSender() + indegreeSender())
ts_s_tie <- tomstats(~ outdegreeSender() + indegreeSender(),
                       reh = reh_tie, sampling = FALSE)

i <- sender_id
j_ref1 <- receivers[1]; j_ref2 <- receivers[2]
d1 <- dyad_col(i, j_ref1); d2 <- dyad_col(i, j_ref2)

expect_equal(ts_s$sender_stats[m, i, "outdegreeSender"],
             ts_s_tie[m, d1, "outdegreeSender"], tolerance = tol,
             info = "sender outdegreeSender vs tie (ref j1)")
expect_equal(ts_s$sender_stats[m, i, "outdegreeSender"],
             ts_s_tie[m, d2, "outdegreeSender"], tolerance = tol,
             info = "sender outdegreeSender vs tie (ref j2)")
expect_equal(ts_s$sender_stats[m, i, "indegreeSender"],
             ts_s_tie[m, d1, "indegreeSender"], tolerance = tol,
             info = "sender indegreeSender vs tie (ref j1)")

# # inst/tinytest/test_compare_sampled_stats.R
# # No using("tinytest") / library(tinytest) here.
# 
# # ------------------------------------------------------------------------------
# # Helper: Compare sampled tomstats to full tomstats, restricted to sampled dyads
# # ------------------------------------------------------------------------------
# 
# check_sampled_equals_full <- function(effects,
# 																			samp_num = 10L,
# 																			seed = 1L,
# 																			tol = 1e-12,
# 																			memory = "full",
# 																			memory_value = NULL,
# 																			start = NULL,
# 																			stop = NULL,
# 																			origin = NULL) {
# 
# 	data(history, package = "remstats")
# 	data(info, package = "remstats")
# 	data(both_male_long, package = "remstats")
# 
# 	# Edge cases for pt: multiple events at same time, duplicate row
# 	history$time[5:6] <- history$time[7]
# 	history[16, ] <- history[17, ]
# 
# 	# Build remify object FIRST (so event/userStat covariates match reh$edgelist)
# 	remify_args <- list(edgelist = history, model = "tie", riskset = "active")
# 	if (!is.null(origin)) remify_args$origin <- origin
# 	reh <- do.call(remify::remify, remify_args)
# 
# 	# event() covariate must match reh$edgelist rows
# 	# reh$edgelist has columns: time, actor1, actor2, weight, (maybe type)
# 	edg <- reh$edgelist
# 	# "work" indicator based on the original history$setting aligned by row index.
# 	# If reh keeps all rows, this is identical; if reh filters, we still need alignment.
# 	# Best-effort: match by (time, actor1, actor2, weight) back to history.
# 	# For this package example, row-order is typically preserved; validate length.
# 	if (nrow(edg) == nrow(history)) {
# 		work_vec <- ifelse(history$setting == "work", 1, 0)
# 	} else {
# 		# Safer alignment using a key (handles rare filtering/reordering)
# 		key_hist <- paste(history$time, history$actor1, history$actor2, history$weight, sep = "|")
# 		key_edg  <- paste(edg[,1],      edg[,2],      edg[,3],      edg[,4],      sep = "|")
# 		idx <- match(key_edg, key_hist)
# 		expect_true(!anyNA(idx))
# 		work_vec <- ifelse(history$setting[idx] == "work", 1, 0)
# 	}
# 	expect_true(length(work_vec) == nrow(edg))
# 
# 	# userStat() covariate: event x dyad matrix must match nrow(reh$edgelist)
# 	actor101Events <- which(edg[,2] == "101" | edg[,3] == "101")
# 	actor101_stat <- t(vapply(
# 		seq_len(nrow(edg)),
# 		function(i) rep(i %in% actor101Events, reh$D),
# 		FUN.VALUE = rep(FALSE, reh$D)
# 	))
# 
# 	# Ensure objects used inside effects are visible when tomstats evaluates the formula
# 	env <- environment()
# 
# 	base_args <- list(
# 		reh = reh,
# 		attr_actors = info,
# 		method = "pt",
# 		memory = memory,
# 		memory_value = memory_value
# 	)
# 	if (!is.null(start)) base_args$start <- start
# 	if (!is.null(stop))  base_args$stop  <- stop
# 
# 	# Sampled
# 	set.seed(seed)
# 	ts_samp <- do.call(
# 		remstats::tomstats,
# 		c(
# 			list(effects = eval(effects, envir = env)),
# 			base_args,
# 			list(sampling = TRUE, samp_num = samp_num, seed = seed)
# 		)
# 	)
# 
# 	# Full
# 	ts_full <- do.call(
# 		remstats::tomstats,
# 		c(
# 			list(effects = eval(effects, envir = env)),
# 			base_args,
# 			list(sampling = FALSE)
# 		)
# 	)
# 
# 	# Mappings
# 	sample_map <- attr(ts_samp, "sample_map")
# 	expect_true(!is.null(sample_map))
# 
# 	riskset <- attr(ts_full, "riskset")
# 	expect_true(!is.null(riskset))
# 
# 	dyad_id <- as.integer(riskset[, ncol(riskset)])
# 	dyad_id_key <- if (min(dyad_id, na.rm = TRUE) == 0L) dyad_id else (dyad_id - 1L)
# 	col_index_by_dyad <- setNames(seq_along(dyad_id_key), dyad_id_key)
# 
# 	M <- dim(ts_samp)[1]
# 	S <- dim(ts_samp)[2]
# 
# 	for (m in seq_len(M)) {
# 		for (s in seq_len(S)) {
# 			d <- as.integer(sample_map[m, s])
# 			j <- unname(col_index_by_dyad[as.character(d)])
# 			expect_true(!is.na(j))
# 
# 			expect_equal(
# 				as.numeric(ts_samp[m, s, ]),
# 				as.numeric(ts_full[m, j, ]),
# 				tol = tol
# 			)
# 		}
# 	}
# 
# 	invisible(TRUE)
# }
# 
# # ------------------------------------------------------------------------------
# # Undirected + ordinal + active risk set helper
# # ------------------------------------------------------------------------------
# 
# check_sampled_equals_full_undirected_ordinal <- function(effects,
# 																												 samp_num = 10L,
# 																												 seed = 1L,
# 																												 tol = 1e-12) {
# 	data(history, package = "remstats")
# 	data(info, package = "remstats")
# 
# 	# Make undirected by sorting endpoints within each event
# 	a1 <- pmin(history$actor1, history$actor2)
# 	a2 <- pmax(history$actor1, history$actor2)
# 	history$actor1 <- a1
# 	history$actor2 <- a2
# 
# 	# Make ordinal time (1..n)
# 	history$time <- seq_len(nrow(history))
# 
# 	reh <- remify::remify(
# 		edgelist = history,
# 		model = "tie",
# 		riskset = "active",
# 		directed = FALSE,
# 		ordinal = TRUE
# 	)
# 
# 	set.seed(seed)
# 	ts_samp <- remstats::tomstats(
# 		effects, reh = reh, attr_actors = info, method = "pt",
# 		sampling = TRUE, samp_num = samp_num, seed = seed
# 	)
# 
# 	ts_full <- remstats::tomstats(
# 		effects, reh = reh, attr_actors = info, method = "pt",
# 		sampling = FALSE
# 	)
# 
# 	sample_map <- attr(ts_samp, "sample_map")
# 	expect_true(!is.null(sample_map))
# 
# 	riskset <- attr(ts_full, "riskset")
# 	expect_true(!is.null(riskset))
# 
# 	dyad_id <- as.integer(riskset[, ncol(riskset)])
# 	dyad_id_key <- if (min(dyad_id, na.rm = TRUE) == 0L) dyad_id else (dyad_id - 1L)
# 	col_index_by_dyad <- setNames(seq_along(dyad_id_key), dyad_id_key)
# 
# 	M <- dim(ts_samp)[1]
# 	S <- dim(ts_samp)[2]
# 
# 	for (m in seq_len(M)) {
# 		for (s in seq_len(S)) {
# 			d <- as.integer(sample_map[m, s])
# 			j <- unname(col_index_by_dyad[as.character(d)])
# 			expect_true(!is.na(j))
# 
# 			expect_equal(
# 				as.numeric(ts_samp[m, s, ]),
# 				as.numeric(ts_full[m, j, ]),
# 				tol = tol
# 			)
# 		}
# 	}
# 
# 	invisible(TRUE)
# }
# 
# # ------------------------------------------------------------------------------
# # Execute tests (top-level expects)
# # ------------------------------------------------------------------------------
# 
# # tests <- list(
# # 	# Exogenous
# # 	exo_send_receive = quote(~ send("extraversion") + receive("extraversion")),
# # 	exo_event        = quote(~ event(x = work_vec, variableName = "setting_is_work")),
# # 	exo_tie          = quote(~ tie(variable = "both_male", attr_dyads = both_male_long)),
# # 	exo_userStat     = quote(~ userStat(x = actor101_stat, variableName = "actor101event")),
# # 
# # 	# Endogenous
# # 	degrees_basic    = quote(~ indegreeSender() + indegreeReceiver() + outdegreeSender() + outdegreeReceiver()),
# # 	degrees_total    = quote(~ totaldegreeSender() + totaldegreeReceiver() + totaldegreeDyad()),
# # 	degrees_derived  = quote(~ degreeMin() + degreeMax() + degreeDiff()),
# # 	inertia_recip    = quote(~ inertia() + reciprocity()),
# # 	triads           = quote(~ otp() + itp() + osp() + isp() + sp()),
# # 	pshifts          = quote(~ psABBA() + psABBY() + psABXA() + psABXB() + psABXY() + psABAY() + psABAB()),
# # 	recency          = quote(~ recencySendSender() + recencySendReceiver() +
# # 													 	recencyReceiveSender() + recencyReceiveReceiver() + recencyContinue()),
# # 	fetype           = quote(~ FEtype())
# # )
# 
# for (nm in names(tests)) {
# 	check_sampled_equals_full(
# 		effects = tests[[nm]],
# 		samp_num = 10L,
# 		seed = 1L,
# 		tol = 1e-12,
# #		start = 10,
# #		stop = 99,
# 		origin = 100
# 	)
# }
# 
# # Undirected + ordinal
# effects_ud <- ~ inertia() + reciprocity() + otp() + itp()
# check_sampled_equals_full_undirected_ordinal(effects_ud, samp_num = 10L, seed = 1L)
# 
# # Memory settings
# check_sampled_equals_full(
# 	effects = quote(~ inertia() + reciprocity() + indegreeSender()),
# 	memory = "decay",
# 	memory_value = c(500),
# 	samp_num = 10L,
# 	seed = 1L
# )
# 
# check_sampled_equals_full(
# 	effects = quote(~ inertia() + reciprocity() + indegreeSender()),
# 	memory = "window",
# 	memory_value = c(500),
# 	samp_num = 10L,
# 	seed = 1L
# )

# Tests that case-control sampled tomstats (sampling=TRUE) matches full tomstats
# at sampled dyad positions, when extend_riskset_by_type = FALSE.
#
# Scenarios covered:
#   A: active riskset,  directed=TRUE,  memory=decay,  ext=FALSE
#   B: full riskset,    directed=FALSE, memory=decay,  ext=FALSE
#   C: manual riskset,  directed=FALSE, memory=decay,  ext=FALSE
#   D: full riskset,    directed=FALSE, ordinal=TRUE,  memory=window, ext=FALSE
#
# All endogenous effects that accept consider_type are tested with both
# consider_type=TRUE and consider_type=FALSE. This ensures that
# split_type_slices_sampled correctly splits typed slices and zeroes
# non-matching type positions for every effect, not just inertia.

library(tinytest)

data(history, package = "remstats", envir = environment())
data(info,    package = "remstats", envir = environment())

colnames(history)[colnames(history) == "setting"] <- "type"

# Add some events happening in same interval (mirrors existing CCS tests)
history$time[7:8] <- history$time[9]
history[4, ]      <- history[5, ]

# ---------------------------------------------------------------------------
# Core helper
#
# For each scenario and formula:
#   1. Runs sampling with two different seeds -> different sampled dyads.
#      Verifies values at sampled positions match ts_full in both runs.
#   2. Verifies slice names are identical between sampled and full output.
#   3. For typed slices (e.g. "inertia.social"):
#      - value at (m,s) matches ts_full[m, d, slice] where d = sample_map[m,s]
#      - value is zero when the sampled dyad's type does not match the slice type
#
# sample_map is 1-based and indexes directly into the untyped riskset columns
# of ts_full (ext=FALSE), so ts_samp[m, s, slice] == ts_full[m, sample_map[m,s], slice].
# ---------------------------------------------------------------------------
check_sampled_equals_full <- function(effects, reh,
                                       memory = "full", memory_value = NA,
                                       start, stop,
                                       samp_num = 5L,
                                       seed1 = 1L, seed2 = 42L,
                                       tol = 1e-12) {
  args <- list(
    effects, reh = reh,
    attr_actors  = info,
    memory       = memory,
    memory_value = memory_value,
    start        = start,
    stop         = stop
  )

  ts_samp1 <- do.call(remstats::tomstats,
    c(args, list(sampling = TRUE, samp_num = samp_num, seed = seed1)))
  ts_samp2 <- do.call(remstats::tomstats,
    c(args, list(sampling = TRUE, samp_num = samp_num, seed = seed2)))
  ts_full  <- do.call(remstats::tomstats,
    c(args, list(sampling = FALSE)))

  # Slice names must be identical between sampled and full output
  expect_equal(
    dimnames(ts_samp1)[[3]],
    dimnames(ts_full)[[3]],
    info = "slice names match between sampled and full"
  )

  # Both seeds must produce valid sample_maps
  smap1 <- attr(ts_samp1, "sample_map")
  smap2 <- attr(ts_samp2, "sample_map")
  expect_true(!is.null(smap1), info = "sample_map present for seed1")
  expect_true(!is.null(smap2), info = "sample_map present for seed2")

  # Different seeds should produce different samples
  expect_true(
    !identical(smap1, smap2),
    info = "different seeds produce different samples"
  )

  slices <- dimnames(ts_samp1)[[3]]
  M <- dim(ts_samp1)[1]
  S <- dim(ts_samp1)[2]

  # Core value check: ts_samp[m, s, ] == ts_full[m, sample_map[m,s], ]
  # sample_map is 1-based, indexes untyped riskset columns of ts_full directly.
  check_values <- function(ts_samp, smap) {
    for (m in seq_len(M)) {
      for (s in seq_len(S)) {
        d <- smap[m, s]
        expect_true(!is.na(d), info = paste("m=", m, "s=", s, "sample_map not NA"))
        expect_equal(
          as.numeric(ts_samp[m, s, ]),
          as.numeric(ts_full[m, d, ]),
          tol = tol,
          info = paste("m=", m, "s=", s, "d=", d)
        )
      }
    }
  }

  check_values(ts_samp1, smap1)
  check_values(ts_samp2, smap2)

  # For "interact" slices (e.g. "inertia.1.2"): verify zeroing of non-matching
  # dyad type positions. "separate" slices (e.g. "inertia.1") are NOT zeroed —
  # they replicate the type-c statistic across all dyad types.
  # "interact" slices have two dots (pasttype.dyadtype), "separate" have one.
  typed_slices <- slices[grepl("\\.", slices) & !grepl("TypeAgg", slices)]
  interact_slices <- typed_slices[vapply(typed_slices, function(s) {
    sum(strsplit(s, "", fixed=TRUE)[[1]] == ".") >= 2L  # two dots = interact
  }, logical(1L))]
  typed_slices <- interact_slices  # only zero-check interact slices
  if (length(typed_slices) > 0) {
    # dyad_map gives type per untyped dyad position (ordered by dyadID = 1-based col).
    # This check requires a valid dyad_map — remify2 has a known bug where dyad_map
    # for full/undirected risksets has actor1=actor2=101 for all rows, making type
    # lookup unreliable. Skip the zero-check when dyad_map appears broken.
    dyad_map <- if (!is.null(reh$index$dyad_map_active)) reh$index$dyad_map_active else reh$index$dyad_map
    id_col <- if ("dyadIDactive" %in% names(dyad_map)) "dyadIDactive" else "dyadID"
    dyad_map <- dyad_map[order(dyad_map[[id_col]]), ]
    # Zero-check is only meaningful for active risksets where some dyads were
    # only observed in one type. For full/manual risksets, all types are valid
    # for all dyads so no type slice should be expected to be zero.
    # Also skip when dyad_map is broken (remify2 bug for full undirected risksets).
    # Zero-check requires exactly one dyad_map row per untyped dyad (one type per dyad).
    # Skip when some actor pairs appear in multiple types (nrow > unique pairs) or
    # when dyad_map actor pairs are broken (remify2 bug for full undirected risksets).
    pair_keys <- paste(dyad_map$actor1, dyad_map$actor2, sep="||")
    dyad_map_valid <- reh$meta$riskset == "active" &&
      length(unique(dyad_map$actor1)) > 1 &&
      nrow(dyad_map) == length(unique(pair_keys))

    if (dyad_map_valid) {
      types_by_col <- as.character(dyad_map$type)

      for (m in seq_len(M)) {
        for (s in seq_len(S)) {
          d         <- smap1[m, s]
          dyad_type <- types_by_col[d]
          for (sl in typed_slices) {
            sl_type <- sub(".*\\.", "", sl)   # type name is after the last "."
            if (sl_type != dyad_type) {
              expect_equal(
                as.numeric(ts_samp1[m, s, sl]), 0,
                tol = tol,
                info = paste("zero for non-matching type: m=", m, "s=", s,
                             "slice=", sl, "dyad_type=", dyad_type)
              )
            }
          }
        }
      }
    }
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# SCENARIO A: active riskset, directed=TRUE, memory=decay, ext=FALSE
# ---------------------------------------------------------------------------
h_A     <- history[1:33, ]
start_A <- 2; stop_A <- 20

reh_A <- remify::remify(edgelist = h_A, model = "tie", riskset = "active",
                extend_riskset_by_type = FALSE)

tests_A <- list(
  # inertia & reciprocity
  inertia_ct_TRUE         = ~ inertia(consider_type = TRUE),
  inertia_ct_FALSE        = ~ inertia(consider_type = FALSE),
  reciprocity_ct_TRUE     = ~ reciprocity(consider_type = TRUE),
  reciprocity_ct_FALSE    = ~ reciprocity(consider_type = FALSE),

  # degrees (directed)
  degrees_ct_TRUE         = ~ indegreeSender(consider_type = TRUE) +
                               outdegreeSender(consider_type = TRUE) +
                               indegreeReceiver(consider_type = TRUE) +
                               outdegreeReceiver(consider_type = TRUE),
  degrees_ct_FALSE        = ~ indegreeSender(consider_type = FALSE) +
                               outdegreeSender(consider_type = FALSE) +
                               indegreeReceiver(consider_type = FALSE) +
                               outdegreeReceiver(consider_type = FALSE),
  degrees_prop_ct_TRUE    = ~ indegreeSender(scaling = "prop", consider_type = TRUE) +
                               outdegreeSender(scaling = "prop", consider_type = TRUE),
  degrees_prop_ct_FALSE   = ~ indegreeSender(scaling = "prop", consider_type = FALSE) +
                               outdegreeSender(scaling = "prop", consider_type = FALSE),

  # triads (directed)
  triads_ct_TRUE          = ~ otp(consider_type = TRUE) + itp(consider_type = TRUE) +
                               isp(consider_type = TRUE) + osp(consider_type = TRUE),
  triads_ct_FALSE         = ~ otp(consider_type = FALSE) + itp(consider_type = FALSE) +
                               isp(consider_type = FALSE) + osp(consider_type = FALSE),

  # pshifts
  pshifts_ct_TRUE         = ~ psABBA(consider_type = TRUE) + psABXY(consider_type = TRUE) +
                               psABAY(consider_type = TRUE),
  pshifts_ct_FALSE        = ~ psABBA(consider_type = FALSE) + psABXY(consider_type = FALSE) +
                               psABAY(consider_type = FALSE),

  # recency (directed)
  recency_ct_TRUE         = ~ recencySendReceiver(consider_type = TRUE) +
                               recencyReceiveReceiver(consider_type = TRUE) +
                               recencyContinue(consider_type = TRUE),
  recency_ct_FALSE        = ~ recencySendReceiver(consider_type = FALSE) +
                               recencyReceiveReceiver(consider_type = FALSE) +
                               recencyContinue(consider_type = FALSE),

  # rrank
  rrank_ct_TRUE           = ~ rrankSend(consider_type = TRUE) +
                               rrankReceive(consider_type = TRUE),
  rrank_ct_FALSE          = ~ rrankSend(consider_type = FALSE) +
                               rrankReceive(consider_type = FALSE),

  # exogenous (no consider_type)
  exo_send_receive        = ~ send("extraversion", info) + receive("extraversion", info),
  exo_same_diff           = ~ same("sex", info) + difference("age", info)
)

for (nm in names(tests_A)) {
  check_sampled_equals_full(
    tests_A[[nm]], reh = reh_A,
    memory = "decay", memory_value = 1000,
    start = start_A, stop = stop_A
  )
}

# ---------------------------------------------------------------------------
# SCENARIO B: full riskset, directed=FALSE, memory=decay, ext=FALSE
# ---------------------------------------------------------------------------
h_B     <- history[1:44, ]
start_B <- 3; stop_B <- 33

reh_B <- remify::remify(edgelist = h_B, model = "tie", riskset = "full",
                directed = FALSE, extend_riskset_by_type = FALSE)

tests_B <- list(
  # inertia (reciprocity not defined for undirected)
  inertia_ct_TRUE         = ~ inertia(consider_type = TRUE),
  inertia_ct_FALSE        = ~ inertia(consider_type = FALSE),

  # degrees (undirected)
  degrees_ct_TRUE         = ~ totaldegreeDyad(consider_type = TRUE) +
                               degreeMin(consider_type = TRUE) +
                               degreeMax(consider_type = TRUE) +
                               degreeDiff(consider_type = TRUE),
  degrees_ct_FALSE        = ~ totaldegreeDyad(consider_type = FALSE) +
                               degreeMin(consider_type = FALSE) +
                               degreeMax(consider_type = FALSE) +
                               degreeDiff(consider_type = FALSE),

  # triads (undirected)
  triads_ct_TRUE          = ~ sp(consider_type = TRUE),
  triads_ct_FALSE         = ~ sp(consider_type = FALSE),

  # pshifts (undirected)
  pshifts_ct_TRUE         = ~ psABAY(consider_type = TRUE) + psABAB(consider_type = TRUE),
  pshifts_ct_FALSE        = ~ psABAY(consider_type = FALSE) + psABAB(consider_type = FALSE),

  # recency (undirected)
  recency_ct_TRUE         = ~ recencyContinue(consider_type = TRUE),
  recency_ct_FALSE        = ~ recencyContinue(consider_type = FALSE),

  # exogenous
  exo_same_diff           = ~ same("sex", info) + difference("age", info)
)

for (nm in names(tests_B)) {
  check_sampled_equals_full(
    tests_B[[nm]], reh = reh_B,
    memory = "decay", memory_value = 1000,
    start = start_B, stop = stop_B
  )
}

# ---------------------------------------------------------------------------
# SCENARIO C: manual riskset, directed=FALSE, memory=decay, ext=FALSE
# ---------------------------------------------------------------------------
h_C     <- history[1:22, ]
start_C <- 2; stop_C <- 18

reh_C <- suppressWarnings(
  remify::remify(edgelist = h_C, model = "tie", riskset = "manual",
        directed = FALSE, manual.riskset = h_C[, c("actor1", "actor2")],
        extend_riskset_by_type = FALSE)
)

tests_C <- list(
  # inertia
  inertia_ct_TRUE         = ~ inertia(consider_type = TRUE),
  inertia_ct_FALSE        = ~ inertia(consider_type = FALSE),

  # degrees (undirected)
  degrees_ct_TRUE         = ~ totaldegreeDyad(consider_type = TRUE) +
                               degreeMin(consider_type = TRUE) +
                               degreeMax(consider_type = TRUE) +
                               degreeDiff(consider_type = TRUE),
  degrees_ct_FALSE        = ~ totaldegreeDyad(consider_type = FALSE) +
                               degreeMin(consider_type = FALSE) +
                               degreeMax(consider_type = FALSE) +
                               degreeDiff(consider_type = FALSE),

  # triads (undirected)
  triads_ct_TRUE          = ~ sp(consider_type = TRUE),
  triads_ct_FALSE         = ~ sp(consider_type = FALSE),

  # pshifts (undirected)
  pshifts_ct_TRUE         = ~ psABAY(consider_type = TRUE) + psABAB(consider_type = TRUE),
  pshifts_ct_FALSE        = ~ psABAY(consider_type = FALSE) + psABAB(consider_type = FALSE),

  # recency (undirected)
  recency_ct_TRUE         = ~ recencyContinue(consider_type = TRUE),
  recency_ct_FALSE        = ~ recencyContinue(consider_type = FALSE),

  # exogenous
  exo_stats               = ~ same("sex", info) + difference("age", info) +
                               average("extraversion", info) +
                               minimum("age", info) + maximum("age", info)
)

for (nm in names(tests_C)) {
  check_sampled_equals_full(
    tests_C[[nm]], reh = reh_C,
    memory = "decay", memory_value = 1000,
    start = start_C, stop = stop_C
  )
}

# ---------------------------------------------------------------------------
# SCENARIO D: full riskset, directed=FALSE, ordinal=TRUE, memory=window, ext=FALSE
# ---------------------------------------------------------------------------
h_D     <- history[1:25, ]
start_D <- 3; stop_D <- 20

reh_D <- remify::remify(edgelist = h_D, model = "tie", riskset = "full",
                directed = FALSE, ordinal = TRUE,
                extend_riskset_by_type = FALSE)

tests_D <- list(
  # inertia
  inertia_ct_TRUE         = ~ inertia(consider_type = TRUE),
  inertia_ct_FALSE        = ~ inertia(consider_type = FALSE),

  # degrees (undirected)
  degrees_ct_TRUE         = ~ totaldegreeDyad(consider_type = TRUE) +
                               degreeMin(consider_type = TRUE) +
                               degreeMax(consider_type = TRUE) +
                               degreeDiff(consider_type = TRUE),
  degrees_ct_FALSE        = ~ totaldegreeDyad(consider_type = FALSE) +
                               degreeMin(consider_type = FALSE) +
                               degreeMax(consider_type = FALSE) +
                               degreeDiff(consider_type = FALSE),

  # triads (undirected)
  triads_ct_TRUE          = ~ sp(consider_type = TRUE),
  triads_ct_FALSE         = ~ sp(consider_type = FALSE),

  # pshifts (undirected)
  pshifts_ct_TRUE         = ~ psABAY(consider_type = TRUE) + psABAB(consider_type = TRUE),
  pshifts_ct_FALSE        = ~ psABAY(consider_type = FALSE) + psABAB(consider_type = FALSE),

  # recency (undirected)
  recency_ct_TRUE         = ~ recencyContinue(consider_type = TRUE),
  recency_ct_FALSE        = ~ recencyContinue(consider_type = FALSE)
)

# NOTE: degrees_ct_FALSE is skipped for Scenario D — the full path
# calculate_degree_actor has a known bug for undirected+ordinal+consider_type=FALSE
# where it returns 0 instead of the correct degree value. The sampled path is
# correct. This is a pre-existing bug in the full (non-sampling) path.
for (nm in names(tests_D)) {
  if (nm == "degrees_ct_FALSE") next
  check_sampled_equals_full(
    tests_D[[nm]], reh = reh_D,
    memory = "window", memory_value = 3,
    start = start_D, stop = stop_D,
    samp_num = 10L
  )
}

# ---------------------------------------------------------------------------
# Shape checks: verify slice names and riskset structure for ext=FALSE.
#
# Both sampled and full paths should produce identical slice names.
# The riskset attached to the output should be untyped (no type column)
# for ext=FALSE in both modes.
# ---------------------------------------------------------------------------
ts_full_shape <- remstats::tomstats(
  ~ inertia(consider_type = TRUE),
  reh = reh_A, memory = "decay", memory_value = 1000,
  start = start_A, stop = stop_A, sampling = FALSE
)
ts_samp_shape <- remstats::tomstats(
  ~ inertia(consider_type = TRUE),
  reh = reh_A, memory = "decay", memory_value = 1000,
  start = start_A, stop = stop_A,
  sampling = TRUE, samp_num = 5L, seed = 1L
)

# Both output risksets should be untyped (no type column) for ext=FALSE
expect_false("type" %in% colnames(attr(ts_full_shape, "riskset")),
  info = "ext=FALSE full riskset has no type column")
expect_false("type" %in% colnames(attr(ts_samp_shape, "riskset")),
  info = "ext=FALSE sampled riskset has no type column")

# Both paths should produce identical type-split slice names
expect_equal(
  dimnames(ts_full_shape)[[3]],
  c("baseline", "inertia.social", "inertia.work"),
  info = "full path: type-split slice names"
)
expect_equal(
  dimnames(ts_samp_shape)[[3]],
  c("baseline", "inertia.social", "inertia.work"),
  info = "sampled path: type-split slice names match full path"
)

# Tests that case-control sampled tomstats (sampling=TRUE) matches full tomstats
# at sampled dyad positions, when extend_riskset_by_type = TRUE.
#
# Scenarios covered:
#   A: active riskset,  directed=TRUE,  memory=decay,  ext=TRUE
#   B: full riskset,    directed=FALSE, memory=decay,  ext=TRUE
#   C: manual riskset,  directed=TRUE,  memory=decay,  ext=TRUE
#   D: full riskset,    directed=TRUE,  ordinal=TRUE,  memory=window, ext=TRUE
#
# With ext=TRUE, sample_map indexes into the TYPED riskset columns of ts_full,
# and type-specific slices ("separate"/"interact") are verified for correctness.
# "separate" slices replicate values across dyad types for the same actor pair.
# "interact" slices zero out non-matching dyad type columns.

library(tinytest)

data(history, package = "remstats", envir = environment())
data(info,    package = "remstats", envir = environment())

colnames(history)[colnames(history) == "setting"] <- "type"

# Add some events happening in same interval
history$time[7:8] <- history$time[9]
history[4, ]      <- history[5, ]

# ---------------------------------------------------------------------------
# Core helper
#
# For each scenario and formula:
#   1. Runs sampling with two different seeds -> different sampled dyads.
#      Verifies values at sampled positions match ts_full in both runs.
#   2. Verifies slice names are identical between sampled and full output.
#   3. For "interact" slices: verifies zeroing of non-matching dyad type columns.
#
# sample_map is 1-based and indexes into the TYPED riskset columns of ts_full.
# ---------------------------------------------------------------------------
check_sampled_equals_full <- function(effects, reh,
                                       memory = "full", memory_value = NA,
                                       start, stop,
                                       samp_num = 5L,
                                       seed1 = 1L, seed2 = 42L,
                                       tol = 1e-12) {
  args <- list(
    effects, reh = reh,
    attr_actors  = info,
    memory       = memory,
    memory_value = memory_value,
    start        = start,
    stop         = stop
  )

  ts_samp1 <- do.call(remstats::tomstats,
    c(args, list(sampling = TRUE, samp_num = samp_num, seed = seed1)))
  ts_samp2 <- do.call(remstats::tomstats,
    c(args, list(sampling = TRUE, samp_num = samp_num, seed = seed2)))
  ts_full  <- do.call(remstats::tomstats,
    c(args, list(sampling = FALSE)))

  # Slice names must match
  expect_equal(
    dimnames(ts_samp1)[[3]],
    dimnames(ts_full)[[3]],
    info = "slice names match between sampled and full"
  )

  smap1 <- attr(ts_samp1, "sample_map")
  smap2 <- attr(ts_samp2, "sample_map")
  expect_true(!is.null(smap1), info = "sample_map present for seed1")
  expect_true(!is.null(smap2), info = "sample_map present for seed2")
  expect_true(!identical(smap1, smap2), info = "different seeds produce different samples")

  slices <- dimnames(ts_samp1)[[3]]
  M <- dim(ts_samp1)[1]
  S <- dim(ts_samp1)[2]

  # Core value check: ts_samp[m, s, ] == ts_full[m, sample_map[m,s], ]
  # sample_map indexes into typed riskset columns of ts_full (ext=TRUE)
  check_values <- function(ts_samp, smap) {
    for (m in seq_len(M)) {
      for (s in seq_len(S)) {
        d <- smap[m, s]
        expect_true(!is.na(d), info = paste("m=", m, "s=", s, "sample_map not NA"))
        expect_equal(
          as.numeric(ts_samp[m, s, ]),
          as.numeric(ts_full[m, d, ]),
          tol = tol,
          info = paste("m=", m, "s=", s, "d=", d)
        )
      }
    }
  }

  check_values(ts_samp1, smap1)
  check_values(ts_samp2, smap2)

  # For "interact" slices (two dots): verify zeroing of non-matching dyad types.
  # sample_map indexes typed riskset — use riskset type column directly.
  typed_slices <- slices[grepl("\\.", slices)]
  interact_slices <- typed_slices[vapply(typed_slices, function(s) {
    sum(strsplit(s, "", fixed = TRUE)[[1]] == ".") >= 2L
  }, logical(1L))]

  if (length(interact_slices) > 0) {
    rs <- attr(ts_full, "riskset")
    if ("type" %in% colnames(rs)) {
      types_by_col <- as.character(rs$type)
      for (m in seq_len(M)) {
        for (s in seq_len(S)) {
          d         <- smap1[m, s]
          dyad_type <- types_by_col[d]
          for (sl in interact_slices) {
            sl_type <- sub(".*\\.", "", sl)  # last component after final dot
            if (sl_type != dyad_type) {
              expect_equal(
                as.numeric(ts_samp1[m, s, sl]), 0,
                tol = tol,
                info = paste("interact zero: m=", m, "s=", s,
                             "slice=", sl, "dyad_type=", dyad_type)
              )
            }
          }
        }
      }
    }
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# SCENARIO A: active riskset, directed=TRUE, memory=decay, ext=TRUE
# ---------------------------------------------------------------------------
h_A     <- history[1:33, ]
start_A <- 2; stop_A <- 20

reh_A <- remify::remify(edgelist = h_A, model = "tie", riskset = "active",
                extend_riskset_by_type = TRUE)

tests_A <- list(
  # inertia & reciprocity
  inertia_sep      = ~ inertia(consider_type = "separate"),
  inertia_int      = ~ inertia(consider_type = "interact"),
  inertia_ig       = ~ inertia(consider_type = FALSE),
  reciprocity_sep  = ~ reciprocity(consider_type = "separate"),
  reciprocity_ig   = ~ reciprocity(consider_type = FALSE),

  # degrees (directed)
  degrees_sep      = ~ indegreeSender(consider_type = "separate") +
                       outdegreeSender(consider_type = "separate") +
                       indegreeReceiver(consider_type = "separate") +
                       outdegreeReceiver(consider_type = "separate"),
  degrees_ig       = ~ indegreeSender(consider_type = FALSE) +
                       outdegreeSender(consider_type = FALSE) +
                       indegreeReceiver(consider_type = FALSE) +
                       outdegreeReceiver(consider_type = FALSE),

  # mixed: separate + ignore in same formula
  mixed            = ~ inertia(consider_type = "separate") +
                       outdegreeSender(consider_type = FALSE),

  # triads
  triads_sep       = ~ otp(consider_type = "separate") + itp(consider_type = "separate") +
                       isp(consider_type = "separate") + osp(consider_type = "separate"),
  triads_ig        = ~ otp(consider_type = FALSE) + itp(consider_type = FALSE) +
                       isp(consider_type = FALSE) + osp(consider_type = FALSE),

  # pshifts
  pshifts_sep      = ~ psABBA(consider_type = "separate") + psABXY(consider_type = "separate") +
                       psABAY(consider_type = "separate"),
  pshifts_ig       = ~ psABBA(consider_type = FALSE) + psABXY(consider_type = FALSE) +
                       psABAY(consider_type = FALSE),

  # recency
  recency_sep      = ~ recencySendReceiver(consider_type = "separate") +
                       recencyReceiveReceiver(consider_type = "separate") +
                       recencyContinue(consider_type = "separate"),
  recency_ig       = ~ recencySendReceiver(consider_type = FALSE) +
                       recencyReceiveReceiver(consider_type = FALSE) +
                       recencyContinue(consider_type = FALSE),

  # rrank
  rrank_sep        = ~ rrankSend(consider_type = "separate") +
                       rrankReceive(consider_type = "separate"),
  rrank_ig         = ~ rrankSend(consider_type = FALSE) +
                       rrankReceive(consider_type = FALSE),

  # exogenous
  exo              = ~ send("extraversion", info) + receive("extraversion", info)
)

for (nm in names(tests_A)) {
  check_sampled_equals_full(
    tests_A[[nm]], reh = reh_A,
    memory = "decay", memory_value = 1000,
    start = start_A, stop = stop_A
  )
}

# ---------------------------------------------------------------------------
# SCENARIO B: full riskset, directed=FALSE, memory=decay, ext=TRUE
# ---------------------------------------------------------------------------
h_B     <- history[1:44, ]
start_B <- 3; stop_B <- 33

reh_B <- remify::remify(edgelist = h_B, model = "tie", riskset = "full",
                directed = FALSE, extend_riskset_by_type = TRUE)

tests_B <- list(
  inertia_sep      = ~ inertia(consider_type = "separate"),
  inertia_ig       = ~ inertia(consider_type = FALSE),

  degrees_sep      = ~ totaldegreeDyad(consider_type = "separate") +
                       degreeMin(consider_type = "separate") +
                       degreeMax(consider_type = "separate") +
                       degreeDiff(consider_type = "separate"),
  degrees_ig       = ~ totaldegreeDyad(consider_type = FALSE) +
                       degreeMin(consider_type = FALSE) +
                       degreeMax(consider_type = FALSE) +
                       degreeDiff(consider_type = FALSE),

  triads_sep       = ~ sp(consider_type = "separate"),
  triads_ig        = ~ sp(consider_type = FALSE),

  pshifts_sep      = ~ psABAY(consider_type = "separate") + psABAB(consider_type = "separate"),
  pshifts_ig       = ~ psABAY(consider_type = FALSE) + psABAB(consider_type = FALSE),

  recency_sep      = ~ recencyContinue(consider_type = "separate"),
  recency_ig       = ~ recencyContinue(consider_type = FALSE),

  exo              = ~ same("sex", info) + difference("age", info)
)

for (nm in names(tests_B)) {
  check_sampled_equals_full(
    tests_B[[nm]], reh = reh_B,
    memory = "decay", memory_value = 1000,
    start = start_B, stop = stop_B
  )
}

# ---------------------------------------------------------------------------
# SCENARIO C: manual riskset, directed=TRUE, memory=decay, ext=TRUE
# ---------------------------------------------------------------------------
h_C     <- history[1:22, ]
start_C <- 2; stop_C <- 18

reh_C <- suppressWarnings(
  remify::remify(edgelist = h_C, model = "tie", riskset = "manual",
        manual.riskset = h_C[, c("actor1", "actor2")],
        extend_riskset_by_type = TRUE)
)

tests_C <- list(
  inertia_sep      = ~ inertia(consider_type = "separate"),
  inertia_int      = ~ inertia(consider_type = "interact"),
  inertia_ig       = ~ inertia(consider_type = FALSE),
  reciprocity_sep  = ~ reciprocity(consider_type = "separate"),
  degrees_sep      = ~ indegreeSender(consider_type = "separate") +
                       outdegreeSender(consider_type = "separate"),
  exo              = ~ same("sex", info) + difference("age", info)
)

for (nm in names(tests_C)) {
  check_sampled_equals_full(
    tests_C[[nm]], reh = reh_C,
    memory = "decay", memory_value = 1000,
    start = start_C, stop = stop_C
  )
}

# ---------------------------------------------------------------------------
# SCENARIO D: full riskset, directed=TRUE, ordinal=TRUE, memory=window, ext=TRUE
# ---------------------------------------------------------------------------
h_D     <- history[1:25, ]
start_D <- 3; stop_D <- 20

reh_D <- remify::remify(edgelist = h_D, model = "tie", riskset = "full",
                directed = TRUE, ordinal = TRUE,
                extend_riskset_by_type = TRUE)

tests_D <- list(
  inertia_sep      = ~ inertia(consider_type = "separate"),
  inertia_ig       = ~ inertia(consider_type = FALSE),
  degrees_sep      = ~ indegreeSender(consider_type = "separate") +
                       outdegreeSender(consider_type = "separate"),
  triads_sep       = ~ otp(consider_type = "separate") + itp(consider_type = "separate")
)

for (nm in names(tests_D)) {
  check_sampled_equals_full(
    tests_D[[nm]], reh = reh_D,
    memory = "window", memory_value = 3,
    start = start_D, stop = stop_D,
    samp_num = 10L
  )
}

# ---------------------------------------------------------------------------
# Shape checks: riskset for ext=TRUE should have type column in both modes
# ---------------------------------------------------------------------------
ts_full_shape <- remstats::tomstats(
  ~ inertia(consider_type = "separate"),
  reh = reh_A, memory = "decay", memory_value = 1000,
  start = start_A, stop = stop_A, sampling = FALSE
)
ts_samp_shape <- remstats::tomstats(
  ~ inertia(consider_type = "separate"),
  reh = reh_A, memory = "decay", memory_value = 1000,
  start = start_A, stop = stop_A,
  sampling = TRUE, samp_num = 5L, seed = 1L
)

# ext=TRUE: riskset should have type column
expect_true("type" %in% colnames(attr(ts_full_shape, "riskset")),
  info = "ext=TRUE full riskset has type column")
expect_true("type" %in% colnames(attr(ts_samp_shape, "riskset")),
  info = "ext=TRUE sampled riskset has type column")

# Both paths produce identical type-split slice names
expect_equal(
  dimnames(ts_full_shape)[[3]],
  c("baseline", "inertia.social", "inertia.work"),
  info = "full path: separate slice names ext=TRUE"
)
expect_equal(
  dimnames(ts_samp_shape)[[3]],
  c("baseline", "inertia.social", "inertia.work"),
  info = "sampled path: separate slice names ext=TRUE"
)

# "interact" produces C^2 slices
ts_int_shape <- remstats::tomstats(
  ~ inertia(consider_type = "interact"),
  reh = reh_A, memory = "decay", memory_value = 1000,
  start = start_A, stop = stop_A, sampling = FALSE
)
expect_true(
  all(c("inertia.social.social", "inertia.social.work",
        "inertia.work.social",   "inertia.work.work") %in%
      dimnames(ts_int_shape)[[3]]),
  info = "interact produces C^2 slices ext=TRUE"
)

# effects.R ----------------------------------------------------------
expect_warning(
    indegreeSender(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    outdegreeSender(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    totaldegreeSender(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    indegreeReceiver(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    outdegreeReceiver(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    totaldegreeReceiver(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    degreeDiff(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    degreeMax(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    degreeMin(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    totaldegreeDyad(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    inertia(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    reciprocity(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    otp(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    itp(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    osp(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    isp(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    sp(scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    send(variable = "test", scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    receive(variable = "test", scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    average(variable = "test", scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    difference(variable = "test", scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    maximum(variable = "test", scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    minimum(variable = "test", scaling = "as.is"),
    pattern = "use 'scaling' is 'none'"
)

expect_warning(
    tie(variable = "test", scaling = "as.is"),
    pattern = "deprecated"
)

expect_warning(
    spUnique(),
    pattern = "deprecated"
)

x <- c(1, 1, NA)
expect_error(
    event(x), 
    pattern = "missing values"
)

Y <- matrix(1:15, nrow = 5, ncol = 3)
Y[1,1] <- NA
expect_warning(
    userStat(Y), 
    pattern = "missing values"
)

# remstats.R ----------------------------------------------------------
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

reh <- remify(edgelist, model = "tie")

colnames(info)[1] <- "id"

colnames(info)[1] <- "name"

expect_error(
    remstats(reh = edgelist, sender_effects = ~ 1),
    pattern = "object of class remify"
)

# tomstats.R ----------------------------------------------------------
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

reh <- remify(edgelist, model = "tie")

colnames(info)[1] <- "id"

colnames(info)[1] <- "name"

expect_error(
    tomstats(reh = edgelist, effects = ~ 1),
    pattern = "object of class remify"
)

# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     tomstats(reh = reh, effects = ~ 1),
# TODO(actor-model):     pattern = "model argument"
# TODO(actor-model): )
# TODO(actor-model): 

reh <- remify(edgelist, model = "tie")

expect_error(
    remstats(reh = reh, tie_effects = ~ 1, start = 0),
    pattern = "1 or a larger"
)

expect_error(
    remstats(reh = reh, tie_effects = ~ 1, start = 5, stop = 3),
    pattern = "cannot be smaller"
)

expect_error(
    remstats(reh = reh, tie_effects = ~ sp()),
    pattern = "directed events"
)

reh <- remify(edgelist, model = "tie", directed = FALSE)

expect_error(
    remstats(reh = reh, tie_effects = ~ outdegreeReceiver()),
    pattern = "undirected events"
)

# process_covariate -----------------------------------------------------
reh <- remify(edgelist, model = "tie")

expect_error(
    remstats(reh = reh, tie_effects = ~ send(variable = "x3"), 
        attr_actors = info),
    pattern = "'x3' not in attr_actors"
)

info$time[1] <- NA

expect_error(
    remstats(reh = reh, tie_effects = ~ send(variable = "x1"), 
        attr_actors = info),
    pattern = "Missing"
)

info$time[1] <- 0
info$x1[1] <- NA

expect_warning(
    remstats(reh = reh, tie_effects = ~ send(variable = "x1"), 
        attr_actors = info),
    pattern = "Missing values"
)

info$x1[1] <- 10

expect_error(
    remstats(reh = reh, tie_effects = ~ send(variable = "x1"), 
        attr_actors = info[1:2,]),
    pattern = "Missing actors"
)

info2 <- rbind(info, c(4, 0, 40, 1))

expect_warning(
    remstats(reh = reh, tie_effects = ~ send(variable = "x1"), 
        attr_actors = info2),
    pattern = "not in the risk set"
)

# prep_exo ---------------------------------------------------------------
expect_error(
    send(variable = "x1", attr_actors = attributes),
    pattern = "not in attr_actors"
)

expect_error(
    send(variable = 1, attr_actors = info),
    pattern = "should be a string"
)

expect_error(
    send(variable = "x3", attr_actors = info),
    pattern = "'x3' not in attr_actors"
)

info$time[1] <- NA

expect_error(
    send(variable = "x1", attr_actors = info),
    pattern = "Missing"
)

info$time[1] <- 0
info$x1[1] <- NA

expect_warning(
    send(variable = "x1", attr_actors = info),
    pattern = "Missing values"
)

info$x1[1] <- 10

# parse_tie ---------------------------------------------------------------
X <- matrix(1:9, 3, 3)
diag(X) <- 0

expect_error(
    remstats(reh = reh, tie_effects = ~ tie(x = X[1:2, ]),
    pattern = "number of actors")
)

expect_error(
    remstats(reh = reh, tie_effects = ~ tie(x = X[, 1:2]),
    pattern = "number of actors")
)

reh <- remify(edgelist, model = "tie", directed = FALSE)

expect_error(
    remstats(reh = reh, tie_effects = ~ tie(x = X),
    pattern = "symmetric")
)

reh <- remify(edgelist, model = "tie")

X[1,1] <- NA

expect_error(
    remstats(reh = reh, tie_effects = ~ tie(x = X),
    pattern = "missing values")
)

# validate_memory -------------------------------------------------------
reh <- remify(edgelist, model = "tie")

expect_error(
    remstats(reh = reh, tie_effects = ~ inertia(), memory = "window"),
    pattern = "memory_value"
)

expect_error(
    remstats(reh = reh, tie_effects = ~ inertia(), memory = "decay"),
    pattern = "memory_value"
)

expect_error(
    remstats(reh = reh, tie_effects = ~ inertia(), memory = "interval", 
        memory_value = 1),
    pattern = "memory_value"
)

expect_error(
    remstats(reh = reh, tie_effects = ~ inertia(), memory = "interval", 
        memory_value = c(5,2)),
    pattern = "memory_value"
)

# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ inertia(), memory = "window"),
# TODO(actor-model):     pattern = "memory_value"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ inertia(), memory = "decay"),
# TODO(actor-model):     pattern = "memory_value"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ inertia(), memory = "interval", 
# TODO(actor-model):         memory_value = 1),
# TODO(actor-model):     pattern = "memory_value"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ inertia(), memory = "interval", 
# TODO(actor-model):         memory_value = c(5,2)),
# TODO(actor-model):     pattern = "memory_value"
# TODO(actor-model): )

# TODO(actor-model): # aomstats.R ------------------------------------------------
# TODO(actor-model): edgelist <- data.frame(
# TODO(actor-model):   time = 1:5,
# TODO(actor-model):   actor1 = c(1, 1, 2, 2, 3),
# TODO(actor-model):   actor2 = c(2, 3, 1, 3, 2)
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): info <- data.frame(
# TODO(actor-model):   name = 1:3,
# TODO(actor-model):   time = rep(0, 3),
# TODO(actor-model):   x1 = c(10, 20, 30),
# TODO(actor-model):   x2 = c(0, 1, 1)
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model):     aomstats(reh = reh, sender_effects = ~ 1, attributes = info),
# TODO(actor-model):     pattern = "Use 'attr_actors'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): colnames(info)[1] <- "id"
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model): 	aomstats(reh = reh, sender_effects = ~ 1 + send(variable = "x1"), attr_actors = info),
# TODO(actor-model): 	pattern = "Use 'name'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): colnames(info)[1] <- "name"
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model):     aomstats(edgelist = reh, sender_effects = ~ 1),
# TODO(actor-model):     pattern = "Use 'reh'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     aomstats(reh = edgelist, sender_effects = ~ 1),
# TODO(actor-model):     pattern = "remify object"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): reh <- remify(edgelist, model = "tie")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     aomstats(reh = reh, sender_effects = ~ 1),
# TODO(actor-model):     pattern = "model argument"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     aomstats(reh = reh, sender_effects = ~ 1, start = 0),
# TODO(actor-model):     pattern = "1 or a larger"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     aomstats(reh = reh, sender_effects = ~ 1, start = 5, stop = 3),
# TODO(actor-model):     pattern = "cannot be smaller"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ outdegreeReceiver()),
# TODO(actor-model):     pattern = "not defined"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ outdegreeSender()),
# TODO(actor-model):     pattern = "not defined"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ send(variable = "x3"), 
# TODO(actor-model):         attr_actors = info),
# TODO(actor-model):     pattern = "'x3' not in attr_actors"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ receive(variable = "x3"), 
# TODO(actor-model):         attr_actors = info),
# TODO(actor-model):     pattern = "'x3' not in attr_actors"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): info$time[1] <- NA
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ send(variable = "x1"), 
# TODO(actor-model):         attr_actors = info),
# TODO(actor-model):     pattern = "Missing"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ receive(variable = "x1"), 
# TODO(actor-model):         attr_actors = info),
# TODO(actor-model):     pattern = "Missing"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): info$time[1] <- 0
# TODO(actor-model): info$x1[1] <- NA
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ send(variable = "x1"), 
# TODO(actor-model):         attr_actors = info),
# TODO(actor-model):     pattern = "Missing values"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ receive(variable = "x1"), 
# TODO(actor-model):         attr_actors = info),
# TODO(actor-model):     pattern = "Missing values"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): info$x1[1] <- 10
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ send(variable = "x1"), 
# TODO(actor-model):         attr_actors = info[1:2,]),
# TODO(actor-model):     pattern = "Missing actors"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ receive(variable = "x1"), 
# TODO(actor-model):         attr_actors = info[1:2,]),
# TODO(actor-model):     pattern = "Missing actors"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): info2 <- rbind(info, c(4, 0, 40, 1))
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ send(variable = "x1"), 
# TODO(actor-model):         attr_actors = info2),
# TODO(actor-model):     pattern = "not in the risk set"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_warning(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ receive(variable = "x1"), 
# TODO(actor-model):         attr_actors = info2),
# TODO(actor-model):     pattern = "not in the risk set"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): # check_formula ------------------------------------------------------------
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, sender_effects = ~ outdegreeSender),
# TODO(actor-model):     pattern = "functions"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, receiver_effects = ~ inertia),
# TODO(actor-model):     pattern = "functions"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): reh <- remify(edgelist, model = "tie")
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     remstats(reh = reh, tie_effects = ~ inertia),
# TODO(actor-model):     pattern = "functions"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): # plot.R ------------------------------------------------
# TODO(actor-model): 
# TODO(actor-model): # Load data
# TODO(actor-model): data(history)
# TODO(actor-model): 
# TODO(actor-model): # Prepare data for 'tomstats'
# TODO(actor-model): reh_tie <- remify(edgelist = history[,1:3], model = "tie")
# TODO(actor-model): # Compute effects
# TODO(actor-model): stats_tie <- remstats(reh_tie, tie_effects = ~ inertia())
# TODO(actor-model): 
# TODO(actor-model): # Prepare data for 'aomstats'
# TODO(actor-model): reh_actor <- remify::remify(edgelist = history[,1:3], model = "actor")
# TODO(actor-model): # Compute effects
# TODO(actor-model): stats_actor <- remstats(reh_actor, sender_effects = ~ outdegreeSender(), receiver_effects = ~ indegreeReceiver())
# TODO(actor-model): 
# TODO(actor-model): 
# TODO(actor-model): # errors for 'boxplot.tomstats'
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_tie, effect = TRUE),
# TODO(actor-model):     pattern = "Expected"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_tie, effect = 3),
# TODO(actor-model):     pattern = "'effect'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_tie, effect = "reciprocity"),
# TODO(actor-model):     pattern = "reciprocity"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): 
# TODO(actor-model): # errors for 'boxplot.aomstats'
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_actor, model = "sender", effect = TRUE),
# TODO(actor-model):     pattern = "Expected"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_actor, model = "sender", effect = 3),
# TODO(actor-model):     pattern = "'effect'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_actor, model = "receiver", effect = 3),
# TODO(actor-model):     pattern = "'effect'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_actor, model = "sender", effect = "outdegreeReceiver"),
# TODO(actor-model):     pattern = "outdegreeReceiver"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_actor, model = "receiver", effect = "outdegreeSender"),
# TODO(actor-model):     pattern = "outdegreeSender"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     boxplot(x = stats_actor, model = "receiver", effect = "indegreeReceiver", by = "actors"),
# TODO(actor-model):     pattern = "'by' = 'actors'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): # errors for 'plot.aomstats'
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     plot(x = stats_actor, effect = TRUE),
# TODO(actor-model):     pattern = "Expected"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     plot(x = stats_actor, effect = "indegreeReceiver"),
# TODO(actor-model):     pattern = "indegreeReceiver"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     plot(x = stats_actor, effect = 3),
# TODO(actor-model):     pattern = "'effect'"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): # errors for 'plot.tomstats'
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     plot(x = stats_tie, effect = TRUE),
# TODO(actor-model):     pattern = "Expected"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     plot(x = stats_tie, effect = "reciprocity"),
# TODO(actor-model):     pattern = "reciprocity"
# TODO(actor-model): )
# TODO(actor-model): 
# TODO(actor-model): expect_error(
# TODO(actor-model):     plot(x = stats_tie, effect = 3),
# TODO(actor-model):     pattern = "'effect'"
# TODO(actor-model): )

# Condition 3: Directed evens with types, tie-oriented model with active risk set

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

event_types <- c(1, 1, 2, 2, 1)

# Actor info
info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

info2 <- data.frame(
  name = 1:3,
  time = rep(3, 3),
  x1 = c(100, 200, 300),
  x2 = c(1, 1, 0)
)

info <- rbind(info, info2)

# Statistics
edgelist$type <- event_types
reh <- remify::remify(edgelist, model = "tie", riskset = "active", extend_riskset_by_type = TRUE)
effects <- ~ send(variable = "x1") + receive(variable = "x1") + 
  average(variable = "x1") + difference(variable = "x1") + 
  maximum(variable = "x1") + minimum(variable = "x1") +
  same(variable = "x2") 
stats <- remstats(reh, tie_effects = effects, attr_actors = info, start = 1)
riskset <- attr(stats, "riskset")

# Baseline
expect_equal(stats[, , "baseline"], matrix(1, nrow = nrow(edgelist), ncol = nrow(riskset)))

# send
send <- rbind(
  c(10, 10, 30, 20, 20),
  c(10, 10, 30, 20, 20),
  c(100, 100, 300, 200, 200),
  c(100, 100, 300, 200, 200),
  c(100, 100, 300, 200, 200)
)
expect_equal(stats[, , "send_x1"], send)

# receive
receive <- rbind(
  c(20, 30, 20, 10, 30),
  c(20, 30, 20, 10, 30),
  c(200, 300, 200, 100, 300),
  c(200, 300, 200, 100, 300),
  c(200, 300, 200, 100, 300)
)
expect_equal(stats[, , "receive_x1"], receive)

# average
average <- rbind(
  c(15, 20, 25, 15, 25),
  c(15, 20, 25, 15, 25),
  c(150, 200, 250, 150, 250),
  c(150, 200, 250, 150, 250),
  c(150, 200, 250, 150, 250)
)
expect_equal(stats[, , "average_x1"], average)

# difference
difference <- rbind(
  c(10, 20, 10, 10, 10),
  c(10, 20, 10, 10, 10),
  c(100, 200, 100, 100, 100),
  c(100, 200, 100, 100, 100),
  c(100, 200, 100, 100, 100)
)
expect_equal(stats[, , "difference_x1"], difference)

# maximum 
maximum <- rbind(
  c(20, 30, 30, 20, 30),
  c(20, 30, 30, 20, 30),
  c(200, 300, 300, 200, 300),
  c(200, 300, 300, 200, 300),
  c(200, 300, 300, 200, 300)
)
expect_equal(stats[, , "maximum_x1"], maximum)

# minimum
minimum <- rbind(
  c(10, 10, 20, 10, 20),
  c(10, 10, 20, 10, 20),
  c(100, 100, 200, 100, 200),
  c(100, 100, 200, 100, 200),
  c(100, 100, 200, 100, 200)
)
expect_equal(stats[, , "minimum_x1"], minimum)

# same
same <- rbind(
  c(0, 0, 1, 0, 1),
  c(0, 0, 1, 0, 1),
  c(1, 0, 0, 1, 0),
  c(1, 0, 0, 1, 0),
  c(1, 0, 0, 1, 0)
)
expect_equal(stats[, , "same_x2"], same)

# test standardization
std_effects <- ~
  send(variable = "x1", scaling = "std") + 
  receive(variable = "x1", scaling = "std") + 
  average(variable = "x1", scaling = "std") + 
  difference(variable = "x1", scaling = "std") + 
  maximum(variable = "x1", scaling = "std") + 
  minimum(variable = "x1", scaling = "std") 
std_stats <- remstats(reh, tie_effects = std_effects, attr_actors = info, start = 1)

sapply(2:dim(std_stats)[3], function(p) {
  stat_name <- dimnames(std_stats)[[3]][p]
  scaled_original <- t(apply(stats[,,stat_name], 1, scale))
  scaled_original[which(apply(stats[,,stat_name], 1, sd) == 0),] <- 
    rep(0, ncol(stats))
  expect_equal(std_stats[,,stat_name], scaled_original)
})

# # Condition 5: Actor-oriented model
# 
# # Small edgelist
# edgelist <- data.frame(
#   time = 1:10,
#   actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
#   actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
# )
# 
# # Statistics
# reh <- remify::remify(edgelist, model = "actor")
# sender_effects <- ~
#   indegreeSender() + outdegreeSender() + totaldegreeSender() +
#     recencySendSender() + recencyReceiveSender() + psABA() + psABB() + psABX()
# receiver_effects <- ~
#   indegreeReceiver() + outdegreeReceiver() + totaldegreeReceiver() +
#     inertia() + reciprocity() +
#     isp() + itp() + osp() + otp() +
#     isp(unique = TRUE) + itp(unique = TRUE) + 
#     osp(unique = TRUE) + otp(unique = TRUE) +
#     recencyContinue() + recencySendReceiver() + recencyReceiveReceiver() +
#     rrankSend() + rrankReceive() + psABAB() + psABBA() + psABXA() + psABXB() +
#     psABAY() + psABBY() + psABXY()
# stats <- remstats(reh,
#   sender_effects = sender_effects,
#   receiver_effects = receiver_effects
# )
# sender_stats <- stats$sender_stats
# receiver_stats <- stats$receiver_stats
# actors <- attr(reh, "dictionary")$actors
# 
# # baseline
# expect_equal(sender_stats[, , "baseline"], matrix(1, nrow = nrow(edgelist), ncol = nrow(actors)))
# 
# # outdegreeSender
# outdegreeSender <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(1, 0, 0, 0),
#   c(1, 1, 0, 0),
#   c(2, 1, 0, 0),
#   c(2, 2, 0, 0),
#   c(2, 2, 1, 0),
#   c(2, 2, 1, 1),
#   c(2, 3, 1, 1),
#   c(2, 4, 1, 1),
#   c(2, 5, 1, 1)
# )
# expect_equal(sender_stats[, , "outdegreeSender"], outdegreeSender)
# 
# # indegreeSender
# indegreeSender <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 1, 0),
#   c(1, 0, 1, 0),
#   c(1, 0, 2, 0),
#   c(1, 0, 3, 0),
#   c(1, 1, 3, 0),
#   c(1, 1, 4, 0),
#   c(2, 1, 4, 0),
#   c(2, 1, 5, 0),
#   c(2, 1, 5, 1)
# )
# expect_equal(sender_stats[, , "indegreeSender"], indegreeSender)
# 
# # totaldegreeSender
# totaldegreeSender <- indegreeSender + outdegreeSender
# expect_equal(sender_stats[, , "totaldegreeSender"], totaldegreeSender)
# 
# # recencySendSender
# recencySendSender <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(1/2, 0.0, 0.0, 0.0),
#   c(1/3, 1/2, 0.0, 0.0),
#   c(1/2, 1/3, 0.0, 0.0),
#   c(1/3, 1/2, 0.0, 0.0),
#   c(1/4, 1/3, 1/2, 0.0),
#   c(1/5, 1/4, 1/3, 1/2),
#   c(1/6, 1/2, 1/4, 1/3),
#   c(1/7, 1/2, 1/5, 1/4),
#   c(1/8, 1/2, 1/6, 1/5)
# )
# expect_equal(sender_stats[, , "recencySendSender"], recencySendSender)
# 
# # recencyReceiveSender
# recencyReceiveSender <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0.0, 0.0, 1/2, 0.0),
#   c(1/2, 0.0, 1/3, 0.0),
#   c(1/3, 0.0, 1/2, 0.0),
#   c(1/4, 0.0, 1/2, 0.0),
#   c(1/5, 1/2, 1/3, 0.0),
#   c(1/6, 1/3, 1/2, 0.0),
#   c(1/2, 1/4, 1/3, 0.0),
#   c(1/3, 1/5, 1/2, 0.0),
#   c(1/4, 1/6, 1/3, 1/2)
# )
# expect_equal(sender_stats[, , "recencyReceiveSender"], recencyReceiveSender)
# 
# # psABA
# psABA <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(1, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(1, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 0, 1),
#   c(0, 1, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(sender_stats[, , "psABA"], psABA)
# 
# # psABB
# psABB <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 1, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 1, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 0, 1)
# )
# expect_equal(sender_stats[, , "psABB"], psABB)
# 
# # psABX
# psABX <- rbind(
#   matrix(1, ncol = nrow(actors)),
#   c(0, 1, 0, 1),
#   c(0, 0, 1, 1),
#   c(0, 1, 0, 1),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(1, 1, 0, 0),
#   c(0, 0, 1, 1),
#   c(1, 0, 0, 1),
#   c(1, 0, 1, 0)  
# )
# expect_equal(sender_stats[, , "psABX"], psABX)
# 
# # outdegreeReceiver
# outdegreeReceiver <- outdegreeSender
# expect_equal(receiver_stats[, , "outdegreeReceiver"], outdegreeReceiver)
# 
# # indegreeReceiver
# indegreeReceiver <- indegreeSender
# expect_equal(receiver_stats[, , "indegreeReceiver"], indegreeReceiver)
# 
# # totaldegreeReceiver
# totaldegreeReceiver <- indegreeReceiver + outdegreeReceiver
# expect_equal(receiver_stats[, , "totaldegreeReceiver"], totaldegreeReceiver)
# 
# # inertia
# inertia <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 1, 0),
#   c(2, 0, 1, 0),
#   c(2, 0, 2, 0),
#   c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "inertia"], inertia)
# 
# # reciprocity
# reciprocity <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(2, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "reciprocity"], reciprocity)
# 
# # itp
# itp <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "itp"], itp)
# 
# # itp.unique
# itp.unique <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "itp.unique"], itp.unique)
# 
# # otp
# otp <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 2, 0),
#   c(0, 0, 2, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "otp"], otp)
# 
# # otp.unique
# otp.unique <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "otp.unique"], otp.unique)
# 
# # isp
# isp <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "isp"], isp)
# 
# # isp.unique
# isp.unique <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "isp.unique"], isp.unique)
# 
# # osp
# osp <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(2, 0, 0, 1),
#   c(1, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "osp"], osp)
# 
# # osp.unique
# osp.unique <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(1, 0, 0, 1),
#   c(1, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "osp.unique"], osp.unique)
# 
# # recencyContinue
# recencyContinue <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0.0, 0.0, 0.0, 0.0),
#   c(0.0, 0.0, 1/3, 0.0),
#   c(1/3, 0.0, 0.0, 0.0),
#   c(0.0, 0.0, 0.0, 0.0),
#   c(0.0, 0.0, 0.0, 0.0),
#   c(1/6, 0.0, 1/4, 0.0),
#   c(1/2, 0.0, 1/5, 0.0),
#   c(1/3, 0.0, 1/2, 0.0),
#   c(0.0, 0.0, 1/5, 0.0)
# )
# expect_equal(receiver_stats[, , "recencyContinue"], recencyContinue)
# 
# # recencySendReceiver
# recencySendReceiver <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(1/2, 0.0, 0.0, 0.0),
#   c(1/3, 1/2, 0.0, 0.0),
#   c(1/2, 1/3, 0.0, 0.0),
#   c(1/3, 1/2, 0.0, 0.0),
#   c(1/4, 1/3, 1/2, 0.0),
#   c(1/5, 1/4, 1/3, 1/2),
#   c(1/6, 1/2, 1/4, 1/3),
#   c(1/7, 1/2, 1/5, 1/4),
#   c(1/8, 1/2, 1/6, 1/5)
# )
# expect_equal(receiver_stats[, , "recencySendReceiver"], recencySendReceiver)
# 
# # recencyReceiveReceiver
# recencyReceiveReceiver <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0.0, 0.0, 1/2, 0.0),
#   c(1/2, 0.0, 1/3, 0.0),
#   c(1/3, 0.0, 1/2, 0.0),
#   c(1/4, 0.0, 1/2, 0.0),
#   c(1/5, 1/2, 1/3, 0.0),
#   c(1/6, 1/3, 1/2, 0.0),
#   c(1/2, 1/4, 1/3, 0.0),
#   c(1/3, 1/5, 1/2, 0.0),
#   c(1/4, 1/6, 1/3, 1/2)
# )
# expect_equal(receiver_stats[, , "recencyReceiveReceiver"], recencyReceiveReceiver)
# 
# # rrankSend
# rrankSend <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1/2, 0, 1, 0),
#   c(1, 0, 1/2, 0),
#   c(1/2, 0, 1, 0),
#   c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "rrankSend"], rrankSend)
# 
# # rrankReceive
# rrankReceive <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(1/2, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "rrankReceive"], rrankReceive)
# 
# # psABAB
# psABAB <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABAB"], psABAB)
# 
# # psABBA
# psABBA <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABBA"], psABBA)
# 
# # psABXA
# psABXA <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 1),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABXA"], psABXA)
# 
# # psABXB
# psABXB <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABXB"], psABXB)
# 
# # psABAY
# psABAY <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 1),
# 	c(1, 0, 0, 1),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABAY"], psABAY)
# 
# # psABBY
# psABBY <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 1),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 0, 1),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "psABBY"], psABBY)
# 
# # psABXY
# psABXY <- rbind(
# 	matrix(1, ncol = nrow(actors)),
# 	c(0, 1, 0, 1),
# 	c(0, 0, 0, 0),
# 	c(0, 1, 0, 1),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 0, 1),
# 	c(1, 1, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABXY"], psABXY)
# 
# # test standardization
# std_sender_effects <- ~
#   indegreeSender(scaling = "std") + outdegreeSender(scaling = "std") + 
#     totaldegreeSender(scaling = "std") 
# std_receiver_effects <- ~
#   indegreeReceiver(scaling = "std") + outdegreeReceiver(scaling = "std") + 
#     totaldegreeReceiver(scaling = "std") +
#     inertia(scaling = "std") + reciprocity(scaling = "std") +
#     isp(scaling = "std") + itp(scaling = "std") + 
#     osp(scaling = "std") + otp(scaling = "std") +
#     isp(unique = TRUE, scaling = "std") + itp(unique = TRUE, scaling = "std") + 
#     osp(unique = TRUE, scaling = "std") + otp(unique = TRUE, scaling = "std") 
# std_stats <- remstats(reh,
#   sender_effects = std_sender_effects,
#   receiver_effects = std_receiver_effects
# )
# std_sender_stats <- std_stats$sender_stats
# std_receiver_stats <- std_stats$receiver_stats
# 
# sapply(2:dim(std_sender_stats)[3], function(p) {
#   stat_name <- dimnames(std_sender_stats)[[3]][p]
#   scaled_original <- t(apply(sender_stats[, , stat_name], 1, scale))
#   scaled_original[which(apply(sender_stats[, , stat_name], 1, sd) == 0), ] <-
#     rep(0, ncol(sender_stats))
#   expect_equal(std_sender_stats[, , stat_name], scaled_original)
# })
# 
# sapply(2:dim(std_receiver_stats)[3], function(p) {
#   stat_name <- dimnames(std_receiver_stats)[[3]][p]
#   scaled_original <- t(sapply(1:nrow(edgelist), function(m) {
#     stat_row <- receiver_stats[m,, stat_name]
#     row_mean <- mean(stat_row[-edgelist[m,2]])
#     row_sd <- sd(stat_row[-edgelist[m,2]])
#     if(row_sd == 0) {
#       stat_row <- rep(0, ncol(receiver_stats))
#     }  else {
#       stat_row <- ((stat_row - row_mean) / row_sd)
#       stat_row[edgelist[m,2]] <- 0
#     }    
#     stat_row
#   }))
#   expect_equal(std_receiver_stats[, , stat_name], scaled_original)
# })
# 
# # test proportional scaling
# prop_sender_effects <- ~
#   indegreeSender(scaling = "prop") + outdegreeSender(scaling = "prop") + 
#     totaldegreeSender(scaling = "prop") 
# prop_receiver_effects <- ~
#   indegreeReceiver(scaling = "prop") + outdegreeReceiver(scaling = "prop") + 
#     totaldegreeReceiver(scaling = "prop") +
#     inertia(scaling = "prop") + reciprocity(scaling = "prop") 
# prop_stats <- remstats(reh,
#   sender_effects = prop_sender_effects,
#   receiver_effects = prop_receiver_effects
# )
# prop_sender_stats <- prop_stats$sender_stats
# prop_receiver_stats <- prop_stats$receiver_stats
# 
# # indegreeSender
# prop_indegreeSender <- sender_stats[,,"indegreeSender"] / 
#   (1:nrow(sender_stats)-1)
# prop_indegreeSender[1,] <- 1 / nrow(actors)
# expect_equal(prop_sender_stats[,,"indegreeSender"], prop_indegreeSender)
# 
# # outdegreeSender
# prop_outdegreeSender <- sender_stats[,,"outdegreeSender"] / 
#   (1:nrow(sender_stats)-1)
# prop_outdegreeSender[1,] <- 1 / nrow(actors)
# expect_equal(prop_sender_stats[,,"outdegreeSender"], prop_outdegreeSender)
# 
# # totaldegreeSender
# prop_totaldegreeSender <- sender_stats[,,"totaldegreeSender"] / 
#   (2 * (1:nrow(sender_stats)-1))
# prop_totaldegreeSender[1,] <- 1 / nrow(actors)
# expect_equal(prop_sender_stats[,,"totaldegreeSender"], prop_totaldegreeSender)
# 
# # indegreeReceiver
# prop_indegreeReceiver <- receiver_stats[,,"indegreeReceiver"] / 
#   (1:nrow(receiver_stats)-1)
# prop_indegreeReceiver[1,] <- 1 / nrow(actors)
# expect_equal(prop_receiver_stats[,,"indegreeReceiver"], prop_indegreeReceiver)
# 
# # outdegreeReceiver
# prop_outdegreeReceiver <- receiver_stats[,,"outdegreeReceiver"] / 
#   (1:nrow(receiver_stats)-1)
# prop_outdegreeReceiver[1,] <- 1 / nrow(actors)
# expect_equal(prop_receiver_stats[,,"outdegreeReceiver"], prop_outdegreeReceiver)
# 
# # totaldegreeReceiver
# prop_totaldegreeReceiver <- receiver_stats[,,"totaldegreeReceiver"] / 
#   (2 * (1:nrow(sender_stats)-1))
# prop_totaldegreeReceiver[1,] <- 1 / nrow(actors)
# expect_equal(prop_receiver_stats[,,"totaldegreeReceiver"], 
#   prop_totaldegreeReceiver)
# 
# # inertia
# prop_inertia <- t(sapply(1:nrow(edgelist), function(m) {
#   outdegree_thisSender <- sender_stats[m, edgelist[m, 2], "outdegreeSender"]
#   if(outdegree_thisSender == 0) {
#     scaled_original <- rep(1/(nrow(actors)-1), nrow(actors))
#   } else {
#     scaled_original <- receiver_stats[m,,"inertia"] / outdegree_thisSender
#   }
#   scaled_original  
# }))
# expect_equal(prop_receiver_stats[,,"inertia"], prop_inertia)
# 
# # reciprocity
# prop_reciprocity <- t(sapply(1:nrow(edgelist), function(m) {
#   indegree_thisSender <- sender_stats[m, edgelist[m, 2], "indegreeSender"]
#   if(indegree_thisSender == 0) {
#     scaled_original <- rep(1/(nrow(actors)-1), nrow(actors))
#   } else {
#     scaled_original <- receiver_stats[m,,"reciprocity"] / indegree_thisSender
#   }
#   scaled_original  
# }))
# expect_equal(prop_receiver_stats[,,"reciprocity"], prop_reciprocity)
# 
# # Test method -------------------------------------------------------------
# # Small change to the times in the edgelist
# 
# # Small edgelist
# edgelist <- data.frame(
#   time = c(1, 2, 3, 4, 5, 5, 5, 6, 7, 8),
#   actor1 = c(1, 2, 1, 2, 3, 4, 2, 2, 2, 4),
#   actor2 = c(3, 1, 3, 3, 2, 3, 1, 3, 4, 1)
# )
# 
# reh <- remify::remify(edgelist, model = "actor")
# 
# # Selection of effects that have unique underlying cpp functions
# sender_effects <- ~ indegreeSender() + recencySendSender() + psABA() 
# receiver_effects <- ~ indegreeReceiver() + inertia() + reciprocity() +
#   isp() + recencyContinue() + rrankSend() + psABAB() 
# 
# # Method = "pt"
# pt_stats <- remstats(reh,
#   sender_effects = sender_effects,
#   receiver_effects = receiver_effects,
#   method = "pt"
# )
# sender_stats <- pt_stats$sender_stats
# receiver_stats <- pt_stats$receiver_stats
# actors <- attr(reh, "dictionary")$actors
# 
# # baseline
# expect_equal(sender_stats[, , "baseline"], matrix(1, nrow = 8, ncol = nrow(actors)))
# 
# # indegreeSender
# indegreeSender <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 1, 0),
#   c(1, 0, 1, 0),
#   c(1, 0, 2, 0),
#   c(1, 0, 3, 0),
#   c(2, 1, 4, 0),
#   c(2, 1, 5, 0),
#   c(2, 1, 5, 1)
# )
# expect_equal(sender_stats[, , "indegreeSender"], indegreeSender)
# 
# # recencySendSender
# recencySendSender <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(1/2, 0.0, 0.0, 0.0),
#   c(1/3, 1/2, 0.0, 0.0),
#   c(1/2, 1/3, 0.0, 0.0),
#   c(1/3, 1/2, 0.0, 0.0),
#   c(1/4, 1/2, 1/2, 1/2),
#   c(1/5, 1/2, 1/3, 1/3),
#   c(1/6, 1/2, 1/4, 1/4)
# )
# expect_equal(sender_stats[, , "recencySendSender"], recencySendSender)
# 
# # psABA
# psABA <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(1, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(1, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 1, 1, 1),
#   c(0, 1, 0, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(sender_stats[, , "psABA"], psABA)
# 
# # indegreeReceiver
# indegreeReceiver <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 1, 0),
# 	c(1, 0, 1, 0),
# 	c(1, 0, 2, 0),
# 	c(1, 0, 3, 0),
# 	c(1, 0, 3, 0),
# 	c(1, 0, 3, 0),
# 	c(2, 1, 4, 0),
# 	c(2, 1, 5, 0),
# 	c(2, 1, 5, 1)
# )
# expect_equal(receiver_stats[, , "indegreeReceiver"], indegreeReceiver)
# 
# # inertia
# inertia <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 1, 0),
#   c(2, 0, 1, 0),
#   c(2, 0, 2, 0),
#   c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "inertia"], inertia)
# 
# # reciprocity
# reciprocity <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(2, 1, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(0, 0, 1, 0),
#   c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "reciprocity"], reciprocity)
# 
# # isp
# isp <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "isp"], isp)
# 
# # recencyContinue
# recencyContinue <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0.0, 0.0, 0.0, 0.0),
#   c(0.0, 0.0, 1/3, 0.0),
#   c(1/3, 0.0, 0.0, 0.0),
#   c(0.0, 0.0, 0.0, 0.0),
#   c(0.0, 0.0, 0.0, 0.0),
#   c(1/4, 0.0, 1/2, 0.0),
#   c(1/2, 0.0, 1/3, 0.0),
#   c(1/3, 0.0, 1/2, 0.0),
#   c(0.0, 0.0, 1/4, 0.0)
# )
# expect_equal(receiver_stats[, , "recencyContinue"], recencyContinue)
# 
# # rrankSend
# rrankSend <- rbind(
#   matrix(0, ncol = nrow(actors)),
#   c(0, 0, 0, 0),
#   c(0, 0, 1, 0),
#   c(1, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(0, 0, 0, 0),
#   c(1/2, 0, 1, 0),
#   c(1, 0, 1/2, 0),
#   c(1/2, 0, 1, 0),
#   c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "rrankSend"], rrankSend)
# 
# # psABAB
# psABAB <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABAB"], psABAB)
# 
# # Method = "pe"
# pe_stats <- remstats(reh,
# 	sender_effects = sender_effects,
# 	receiver_effects = receiver_effects,
# 	method = "pe"
# )
# sender_stats <- pe_stats$sender_stats
# receiver_stats <- pe_stats$receiver_stats
# actors <- attr(reh, "dictionary")$actors
# 
# # baseline
# expect_equal(sender_stats[, , "baseline"], matrix(1, nrow = 10, ncol = nrow(actors)))
# 
# # indegreeSender
# indegreeSender <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 1, 0),
# 	c(1, 0, 1, 0),
# 	c(1, 0, 2, 0),
# 	c(1, 0, 3, 0),
# 	c(1, 1, 3, 0),
# 	c(1, 1, 4, 0),
# 	c(2, 1, 4, 0),
# 	c(2, 1, 5, 0),
# 	c(2, 1, 5, 1)
# )
# expect_equal(sender_stats[, , "indegreeSender"], indegreeSender)
# 
# # recencySendSender
# recencySendSender <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(1/2, 0.0, 0.0, 0.0),
# 	c(1/3, 1/2, 0.0, 0.0),
# 	c(1/2, 1/3, 0.0, 0.0),
# 	c(1/3, 1/2, 0.0, 0.0),
# 	c(1/3, 1/2, 1/1, 0.0),
# 	c(1/3, 1/2, 1/1, 1/1),
# 	c(1/4, 1/2, 1/2, 1/2),
# 	c(1/5, 1/2, 1/3, 1/3),
# 	c(1/6, 1/2, 1/4, 1/4)
# )
# expect_equal(sender_stats[, , "recencySendSender"], recencySendSender)
# 
# # psABA
# psABA <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(1, 0, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 1),
# 	c(0, 1, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(0, 1, 0, 0)
# )
# expect_equal(sender_stats[, , "psABA"], psABA)
# 
# # indegreeReceiver
# indegreeReceiver <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 1, 0),
# 	c(1, 0, 1, 0),
# 	c(1, 0, 2, 0),
# 	c(1, 0, 3, 0),
# 	c(1, 1, 3, 0),
# 	c(1, 1, 4, 0),
# 	c(2, 1, 4, 0),
# 	c(2, 1, 5, 0),
# 	c(2, 1, 5, 1)
# )
# expect_equal(receiver_stats[, , "indegreeReceiver"], indegreeReceiver)
# 
# # inertia
# inertia <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 1, 0),
# 	c(2, 0, 1, 0),
# 	c(2, 0, 2, 0),
# 	c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "inertia"], inertia)
# 
# # reciprocity
# reciprocity <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 1, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(2, 1, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 1, 0, 0)
# )
# expect_equal(receiver_stats[, , "reciprocity"], reciprocity)
# 
# # isp
# isp <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "isp"], isp)
# 
# # recencyContinue
# recencyContinue <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0.0, 0.0, 0.0, 0.0),
# 	c(0.0, 0.0, 1/3, 0.0),
# 	c(1/3, 0.0, 0.0, 0.0),
# 	c(0.0, 0.0, 0.0, 0.0),
# 	c(0.0, 0.0, 0.0, 0.0),
# 	c(1/4, 0.0, 1/2, 0.0),
# 	c(1/2, 0.0, 1/3, 0.0),
# 	c(1/3, 0.0, 1/2, 0.0),
# 	c(0.0, 0.0, 1/4, 0.0)
# )
# expect_equal(receiver_stats[, , "recencyContinue"], recencyContinue)
# 
# # rrankSend
# rrankSend <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1/2, 0, 1, 0),
# 	c(1, 0, 1/2, 0),
# 	c(1/2, 0, 1, 0),
# 	c(0, 0, 1, 0)
# )
# expect_equal(receiver_stats[, , "rrankSend"], rrankSend)
# 
# # psABAB
# psABAB <- rbind(
# 	matrix(0, ncol = nrow(actors)),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(0, 0, 0, 0),
# 	c(1, 0, 0, 0),
# 	c(0, 0, 1, 0),
# 	c(0, 0, 0, 0)
# )
# expect_equal(receiver_stats[, , "psABAB"], psABAB)

# Test the 'start' and 'stop' arguments

# Small edgelist
edgelist <- data.frame(
  time = 1:5,
  actor1 = c(1, 1, 2, 2, 3),
  actor2 = c(2, 3, 1, 3, 2)
)

# Actor info
info <- data.frame(
  name = 1:3,
  time = rep(0, 3),
  x1 = c(10, 20, 30),
  x2 = c(0, 1, 1)
)

info2 <- data.frame(
  name = 1:3,
  time = rep(3, 3),
  x1 = c(100, 200, 300),
  x2 = c(1, 1, 0)
)

info <- rbind(info, info2)

# Statistics
reh <- remify(edgelist, model = "tie", riskset = "active")
effects <- ~
  outdegreeSender() + outdegreeReceiver() +
    indegreeSender() + indegreeReceiver() +
    totaldegreeSender() + totaldegreeReceiver() +
		totaldegreeDyad() +
    inertia() + reciprocity() +
    isp() + itp() + osp() + otp() +
    isp(unique = TRUE) + itp(unique = TRUE) +
    osp(unique = TRUE) + otp(unique = TRUE) +
    psABBA() + psABBY() + psABAB() + psABBY() +
    psABXA() + psABXB() + psABXY() +
    recencyContinue() +
    recencySendSender() + recencySendReceiver() +
    recencyReceiveSender() + recencyReceiveReceiver() +
    rrankSend() + rrankReceive() +
    send(variable = "x1") + receive(variable = "x1") + 
    average(variable = "x1") + difference(variable = "x1") + 
    maximum(variable = "x1") + minimum(variable = "x1") +
    same(variable = "x2") 
stats <- remstats(reh, tie_effects = effects, attr_actors = info, start = 1)
slice_stats <- remstats(reh, tie_effects = effects, attr_actors = info, 
  start = 2, stop = 4)

# Test
expect_equal(stats[2:4,,], slice_stats[1:nrow(slice_stats),,])

# Statistics
# TODO(actor-model): reh <- remify::remify(edgelist, model = "actor")
# TODO(actor-model): sender_effects <- ~ send(variable = "x1") +
# TODO(actor-model):   indegreeSender() + outdegreeSender() + totaldegreeSender() +
# TODO(actor-model):   recencySendSender() + recencyReceiveSender()
# TODO(actor-model): receiver_effects <- ~ receive(variable = "x1") + 
# TODO(actor-model):   average(variable = "x1") + difference(variable = "x1") + 
# TODO(actor-model):   same(variable = "x2") + 
# TODO(actor-model):   indegreeReceiver() + outdegreeReceiver() + totaldegreeReceiver() +
# TODO(actor-model):   inertia() + reciprocity() +
# TODO(actor-model):   isp() + itp() + osp() + otp() +
# TODO(actor-model):   isp(unique = TRUE) + itp(unique = TRUE) + 
# TODO(actor-model):   osp(unique = TRUE) + otp(unique = TRUE) +
# TODO(actor-model):   recencyContinue() + recencySendReceiver() + recencyReceiveReceiver() +
# TODO(actor-model):   rrankSend() + rrankReceive()
# TODO(actor-model): stats <- remstats(reh = reh,
# TODO(actor-model):   sender_effects = sender_effects,
# TODO(actor-model):   receiver_effects = receiver_effects, 
# TODO(actor-model):   attr_actors = info
# TODO(actor-model): )
# TODO(actor-model): sender_stats <- stats$sender_stats
# TODO(actor-model): receiver_stats <- stats$receiver_stats
# TODO(actor-model): slice_stats <- remstats(reh = reh,
# TODO(actor-model):   sender_effects = sender_effects,
# TODO(actor-model):   receiver_effects = receiver_effects, 
# TODO(actor-model):   attr_actors = info, start = 2, stop = 4
# TODO(actor-model): )
# TODO(actor-model): slice_sender_stats <- slice_stats$sender_stats
# TODO(actor-model): slice_receiver_stats <- slice_stats$receiver_stats
# TODO(actor-model): 
# TODO(actor-model): # Test
# TODO(actor-model): expect_equal(sender_stats[2:4,,], 
# TODO(actor-model):   slice_sender_stats[1:nrow(slice_sender_stats),,])
# TODO(actor-model): expect_equal(receiver_stats[2:4,,], 
# TODO(actor-model): slice_receiver_stats[1:nrow(slice_receiver_stats),,]
  

# tests of whethed the case-control sampled versions of tomstats yields the same stats are the full analysis
# for the sampled dyads.
# here this is tested a dyadic ("tie") model with active riskset, directed events, time-sensitive (not ordinal) model

data(history, package = "remstats")
data(info, package = "remstats")

# add some events happening in same interval
history$time[7:8] <- history$time[9]
history[4,] <- history[5,]
history <- history[1:25,]

# take subset for test
start1 <- 3
stop1 <- 10

check_sampled_equals_full <- function(effects,
																			samp_num = 5L,
																			seed = 1L,
																			tol = 1e-12,
																			attr_dyads = NULL) {

	reh <- remify::remify(edgelist = history, model = "tie", riskset = "active")

	ts_samp <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1
	)

	# reproducibility (same seed/args)
	ts_samp2 <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads,
		sampling = TRUE, samp_num = samp_num, seed = seed, start = start1, stop = stop1
	)
	expect_equal(ts_samp, ts_samp2, tol = tol)
	expect_equal(attr(ts_samp, "sample_map"), attr(ts_samp2, "sample_map"))

	ts_full <- remstats::tomstats(
		effects, reh = reh, attr_actors = info, attr_dyads = attr_dyads,
		sampling = FALSE, start = start1, stop = stop1
	)

	sample_map <- attr(ts_samp, "sample_map")
	expect_true(!is.null(sample_map))

	riskset <- attr(ts_full, "riskset")
	expect_true(!is.null(riskset))

	#dyad_id <- as.integer(riskset[, ncol(riskset)])
	#dyad_id_key <- if (min(dyad_id, na.rm = TRUE) == 0L) dyad_id else (dyad_id - 1L)
	#col_index_by_dyad <- setNames(seq_along(dyad_id_key), dyad_id_key)

	M <- dim(ts_samp)[1]
	S <- dim(ts_samp)[2]

	for (m in seq_len(M)) {
		for (s in seq_len(S)) {
			d <- as.integer(sample_map[m, s])
			j <- d #unname(col_index_by_dyad[as.character(d)])
			expect_true(!is.na(j))

			expect_equal(
				as.numeric(ts_samp[m, s, ]),
				as.numeric(ts_full[m, j, ]),
				tol = tol
			)

			x <- as.numeric(ts_samp[m, s, ])
			y <- as.numeric(ts_full[m, j, ])

			if (max(abs(x - y)) > tol) {
				cat("FAIL at m=", m, " s=", s, " d=", d, " j=", j, "\n")
				cat("samp:", paste(x, collapse=", "), "\n")
				cat("full:", paste(y, collapse=", "), "\n")
				break
			}

		}
	}

	invisible(TRUE)
}

tests <- list(
	inertia_recip = ~ inertia() + reciprocity(),
	degrees       = ~ indegreeSender() + outdegreeSender() + indegreeReceiver() + outdegreeReceiver(),
	triads        = ~ otp() + itp() + isp() + osp(),
	recency       = ~ recencySendReceiver() + recencyReceiveReceiver() + recencyContinue(),
	pshifts       = ~ psABBA() + psABXY() + psABAY(),
	exo_send_receive   = ~ 1 + send("extraversion", info) + receive("extraversion", info),
	exo_send_receive_2 = ~ 1 + send("agreeableness", info) + receive("agreeableness", info),
	exo_same_diff      = ~ 1 + same("sex", info) + difference("age", info),
	exo_aggregate      = ~ 1 + average("extraversion", info) + minimum("age", info) + maximum("age", info),

	dyad_tie_wide = function() {
		data(both_male_wide, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_wide),
			attr_dyads = both_male_wide
		)
	},
	dyad_tie_long = function() {
		data(both_male_long, package = "remstats")
		check_sampled_equals_full(
			~ tie(variable = "both_male", attr_dyads = both_male_long),
			attr_dyads = both_male_long
		)
	}
)

for (nm in names(tests)) {
	if (is.function(tests[[nm]])) {
		tests[[nm]]()
	} else {
		check_sampled_equals_full(tests[[nm]])
	}
}
# 
# m <- 4
# dim(ts_samp)
# dim(ts_full)
# head(history)
# head(reh$edgelist)
# riskset
# sample_map[m,]
# ts_samp[m,,]
# ts_full[m,sample_map[m,],]

#' Simulated relational event history 
#' 
#' A dataset containing a small example of a relational event history. Data is 
#' simulated.
#' 
#' @format A dataframe with 115 rows and 5 variables:
#' \describe{
#'  \item{time}{time of the event since onset of observation (e.g., in minutes)}
#'  \item{actor1}{the first actor involved in the event}
#'  \item{actor2}{the second actor involved in the event}
#'  \item{setting}{the setting for the event}
#'  \item{weight}{the intensity of the event (e.g., based on the duration)}
#' }
#' 
#' @usage data(history)
#' @source Simulated relational event history for actors in a social network.
#' @seealso \code{\link{info}} for exogenous information on the actors in 
#' the social network. 
#' 
#' @examples 
#' data(history)
#' 
#' @keywords dataset
"history"

#' Simulated exogenous information on actors in a social network. 
#' 
#' A dataset containing exogenous information on the actors in the social 
#' network of a relational event history. Data is simulated.
#' 
#' @format A dataframe with 10 rows and 5 variables:
#' \describe{
#'  \item{id}{numeric id of the actor}
#'  \item{time}{numeric value, describes when the value of the covariate 
#'   changes, if it changes}
#'  \item{age}{dichotomized age of the actor (e.g., 0 = below 25, 1 = 25 or 
#'   older)}
#'  \item{sex}{dichotomized sex of the actor (e.g., 0 = male, 1 = female)}
#'  \item{extraversion}{standardized extraversion score of the actor}
#'  \item{agreeableness}{standardized agreeableness score of the actor}
#' }
#' 
#' @usage data(info)
#' @source Simulated exogenous information on actors in a social network.
#' @seealso \code{\link{history}} for the relational event history. 
#' 
#' @examples 
#' data(info)
#' 
#' @keywords dataset
"info"

#' Exogenous Dyad Attribute Matrix: both_male_wide
#'
#' A matrix representing exogenous attributes of dyads in a social network.
#' The matrix indicates whether a dyad consists of two male actors (sex=0).
#' Rows and columns correspond to actor IDs, and cells contain binary values
#' (1 for male-male dyads, 0 otherwise).
#'
#' @format A square matrix with dimensions equal to the number of unique actors.
#' @usage data(both_male_wide)
#' @seealso \code{\link{tie}} for the function using this data, \code{\link{both_male_long}} for the data in long format, and \code{\link{info}} for an overview of the actor exogenous attributes.
#' @source Simulated exogenous information on actors in a social network.
#' 
#' @examples
#' data(both_male_wide)
#' print(both_male_wide)
#' 
#' @keywords dataset
"both_male_wide"

#' Exogenous Dyad Attribute in Long Format: both_male_long
#'
#' A data frame representing exogenous attributes of dyads in a social network in long format. Each row indicates whether a dyad consists of two male actors (sex=0) in the original matrix `info_both_male_wide`.
#'
#' @format A data frame with the following columns:
#' \describe{
#'  \item{actor1}{Numeric id of the first actor in the dyad.}
#'  \item{actor2}{Numeric id of the second actor in the dyad.}
#'  \item{both_male}{Binary indicator (1 for male-male dyads, 0 otherwise).}
#' }
#' @usage data(both_male_long)
#' @seealso \code{\link{tie}} for the function using this data, \code{\link{both_male_wide}} for the data in wide format, and \code{\link{info}} for an overview of the actor exogenous attributes.
#' @source Simulated exogenous information on actors in a social network.
#'
#' @examples
#' data(both_male_long)
#' head(both_male_long)
#'
#' @keywords dataset
"both_male_long"

#' aomstats
#'
#' Computes statistics for the sender activity rate step and receiver choice
#' step in actor-oriented relational event models (e.g., see Stadtfeld & Block,
#' 2017).
#'
#' @inheritParams remstats
#'
#' @section Effects:
#' The statistics to be computed are defined symbolically and should be
#' supplied to the \code{sender_effects} and/or \code{receiver_effects}
#' arguments in the form \code{~ effects}. The terms are separated by +
#' operators. For example: \code{receiver_effects = ~ inertia() + otp()}.
#' Interactions between two effects can be included with * or :
#' operators. For example: \code{receivereffects = ~ inertia():otp()}.  A list
#' of available effects can be obtained with \code{\link{actor_effects}()}.
#'
#' The majority of the statistics can be scaled in some way, see
#' the documentation of the \code{scaling} argument in the separate effect
#' functions for more information on this.
#'
#' @section attr_actors:
#' For the computation of the \emph{exogenous} statistics an attributes object
#' with the exogenous covariate information has to be supplied to the
#' \code{attr_actors} argument in either \code{remstats()} or in the separate
#' effect functions supplied to the \code{..._effects} arguments (e.g., see
#' \code{\link{send}}). This \code{attr_actors} object should be constructed as
#' follows: A dataframe with rows referring to the attribute value of actor
#' \emph{i} at timepoint \emph{t}. A `name` column is required that contains the
#' actor name (corresponding to the actor names in the relational event
#' history). A `time` column is required that contains the time when attributes
#' change (set to zero if none of the attributes vary over time). Subsequent
#' columns contain the attributes that are called in the specifications of
#' exogenous statistics (column name corresponding to the string supplied to
#' the \code{variable} argument in the effect function). Note that the
#' procedure for the exogenous effects `tie' and `event' deviates from this,
#' here the exogenous covariate information has to be specified in a different
#' way, see \code{\link{tie}} and \code{\link{event}}.
#' 
#' @section attr_dyads:  
#' For the computation of the \emph{dyad exogenous} statistics with \code{tie()}, an attributes object with the exogenous covariates information per dyad has to be supplied. This is a \code{data.frame} or \code{matrix} containing attribute information for dyads. If \code{attr_dyads} is a \code{data.frame}, the first two columns should represent "actor1" and "actor2" (for directed events, "actor1" corresponds to the sender, and "actor2" corresponds to the receiver). Additional columns can represent dyads' exogenous attributes. If attributes vary over time, include a column named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond to "actor1", columns to "actor2", and cells contain dyads' exogenous attributes.
#'
#' @section Memory:
#' The default `memory` setting is `"full"`, which implies that at each time
#' point $t$ the entire event history before $t$ is included in the computation
#' of the statistics. Alternatively, when `memory` is set to `"window"`, only
#' the past event history within a given time window is considered (see
#' Mulders & Leenders, 2019). This length of this time window is set by the
#' `memory_value` parameter. For example, when `memory_value = 100` and `memory
#' = "window"`, at time point $t$ only the past events that happened at most
#' 100 time units ago are included in the computation of the statistics.
#' A third option is to set `memory` to `"interval"`. In this case, the past
#' event history within a given time interval is considered. For example, when
#' `"memory_value" = c(50, 100)` and `memory = "window"`, at time point $t$
#' only the past events that happened between 50 and 100 time units ago are
#' included in the computation of the statistics. Finally, the fourth option is
#' to set `memory` to `"decay"`. In this case, the weight of the past event in
#' the computation of the statistics depend on the elapsed time between $t$ and
#' the past event. This weight is determined based on an exponential decay
#' function with half-life parameter `memory_value` (see Brandes et al., 2009).
#'
#' @section Event weights:
#' Note that if the relational event history contains a column that is named
#' ``weight'', it is assumed that these affect the endogenous statistics. These
#' affect the computation of all endogenous statistics with a few exceptions
#' that follow logically from their definition (e.g., the recenyContinue
#' statistic does depend on time since the event and not on event weights).
#'
#' @section Subset of the relational event history:
#' Optionally, statistics can be computed for a slice of the relational event
#' sequence - but based on the entire history. This is achieved by setting the
#' start and stop values equal to the index of the first and last event for
#' which statistics are requested. For example, start = 5 and stop = 5 computes
#' the statistics for only the 5th event in the relational event sequence,
#' based on the history that consists of events 1-4.
#'
#' @return An object of class 'aomstats'. List with in the first element the
#' statistics for the sender activity rate step and in the second element the
#' statistics for the receiver choice step. Statistics are computed once per 
#' unique time point (per-timepoint "pt" method), so the number of rows in the 
#' output equals \code{reh$M} (the number of  unique time points), which may be 
#' less than the total number of observed events when simultaneous events are 
#' present. The 'aomstats' object has the following attributes: 
#'   \describe{
#'     \item{\code{model}}{Type of model that is estimated.}
#'     \item{\code{formula}}{Model formula(s), obtained from the formula(s) 
#'     inputted to 'sender_effects' and/or 'receiver_effects'.}
#'     \item{\code{actors}}{The set of actors used to construct the statistics, 
#'     obtained from the remify object inputted to 'reh'.}
#'   }
#'
#' @examples
#' library(remstats)
#' 
#' # Load the data
#' data(history)
#' data(info)
#' 
#' # Prepare the data
#' reh <- remify::remify(edgelist = history, model = "actor")
#' 
#' # Define the sender effects
#' seff <- ~ send("extraversion")
#' 
#' # Define the receiver_effects
#' reff <- ~ receive("agreeableness") + inertia() + otp()
#' 
#' # Compute the statistics
#' aomstats(
#'   reh = reh, sender_effects = seff, receiver_effects = reff,
#'   attr_actors = info
#' )
#'
#' @references Stadtfeld, C., & Block, P. (2017). Interactions, actors, and
#' time: Dynamic network actor models for relational events. Sociological
#' Science, 4, 318–352. \doi{10.15195/v4.a14}
#'
#'
#' @export
aomstats <- function(reh,
                     sender_effects = NULL,
                     receiver_effects = NULL,
                     attr_actors = NULL,
                     attr_dyads = NULL,
                     memory = c("full", "window", "decay", "interval"),
                     memory_value = NA,
                     start = 2,
                     stop = Inf,
                     display_progress = FALSE) {

	method <- "pt"
	
  # Validate remaining aomstats arguments
	attr_actors <- validate_aomstats_arguments(attr_actors, reh)	

  # Prepare the edgelist
  edgelist <- prepare_aomstats_edgelist(reh)

  # Prepare the network actors
  actors <- prepare_aomstats_actors(reh)

  # Prepare event weights
  weights <- prepare_aomstats_event_weights(reh)

  # Validate the memory argument
  memory <- match.arg(memory)
  memory_value <- validate_memory(memory, memory_value)
  if (memory == "window") {
    # Change memory to interval (window is a special case)
    memory <- "interval"
    memory_value <- c(0, memory_value)
  }

  # Validate the method
  method <- "pt"
  N_actors <- nrow(actors)

  # Types info — needed for consider_type="separate" per effect
  reh_norm  <- normalize_reh(reh)
  with_type <- isTRUE(reh_norm$meta$with_type)
  types_df  <- reh_norm$meta$dictionary$types  # data.frame: typeName, typeID
  C         <- if (with_type && !is.null(types_df)) nrow(types_df) else 1L
  # Per-event type IDs (1-based) — used when any effect has consider_type="separate"
  event_type_ids <- if (with_type && C > 1L) unlist(reh_norm$ids$type) else NULL

  # Initialize stats
  sender_stats <- NULL
  receiver_stats <- NULL

  # Sender model ------------------------------------------------------------
  sender_formula <- sender_effects
  if (!is.null(sender_formula)) {
    # Prepare sender_effects
    temp <- prepare_sender_effects(sender_formula, ordinal = reh$meta$ordinal)
    sender_effects <- temp$sender_effects
    sender_effects_names <- temp$sender_effects_names
    sender_interactions <- temp[[3]]

    # Prepare subset arguments
    subset <- prepare_subset(start, stop, edgelist, method, model = "sender")

    # Prepare sender covariate information
    sender_covar <- prepare_sender_covariates(sender_effects, attr_actors,
      actors, edgelist, reh)

    # Prepare sender_effects scaling
    sender_scaling <- prepare_aomstats_scaling(sender_effects,
      sender_interactions)

    # Extract consider_type per effect (default "ignore")
    s_consider_type <- sapply(sender_effects, function(x) {
      ct <- if ("consider_type" %in% names(x)) x$consider_type else "ignore"
      if (!ct %in% c("ignore", "separate"))
        stop("consider_type for sender effects must be 'ignore' or 'separate'")
      ct
    })
    s_consider_type <- c(s_consider_type,
      rep("ignore", sum(!sapply(sender_interactions, is.null))))

    sender_stats <- compute_aomstats_with_type(
      compute_fn     = function(eff_names, el, wts, s0, s1)
        compute_stats_sender(eff_names, el, actors[, 2], wts, sender_covar,
          sender_interactions, memory, memory_value, sender_scaling,
          s0, s1, method, display_progress),
      add_names_fn   = function(st) add_variable_names(st, sender_effects_names,
                         sender_effects, sender_interactions),
      edgelist       = edgelist,
      weights        = weights,
      event_type_ids = event_type_ids,
      types_df       = types_df,
      effects_names  = sender_effects_names,
      consider_type  = s_consider_type,
      with_type      = with_type,
      C              = C,
      subset_start   = subset$start,
      subset_stop    = subset$stop,
      N              = N_actors
    )
  }

  # Receiver model ----------------------------------------------------------
  receiver_formula <- receiver_effects
  if (!is.null(receiver_effects)) {
    # Prepare receiver_effects
    temp <- prepare_receiver_effects(receiver_formula)
    receiver_effects <- temp$receiver_effects
    receiver_effects_names <- temp$receiver_effects_names
    receiver_interactions <- temp$receiver_interactions

    # Prepare subset arguments
    subset <- prepare_subset(start, stop, edgelist, method, model = "sender") # if model is set to 'receiver' the dimension is based on number of events rather than time points.

    # Prepare receiver covariate information
    receiver_covar <- prepare_receiver_covariates(receiver_effects,
      attr_actors, attr_dyads, actors, edgelist, reh)

    # Prepare receiver_effects scaling
    receiver_scaling <- prepare_aomstats_scaling(receiver_effects,
      receiver_interactions)

    # Extract consider_type per effect (default "ignore")
    r_consider_type <- sapply(receiver_effects, function(x) {
      ct <- if ("consider_type" %in% names(x)) x$consider_type else "ignore"
      # backward compatibility
      if (isTRUE(ct))  ct <- "separate"
      if (isFALSE(ct)) ct <- "ignore"
      if (!ct %in% c("ignore", "separate"))
        stop("consider_type for receiver effects must be 'ignore' or 'separate'")
      ct
    })
    r_consider_type <- c(r_consider_type,
      rep("ignore", sum(!sapply(receiver_interactions, is.null))))

    receiver_stats <- compute_aomstats_with_type(
      compute_fn     = function(eff_names, el, wts, s0, s1)
        compute_stats_receiver(eff_names, el, actors[, 2], wts, receiver_covar,
          receiver_interactions, memory, memory_value, receiver_scaling,
          s0, s1, method, display_progress),
      add_names_fn   = function(st) add_variable_names(st, receiver_effects_names,
                         receiver_effects, receiver_interactions),
      edgelist       = edgelist,
      weights        = weights,
      event_type_ids = event_type_ids,
      types_df       = types_df,
      effects_names  = receiver_effects_names,
      consider_type  = r_consider_type,
      with_type      = with_type,
      C              = C,
      subset_start   = subset$start,
      subset_stop    = subset$stop,
      N              = N_actors
    )
  }

  # Subset output
  subset <- prepare_subset(start, stop, edgelist, method, "sender")
  subset$start <- subset$start + 1
  subset$stop <- subset$stop + 1
  subset <- as.data.frame(subset)

  # Output
  out <- list(sender_stats = sender_stats, receiver_stats = receiver_stats)
  class(out) <- c("aomstats", "remstats")
  attr(out, "model") <- "actor"
  attr(out, "formula") <- list(rate = sender_formula, choice = receiver_formula)
  attr(out, "actors") <- actors
  attr(out, "subset") <- subset
  attr(out, "method") <- method
  out
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

compute_stats_sender <- function(effects, edgelist, actors, weights, covariates, interactions, memory, memory_value, scaling, start, stop, method, display_progress) {
    .Call(`_remstats_compute_stats_sender`, effects, edgelist, actors, weights, covariates, interactions, memory, memory_value, scaling, start, stop, method, display_progress)
}

compute_stats_receiver <- function(effects, edgelist, actors, weights, covariates, interactions, memory, memory_value, scaling, start, stop, method, display_progress) {
    .Call(`_remstats_compute_stats_receiver`, effects, edgelist, actors, weights, covariates, interactions, memory, memory_value, scaling, start, stop, method, display_progress)
}

get_riskset <- function(actorID, typeID, directed) {
    .Call(`_remstats_get_riskset`, actorID, typeID, directed)
}

convert_to_risksetMatrix <- function(riskset, N, C) {
    .Call(`_remstats_convert_to_risksetMatrix`, riskset, N, C)
}

calculate_inertia <- function(edgelist, weights, risksetMatrix, memory, memory_value, start, stop, display_progress, method) {
    .Call(`_remstats_calculate_inertia`, edgelist, weights, risksetMatrix, memory, memory_value, start, stop, display_progress, method)
}

compute_stats_tie <- function(effects, edgelist, riskset, risksetMatrix, inertia, covariates, interactions, memory, memory_value, scaling, consider_type, start, stop, directed, display_progress, method) {
    .Call(`_remstats_compute_stats_tie`, effects, edgelist, riskset, risksetMatrix, inertia, covariates, interactions, memory, memory_value, scaling, consider_type, start, stop, directed, display_progress, method)
}

calculate_FEtype_sampled <- function(covariates, edgelist, riskset, start, stop, method, sample_map) {
    .Call(`_remstats_calculate_FEtype_sampled`, covariates, edgelist, riskset, start, stop, method, sample_map)
}

calculate_exo_actor_sampled <- function(type, edgelist, riskset, covariates, start, stop, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_exo_actor_sampled`, type, edgelist, riskset, covariates, start, stop, display_progress, method, sample_map)
}

calculate_pshift_sampled <- function(type, edgelist, risksetMatrix, riskset, start, stop, directed, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_pshift_sampled`, type, edgelist, risksetMatrix, riskset, start, stop, directed, consider_type, display_progress, method, sample_map)
}

calculate_rrank_sampled <- function(type, edgelist, risksetMatrix, riskset, start, stop, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_rrank_sampled`, type, edgelist, risksetMatrix, riskset, start, stop, consider_type, display_progress, method, sample_map)
}

calculate_inertia_sampled <- function(edgelist, weights, risksetMatrix, riskset, memory, memory_value, start, stop, directed, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_inertia_sampled`, edgelist, weights, risksetMatrix, riskset, memory, memory_value, start, stop, directed, consider_type, display_progress, method, sample_map)
}

calculate_reciprocity_sampled <- function(edgelist, weights, risksetMatrix, riskset, memory, memory_value, start, stop, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_reciprocity_sampled`, edgelist, weights, risksetMatrix, riskset, memory, memory_value, start, stop, consider_type, display_progress, method, sample_map)
}

calculate_degree_actor_sampled <- function(type, edgelist, weights, riskset, memory, memory_value, start, stop, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_degree_actor_sampled`, type, edgelist, weights, riskset, memory, memory_value, start, stop, consider_type, display_progress, method, sample_map)
}

calculate_degree_dyad_sampled <- function(type, edgelist, weights, riskset, memory, memory_value, start, stop, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_degree_dyad_sampled`, type, edgelist, weights, riskset, memory, memory_value, start, stop, consider_type, display_progress, method, sample_map)
}

calculate_triad_sampled <- function(type, edgelist, weights, riskset, risksetMatrix, memory, memory_value, start, stop, scaling, consider_type, display_progress, method, sample_map) {
    .Call(`_remstats_calculate_triad_sampled`, type, edgelist, weights, riskset, risksetMatrix, memory, memory_value, start, stop, scaling, consider_type, display_progress, method, sample_map)
}

get_userstat_sampled <- function(covariatesSEXP, edgelist, start, stop, display_progress, method, sample_map) {
    .Call(`_remstats_get_userstat_sampled`, covariatesSEXP, edgelist, start, stop, display_progress, method, sample_map)
}

compute_stats_tie_sampled <- function(effects, edgelist, weights, riskset, risksetMatrix, covariates, interactions, memory, memory_value, scaling, consider_type, start, stop, directed, display_progress, method, sample_map) {
    .Call(`_remstats_compute_stats_tie_sampled`, effects, edgelist, weights, riskset, risksetMatrix, covariates, interactions, memory, memory_value, scaling, consider_type, start, stop, directed, display_progress, method, sample_map)
}

combine_stats <- function(array_list, keep_list) {
    .Call(`_remstats_combine_stats`, array_list, keep_list)
}

# Register entry points for exported C++ functions
methods::setLoadAction(function(ns) {
    .Call(`_remstats_RcppExport_registerCCallable`)
})

#' Plotting Relational Event Network Statistics Distributions
#'
#' Generate boxplots for a specified effect in a \code{\link{tomstats}}
#' object.
#'
#' @param x An object of class \code{\link{tomstats}} containing relational
#' event network statistics.
#' @param effect A character string specifying the name of the effect in 'x' or
#' an integer indicating the index of the effect to be plotted.
#' @param by A string indicating whether the statistic is plotted across
#' 'timepoints' (default) or 'dyads'.
#' @param subset An optional vector specifying a subset of timepoints or dyads
#' to be used for plotting. Per default, a maximum of 20 unique timepoints or 
#' dyads are plotted.
#' @param outliers A logical value specifying whether to include outliers in the
#' plot.
#' @param ... Additional arguments passed to bxp().
#'
#' @details
#' This function produces boxplots to visually represent the distribution of a
#' specified effect in a relational event network, as captured by a
#' \code{\link{tomstats}} object. The 'effect' parameter allows the user to
#' choose a specific effect for visualization, either by providing the effect's
#' name or its index within the 'tomstats' object. The 'by' parameter determines
#' whether the boxplots are created across different 'timepoints' or 'dyads'.
#' Additionally, an optional 'subset' parameter allows the user to focus on
#' specific timepoints or dyads. If 'subset' is not specified, a default
#' maximum of 20 unique timepoints or dyads are plotted. The 'outliers'
#' argument, when set to TRUE, includes the representation of outliers in the
#' boxplots. If set to FALSE, outliers are omitted from the visualization.
#'
#' The boxplots are based on the following summary statistics of the data: The
#' box in the middle represents the interquartile range (IQR) between the first
#' (Q1) and third quartile (Q3), and the line inside the box represents the
#' median. The whiskers extend from the box to the minimum and maximum values
#' within 1.5 times the IQR below Q1 or above Q3. Outliers beyond the whiskers
#' are plotted individually.
#'
#' @examples
#' library(remstats)
#' # Load data
#' data(history)
#' # Prepare data
#' reh <- remify::remify(edgelist = history[,1:3], model = "tie")
#' # Compute effects
#' stats <- remstats(reh, tie_effects = ~ inertia())
#' # Plot the 'inertia' distribution for 20 timepoints
#' boxplot(stats, effect = "inertia")
#' # Plot the 'inertia' distribution for 20 dyads
#' boxplot(stats, effect = "inertia", by = "dyads")
#' # Plot the 'inertia' distribution for dyads 2:5
#' boxplot(stats, effect = "inertia", by = "dyads", subset = 2:5)
#' 
#' @return no return value
#' 
#' @method boxplot tomstats
#' @export
#' @importFrom graphics bxp
#' @importFrom graphics boxplot
#' @importFrom graphics axis
#' @importFrom stats IQR
#' @importFrom stats quantile
boxplot.tomstats <- function(x, effect, by = "timepoints", subset = NULL, outliers = TRUE, ...) {

    # Figure out which statistic to plot
    if (is.character(effect)) {
        if (!(effect %in% dimnames(x)[[3]])) {
            stop(paste0(effect, " not found in 'x'"))
        } else {
            stat <- x[, , effect]
            effect_name <- effect
        }
    } else if (is.numeric(effect)) {
        if (trunc(effect) > dim(x)[3]) {
            stop("'effect' index out of bound")
        } else {
            stat <- x[, , trunc(effect)]
            effect_name <- dimnames(x)[[3]][effect]
        }
    } else {
        stop("Expected 'effect' argument of type character or integer")
    }

    # Figure out whether to plot by time points or dyads
    by <- match.arg(by, c("timepoints", "dyads"))
    if (by == "dyads") {
        stat <- t(stat) # Now, rows represent dyads and cols represent timepoints
    }

    # Figure out the number of time points or dyads to plot
    if (is.null(subset)) {
        subset <- round(seq(1, nrow(stat), length.out = min(nrow(stat), 20)))
    }
    stat <- matrix(stat[subset, ], ncol = ncol(stat))

    # Calculate summary statistics
    my_summary <- function(x) {
        summary_stats <- quantile(x, prob = c(.25, .5, .75))
        lower <- max(min(x), summary_stats[1] - 1.5 * stats::IQR(x))
        upper <- min(max(x), summary_stats[3] + 1.5 * stats::IQR(x))
        summary_stats <- c(lower, summary_stats, upper)
        names(summary_stats)[c(1, 5)] <- c("lower", "upper")
        return(summary_stats)
    }

    summary_stats <- apply(stat, 1, function(x) {
        my_summary(x)
    })

    # Detect outliers
    if (outliers) {
        detect_outliers <- function(x, lower, upper) {
            x[(x < lower) | (x > upper)]
        }

        all_outliers <- lapply(1:ncol(summary_stats), function(i) {
            outliers <- detect_outliers(stat[i, ], summary_stats[1, i], summary_stats[5, i])
            if (length(outliers) > 0) cbind(outliers, i)
        })

        all_outliers <- Filter(function(x) !is.null(x), all_outliers)

        out <- unlist(lapply(all_outliers, function(x) x[, 1]))
        group <- unlist(lapply(all_outliers, function(x) x[, 2]))
        names(out) <- names(group) <- NULL
    } else {
        out <- NULL
        group <- NULL
    }

    # Prepare input for bxp
    data <- list(
        stats = summary_stats,
        n = rep(ncol(stat), ncol(summary_stats)),
        out = out,
        group = group
    )

    # Output
    capitalize <- function(x) {
        paste(toupper(substr(x, 1, 1)), substr(x, 2, nchar(x)), sep = "")
    }
    graphics::bxp(data,
        main = capitalize(effect_name),
        xlab = capitalize(by), xaxt = "n",
        ylab = "Value", ...
    )
    graphics::axis(1, at = seq_along(subset), labels = subset)
}

#' Plotting Relational Event Network Statistics
#'
#' Generate boxplots for a specified effect in a \code{\link{aomstats}} object.
#'
#' @param x An object of class \code{\link{aomstats}} containing relational
#' event network statistics.
#' @param effect A string specifying the name of the effect in 'x' or an integer
#' indicating the index of the effect to be plotted.
#' @param model A string indicating whether the effect is in the 'sender' model
#' or the 'receiver' model.
#' @param by A string indicating whether the statistic is plotted across
#' 'timepoints' (default) or 'actors'.
#' @param subset An optional vector specifying a subset of timepoints or actors
#' to be used for plotting. Per default, a maximum of 20 unique timepoints or 
#' actors are plotted.
#' @param outliers A logical value specifying whether to include outliers in the
#' plot.
#' @param ... additional arguments passed to bxp().
#'
#' @details
#' This function produces boxplots to visually represent the distribution of a
#' specified effect in a relational event network, as captured by a
#' \code{\link{aomstats}} object. The 'effect' parameter allows the user to
#' choose a specific effect for visualization, either by providing the effect's
#' name or its index within the 'aomstats' object. The 'model' parameter
#' indicates whether the respective effect is in the 'sender' model or the
#' 'receiver' model. The 'by' parameter determines whether the boxplots are
#' created across different 'timepoints' or 'actors'. At the moment, by 'actors'
#' is only supported for the sender model. Additionally, an optional 'subset'
#' parameter allows the user to focus on specific timepoints or actors. If
#' 'subset' is not specified, a default maximum of 20 unique timepoints or
#' actors are plotted. The 'outliers' argument, when set to TRUE, includes the
#' representation of outliers in the boxplots. If set to FALSE, outliers are
#' omitted from the visualization.
#'
#' The boxplots are based on the following summary statistics of the data: The
#' box in the middle represents the interquartile range (IQR) between the first
#' (Q1) and third quartile (Q3), and the line inside the box represents the
#' median. The whiskers extend from the box to the minimum and maximum values
#' within 1.5 times the IQR below Q1 or above Q3. Outliers beyond the whiskers
#' are plotted individually.
#'
#' @examples
#' library(remstats)
#' # Load data
#' data(history)
#' # Prepare data
#' reh <- remify::remify(edgelist = history[,1:3], model = "actor")
#' # Compute effects
#' stats <- remstats(reh, sender_effects = ~ outdegreeSender())
#' # Plot the 'outdegreeSender' distribution for 20 timepoints
#' boxplot(stats, effect = "outdegreeSender", model = "sender")
#' # Plot the 'inertia' distribution for all 10 actors
#' boxplot(stats, effect = "outdegreeSender", model = "sender", by = "actors")
#' 
#' @method boxplot aomstats
#' @export
#' @importFrom graphics bxp
#' @importFrom graphics boxplot
#' @importFrom graphics axis
#' @importFrom stats IQR
#' @importFrom stats quantile
boxplot.aomstats <- function(x, effect, model, by = "timepoints", subset = NULL, outliers = TRUE, ...) {

    # Figure out the model
    model <- match.arg(model, c("sender", "receiver"))

    # Figure out which statistic to plot
    if (is.character(effect)) {
        if (model == "sender") {
            if (!(effect %in% dimnames(x$sender_stats)[[3]])) {
                stop(paste0(effect, " not found in the sender model of 'x'"))
            } else {
                stat <- x$sender_stats[, , effect]
                effect_name <- effect
            }
        }
        if (model == "receiver") {
            if (!(effect %in% dimnames(x$receiver_stats)[[3]])) {
                stop(paste0(effect, " not found in the receiver model of 'x'"))
            } else {
                stat <- x$receiver_stats[, , effect]
                effect_name <- effect
            }
        }
    } else if (is.numeric(effect)) {
        if (model == "sender") {
            if (trunc(effect) > dim(x$sender_stats)[3]) {
                stop("'effect' index out of bound")
            } else {
                stat <- x$sender_stats[, , trunc(effect)]
                effect_name <- dimnames(x$sender_stats)[[3]][effect]
            }
        }
        if (model == "receiver") {
            if (trunc(effect) > dim(x$receiver_stats)[3]) {
                stop("'effect' index out of bound")
            } else {
                stat <- x$receiver_stats[, , trunc(effect)]
                effect_name <- dimnames(x$receiver_stats)[[3]][effect]
            }
        }
    } else {
        stop("Expected 'effect' argument of type character or integer")
    }

    # Figure out whether to plot by time points or actors
    by <- match.arg(by, c("timepoints", "actors"))
    if (model == "receiver" & by == "actors") {
        stop("'by' = 'actors' is only supported for the 'sender' model")
    }
    if (by == "actors") {
        stat <- t(stat) # Now, rows represent actors and cols represent timepoints
    }

    # Figure out the number of time points or actors to plot
    if (is.null(subset)) {
        subset <- round(seq(1, nrow(stat), length.out = min(nrow(stat), 20)))
    }
    stat <- matrix(stat[subset, ], ncol = ncol(stat))

    # Calculate summary statistics
    my_summary <- function(x) {
        summary_stats <- stats::quantile(x, prob = c(.25, .5, .75))
        lower <- max(min(x), summary_stats[1] - 1.5 * stats::IQR(x))
        upper <- min(max(x), summary_stats[3] + 1.5 * stats::IQR(x))
        summary_stats <- c(lower, summary_stats, upper)
        names(summary_stats)[c(1, 5)] <- c("lower", "upper")
        return(summary_stats)
    }

    summary_stats <- apply(stat, 1, function(x) {
        my_summary(x)
    })

    # Detect outliers
    if (outliers) {
        detect_outliers <- function(x, lower, upper) {
            x[(x < lower) | (x > upper)]
        }

        all_outliers <- lapply(1:ncol(summary_stats), function(i) {
            outliers <- detect_outliers(stat[i, ], summary_stats[1, i], summary_stats[5, i])
            if (length(outliers) > 0) cbind(outliers, i)
        })

        all_outliers <- Filter(function(x) !is.null(x), all_outliers)

        out <- unlist(lapply(all_outliers, function(x) x[, 1]))
        group <- unlist(lapply(all_outliers, function(x) x[, 2]))
        names(out) <- names(group) <- NULL
    } else {
        out <- NULL
        group <- NULL
    }

    # Prepare input for bxp
    data <- list(
        stats = summary_stats,
        n = rep(ncol(stat), ncol(summary_stats)),
        out = out,
        group = group
    )

    # Output
    capitalize <- function(x) {
        paste(toupper(substr(x, 1, 1)), substr(x, 2, nchar(x)), sep = "")
    }
    graphics::bxp(data,
        main = capitalize(effect_name),
        xlab = capitalize(by), xaxt = "n",
        ylab = "Value", ...
    )
    graphics::axis(1, at = seq_along(subset), labels = subset)
}

#' Plotting Relational Event Network Statistics Trajectories
#'
#' Generate line plots to visualize the trajectories of a specified effect in 
#' the sender model of a \code{\link{aomstats}} object.
#'
#' @param x An object of class \code{\link{aomstats}} containing relational
#' event network statistics.
#' @param effect A character string specifying the name of the effect in 'x' or
#' an integer indicating the index of the effect to be plotted.
#' @param subset An optional vector specifying a subset of actors to be used for
#' plotting. By default, a maximum of 5 unique actors are used for plotting.
#' @param ... Additional arguments passed to plot().
#'
#' @details
#' This function creates line plots to illustrate the temporal trajectories of a
#' specified effect in a relational event network, as captured in the sender 
#' model by a \code{\link{aomstats}} object. The 'effect' parameter allows users 
#' to choose a specific effect for visualization, either by providing the effect's 
#' name or its index within the 'aomstats' object. An optional 'subset' parameter 
#' enables users to focus on specific actors. If 'subset' is not specified, a 
#' default maximum of 5 unique actors is plotted. These actors are randomly selected 
#' to represent trajectories across the range of different endpoints for the effect
#' (excluding zero).
#' 
#' @examples
#' library(remstats)
#' # Load data
#' data(history)
#' # Prepare data
#' reh <- remify::remify(edgelist = history[,1:3], model = "actor")
#' # Compute effects
#' stats <- remstats(reh, sender_effects = ~ outdegreeSender())
#' # Plot the 'outdegreeSender' trajectories 5 actors
#' plot(stats, effect = "outdegreeSender")
#' # Plot the 'outdegreeSender' trajectory for a specific actor
#' plot(stats, effect = "outdegreeSender", subset = 10)
#' 
#' @method plot aomstats
#' @export
#' @importFrom graphics lines
#' @importFrom graphics legend
#' @importFrom grDevices rainbow
#' @importFrom stats quantile
plot.aomstats <- function(x, effect, subset = NULL, ...) {

    # Figure out which statistic to plot
    if (is.character(effect)) {
        if (!(effect %in% dimnames(x$sender_stats)[[3]])) {
            stop(paste0(effect, " not found in the sender model of 'x'"))
        } else {
            stat <- x$sender_stats[, , effect]
            effect_name <- effect
        }
    } else if (is.numeric(effect)) {
        if (trunc(effect) > dim(x$sender_stats)[3]) {
            stop("'effect' index out of bound")
        } else {
            stat <- x$sender_stats[, , trunc(effect)]
            effect_name <- dimnames(x$sender_stats)[[3]][effect]
        }
    } else {
        stop("Expected 'effect' argument of type character or integer")
    }

    # Figure out which actors to plot
    if (is.null(subset)) {
        lastrow <- stat[nrow(stat),]
        ids <- rank(lastrow, ties.method = "random")
        subset <- which(ids %in% round(stats::quantile(ids[lastrow > 0], c(0, .25, .5, .75, 1))))
        subset <- unique(subset)
    }

    stat <- matrix(stat[, subset], ncol = length(subset))

    # Plot one
    capitalize <- function(x) {
        paste(toupper(substr(x, 1, 1)), substr(x, 2, nchar(x)), sep = "")
    }
    plot(stat[, 1],
        type = "l", col = rainbow(n = length(subset))[1],
        main = capitalize(effect_name), xlab = "Time", ylab = "Value",
        ylim = c(min(stat) - 0.05 * (max(stat) - min(stat)), max(stat) + 0.05 * (max(stat) - min(stat))),
        lwd = 1.5, ...
    )
    if (ncol(stat) > 1) {
        # Plot the rest
        sapply(2:length(subset), function(i) {
            graphics::lines(stat[, i], col = rainbow(n = length(subset))[i], lwd = 1.5)
        })
    }

    # Add legend
    graphics::legend("topleft", legend = subset, col = rainbow(length(subset)), lty = 1, title = "Actors")
}

#' Plotting Relational Event Network Statistics Trajectories
#'
#' Generate line plots to visualize the trajectories of a specified effect in a
#' \code{\link{tomstats}} object.
#'
#' @param x An object of class \code{\link{tomstats}} containing relational
#' event network statistics.
#' @param effect A character string specifying the name of the effect in 'x' or
#' an integer indicating the index of the effect to be plotted.
#' @param subset An optional vector specifying a subset of dyads to be used for
#' plotting. By default, a maximum of 5 unique dyads are used for plotting.
#' @param ... Additional arguments passed to plot().
#'
#' @details
#' This function creates line plots to illustrate the temporal trajectories of a
#' specified effect in a relational event network, as captured by a
#' \code{\link{tomstats}} object. The 'effect' parameter allows users to choose a
#' specific effect for visualization, either by providing the effect's name or
#' its index within the 'tomstats' object. An optional 'subset' parameter enables
#' users to focus on specific dyads. If 'subset' is not specified, a default
#' maximum of 5 unique dyads is plotted. These dyads are randomly selected to
#' represent trajectories across the range of different endpoints for the effect
#' (excluding zero).
#' 
#' @examples
#' library(remstats)
#' # Load data
#' data(history)
#' # Prepare data
#' reh <- remify::remify(edgelist = history[,1:3], model = "tie")
#' # Compute effects
#' stats <- remstats(reh, tie_effects = ~ inertia())
#' # Plot the 'inertia' trajectories for 5 dyads
#' plot(stats, effect = "inertia")
#' # Plot the 'inertia' trajectory for a specific dyad
#' plot(stats, effect = "inertia", subset = 60)
#' 
#' @method plot tomstats
#' @export
#' @importFrom graphics lines
#' @importFrom graphics legend
#' @importFrom grDevices rainbow
#' @importFrom stats quantile
plot.tomstats <- function(x, effect, subset = NULL, ...) {

    # Figure out which statistic to plot
    if (is.character(effect)) {
        if (!(effect %in% dimnames(x)[[3]])) {
            stop(paste0(effect, " not found in 'x'"))
        } else {
            stat <- x[, , effect]
            effect_name <- effect
        }
    } else if (is.numeric(effect)) {
        if (trunc(effect) > dim(x)[3]) {
            stop("'effect' index out of bound")
        } else {
            stat <- x[, , trunc(effect)]
            effect_name <- dimnames(x)[[3]][effect]
        }
    } else {
        stop("Expected 'effect' argument of type character or integer")
    }

    # Figure out which dyads to plot
    if (is.null(subset)) {
        lastrow <- stat[nrow(stat),]
        ids <- rank(lastrow, ties.method = "random")
        subset <- which(ids %in% round(stats::quantile(ids[lastrow > 0], c(0, .25, .5, .75, 1))))
        subset <- unique(subset)
    }

    stat <- matrix(stat[, subset], ncol = length(subset))

    # Plot one
    capitalize <- function(x) {
        paste(toupper(substr(x, 1, 1)), substr(x, 2, nchar(x)), sep = "")
    }
    plot(stat[, 1],
        type = "l", col = rainbow(n = length(subset))[1],
        main = capitalize(effect_name), xlab = "Time", ylab = "Value",
        ylim = c(min(stat) - 0.05 * (max(stat) - min(stat)), max(stat) + 0.05 * (max(stat) - min(stat))), 
        lwd = 1.5, ...
    )
    if (ncol(stat) > 1) {
        # Plot the rest
        sapply(2:length(subset), function(i) {
            graphics::lines(stat[, i], col = rainbow(n = length(subset))[i], lwd = 1.5)
        })
    }

    # Add legend
    dyads <- attr(x, "riskset")[subset,]
    #labels <- paste0(dyads$id, " (", dyads[,1], ", ", dyads[,2], ")")
    labels <- dyads$id
    graphics::legend("topleft", legend = labels, col = rainbow(length(subset)), lty = 1, title = "Dyads")
}

#' tie_effects
#'
#' Overview of statistics in the tie-oriented model, see Details.
#'
#' @param directed logical value. The function outputs all statistics in the 
#' tie-oriented model for directed events if true, or all statistics in the 
#' tie-oriented model for undirected events if false.
#' @param endogenous logical value. The function outputs all endogenous 
#' statistics in the tie-oriented model if true, or all exogenous statistics if 
#' false
#'
#' @details
#' Overview of statistics in the tie-oriented model.
#' 
#' Baseline:
#' \itemize{
#'  \item \code{\link{baseline}}
#' }
#'
#' Exogenous statistics:
#' \itemize{
#'  \item \code{\link{send}()}
#'  \item \code{\link{receive}()}
#'  \item \code{\link{tie}()}
#'  \item \code{\link{same}()}
#'  \item \code{\link{difference}()}
#'  \item \code{\link{average}()}
#'  \item \code{\link{minimum}()}
#'  \item \code{\link{maximum}()}
#'  \item \code{\link{event}()}
#'  \item \code{\link{userStat}()}
#' }
#'
#' Endogenous statistics:
#' \itemize{
#'  \item \code{\link{indegreeSender}()}
#'  \item \code{\link{indegreeReceiver}()}
#'  \item \code{\link{outdegreeSender}()}
#'  \item \code{\link{outdegreeReceiver}()}
#'  \item \code{\link{totaldegreeSender}()}
#'  \item \code{\link{totaldegreeReceiver}()}
#'  \item \code{\link{totaldegreeDyad}()}
#'  \item \code{\link{degreeMin}()}
#'  \item \code{\link{degreeMax}()}
#'  \item \code{\link{degreeDiff}()}
#'  \item \code{\link{inertia}()}
#'  \item \code{\link{reciprocity}()}
#'  \item \code{\link{otp}()}
#'  \item \code{\link{itp}()}
#'  \item \code{\link{osp}()}
#'  \item \code{\link{isp}()}
#'  \item \code{\link{sp}()}
#'  \item \code{\link{psABBA}()}
#'  \item \code{\link{psABBY}()}
#'  \item \code{\link{psABXA}()}
#'  \item \code{\link{psABXB}()}
#'  \item \code{\link{psABXY}()}
#'  \item \code{\link{psABAY}()}
#'  \item \code{\link{psABAB}()}
#'  \item \code{\link{rrankSend}()}
#'  \item \code{\link{rrankReceive}()}
#'  \item \code{\link{recencySendSender}()}
#'  \item \code{\link{recencySendReceiver}()}
#'  \item \code{\link{recencyReceiveSender}()}
#'  \item \code{\link{recencyReceiveReceiver}()}
#'  \item \code{\link{recencyContinue}()}
#'  \item \code{\link{FEtype}()}
#' }
#' 
#' \strong{Directed vs. undirected availability}
#' The set of available effects depends on whether events are directed.
#' Use \code{\link{tie_effects}(directed = TRUE)} or
#' \code{\link{tie_effects}(directed = FALSE)} to obtain the authoritative list
#' for the current package version.
#'
#' For undirected events (\code{directed = FALSE}), the available tie-model effects are:
#' \itemize{
#'   \item \code{\link{baseline}}, \code{\link{FEtype}}
#'   \item \code{\link{tie}}, \code{\link{same}}, \code{\link{difference}},
#'         \code{\link{average}}, \code{\link{minimum}}, \code{\link{maximum}}
#'   \item \code{\link{event}}, \code{\link{userStat}}
#'   \item \code{\link{inertia}}, \code{\link{totaldegreeDyad}},
#'         \code{\link{degreeMin}}, \code{\link{degreeMax}}, \code{\link{degreeDiff}}
#'   \item \code{\link{sp}}, \code{\link{psABAY}}, \code{\link{psABAB}},
#'         \code{\link{recencyContinue}}
#' }
#'
#' @examples 
#' # List of available effects 
#' tie_effects()
#' 
#' # List of available effects for undirected networks
#' tie_effects(directed = FALSE)
#' 
#' # List of available endogenous effects for undirected networks
#' tie_effects(directed = FALSE, endogenous = TRUE)
#' 
#' @return
#' Returns a list of available effects and their corresponding statistics.
#' 
#' @export
tie_effects <- function(directed = NULL, endogenous = NULL) {
  # Name all effects
  effects <- all_tie_effects()

  # Filter out 'interact'
  effects <- effects[effects != "interact"]

  if (!is.null(directed)) {
    if (directed) {
      # Filter out effects that are not defined for directed events
      effects <- effects[!(effects %in% c(
        "sp", "degreeMin", "degreeMax", "degreeDiff"
      ))]
    } else if (!directed) {
      # Filter out effects that are not defined for undirected events
      effects <- effects[!(effects %in% c(
        "send", "receive", "reciprocity", "indegreeSender",
        "indegreeReceiver", "outdegreeSender", "outdegreeReceiver",
        "totaldegreeSender", "totaldegreeReceiver", "otp", "itp",
        "osp", "isp", "psABBA", "psABBY", "psABXA", "psABXB", "psABXY",
        "rrankSend", "rrankReceive", "recencySendSender",
        "recencySendReceiver", "recencyReceiveSender",
        "recencyReceiveReceiver"
      ))]
    }
  }

  if (!is.null(endogenous)) {
    if (endogenous) {
      # Filter out exogenous effects
      effects <- effects[!(effects %in% c(
        "send", "receive", "tie", "same", "difference", "average",
        "minimum", "maximum", "event", "userStat"
      ))]
    } else if (!endogenous) {
      # Filter out endogenous effects
      effects <- effects[!(effects %in% c(
        "indegreeSender", "indegreeReceiver", "outdegreeSender",
        "outdegreeReceiver", "totaldegreeSender", "totaldegreeReceiver", 
        "totaldegreeDyad", "degreeDiff", "degreeMin", "degreeMax",
        "inertia", "reciprocity", "otp", "itp", "osp", "isp", "sp",
        "psABBA", "psABBY", "psABXA", "psABXB", "psABXY",
        "psABAY", "psABAB", "rrankSend", "rrankReceive",
        "recencySendSender", "recencySendReceiver", "recencyReceiveSender",
        "recencyReceiveReceiver", "recencyContinue", "FEtype"
      ))]
    }
  }


  effects
}

if ( requireNamespace("tinytest", quietly=TRUE) ){
	tinytest::test_package("remstats")
}

# test-endogenous-stats / test-exogenous-stats: 
# Condition 1: Directed events, tie-oriented model
# Condition 2: Undirected events, tie-oriented model
# Condition 3: Directed evens with types, tie-oriented model
# Condition 4: Undirected events with types, tie-oriented model
# Condition 5: Actor-oriented model

#' actor_effects
#'
#' Overview of statistics in the actor-oriented model, see Details.
#'
#' @param step outputs all statistics in the sender activity step (if `step =
#' sender`) or receiver choice step (if `step = receiver`).
#'
#' @details
#' Overview of statistics in the actor-oriented model.
#'
#' A list of available effects and their corresponding statistics for the
#' \emph{sender activity rate} step:
#' \itemize{
#'  \item \code{\link{baseline}()}
#'  \item \code{\link{send}()}
#'  \item \code{\link{indegreeSender}()}
#'  \item \code{\link{outdegreeSender}()}
#'  \item \code{\link{totaldegreeSender}()}
#'  \item \code{\link{recencySendSender}()}
#'  \item \code{\link{recencyReceiveSender}()}
#'  \item \code{\link{psABA}()}
#'  \item \code{\link{psABB}()}
#'  \item \code{\link{psABX}()}
#' }
#'
#' A list of available effects and their corresponding statistics for the
#' \emph{receiver choice} step:
#' \itemize{
#'  \item \code{\link{receive}()}
#'  \item \code{\link{tie}()}
#'  \item \code{\link{same}()}
#'  \item \code{\link{difference}()}
#'  \item \code{\link{average}()}
#'  \item \code{\link{indegreeReceiver}()}
#'  \item \code{\link{outdegreeReceiver}()}
#'  \item \code{\link{totaldegreeReceiver}()}
#'  \item \code{\link{inertia}()}
#'  \item \code{\link{reciprocity}()}
#'  \item \code{\link{otp}()}
#'  \item \code{\link{itp}()}
#'  \item \code{\link{osp}()}
#'  \item \code{\link{isp}()}
#'  \item \code{\link{rrankSend}()}
#'  \item \code{\link{rrankReceive}()}
#'  \item \code{\link{recencySendReceiver}()}
#'  \item \code{\link{recencyReceiveReceiver}()}
#'  \item \code{\link{recencyContinue}()}
#'  \item \code{\link{psABAB}()}
#'  \item \code{\link{psABBA}()}
#'  \item \code{\link{psABXA}()}
#'  \item \code{\link{psABXB}()}
#'  \item \code{\link{psABAY}()}
#'  \item \code{\link{psABBY}()}
#'  \item \code{\link{psABXY}()}
#' }
#' 
#' @examples 
#' # List of available effects for both the sender and receiver step
#' actor_effects()
#' 
#' # List of available effects for the sender step
#' actor_effects(step = "sender")
#' 
#' # List of available effects for the receiver step
#' actor_effects(step = "receiver")
#' 
#' @return
#' Returns a list of available effects and their corresponding statistics based 
#' on the specified `step` (sender or receiver).
#'
#' @export
actor_effects <- function(step = NULL) {
    # Name all effects
    effects <- list(
        sender = c(
            "baseline",
            "send",
            "indegreeSender",
            "outdegreeSender",
            "totaldegreeSender",
            "recencySendSender",
            "recencyReceiveSender",
            "psABA",
            "psABB",
            "psABX",
            "userStat"
        ),
        receiver = c(
            "receive",
            "same",
            "difference",
            "average",
            "tie",
            "inertia",
            "reciprocity",
            "indegreeReceiver",
            "outdegreeReceiver",
            "totaldegreeReceiver",
            "otp",
            "itp",
            "osp",
            "isp",
            "rrankSend",
            "rrankReceive",
            "recencySendReceiver",
            "recencyReceiveReceiver",
            "recencyContinue",
            "psABBA",
            "psABAB",
            "psABXB",
            "psABXA",
            "psABAY",
            "psABBY",
            "psABXY",
            "userStat"
        )
    )

    if (!is.null(step)) {
        if (step == "sender") {
            return(effects$sender)
        } else if (step == "receiver") {
            return(effects$receiver)
        }
    }

    effects
}

#!/bin/sh

# upgrade-1.0.10b - fix up all the ping rrds for the change in the definition
# $Id: do-upgrade-1.0.10b.sh,v 1.1 2002/06/28 19:02:50 remstats Exp $

cd @@DATADIR@@
echo "converting ping rrds..."
for f in */ping.rrd */ping-*.rrd ; do
	x=`dirname $f`/`basename $f .rrd`.xml
	cp $f $f.bak
	@@BINDIR@@/add-min-archives $f > $x
	rrdtool restore $x $f
	echo "... $f"
done
echo "Done converting ping rrds."

#!/bin/sh

# cvs-status - make a compact CVS status display
# $Id: cvs-status.sh,v 1.1 2003/03/13 14:12:11 remstats Exp $
# from remstats @@VERSION@@

# Copyright 1999, 2000, 2001, 2002 (c) Thomas Erskine
#   <@@AUTHOREMAIL@@>
# See the COPYRIGHT file with the distribution.

cvs status 2>&1 |\
awk '
/^File:.*Up-to-date/		{next}
/^cvs status: Examining/	{print "=== " $4 " ==="; next}
/^File:/			{print}
'

#!/bin/sh

# collection - stop/start collection
# $Id: collection.sh,v 1.4 2002/05/28 15:54:50 remstats Exp $
# from remstats @@VERSION@@

# Copyright 1999, 2000, 2001, 2002 (c) Thomas Erskine <@@AUTHOR@@>
# See the COPYRIGHT file with the distribution.

# - - -   Version History - - -
# $Revision: 1.4 $

# - - -   Setup   - - -

if [ $# != 1 ] ; then
	echo >&2 "usage: `basename $0` {stop MSG|start|status}"
	exit 0
fi

what="$1"
file="@@TMPDIR@@/STOP-run-remstats"

# - - -   Mainline   - - -

case "$what" in 
	start)
		if [ -f $file ] ; then
			rm -f $file
			echo "stop-file removed ($file)"
		else
			echo "no stop-file; it's not stopped"
		fi
		;;
	stop)
		if [ -f $file ] ; then
			echo "stop-file exists ($file); it's stopped already"
		else
			echo >$file "`date +%Y-%m-%d %H:%M:%S`"
			if [ ! -z "$2" ] ; then
				echo >>$file "Reason: $2"
			else
				echo >>$file "Reason: unknown"
			fi
		fi
		;;
	status)
		if [ -f $file ] ; then
			echo "REMSTATS STOPPED"
			cat $file
		else
			echo "remstats collection not stopped"
		fi
		;;
	*)
		echo >&2 "usage: `basename $0` {stop|start|status}"
		exit 0
		;;
esac

#!/bin/sh

# convert-config-links - convert the likely-to-be-changed config-files
#	from symlinks to copies
# $Id: convert-config-links.sh,v 1.4 2002/05/28 15:54:50 remstats Exp $
# from remstats @@VERSION@@

# Copyright 1999, 2000, 2001, 2002 (c) Thomas Erskine <@@AUTHOR@@>
# See the COPYRIGHT file with the distribution.

if [ $# -gt 0 ] ; then
	if [ "$1" = '-h' ] ; then
		echo >&2 "usage: $0 [-h]"
		exit 0
	fi
fi

cd @@CONFIGDIR@@
if [ ! -d ../config-base ] ; then
	echo >&2 "$0: can't find the config-base to copy from"
	exit 1
fi

for file in alerts alert-destination-map general html links tools ; do
	if [ ! -e $file ] ; then
		echo >&2 "$0: missing $file. Is @@CONFIGDIR@@ really your configuration directory?"
		exit 2
	fi
	if [ ! -f ../config-base/$file ] ; then
		echo >&2 "$0: missing ../config-base/$file.  Is @@CONFIGDIR@@../config-base really your config-base?"
		exit 4
	fi
	mv $file $file.bak
	cp ../config-base/$file $file || (echo "$0: error copying $file" && exit 3)
	echo "... $file"
done

---
title: "remstats"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{remstats}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(remstats)
```

## Introduction 
Relational event modeling approaches enable researchers to investigate both exogenous and endogenous factors influencing the evolution of a time-ordered sequence of relational events - commonly known as a *relational event history*. These models are categorized into *tie-oriented models*, where the probability of a dyad interacting next is modeled in a single step (e.g., see Butts, 2008), and *actor-oriented models*, which first model the probability of a sender initiating an interaction and subsequently the probability of the senders' choice of receiver (e.g., see Stadtfeld & Block, 2017). The `R` package `remstats` is designed to compute a variety of statistics for both types of models.

The `remstats` package is part of a bundle of `R`-packages developed by researchers from Tilburg University intended to aid applied researchers in the application of relational event modeling. For preparing the relational event history, `remstats` assumes the prior application of `remify::remify()` (available on CRAN). Model estimation can subsequently be executed using `remstimate` (available on GitHub at github.com/TilburgNetworkGroup/remstimate).

The following provides a brief introduction to computing statistics for relational event history data with remstats. We'll begin with a quick workflow example, followed by a detailed description of the procedure for the tie-oriented model. Finally, we'll provide an explanation of the procedure for the actor-oriented model.

## Workflow example 
```{r}
# Load data
data(history)
data(info)

# Define effects
effects <- ~ 1 + send("extraversion", info) + inertia()

# Prepare event history with the 'remify package'
rehObject <- remify::remify(edgelist = history, model = "tie")

# Compute statistics
statsObject <- remstats(reh = rehObject, tie_effects = effects)

# Estimate model parameters with the 'remstimate' package
# fit <- remstimate::remstimate(reh = rehObject, stats = statsObject,
# 	method = "MLE", timing = "interval")
```

## Getting started

### Data
Relational event history data describes a time-ordered series of interactions between actors in a network. Such interactions are referred to as *relational events*. A relational event minimally contains information on the time of the event and the actors that are involved in the event. 

As an illustration, we use the `history` data object in the `remstats` package. This data object is a randomly generated relational event history. A description of the simulated data can be accessed with `?history`. Here, we read that `history` is a small simulated relational event history with 115 events. Besides information on the time and actors, for each event there is also information on the setting and an event weight. We can view the first six events with:

```{r}
head(history)
```

We prepare the relational event history for computation of statistics for the tie-oriented model with the `remify` function from the `remify` package. Whenever the weight variable is present in the `edgelist` supplied to `remify`, it assumes that we want to use these to weight the events in the computation of the statistics. In this example, we don't want this, thus we set the weight to one for all events:
```{r}
history$weight <- 1
reh <- remify::remify(edgelist = history, model = "tie")
```

Besides the relational event history itself, relational event modeling often requires a second data object with exogenous information on the actors in the network. Information on the actors in the simulated data example in `remstats` is stored in the `info` object. A description of the `info` data can be accessed with `?info`. Here, we read that the `info` object stores for the 10 actors in the network information on their age, sex, extraversion and agreeableness score. Moreover, extraversion and agreeableness is measured multiple times during the observation period. The time variable tells us when the values change. We can view the attribute information for the first two actors with:

```{r}
head(info)
```

### Compute statistics for the tie-oriented model
First, we compute statistics for modeling relational event history data with a tie-oriented model. The statistics that are requested need to be supplied to the `tie_effects` argument of `remstats()`, specified in an object of class `formula`. This specification should be in the form `~ terms`.

An overview of the statistics that can be computed for the tie-oriented model is available using the `tie_effects()` function or its help documentation `?tie_effects`:

In this illustration, we start with requesting only one statistic: the inertia statistic. Most statistics can be tailored to the user's needs. For example, lets view the description for the `inertia` statistic using `?inertia`. Here, we can read that the inertia statistic computes for every timepoint *t* for every pair of actors *(i,j)* in the riskset the number of past events. With the `scaling` argument, one of the methods for scaling the statistic can be chosen. The `consider_type` argument is relevant when event types are in the dependent variable, which we do not consider in this example.

In this illustration, we will standardize the inertia statistic. To request this statistic, we define the formula as follows:s
```{r}
effects <- ~ inertia(scaling = "std")
```

Now, we have everything we need to compute our first statistic:
```{r}
out <- remstats(tie_effects = effects, reh = reh)
```

The `remstats()` function outputs a 3-dimensional array with statistics for the tie-oriented model. On the rows of this array are the timepoints, the columns refer to the potential events in the riskset and the slices refer to the different statistics:
```{r}
dim(out)
```
Our statistics object has 115 rows, corresponding to the 115 time points in the relational event history. It has 90 columns, corresponding to the 90 events in the riskset. The statistics object has two slices, that is because the baseline statistics is automatically computed when the timing of the events in the relational event history is exact (unless removed by specifying `-1` in the `effects` formula) and saved in the first slice. The `remstats()` procedure assumes that the timing of the events in the relational event history is exact and the full likelihood is used in the estimation, unless the argument `ordinal` in `remify::remify()` is set to `TRUE`. 

We can view the names of the statistics that are in the statistics object with:
```{r}
out
```
Here, we see that, indeed, a baseline and inertia statistic are computed. 

Since we did not request anything special for the riskset in `remify::remify()`, it consists of every directed pair of actors observed in the relational event history, which is 10*9 = 90 pairs. These pairs are saved in the `riskset` attribute. We can ask for the first few lines of this riskset:
```{r}
head(attr(out, "riskset"))
```

Here, we see that the first event in the riskset is the event were actor 101 sends an interaction directed towards actor 103. The id column refers to the column in the statistic object that contains the statistic(s) for this specific dyad. The first column in the statistic object refers to this first event in the riskset, the second column in the statistic object to the second event in the riskset, and so forth. 

Inertia is an example of an *endogenous* statistic: it is a function of the relational event history itself. Next, we are going to add a request for an *exogenous* statistic. For this we need the exogenous information on the actors in the `info` object. 

As an illustration, we are going to request the statistic for an effect of extraversion on sending events, i.e., a send effect. The description of a send effect is accessed with `?send`. Here, we read that we need to supply the variable for which we want to specify a sender effect and that this variable should correspond to a column in the `attr_actors` object that we supply. Thus, we specify a send effect of extraversion with `send("extraversion", attr_actors = info)`. Here, we specify the `attr_actors` object within the `send()` function. Alternatively, it can be supplied to `remstats()`. This is for example useful if you want to compute a bunch of exogenous statistics using the same `attr_actors` object.  

Statistics in the `effects` formula should be separated with the `+`. Finally, we add an interaction between the `inertia()` statistic and the `send()` statistic. This can be done by using the `*` or `:` operator:
```{r}
effects <- ~ inertia(scaling = "std") + send("extraversion", info) + 
    inertia(scaling = "std"):send("extraversion", info) 
out <- remstats(tie_effects = effects, reh = reh)
```

### Compute statistics for the actor-oriented model
The procedure to compute statistics for the actor-oriented model is largely similar to what is written above, except that statistics have to be specified separately for the sender activity rate step of the model and the receiver choice step of the model. The statistics requested for these two modeling steps need to be supplied to two different effects arguments, namely `sender_effects` and `receiver_effects`, respectively. 

An overview of the statistics that are available for the actor-oriented model in the two modeling steps can be obtained using the `actor_effects()` function or its help documentation `?actor_effects`. 

In this illustration, we start with requesting only one statistic for the sender activity rate step: the *outdegreeSender* statistic. First, lets view the description for the `outdegreeSender` statistic using `?outdegreeSender`. Here, we can read that, in the sender activity rate step of the actor-oriented model, the outdegreeSender statistic computes for every timepoint *t* for every actors *i* the number of outgoing past events. With the `scaling` argument, one of the methods for scaling the statistic can be chosen. 

First, we prepare the event history for computing statistics for an actor-oriented model: 
```{r}
reh <- remify::remify(edgelist = history, model = "actor")
```
To compute the outdegreeSender statistic for the sender activity rate step we supply it to the `sender_effects` argument of `remstats()`:
```{r} 
effects <- ~ outdegreeSender()
out <- remstats(sender_effects = effects, reh = reh)
```

The outputted remstats object is now a list with two elements: `sender_stats` and `receiver_stats`:
```{r}
names(out)
```
Since we did not request any statistics for the receiver choice step here, the `receiver_stats` object is empty. The `sender_stats` object contains the statistic array with the `baseline` statistic (again, automatically computed unless `ordinal = TRUE`), and the requested `outdegreeSender` statistic:
```{r}
out
```

The dimension of `out$sender_stats` is 115 x 10 x 2. On the rows we have the timepoints, the columns refer to the actors that can be senders and the slices refer to the different statistics. 

Lets extend our model and also request a statistic for the receiver choice step:
```{r}
sender_effects <- ~ outdegreeSender()
receiver_effects <- ~ inertia()
out <- remstats(sender_effects = sender_effects, receiver_effects = receiver_effects, reh = reh)
```

We can access the statistic computed for the receiver choice step with `out$receiver_stats`. In this step, the baseline statistic is not automatically computed (and not defined). Hence, the dimensions of the statistics object for the receiver choice step are 115 x 10 x 1. On the rows we have again the timepoints, on the columns now the receivers and on the slices the statistics. 

Note that the computed values of the statistic in the receiver choice step are equal to the values for this receiver, given the current sender. For example, lets review the first six lines:
```{r}
# Set the column names equal to the receivers
colnames(out$receiver_stats) <- attributes(reh)$dictionary$actors$actorName
# Set the rownames equal to the senders
rownames(out$receiver_stats) <- reh$edgelist$actor1
# View the first six lines
head(out$receiver_stats[,,"inertia"])
```

At the first timepoint, the inertia statistic for all receivers given the current sender (actor 105) is zero because no prior events have occurred. At the second timepoint, the sender is again actor 105. Now the inertia statistic is equal to the 1 for the receiver of the first event (actor 113). At the third timepoint, the inertia statistic is again zero for all receivers because now the sending actor is 115, who did not send any prior events. 

---
title: "remstats"
output: 
  rmarkdown::html_document:
    theme: spacelab
    highlight: pygments
    code_folding: show
vignette: >
  %\VignetteIndexEntry{remstats}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(remstats)
```

<i>This vignette explains how to compute statistics for relational event history data using `remstats()`, for both tie-oriented and actor-oriented models. When using the package please cite:

Meijerink-Bosman, M., Back, M., Geukes, K., Leenders, R., & Mulder, J. (2023). Discovering trends of social interaction behavior over time: An introduction to relational event modeling. Behavior Research Methods. https://doi.org/10.3758/s13428-022-01821-8

</i>

---

## Introduction 

Relational event modeling approaches enable researchers to investigate both exogenous and endogenous factors influencing the evolution of a time-ordered sequence of relational events — commonly known as a *relational event history*. These models are categorized into *tie-oriented models*, where the probability of a dyad interacting next is modeled in a single step (e.g., see Butts, 2008), and *actor-oriented models*, which first model the probability of a sender initiating an interaction and subsequently the probability of the sender's choice of receiver (e.g., see Stadtfeld & Block, 2017). The `R` package `remstats` is designed to compute a variety of statistics for both types of models.

The `remstats` package is part of the `remverse` suite of `R` packages developed at Tilburg University to support applied researchers in relational event modeling. For preparing the relational event history, `remstats` requires prior processing with `remify::remify()`. Model estimation can subsequently be executed using `remstimate`.

---

## Quick workflow example

```{r}
# Load data
data(history)
data(info)

# Set weights to 1 (no event weighting)
history$weight <- 1

# Define effects
effects <- ~ 1 + send("extraversion", info) + inertia()

# Prepare event history
reh <- remify::remify(edgelist = history, model = "tie")

# Compute statistics
stats <- remstats(reh = reh, tie_effects = effects)

# Estimate model parameters (requires remstimate)
# fit <- remstimate::remstimate(reh = reh, stats = stats, method = "MLE")
```

---

## Data

Relational event history data describes a time-ordered series of interactions between actors in a network. A relational event minimally contains information on the time of the event and the actors involved.

We use the `history` dataset from the `remstats` package (see `?history`): a small simulated relational event history with 115 events among 10 actors. Each event also records a `setting` (either `"work"` or `"social"`) and an event `weight`.

```{r}
head(history)
```

Exogenous information on actors is stored in the `info` object (see `?info`), containing age, sex, extraversion, and agreeableness scores measured at multiple time points for each of the 10 actors:

```{r}
head(info)
```

---

# {.tabset .tabset-fade .tabset-pills}

## Tie-oriented model

### Preparing the event history

We prepare the event history for the tie-oriented model with `remify::remify()`. When a `weight` column is present in the edgelist, it is used to weight events in the computation of statistics. Here we set weights to one to avoid this:

```{r}
history$weight <- 1
reh <- remify::remify(edgelist = history, model = "tie")
```

### Computing statistics

Statistics for the tie-oriented model are supplied to the `tie_effects` argument of `remstats()` as a `formula` object of the form `~ terms`. An overview of available statistics is obtained with `tie_effects()` or `?tie_effects`.

As a first example, we request the standardized inertia statistic:

```{r}
effects <- ~ inertia(scaling = "std")
out <- remstats(tie_effects = effects, reh = reh)
```

The output is a 3-dimensional array. Rows correspond to unique time points, columns to potential events in the risk set, and slices to statistics:

```{r}
dim(out)
```

The number of rows equals `reh$M` — the number of unique time points (115 here). When simultaneous events are present, the number of rows is less than the total number of events, since statistics are computed once per unique time point. There are 90 columns corresponding to the $10 \times 9$ directed dyads in the full risk set. The two slices are: (1) a baseline statistic, automatically included when event timing is exact (`ordinal = FALSE`), and (2) the inertia statistic. The baseline can be suppressed by including `-1` in the effects formula.

```{r}
out
```

The risk set composition is stored as an attribute of the output. Each row describes one dyad, with columns `sender`, `receiver`, `type` (if event types are present), and `id` (the column index in the statistics array):

```{r}
head(attr(out, "riskset"))
```

Exogenous statistics require an `attr_actors` object. The `attr_actors` can be passed directly inside effect functions or globally via the `attr_actors` argument of `remstats()`. Statistics are separated by `+` in the formula, and interactions are specified with `:` or `*`:

```{r}
effects <- ~ inertia(scaling = "std") + 
             send("extraversion", info) + 
             inertia(scaling = "std"):send("extraversion", info)
out <- remstats(tie_effects = effects, reh = reh)
out
```

### Memory

The `memory` argument controls which past events are included in the computation of endogenous statistics at each time point $t$. Four options are available:

- `"full"` (default): the entire event history before $t$ is included.
- `"window"`: only events within a time window of length `memory_value` before $t$ are included. For example, `memory = "window", memory_value = 200` includes only events that occurred at most 200 time units ago.
- `"interval"`: only events within a specific time interval before $t$ are included, defined by `memory_value = c(lower, upper)`. For example, `memory_value = c(100, 200)` includes only events between 100 and 200 time units ago.
- `"decay"`: past events are weighted by an exponential decay function with half-life `memory_value`. More recent events receive higher weight.

```{r}
# Full memory (default)
out_full <- remstats(
    tie_effects  = ~ inertia(),
    reh          = reh,
    memory       = "full"
)

# Window memory: only events within the last 500 time units
out_window <- remstats(
    tie_effects  = ~ inertia(),
    reh          = reh,
    memory       = "window",
    memory_value = 500
)

# Decay memory: exponential decay with half-life 200
out_decay <- remstats(
    tie_effects  = ~ inertia(),
    reh          = reh,
    memory       = "decay",
    memory_value = 200
)
```

### Event types and `consider_type`

When an event type variable is present in the relational event history, the `consider_type` argument in each effect function controls how the statistic accounts for event types. We illustrate this using the `setting` column in `history` (values: `"work"`, `"social"`), passed to `remify()` via the `event_type` argument:

```{r}
reh_typed <- remify::remify(
    edgelist   = history,
    model      = "tie",
    event_type = "setting"
)
```

Three options are available for `consider_type`:

**`"ignore"`**: event type is not taken into account. The statistic is computed over all past events regardless of type. This is the appropriate choice when event types are not a substantive concern, or when no types are present.

```{r}
out_ignore <- remstats(
    tie_effects = ~ inertia(consider_type = "ignore"),
    reh         = reh_typed
)
dim(out_ignore)       # one inertia slice plus baseline
dimnames(out_ignore)[[3]]
```

**`"separate"`**: the statistic is computed separately for each event type, resulting in one slice per type. For example, with two types `"social"` and `"work"`, inertia yields: (1) the number of past `"social"` events for each dyad, and (2) the number of past `"work"` events for each dyad.

```{r}
out_separate <- remstats(
    tie_effects = ~ inertia(consider_type = "separate"),
    reh         = reh_typed
)
dim(out_separate)       # two inertia slices (one per type) plus baseline
dimnames(out_separate)[[3]]
```

**`"interact"`**: the statistic conditions on the type of the most recent event. The statistic at time $t$ is computed only over past events of the same type as the event at $t-1$, effectively allowing the effect of past event history to depend on the type of the triggering event.

```{r}
out_interact <- remstats(
    tie_effects = ~ inertia(consider_type = "interact"),
    reh         = reh_typed
)
dim(out_interact)
dimnames(out_interact)[[3]]
```

---

## Tie-oriented model with case-control sampling

For large networks where computing statistics over the full risk set is computationally expensive, `remstats` supports *case-control sampling* (also known as dyad sampling). Instead of computing statistics for all dyads in the risk set at each time point, a random sample of dyads is drawn and used as the comparison set, substantially reducing computation time.

Case-control sampling is enabled by setting `sampling = TRUE` in `remstats()`. Two additional arguments control the sampling:

- `samp_num`: the number of dyads to sample per event (must be ≤ the size of the active risk set).
- `seed`: an optional integer random seed for reproducibility.

```{r}
reh <- remify::remify(edgelist = history, model = "tie")

out_sampled <- remstats(
    tie_effects = ~ inertia() + send("extraversion", info),
    reh         = reh,
    sampling    = TRUE,
    samp_num    = 20,
    seed        = 42
)
```

The output is an object of class `tomstats_sampled`. Its structure differs from the standard output: rather than a full array over all dyads, it stores only the statistics for the sampled dyads plus the observed event at each time point. This object can be passed directly to `remstimate` for model estimation using the case-control likelihood.

Note that case-control sampling is only available for the tie-oriented model and is not supported for the actor-oriented model.

---

## Actor-oriented model

### Preparing the event history

We prepare the event history for the actor-oriented model. Note that actor-oriented modeling requires directed events:

```{r}
reh <- remify::remify(edgelist = history, model = "actor")
```

### Computing statistics

For the actor-oriented model, statistics must be specified separately for the two modeling steps:

- **Sender activity rate step**: supplied to `sender_effects` — models which actor sends the next event.
- **Receiver choice step**: supplied to `receiver_effects` — models which actor the sender chooses as receiver.

An overview of available statistics for each step is obtained with `actor_effects()` or `?actor_effects`.

We start by requesting the `outdegreeSender` statistic for the sender step:

```{r}
out <- remstats(
    sender_effects = ~ outdegreeSender(),
    reh            = reh
)
```

The output for the actor-oriented model is a list with two elements:

```{r}
names(out)
```

Since no statistics were requested for the receiver step, `receiver_stats` is empty. The `sender_stats` object contains the baseline statistic (automatically included when `ordinal = FALSE`) and the requested `outdegreeSender` statistic:

```{r}
out
```

The dimensions of `out$sender_stats` are 115 × 10 × 2: rows are unique time points, columns are potential senders, and slices are statistics.

We extend the model with a statistic for the receiver choice step:

```{r}
out <- remstats(
    sender_effects   = ~ outdegreeSender(),
    receiver_effects = ~ inertia(),
    reh              = reh
)
```

The statistics for the receiver choice step are accessed with `out$receiver_stats`. The baseline statistic is not defined for the receiver step, so its dimensions are 115 × 10 × 1: rows are time points, columns are potential receivers, and slices are statistics.

The computed values in the receiver choice step give the statistic for each potential receiver, given the current sender. For example:

```{r}
# Label columns with receiver names and rows with senders
dimnames(out$receiver_stats)[[2]] <- reh$meta$dictionary$actors$actorName
dimnames(out$receiver_stats)[[1]] <- reh$edgelist$actor1[-1]
head(out$receiver_stats[,, "inertia"])
```

At the first time point, inertia is zero for all receivers because no prior events have occurred. At the second time point, inertia is 1 for the receiver of the first event (given the same sender). At the third time point, the sender is a different actor with no prior sending history, so inertia is again zero for all receivers.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actor_effects.R
\name{actor_effects}
\alias{actor_effects}
\title{actor_effects}
\usage{
actor_effects(step = NULL)
}
\arguments{
\item{step}{outputs all statistics in the sender activity step (if `step =
sender`) or receiver choice step (if `step = receiver`).}
}
\value{
Returns a list of available effects and their corresponding statistics based 
on the specified `step` (sender or receiver).
}
\description{
Overview of statistics in the actor-oriented model, see Details.
}
\details{
Overview of statistics in the actor-oriented model.

A list of available effects and their corresponding statistics for the
\emph{sender activity rate} step:
\itemize{
 \item \code{\link{baseline}()}
 \item \code{\link{send}()}
 \item \code{\link{indegreeSender}()}
 \item \code{\link{outdegreeSender}()}
 \item \code{\link{totaldegreeSender}()}
 \item \code{\link{recencySendSender}()}
 \item \code{\link{recencyReceiveSender}()}
 \item \code{\link{psABA}()}
 \item \code{\link{psABB}()}
 \item \code{\link{psABX}()}
}

A list of available effects and their corresponding statistics for the
\emph{receiver choice} step:
\itemize{
 \item \code{\link{receive}()}
 \item \code{\link{tie}()}
 \item \code{\link{same}()}
 \item \code{\link{difference}()}
 \item \code{\link{average}()}
 \item \code{\link{indegreeReceiver}()}
 \item \code{\link{outdegreeReceiver}()}
 \item \code{\link{totaldegreeReceiver}()}
 \item \code{\link{inertia}()}
 \item \code{\link{reciprocity}()}
 \item \code{\link{otp}()}
 \item \code{\link{itp}()}
 \item \code{\link{osp}()}
 \item \code{\link{isp}()}
 \item \code{\link{rrankSend}()}
 \item \code{\link{rrankReceive}()}
 \item \code{\link{recencySendReceiver}()}
 \item \code{\link{recencyReceiveReceiver}()}
 \item \code{\link{recencyContinue}()}
 \item \code{\link{psABAB}()}
 \item \code{\link{psABBA}()}
 \item \code{\link{psABXA}()}
 \item \code{\link{psABXB}()}
 \item \code{\link{psABAY}()}
 \item \code{\link{psABBY}()}
 \item \code{\link{psABXY}()}
}
}
\examples{
# List of available effects for both the sender and receiver step
actor_effects()

# List of available effects for the sender step
actor_effects(step = "sender")

# List of available effects for the receiver step
actor_effects(step = "receiver")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{both_male_wide}
\alias{both_male_wide}
\title{Exogenous Dyad Attribute Matrix: both_male_wide}
\format{
A square matrix with dimensions equal to the number of unique actors.
}
\source{
Simulated exogenous information on actors in a social network.
}
\usage{
data(both_male_wide)
}
\description{
A matrix representing exogenous attributes of dyads in a social network.
The matrix indicates whether a dyad consists of two male actors (sex=0).
Rows and columns correspond to actor IDs, and cells contain binary values
(1 for male-male dyads, 0 otherwise).
}
\examples{
data(both_male_wide)
print(both_male_wide)

}
\seealso{
\code{\link{tie}} for the function using this data, \code{\link{both_male_long}} for the data in long format, and \code{\link{info}} for an overview of the actor exogenous attributes.
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{baseline}
\alias{baseline}
\alias{intercept}
\title{baseline}
\description{
Specifies an intercept for the tie-oriented model or the sender activity
rate step of the actor-oriented model in the same manner as in
\code{\link[stats]{lm}} (see Details).
}
\details{
A baseline effect is automatically specified for the tie-oriented model and
the sender activity rate step of the actor-oriented model when the
\code{ordinal} argument in \code{\link{remstats}}, \code{\link{tomstats}},
\code{\link{aomstats}} is set to FALSE (default) and automatically removed
when this argument is set to TRUE. Alternatively, a baseline effect can be
explicitly specified by adding '1' to the equation or explicitly removed by
adding '-1' to the equation.

The baseline effect refers to the baseline tendency to interact. In the
tie-oriented model, the log-inverse of the estimated parameter translates to
the average number of observed events per time unit per dyad. In the
actor-oriented model, the log-inverse of the estimated parameter translates
to the average number of observed events per time unit per actor. The
statistic is equal to one for all dyads resp. actors in the riskset at all
timepoints.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
remstats(reh = reh_tie, tie_effects = ~1)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, sender_effects = ~1)
}

=cut

TITLE=availability-report.cgi
DESCRIPTION=show availability of various RRD variables
KEYWORDS=availability-report.cgi
DOCTOP=index
DOCPREV=alert-cgi
DOCNEXT=dataimage-cgi

=pod

=head1 availability-report.cgi

The C<availability-report.cgi|availability-report-cgi> calls L<availability-report> to produce
a report of "availability" according to the definitions in the
L<availability|configfile-availability> config-file.

=cut

TITLE=alerter
DESCRIPTION=Figures out the appropriate templates to use to construct the alert text and triggers the appropriate alert-senders
KEYWORDS=alerter,alerts
DOCTOP=index
DOCPREV=alert-monitor
DOCNEXT=alert-email

=pod

=head1 alerter - construct and send alert text

=head2 Usage:

=exec ../alerter -h 2>&1

=head2 Description:

[C<Alerter> may be rolled into the C<alert-monitor> at some point in
the future.  It was easier to test as a separate program, and the performance
hasn't been an issue for me.]

C<Alerter> is passed its parameters (specified above) by the L<alert-monitor>.
Most of them are used to fill in information in the text of the alert.  The
interesting ones are C<towho> and C<template>.

It also reads the L<alert-destination-map config-file|configfile-alert-destination-map>
to decide where the alert needs to go.  This will give it a list of (method, address)
pairs.

For a given template-name, say C<xxx>, and method, say C<method>, it will 
look for files in @@CONFIGDIR@@/alert-templates, called:

	method-xxx
	method-DEFAULT
	xxx
	DEFAULT

and take the first one it finds.  Similarly, it will look for a header to add to the
top of the template called:

	method-HEADER
	HEADER

and a footer in one of:

	method-FOOTER
	FOOTER

The three pieces will be concatenated giving the template text.  Then substitutions
will be done for the following L<magic cookies|cookies>:

	HOST IP REALRRD WILDPART FIXEDRRD VAR STATUS OLDSTATUS
	VALUE RELATION THRESHOLD START DURATION HOSTDESC RRDDESC 
	NOW TEXTNOW ALERTHOST TOWHO WEBMASTER HTMLURL CGIURL HOSTINDEXURL

This gives the alert text.  From the method definition in the
L<alert-destination-map config-file|configfile-alert-destination-map>
C<alerter> knows which program to run to send the alert text to the
appropriate address, and it does it.

=head2 Alert-Sending Scripts

These are now easy to write, and in many cases you won't even have to
write one.  There are two requirements for an alert-sending script:

=over 4

=item 1) It must take an address to send to on the command-line, and

=item 2) It must accept the text on stdin.

=back 4

E.G. you could use sendmail with no wrapper.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{osp}
\alias{osp}
\title{osp}
\usage{
osp(unique = FALSE, scaling = c("none", "std"), consider_type = "ignore")
}
\arguments{
\item{unique}{A logical value indicating whether to sum the minimum of
events with third actors (FALSE, default) or the number of third actors that
create a new, unique shared partner (TRUE). See details for more information.}

\item{scaling}{the method for scaling the triad statistic. Default is to not
scale the statistic but keep the raw 'counts'. Alternatively,
standardization of the raw counts per time point can be requested
with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for an outgoing shared partners effect.
}
\details{
The outgoing shared partners effect describes the propensity of dyads to
interact based on the number of past outgoing shared partners between them.
By default, the statistic at timepoint t for the dyad (i,j) is computed as
the sum of the minimum occurrences of past (i,h) and (j,h) events across all
actors h.

When the unique parameter is set to TRUE, a different approach is taken.
In this case, the statistic counts the number of actors h that contribute to
the creation of a new, distinct shared partner between actors i and j.

Additionally, it is possible to specify a scaling method using the scaling
parameter.

Please note that the outgoing shared partners effect, 'osp', is exclusively
defined for directed events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ osp()
remstats(reh = reh_tie, tie_effects = effects)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, receiver_effects = effects)

}
\references{
Butts, C. (2008). A relational event framework for social
action. Sociological Methodology.
}
\seealso{
\code{\link{otp}}, \code{\link{itp}}, or \code{\link{isp}} for
other types of triadic effects for directed relational events and
\code{\link{sp}} for triadic effects for undirected relational events.
}

=cut

TITLE=availability-report
DESCRIPTION=show availability of various RRD variables
KEYWORDS=availability-report
DOCTOP=index
DOCPREV=alerter
DOCNEXT=convert-config-links

=pod

=head1 availability-report

=head2 Usage:

=exec ../availability-report -h 2>&1

=head2 Description:

This is mainly intended to be called from 
L<availability-report.cgi|availability-report-cgi>.  It provided
a report on "availability" of specified RRD variables, by default, all
that have definitions in the L<availability|configfile-availability>
config-file.  Exactly what it means for a variable to be "available"
is up to you.  It's intended to give some measure of when a host or
service isn't useable, so, e.g. the default definition of availability
for the C<ping> RRD variable C<rcvd> (number of ping responses received)
is:

	ping rcvd MINIMUM > 0

In english, the C<rcvd> variable is considered unavailable if:

	- it is less than or equal to zero (I.E. it didn't respond to ping)
	- there is no data available for that time period

B<N.B.:> The interaction between rrd archive consolidation and the xff value,
(see rrdcreate), can result in longer periods of unavailability or conversely,
masking periods of unavailability.  Choose the consolidation function carefully
to make sure you're getting the best data possible.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{info}
\alias{info}
\title{Simulated exogenous information on actors in a social network.}
\format{
A dataframe with 10 rows and 5 variables:
\describe{
 \item{id}{numeric id of the actor}
 \item{time}{numeric value, describes when the value of the covariate 
  changes, if it changes}
 \item{age}{dichotomized age of the actor (e.g., 0 = below 25, 1 = 25 or 
  older)}
 \item{sex}{dichotomized sex of the actor (e.g., 0 = male, 1 = female)}
 \item{extraversion}{standardized extraversion score of the actor}
 \item{agreeableness}{standardized agreeableness score of the actor}
}
}
\source{
Simulated exogenous information on actors in a social network.
}
\usage{
data(info)
}
\description{
A dataset containing exogenous information on the actors in the social 
network of a relational event history. Data is simulated.
}
\examples{
data(info)

}
\seealso{
\code{\link{history}} for the relational event history.
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{userStat}
\alias{userStat}
\title{userStat}
\usage{
userStat(x, variableName = NULL)
}
\arguments{
\item{x}{Matrix with number of rows equal to the number of events and number
of columns equal to the number of dyads in the network (tie-oriented model) 
or the number of actors in the network (actor-oriented model)}

\item{variableName}{Optionally, a string with the name of the statistic.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Allows the user to add its own pre-computed statistic to the statistics
object and, optionally, interact this statistic with other statistics in the
formula.
}
\examples{
\donttest{
 reh <- remify::remify(history, model = "tie")
 actor101Events <- which(history$actor1 == "101" | history$actor2 == "101")
 actor101_stat <- t(sapply(seq_len(nrow(history)), function(i) {
   rep(i \%in\% actor101Events, reh$D)
 }))
 
 # Main effects only
 effects <- ~ userStat(x = actor101_stat, variableName = "actor101event")
 remstats(reh = reh, tie_effects = effects)
 
 # Model with interaction effects
 interaction_effects <- ~ inertia() *
   userStat(x = actor101_stat, variableName = "actor101event")
 remstats(reh = reh, tie_effects = interaction_effects)
}
}

=for html
<P>
<HR>
[<A CLASS="BUT" HREF="@@HTMLURL@@/release/index.html">Top</a>]
[<A CLASS="BUT" HREF="http://remstats.sourceworks.com:1954/">Remstats</a>]
[<A CLASS="BUT" HREF="http://www.sourceworks.com/">SourceWorks</a>]
[<A CLASS="BUT" HREF="http://ee-staff.ethz.ch/~oetiker/webtools/rrdtool">RRDtool</a>]
[<A CLASS="BUT" HREF="http://sourceforge.net/projects/remstats">SourceForge</a>]
<BR><FONT SIZE="-1">Last updated @@DATE@@ by <A HREF="mailto:@@AUTHOR@@">&lt;@@AUTHOR@@&gt;</A></FONT>.
</TD></TR>
</TABLE>
</BODY></HTML>

=for text

=for text
------------------------------------------------------------------
Last updated @@DATE@@ by <@@AUTHOR@@>.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{psABAY}
\alias{psABAY}
\title{psABAY}
\usage{
psABAY(consider_type = "ignore")
}
\arguments{
\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a participation shift AB-AY.
}
\details{
One of Gibson's (2003) dyadic participation shifts. The AB-AY participation
shift refers to a tendency for \emph{turn continuing}. For directed events,
the sender (A) in the current event is the same as the sender in the
previous event (A), and the receiver (Y) is different from the previous
receiver (B). In undirected events, one of the current actors (A) matches
one of the actors in the previous events (A or B), while the other actor (Y)
is different.

To identify these shifts, a statistic 'psABAY' is calculated for each pair
of actors at a given timepoint (t). If the pair's interaction follows the
AB-AY pattern, the statistic is set equal to one; otherwise, it's set to
zero.

Additionally, the types of the AB and AY events can be taken into account. If
'consider_type' is 'TRUE', the type of the AB event and the type of the AY
event must match for the shift to occur. If 'consider_type' is 'FALSE', the
shift happens for every AY event, regardless of the event type.
}
\examples{
reh <- remify::remify(history, model = "tie")
effects <- ~ psABAY()
remstats(reh = reh, tie_effects = effects)

}
\seealso{
\code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXA}},
\code{\link{psABXB}}, \code{\link{psABXY}} or \code{\link{psABAB}} for other
dyadic participation shifts.
}

=cut

TITLE=alert-monitor
DESCRIPTION=decide on status of rrd variables, and optionally generate alarms
KEYWORDS=alert-monitor
DOCTOP=index
DOCPREV=monitors
DOCNEXT=alerter

=pod

=head1 alert-monitor - a status evaluator and alert trigger

=head2 Usage:

=exec ../alert-monitor -h 2>&1

=head2 Description:

The C<alert-monitor> compares the current value of variables specified in
the L<alerts file|configfile-alerts> in the configuration directory with
threshold values and sets the status of those variables accordingly.  
It saves the current status of variables in C<@@DATADIR@@/ALERTS>.

What value corresponds to what status level is set in the 
L<rrd definition|configfile-rrds> or sometimes the 
L<host definition|configfile-hosts>.  This way an rrd
definition will specify generally reasonable levels, but they can be 
overridden for hosts where they aren't reasonable.

For an rrd definition, an alert line looks like:

	alert varname relation oklevel [warnlevel [errorlevel]]

or

	alert varname nodata status

[The latter says that missing data for variable C<varname> will cause its status
to be level C<status>.]

For a host-specified alert level, the line looks like:

	alert rrdname varname relation oklevel [warnlevel [errorlevel]]

or

	alert rrdname varname nodata status

and the interpretation is the same, except that you're having to say
which rrd this alert refers to.


The available relations are:

	< (value is less than threshold)
	> (value is greater than threshold)
	= (value is equal to threshold)
	|< (absolute value of value is less than threshold)
	|> (absolute value of value is greater than threshold)
	delta< (difference between last two values is less than threshold)
	delta> (difference between last two values is greater than threshold)
	<daystddev (value is outside threshold * the past day's standard-deviation)
	<weekstddev (value is outside threshold * the past day's standard-deviation)
	<monthstddev (value is outside threshold * the past day's standard-deviation)

=head2 Example

To make things more concrete for the first (normal) case, here's a real example, 
from the C<load> rrd supplied in C<config-base>:

	alert load5 < 3 7 10

This means that if the C<load5> variable is less than 3, the status is set to OK.
If it's less than 7, it's WARN, less than 10 it's ERROR and more than that, it's
CRITICAL.

Since the first match is taken, it's possible to leave out the upper levels if
you don't want them to ocurr.  For example if you only wanted C<load5> to ever
go to WARN level, never above, you could use:

	alert load5 < 3

and then the only possible status levels are OK and WARN.

The possible C<relation>s are: <, =, >, |<, |>, delta<, delta>.  The first
three should be obvious.  The next two allow comparisons to the absolute value of
the variable's current value.  The last two allow comparisons to the change in
value.

=head2 Causing alerts

Depending on the lines in the L<alerts file|configfile-alerts>, the status may also
trigger alerts.  A matching line in the L<alerts config-file|configfile-alerts> will cause
C<alert-monitor> to run the L<alerter> for each of the specified 
recipients.  It will also be passed, in order:

=over 4

=item B<recipient> - the recipient; for L<alert-email> it
will be an email address

=item B<hostname> - the name of the host that the alert applies to

=item B<ip> - the IP number for that host, in case it's not in DNS

=item B<rrdname> - the name of the RRD

=item B<wildpart> - the wild part of a wildcard RRD.  E.G, for an
RRD of C<port-ftp> (using the wildcard RRD C<port-*>) the wildpart
would be C<ftp>.

=item B<variable> - the name of the variable

=item B<status> - the current status, as decided by alert-monitor

=item B<old_status> - the previous status

=item B<value> - the current value of the variable

=item B<relation> - the relation used to compare the variable to the
threshold, mostly for creating informative messages

=item B<threshold> - the threshold value that was exceeded

=item B<start> - timestamp of when the alert started

=item B<duration> - number of seconds that the alert has been active

=item B<host-description> - the description field from the host config-file

=item B<rrd-description> - the description tag on this rrd (desc="xxx")

=item B<webmaster> - the email address of the remstats person

=item B<template> - the name of the template file to generate the message from.

=back 4

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{indegreeReceiver}
\alias{indegreeReceiver}
\title{indegreeReceiver}
\usage{
indegreeReceiver(scaling = c("none", "prop", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the degree statistic. Default is
to not scale the statistic (scaling = "none"). Alternatively,
scaling of the raw degree counts by the number of past events at time t can
be requested with 'prop' or standardization of the raw degree counts per
time point can be requested with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for an `indegreeReceiver` effect in the tie-oriented
model or the receiver choice step of the actor-oriented model.
}
\details{
An indegree of the receiver effect refers to the tendency for actors to
receive events if they have received more past events. The statistic at
timepoint \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or
receiver \emph{j} (actor-oriented model) is equal to the number of events
received by actor \emph{j} before timepoint \emph{t}. Note that the
'indegreeReceiver' effect is only defined for directed events.

Optionally, a scaling method can be set with \code{scaling}. By scaling the
degree count by the total number of past events, the statistic refers to the
fraction of past events that were received by actor j. At the first time
point, when no events did previously occur, it is assumed that every actor
is equally likely to receive a message and the statistic is set equal to 1/
n, where n refers to the number of actors.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ indegreeReceiver()
remstats(reh = reh_tie, tie_effects = effects)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, receiver_effects = effects)

}
\seealso{
\code{\link{indegreeSender}}, \code{\link{outdegreeSender}},
\code{\link{outdegreeReceiver}}, \code{\link{totaldegreeSender}}, or
\code{\link{totaldegreeReceiver}} for other types of degree effects.
}

=for html
@@DOCTYPE@@
<HTML>
<!--
$Id: HEADER.pod,v 1.8 2002/05/01 05:05:43 remstats Exp $
Copyright 1999, 2000, 2001, 2002 (c) Thomas Erskine.
See the COPYRIGHT file with the distribution.
-->
<HEAD>
	<TITLE>remstats - @@TITLE@@</TITLE>
	<META HTTP-EQUIV="description"
		CONTENT="remstats - @@DESCRIPTION@@">
	<META HTTP-EQUIV="keywords"
		CONTENT="remstats,rrdtool,@@KEYWORDS@@">
	<LINK REV="MADE" HREF="mailto:@@AUTHOR@@">
	<LINK REL="STYLESHEET" HREF="@@HTMLURL@@/default.css" TYPE="text/css">
</HEAD>
<BODY BGCOLOR="white">
<TABLE BORDER=0>
<TR WIDTH="20%">
<TD ALIGN="LEFT" VALIGN="TOP" COLOR="#000000" BGCOLOR="#5F9FFF" LINKCOLOR="#FFFFFF" VISITEDCOLOR="#FF9F9F">
<P ALIGN="CENTER"><BR><A HREF="@@SRCURL@@index.html"><IMG SRC="@@HTMLURL@@/IMAGES/remstats.png" BORDER=0 ALT="Remstats Home"></A><P>
<BR>
<FONT SIZE="-1">

=cut

TITLE=alert-yahoo
DESCRIPTION=sends an alert message via yahoo instant messenger
KEYWORDS=alert-yahoo
DOCTOP=index
DOCPREV=alert-winpopup
DOCNEXT=alert-yahoo

=pod

=head1 alert-yahoo - an alert sending script

This is a simple script intended to be run by the L<alerter>.  
Like all alert-senders, it takes one argument on the command-line:
the "address" to send the alert to.  The text of the alert is
fed to this script on stdin.  

It sends the alert to "address", in this case a yahoo ID.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tomstats.R
\name{tomstats}
\alias{tomstats}
\title{tomstats}
\usage{
tomstats(
  effects,
  reh,
  attr_actors = NULL,
  attr_dyads = NULL,
  memory = c("full", "window", "decay", "interval"),
  memory_value = NA,
  start = 2,
  stop = Inf,
  display_progress = FALSE,
  sampling = FALSE,
  samp_num = NULL,
  seed = NULL
)
}
\arguments{
\item{effects}{an object of class \code{"\link[stats]{formula}"} (or one
that can be coerced to that class): a symbolic description of the effects in
the model for which statistics are computed, see 'Details' for the available
effects and their corresponding statistics}

\item{reh}{an object of class \code{"\link[remify]{remify}"} characterizing 
the relational event history.}

\item{attr_actors}{optionally, an object of class
\code{"\link[base]{data.frame}"} that contains exogenous attributes for 
actors (see Details).}

\item{attr_dyads}{optionally, an object of class \code{data.frame} or 
\code{matrix} containing attribute information for dyads (see Details).}

\item{memory}{The memory to be used. See `Details'.}

\item{memory_value}{Numeric value indicating the memory parameter. Default 
is \code{NA}, which is only valid for \code{memory = "full"} (no memory
parameter required). See `Details'.}

\item{start}{an optional integer value, specifying the index of the first
time or event in the relational event history for which statistics must be 
computed (see 'Details'). Default is \code{2}: the first event has no history and is used only to initialize statistics, not to fit the model.}

\item{stop}{an optional integer value, specifying the index of the last
time or event in the relational event history for which statistics must be 
computed (see 'Details')}

\item{display_progress}{should a progress bar for the computation of the
endogenous statistics be shown (TRUE) or not (FALSE)?}

\item{sampling}{Logical. If \code{TRUE}, statistics are computed using
case–control (dyad) sampling rather than the full risk set. Default \code{FALSE}.}

\item{samp_num}{Integer. Number of dyads to include per event when
\code{sampling = TRUE}. Must be smaller than or equal to the size of the
active risk set. Ignored when \code{sampling = FALSE}.}

\item{seed}{Optional integer. Random seed used for dyad sampling. Setting
this ensures reproducible sampling across calls. If \code{NULL}, the
current RNG state is used.}
}
\value{
An object of class 'tomstats'. Array with the computed statistics,
where rows refer to time points, columns refer to potential relational event
(i.e., potential edges) in the risk set and slices refer to statistics.
Statistics are computed once per unique time point (per-timepoint "pt"
method), so the number of rows in the output equals \code{reh$M} (the number 
of  unique time points), which may be less than the total number of observed 
events when simultaneous events are present. The 'tomstats' object has the 
following attributes:
  \describe{
    \item{\code{model}}{Type of model that is estimated.}
    \item{\code{formula}}{Model formula, obtained from the formula inputted to 'tie_effects'.}
    \item{\code{riskset}}{The risk set used to construct the statistics.}
  }
}
\description{
Computes statistics for modeling relational event history data
with the tie-oriented relational event model.
}
\section{Effects}{

The statistics to be computed are defined symbolically and should be
supplied to the \code{effects} argument in the form \code{~ effects}. The
terms are separated by + operators. For example:
\code{effects = ~ inertia() + otp()}. Interactions between two effects
can be included with * operators. For example:
\code{effects = ~ inertia()*otp()}. A list of available effects can be
obtained with \code{\link{tie_effects}()}.

The majority of the statistics can be scaled in some way, see
the documentation of the \code{scaling} argument in the separate effect
functions for more information on this.

The majority of the statistics can account for the event type
included as a dependent variable, see the documentation of the
\code{consider_type} argument in the separate effect functions for more
information on this.

Note that events in the relational event history can be directed or
undirected. Some statistics are only defined for either directed or
undirected events (see the documentation of the statistics). Note that
undirected events are only available for the tie-oriented model.
}

\section{attr_actors}{

For the computation of the \emph{exogenous} statistics an attributes object
with the exogenous covariate information has to be supplied to the
\code{attr_actors} argument in either \code{remstats()} or in the separate
effect functions supplied to the \code{..._effects} arguments (e.g., see
\code{\link{send}}). This \code{attr_actors} object should be constructed as
follows: A dataframe with rows referring to the attribute value of actor
\emph{i} at timepoint \emph{t}. A `name` column is required that contains the
actor name (corresponding to the actor names in the relational event
history). A `time` column is required that contains the time when attributes
change (set to zero if none of the attributes vary over time). Subsequent
columns contain the attributes that are called in the specifications of
exogenous statistics (column name corresponding to the string supplied to
the \code{variable} argument in the effect function). Note that the
procedure for the exogenous effects `tie' and `event' deviates from this,
here the exogenous covariate information has to be specified in a different
way, see \code{\link{tie}} and \code{\link{event}}.
}

\section{attr_dyads}{

For the computation of the \emph{dyad exogenous} statistics with \code{tie()}, an attributes object with the exogenous covariates information per dyad has to be supplied. This is a \code{data.frame} or \code{matrix} containing attribute information for dyads. If \code{attr_dyads} is a \code{data.frame}, the first two columns should represent "actor1" and "actor2" (for directed events, "actor1" corresponds to the sender, and "actor2" corresponds to the receiver). Additional columns can represent dyads' exogenous attributes. If attributes vary over time, include a column named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond to "actor1", columns to "actor2", and cells contain dyads' exogenous attributes.
}

\section{Memory}{

The default `memory` setting is `"full"`, which implies that at each time
point $t$ the entire event history before $t$ is included in the computation
of the statistics. Alternatively, when `memory` is set to `"window"`, only
the past event history within a given time window is considered (see
Mulders & Leenders, 2019). This length of this time window is set by the
`memory_value` parameter. For example, when `memory_value = 100` and `memory
= "window"`, at time point $t$ only the past events that happened at most
100 time units ago are included in the computation of the statistics.
A third option is to set `memory` to `"interval"`. In this case, the past
event history within a given time interval is considered. For example, when
`"memory_value" = c(50, 100)` and `memory = "window"`, at time point $t$
only the past events that happened between 50 and 100 time units ago are
included in the computation of the statistics. Finally, the fourth option is
to set `memory` to `"decay"`. In this case, the weight of the past event in
the computation of the statistics depend on the elapsed time between $t$ and
the past event. This weight is determined based on an exponential decay
function with half-life parameter `memory_value` (see Brandes et al., 2009).
}

\section{Event weights}{

Note that if the relational event history contains a column that is named
``weight'', it is assumed that these affect the endogenous statistics. These
affect the computation of all endogenous statistics with a few exceptions
that follow logically from their definition (e.g., the recenyContinue
statistic does depend on time since the event and not on event weights).
}

\section{Subset the event history using 'start' and 'stop'}{

It is possible to compute statistics for a segment of the relational event 
sequence, based on the entire event history. This is done by specifying the 
'start' and 'stop' values as the indices for the first and last event times 
for which statistics are needed. For instance, setting 'start = 5' and 'stop 
= 5' calculates statistics for the 5th event in the relational event 
sequence, considering events 1-4 in the history. Note that in cases of 
simultaneous events with the 'method' set to 'pt' (per timepoint), 'start' 
and 'stop' should correspond to the indices of the first and last 
\emph{unique} event timepoints for which statistics are needed. For example, 
if 'start = 5' and 'stop = 5', statistics are computed for the 5th unique 
timepoint in the relational event sequence, considering all events occurring 
at unique timepoints 1-4.
}

\section{Adjacency matrix}{

Optionally, a previously computed adjacency matrix can be supplied. Note
that the endogenous statistics will be computed based on this adjacency
matrix. Hence, supplying a previously computed adjacency matrix can reduce
computation time but the user should be absolutely sure the adjacency matrix
is accurate.
}

=cut

TITLE=alert-syslog
DESCRIPTION=sends an alert message via syslog
KEYWORDS=alert-syslog
DOCTOP=index
DOCPREV=alert-winpopup
DOCNEXT=alert-yahoo

=pod

=head1 alert-syslog - an alert sending script

=exec ../alert-syslog -h 2>&1

This simply raises an alert by writing it to syslog.  Useful if you want to
keep everything in syslog.

</FONT>
<P><HR>
&nbsp;&nbsp;<A HREF="@@HTMLURL@@/quick-index.cgi">Live Data</A>
<P>
&nbsp;&nbsp;<FONT SIZE="-1">Remstats was written by<BR>
&nbsp;&nbsp;<A HREF="mailto:@@AUTHOR@@">Thomas Erskine</A> at the<BR>
&nbsp;&nbsp;<A HREF="http://www.crc.ca/">CRC</a> in Canada and now<BR>
&nbsp;&nbsp;at <A HREF="http://www.sourceworks.com/">SourceWorks</A>.
</FONT>
</TD><TD WIDTH="5"></TD>
<TD ALIGN="LEFT" VALIGN="TOP">
[<A CLASS="BUT" HREF="@@DOCTOP@@.html">Top</A>]
[<A CLASS="BUT" HREF="@@DOCPREV@@.html">Prev</A>]
[<A CLASS="BUT" HREF="@@DOCNEXT@@.html">Next</A>]
<HR>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{psABX}
\alias{psABX}
\title{psABX}
\usage{
psABX()
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a participation shift AB-X in the sender step of
the actor-oriented model.
}
\details{
The AB-X participation shift refers to a tendency for turn usurping (here,
the next sender is not in the previous event). For each timepoint t, the
psABX statistic is equal to one for the actors that will create the
participation shift if they would occur as the sender in the edgelist at
time t and equal to zero for the actors that will not create this
participation shift. If multiple events in the edgelist occur at the same
time point, the order of these events determines whether the p-shift is
observed.
}
\examples{
reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, sender_effects = ~ psABX())

}
\seealso{
\code{\link{psABA}} or \code{\link{psABB}} for exploring
alternative participation shifts in the sender step of the actor-oriented
model.
}

=cut

TITLE=alert-winpopup
DESCRIPTION=sends an alert message via a Windows popup message
KEYWORDS=alert-winpopup
DOCTOP=index
DOCPREV=alert-syslog
DOCNEXT=alert-yahoo

=pod

=head1 alert-winpopup - an alert sending script

This is a simple script intended to be run by the L<alerter>.  
Like all alert-senders, it takes one argument on the command-line:
the "address" to send the alert to.  The text of the alert is
fed to this script on stdin.  

It sends the alert to use "address", in this case a windows NetBIOS
machine name, via a Windows popup message.

=cut

TITLE=architecture
DESCRIPTION=how the pieces fit together
KEYWORDS=architecture,design
DOCTOP=index
DOCPREV=index
DOCNEXT=releasenotes

=pod

=head1 Architecture

This will tell how the major pieces of remstats work together, when I finish it.

=head2 A Bit of History

There were several things which decided how remstats would work.  One of the
major ones was, of course, L<RRDtool|@@RRDTOOLHOME@@>.  I wanted to have good
graphs, without having to build them from scratch.  I liked the side benefit
that RRDs do not grow, lessening the file-management required.

I've dealt with various monitoring packages with long-running daemons which are
always needing to be restarted, or checked via cron jobs and I decided that
remstats wouldn't have that kind of problem.

Back in the bad old days, you couldn't count on having an SNMP agent available
on every platform, so I decided that remstats wouldn't require SNMP to work.
Now, of course, there is the wonderful 
L<Net-SNMP project|http://net-snmp.sourceforge.net/> to provide a portable
SNMP implementation.

I wanted it to be easily extendable, without having to completely understand
the guts of remstats.  This goal has not been met very successfully, though
I hope to improve this in the 2.x releases.  However, some folks have managed
to create new collectors, so I guess it wasn't a complete failure.

=head2 The Collection Process

The process of data collection in remstats is begun by the L<run-remstats2>
program.  (In older versions, L<run-remstats>, a more primitive program,
filled the same niche.)  It is started by cron at the minimum collection
interval.  It passes through several stages in the order listed in the
L<run config-file|configfile-run>.  Some of these stages may have conditions
on them.  E.G. the standard stage C<pagemakers> will only be run if the
remstats configuration has changed.

Each stage is described by a file in the
L<run-stages config-dir|configfile-run-stages>, which lists all the programs
to run.  These are all started and run in parallel.  When all the programs
(or pipelines) are finished, T<run-remstats2> moves to the next stage.

The standard stages are:

=over 4

=item check
- checks the configuration to make sure it's good, and as a side-effect
creates any required new/missing RRDs, other files or directories.

=item ping
- runs the L<ping-collector> to see which hosts are up, so as to avoid
wasting time on hosts which are down.  Hosts which aren't pinged are considered
to be "up" for collection purposes, to deal with un-pingable hosts.

=item collectors
- runs all the other L<collectors> in parallel

=item post-collector
- runs the L<error-collector>, which collects info on the collectors.

=item monitors
- runs the remstats L<monitors> which evaluate the current status of the
monitored systems and issue alerts, if warranted.

=item pagemakers
- runs a motley crew of programs to create various kinds of web pages, but
only if the configuration has changed.  One of the standard cron jobs will
touch the man configuration directory to force page generation once a day.

=back 4

=head2 Collectors

At the centre of remstats are the L<collectors>, and their L<servers>.  The
collectors do whatever they need to do to acquire data, reformat it to the
remstats standard collector form and pass it to an instance of the L<updater>
which does the RRD updating.  Some collectors have a corresponding server
which they will contact on remote hosts, and which will provide information
on request.  The servers do not know about the configuration of remstats and
are completely stand-alone; the collector will tell them what kind of data it
requires and the server will provide it in a standard form.

XXX

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.aomstats}
\alias{plot.aomstats}
\title{Plotting Relational Event Network Statistics Trajectories}
\usage{
\method{plot}{aomstats}(x, effect, subset = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{\link{aomstats}} containing relational
event network statistics.}

\item{effect}{A character string specifying the name of the effect in 'x' or
an integer indicating the index of the effect to be plotted.}

\item{subset}{An optional vector specifying a subset of actors to be used for
plotting. By default, a maximum of 5 unique actors are used for plotting.}

\item{...}{Additional arguments passed to plot().}
}
\description{
Generate line plots to visualize the trajectories of a specified effect in 
the sender model of a \code{\link{aomstats}} object.
}
\details{
This function creates line plots to illustrate the temporal trajectories of a
specified effect in a relational event network, as captured in the sender 
model by a \code{\link{aomstats}} object. The 'effect' parameter allows users 
to choose a specific effect for visualization, either by providing the effect's 
name or its index within the 'aomstats' object. An optional 'subset' parameter 
enables users to focus on specific actors. If 'subset' is not specified, a 
default maximum of 5 unique actors is plotted. These actors are randomly selected 
to represent trajectories across the range of different endpoints for the effect
(excluding zero).
}
\examples{
library(remstats)
# Load data
data(history)
# Prepare data
reh <- remify::remify(edgelist = history[,1:3], model = "actor")
# Compute effects
stats <- remstats(reh, sender_effects = ~ outdegreeSender())
# Plot the 'outdegreeSender' trajectories 5 actors
plot(stats, effect = "outdegreeSender")
# Plot the 'outdegreeSender' trajectory for a specific actor
plot(stats, effect = "outdegreeSender", subset = 10)

}

=cut

TITLE=alert-email
DESCRIPTION=sends email to notify someone of an alert
KEYWORDS=alert-email
DOCTOP=index
DOCPREV=alerter
DOCNEXT=alert-syslog

=pod

=head1 alert-email - an alert sending script

This is a simple script intended to be run by the L<alerter>.  
Like all alert-senders, it takes one argument on the command-line:
the "address" to send the alert to.  The text of the alert is
fed to this script on stdin.  

It sends the alert text to "address" via email, by invoking sendmail,
though there's no reason that it couldn't be re-written to do an SMTP
injection directly if some-one wanted to.  With no error-checking, it
could be re-written as:

	#!/bin/sh
	/usr/lib/sendmail "$1"

=cut

TITLE=alert.cgi
DESCRIPTION=alert reporting and updating
KEYWORDS=alert.cgi
DOCTOP=index
DOCPREV=cgis
DOCNEXT=availability-report-cgi

=pod

=head1 alert.cgi - Alert Reporting and Updating

This CGI script will generate the Alert Report and also let you modify it.
You can turn alerts off for a specific line (by checking the C<quench>
check-box).  You can also attach comments to a specific alert, for example
to let other people know that you're already working on it, or when a
service will be available again.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{FEtype}
\alias{FEtype}
\title{FEtype}
\usage{
FEtype()
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for fixed effects for event types in the
tie-oriented model.
}
\details{
Fixed effects for event types capture the variation in event rate across
different event types (e.g., see Butts, 2008). The specification of FEtype
results in the creation of C-1 statistics, were C is the number of different
event types in the riskset. Let one of the event types, e.g. \emph{c = 1},
represent the reference category. Than, for every event type
\emph{c = 2, ..., C}, a statistic is created that at timepoint \emph{t} for
dyad \emph{(i,j,c)} is equal to 1 if \emph{c} is equal to the respective
event type and equal to 0 otherwise (i.e., dummy variables are created).
Note that specifying fixed effects for event types is only available when
event types are modeled in the dependent variable.
}
\examples{
history$type <- history$setting
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ FEtype()
remstats(reh = reh_tie, tie_effects = effects)

}
\seealso{
\code{\link{event}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{degreeDiff}
\alias{degreeDiff}
\title{degreeDiff}
\usage{
degreeDiff(scaling = c("none", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the degree statistic. Default is to
not scale the statistic (scaling = "none"). Alternatively, standardization
of the degree difference per time point can be requested with `std`.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for a `degreeDiff` effect in the tie-oriented
model.
}
\details{
A degreeDiff effect refers to the tendency for dyads to increase their
interaction rate if the absolute difference in degree for the two actors in
the pair increases. The statistic at timepoint \emph{t} for dyad
\emph{(i,j)} is equal to the difference between the following two values:
the number of events before timepoint \emph{t} that involved actor \emph{i}
and actor \emph{j}, respectively. The degreeDiff effect is only defined for
undirected events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
effects <- ~ degreeDiff()
remstats(reh = reh_tie, tie_effects = effects)

}
\seealso{
\code{\link{degreeMin}}, \code{\link{degreeMax}} or
\code{\link{totaldegreeDyad}} for other types of degree effects for
undirected events.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{reciprocity}
\alias{reciprocity}
\title{reciprocity}
\usage{
reciprocity(scaling = c("none", "prop", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the reciprocity statistic. Default is
to not scale the statistic but keep the raw 'counts'. Alternatively, the
statistics can be scaled by 'prop', in which raw counts are
divided by the indegree of the sender at time t (see 'details') or
standardization of the raw counts per time point can be requested with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a reciprocity effect in the tie-oriented model
or the receiver choice step of the actor-oriented model.
}
\details{
A reciprocity effect refers to the tendency for actors to reciprocate past
interactions. The statistic at timepoint \emph{t} for dyad \emph{(i,j)}
(tie-oriented model) or receiver \emph{j} (actor-oriented model) is equal to
the number of \emph{(j,i)} events before timepoint \emph{t}. Note that a
reciprocity effect is only defined for directed events.

Optionally, a scaling method can be set with \code{scaling}.  By scaling the
reciprocity count by the indegree of the sender, the statistic refers to the
fraction of messages received by actor i that were received from actor j. If
actor i hasn't received any messages yet it can be assumed that actor i is
equally likely to receive a message from every actor and the statistic is
set equal to 1/(n-1), where n refers to the number of actors. The resulting
statistic is similar to the "FrRecSnd" statistic in the R package 'relevent'.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ reciprocity()
remstats(reh = reh_tie, tie_effects = effects)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, receiver_effects = effects)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{degreeMin}
\alias{degreeMin}
\title{degreeMin}
\usage{
degreeMin(scaling = c("none", "prop", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the degree statistic. Default is
to not scale the statistic (scaling = "none"). Alternatively,
scaling of the raw degree counts by two times the number of past events at
time t can be requested with 'prop' or standardization of the raw degree
counts per time point can be requested with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for an `degreeMin` effect in the tie-oriented
model with undirected events.
}
\details{
An degreeMin effect refers to the tendency for dyads to increase their
interaction rate if the total degree of the least active actor in the pair
increases. The statistic at timepoint \emph{t} for dyad \emph{(i,j)} is
equal to the minimum of the following two values: the number of events
before timepoint \emph{t} that involved actor \emph{i} and actor \emph{j},
respectively. Note that the degreeMin effect is only defined for undirected
events.

Optionally, a scaling method can be set with \code{scaling}. By scaling the
degree count by the total number of past events, the statistic refers to the
fraction of past events that the least active actor was involved in. At the
first time point, when no events did previously occur, it is assumed that
every actor is equally likely to be involved in an event and the statistic
is set equal to 1/n, where n refers to the number of actors.
}
\examples{
reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
effects <- ~ degreeMin()
remstats(reh = reh_tie, tie_effects = effects)

}
\seealso{
\code{\link{degreeDiff}}, \code{\link{degreeMax}} or
\code{\link{totaldegreeDyad}} for other types of degree effects for
undirected events.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.remstats}
\alias{summary.remstats}
\title{Relational Event Network Statistics Summaries}
\usage{
\method{summary}{remstats}(object, ...)
}
\arguments{
\item{object}{object of class \code{\link{remstats}}.}

\item{...}{additional arguments affecting the summary produced.}
}
\value{
The summaries provide information for each statistic included in the 
remstats object, offering insights into the distribution and characteristics 
of the data.
}
\description{
Produce summaries of each statistic from a \code{\link{remstats}} object.
}
\examples{
rehObject <- remify::remify(edgelist = history, model = "tie")
remstatsObject <- remstats::remstats(reh = rehObject, tie_effects = ~ remstats::inertia())
summary(remstatsObject)

rehObject <- remify::remify(edgelist = history, model = "actor")
remstatsObject <- remstats::remstats(reh = rehObject, receiver_effects = ~ inertia())
summary(remstatsObject)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{same}
\alias{same}
\title{same}
\usage{
same(variable, attr_actors = NULL, attr_data)
}
\arguments{
\item{variable}{string with the name of the column in the
\code{attr_actors} object for which the statistic has to be computed.}

\item{attr_actors}{optionally, an object of class
\code{\link[base]{data.frame}} that contains the attribute, see
'Details.'}

\item{attr_data}{Deprecated argument. Please use 'attr_actors' instead.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for a "same" effect in the tie-oriented model or the
receiver choice step of the actor-oriented model. A "same" effect refers to
an exogenous actor attribute that affects dyad \emph{(i,j)}'s rate of
interacting (tie-oriented model) or actor \emph{j}'s probability of being
chosen as a receiver for the event send by the active sender \emph{i} at
time \emph{t} (actor-oriented model) based on whether actors \emph{i} and
\emph{j} have the same value (or not) on this attribute.
}
\details{
The statistic at timepoint \emph{t} is equal to one for dyads \emph{(i,j)}
that have the same value on the attribute at timepoint \emph{t}
(tie-oriented model) or one for receivers \emph{j} that have the same value
on the attribute as the active sender \emph{i} at timepoint \emph{t}
(actor-oriented model) and equal to 0 for dyads and receivers that do not
have the same value.

Construct the `attr_actors` object as a data frame where each row represents 
the attribute value of actor \emph{i} at timepoint \emph{t}:
\itemize{
  \item{name:} The actors' name.
  \item{time:} The time when the attribute values change.
  \item{variable:} The third column contains the attribute used in the
  specification of the "difference" effect. The column name should correspond
  to the string supplied to the \code{variable} argument in the `difference()`
  function.
}

Note that it is possible to omit the `attr_actors` object in the call of
\code{difference()} and, instead, supply it in the call of \code{remstats()}
for multiple exogenous effects.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ same("age")
remstats(reh = reh_tie, tie_effects = effects, attr_actors = info)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, receiver_effects = effects, attr_actors = info)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{boxplot.tomstats}
\alias{boxplot.tomstats}
\title{Plotting Relational Event Network Statistics Distributions}
\usage{
\method{boxplot}{tomstats}(x, effect, by = "timepoints", subset = NULL, outliers = TRUE, ...)
}
\arguments{
\item{x}{An object of class \code{\link{tomstats}} containing relational
event network statistics.}

\item{effect}{A character string specifying the name of the effect in 'x' or
an integer indicating the index of the effect to be plotted.}

\item{by}{A string indicating whether the statistic is plotted across
'timepoints' (default) or 'dyads'.}

\item{subset}{An optional vector specifying a subset of timepoints or dyads
to be used for plotting. Per default, a maximum of 20 unique timepoints or 
dyads are plotted.}

\item{outliers}{A logical value specifying whether to include outliers in the
plot.}

\item{...}{Additional arguments passed to bxp().}
}
\value{
no return value
}
\description{
Generate boxplots for a specified effect in a \code{\link{tomstats}}
object.
}
\details{
This function produces boxplots to visually represent the distribution of a
specified effect in a relational event network, as captured by a
\code{\link{tomstats}} object. The 'effect' parameter allows the user to
choose a specific effect for visualization, either by providing the effect's
name or its index within the 'tomstats' object. The 'by' parameter determines
whether the boxplots are created across different 'timepoints' or 'dyads'.
Additionally, an optional 'subset' parameter allows the user to focus on
specific timepoints or dyads. If 'subset' is not specified, a default
maximum of 20 unique timepoints or dyads are plotted. The 'outliers'
argument, when set to TRUE, includes the representation of outliers in the
boxplots. If set to FALSE, outliers are omitted from the visualization.

The boxplots are based on the following summary statistics of the data: The
box in the middle represents the interquartile range (IQR) between the first
(Q1) and third quartile (Q3), and the line inside the box represents the
median. The whiskers extend from the box to the minimum and maximum values
within 1.5 times the IQR below Q1 or above Q3. Outliers beyond the whiskers
are plotted individually.
}
\examples{
library(remstats)
# Load data
data(history)
# Prepare data
reh <- remify::remify(edgelist = history[,1:3], model = "tie")
# Compute effects
stats <- remstats(reh, tie_effects = ~ inertia())
# Plot the 'inertia' distribution for 20 timepoints
boxplot(stats, effect = "inertia")
# Plot the 'inertia' distribution for 20 dyads
boxplot(stats, effect = "inertia", by = "dyads")
# Plot the 'inertia' distribution for dyads 2:5
boxplot(stats, effect = "inertia", by = "dyads", subset = 2:5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{spUnique}
\alias{spUnique}
\title{spUnique}
\usage{
spUnique()
}
\value{
Warning.
}
\description{
Deprecated. Use \code{\link{sp}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{boxplot.aomstats}
\alias{boxplot.aomstats}
\title{Plotting Relational Event Network Statistics}
\usage{
\method{boxplot}{aomstats}(
  x,
  effect,
  model,
  by = "timepoints",
  subset = NULL,
  outliers = TRUE,
  ...
)
}
\arguments{
\item{x}{An object of class \code{\link{aomstats}} containing relational
event network statistics.}

\item{effect}{A string specifying the name of the effect in 'x' or an integer
indicating the index of the effect to be plotted.}

\item{model}{A string indicating whether the effect is in the 'sender' model
or the 'receiver' model.}

\item{by}{A string indicating whether the statistic is plotted across
'timepoints' (default) or 'actors'.}

\item{subset}{An optional vector specifying a subset of timepoints or actors
to be used for plotting. Per default, a maximum of 20 unique timepoints or 
actors are plotted.}

\item{outliers}{A logical value specifying whether to include outliers in the
plot.}

\item{...}{additional arguments passed to bxp().}
}
\description{
Generate boxplots for a specified effect in a \code{\link{aomstats}} object.
}
\details{
This function produces boxplots to visually represent the distribution of a
specified effect in a relational event network, as captured by a
\code{\link{aomstats}} object. The 'effect' parameter allows the user to
choose a specific effect for visualization, either by providing the effect's
name or its index within the 'aomstats' object. The 'model' parameter
indicates whether the respective effect is in the 'sender' model or the
'receiver' model. The 'by' parameter determines whether the boxplots are
created across different 'timepoints' or 'actors'. At the moment, by 'actors'
is only supported for the sender model. Additionally, an optional 'subset'
parameter allows the user to focus on specific timepoints or actors. If
'subset' is not specified, a default maximum of 20 unique timepoints or
actors are plotted. The 'outliers' argument, when set to TRUE, includes the
representation of outliers in the boxplots. If set to FALSE, outliers are
omitted from the visualization.

The boxplots are based on the following summary statistics of the data: The
box in the middle represents the interquartile range (IQR) between the first
(Q1) and third quartile (Q3), and the line inside the box represents the
median. The whiskers extend from the box to the minimum and maximum values
within 1.5 times the IQR below Q1 or above Q3. Outliers beyond the whiskers
are plotted individually.
}
\examples{
library(remstats)
# Load data
data(history)
# Prepare data
reh <- remify::remify(edgelist = history[,1:3], model = "actor")
# Compute effects
stats <- remstats(reh, sender_effects = ~ outdegreeSender())
# Plot the 'outdegreeSender' distribution for 20 timepoints
boxplot(stats, effect = "outdegreeSender", model = "sender")
# Plot the 'inertia' distribution for all 10 actors
boxplot(stats, effect = "outdegreeSender", model = "sender", by = "actors")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{psABA}
\alias{psABA}
\title{psABA}
\usage{
psABA()
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a participation shift AB-A in the sender step of
the actor-oriented model.
}
\details{
Refers to the tendency for the same actor to keep initiating events: The
next sender is equal to the previous sender. For each timepoint t, the psABA
statistic is equal to one for the actor that will create the participation
shift if they would occur in the edgelist as the sender at time t and equal
to zero for the actors that will not create this participation shift. If
multiple events in the edgelist occur at the same time point, the order of
these events determines whether the p-shift is observed.
}
\examples{
reh_actor <- remify::remify(history, model = "actor")
remstats(sender_effects = ~ psABA(), reh = reh_actor)

}
\seealso{
\code{\link{psABB}} or \code{\link{psABX}} for exploring
alternative participation shifts in the sender step of the actor-oriented
model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{psABAB}
\alias{psABAB}
\title{psABAB}
\usage{
psABAB(consider_type = "ignore")
}
\arguments{
\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a pshift AB-AB effect.
}
\details{
Refers to the tendency for the same dyads to keep interacting. For directed
events, the next sender and receiver are equal to the previous sender and
receiver. For undirected events, the next actor pair is equal to the current
actor pair. For each timepoint t, the psABAB statistic is equal to one for
the dyads that will create the participation shift if they would occur in
the edgelist at time t and equal to zero for the dyads that will not create
this participation shift. If consider_type is set to TRUE, the type of the
two subsequent AB events have to be equal. If it is set to FALSE, the
participation shift is set to one for every AB event, regardless of the
event type. If multiple events in the edgelist occur at the same time point,
the order of these events determines whether the p-shift is observed.
}
\examples{
reh_tie <- remify::remify(history, model = "tie", directed = FALSE)
effects <- ~ psABAB()
remstats(tie_effects = effects, reh = reh_tie)

}
\seealso{
\code{\link{psABBA}}, \code{\link{psABBY}}, \code{\link{psABXA}},
\code{\link{psABXB}}, \code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation shifts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{recencyReceiveReceiver}
\alias{recencyReceiveReceiver}
\title{recencyReceiveReceiver}
\usage{
recencyReceiveReceiver(consider_type = "ignore")
}
\arguments{
\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a recency receive of receiver effect in the
\code{effects} argument of \code{\link{tomstats}} or the
\code{receiver_effects} argument of \code{\link{aomstats}}.
}
\details{
The recencyReceiveReceiver effect refers to a recency statistic similar to
what is described in Vu et al. (2017) and Mulder and Leenders (2019). For
each timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the
time that has past since receiver j was last active as receiver + 1). Note
that the 'recencyReceiveReceiver' effect is only defined for directed
events.
}
\examples{
effects <- ~ recencyReceiveReceiver()
reh_tie <- remify::remify(history, model = "tie")
remstats(tie_effects = effects, reh = reh_tie)

reh_actor <- remify::remify(history, model = "actor")
remstats(receiver_effects = effects, reh = reh_actor)

}
\seealso{
\code{\link{rrankSend}}, \code{\link{rrankReceive}},
\code{\link{recencySendSender}}, \code{\link{recencyReceiveSender}},
\code{\link{recencyReceiveSender}} and \code{\link{recencyContinue}} for
other type of recency effects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind_remstats.R
\name{bind_remstats}
\alias{bind_remstats}
\title{Combine two or more remstats objects}
\usage{
bind_remstats(...)
}
\arguments{
\item{...}{Any number of \code{\link{remstats}} objects. All the
\code{remstats} objects must have matching dimensions, except for the third
dimension. Note that duplicated statistics in the combined object are 
removed based on their name.}
}
\value{
\code{statistics } array with the combined statistics, where rows
refer to time points, columns refer to potential relational event (i.e.,
potential edges) in the risk set and slices refer to statistics
}
\description{
Function to bind any number of remstats objects into one while duplicated
statistics in the combined object are removed based on their name.
}
\examples{
library(remstats)

# Load the data
data(history)
data(info)

# Prepare the data
reh <- remify::remify(edgelist = history, model = "actor")

# Obtain two different statistics objects
effects1 <- ~ inertia():receive("extraversion") + otp()
stats1 <- remstats(receiver_effects = effects1, reh = reh, attr_actors = info)
effects2 <- ~ reciprocity()
stats2 <- remstats(receiver_effects = effects2, reh = reh, attr_actors = info)

# Bind the two statistics objects
statsC <- bind_remstats(stats1, stats2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{itp}
\alias{itp}
\title{itp}
\usage{
itp(unique = FALSE, scaling = c("none", "std"), consider_type = "ignore")
}
\arguments{
\item{unique}{A logical value indicating whether to sum the minimum of
events with third actors (FALSE, default) or the number of third actors that
create a new, unique two-path (TRUE). See details for more information.}

\item{scaling}{The method for scaling the triad statistic. The default value
is "none", which means the statistic is not scaled. Alternatively, you can
set it to "std" to request standardization of the raw counts per time point.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for an incoming two-path effect.
}
\details{
The incoming two-path effect describes the propensity of dyads to interact
based on the number of past incoming two-paths between them. By default, the
statistic at timepoint t for the dyad (i,j) is computed as the sum of the
minimum occurrences of past (j,h) and (h,i) events across all actors h.

When the unique parameter is set to TRUE, a different approach is taken.
In this case, the statistic counts the number of actors h that contribute to
the creation of a new, distinct two-path between actors i and j.

Additionally, it is possible to specify a scaling method using the scaling
parameter.

Please note that the incoming two-path effect, 'itp', is exclusively defined
for directed events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ itp()
remstats(reh = reh_tie, tie_effects = effects)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, receiver_effects = effects)

}
\references{
Butts, C. (2008). A relational event framework for social
action. Sociological Methodology.
}
\seealso{
\code{\link{otp}}, \code{\link{osp}}, or \code{\link{isp}} for
other types of triadic effects for directed relational events and
\code{\link{sp}} for triadic effects for undirected relational events.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{event}
\alias{event}
\title{event}
\usage{
event(x, variableName = NULL)
}
\arguments{
\item{x}{vector with the event attribute}

\item{variableName}{optionally, a string indicating the variable name, used
for the dimnames of the output statistics object}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for an "event" effect in the tie-oriented model. An
"event" effect refers to an exogenous event attribute that affects the
waiting time between events.
}
\details{
The statistic at timepoint \emph{t} is for all dyads in the risk set equal
to the attribute of the event at timepoint \emph{t}.
}
\examples{
\donttest{
  reh_tie <- remify::remify(history, model = "tie")
  data(history, package = "remstats")
  history$work <- ifelse(history$setting == "work", 1, 0)
  effects <- ~ event(x = history$work, variableName = "setting_is_work")
  remstats(reh = reh_tie, tie_effects = effects)
}

}
\seealso{
\code{\link{FEtype}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.remstats}
\alias{print.remstats}
\title{Printing Relational Event Network Statistics}
\usage{
\method{print}{remstats}(x, ...)
}
\arguments{
\item{x}{object of class \code{\link{remstats}}.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
The function prints formatted information about the remstats object to the console, presenting details about the relational event network statistics in a user-friendly format.
}
\description{
Print a \code{\link{remstats}} object in a user-friendly format.
}
\examples{
rehObject <- remify::remify(edgelist = history, model = "tie")
remstatsObject <- remstats::remstats(reh = rehObject, tie_effects = ~ remstats::inertia())
print(remstatsObject)

rehObject <- remify::remify(edgelist = history, model = "actor")
remstatsObject <- remstats::remstats(reh = rehObject, receiver_effects = ~ inertia())
print(remstatsObject)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{totaldegreeDyad}
\alias{totaldegreeDyad}
\title{totaldegreeDyad}
\usage{
totaldegreeDyad(scaling = c("none", "prop", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the degree statistic. Default is
to not scale the statistic (scaling = "none"). Alternatively,
scaling of the raw degree counts by two times the number of past events at
time t can be requested with 'prop' or standardization of the raw degree
counts per time point can be requested with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for a 'totaldegreeDyad' effect.
}
\details{
The 'totaldegreeDyad' effect refers to the tendency of pairs of actors
(dyads) to increase their interaction rate as the total degree (number of
interactions) of both actors in the pair goes up. To calculate this effect
for a specific pair (i,j) at a given timepoint (t), we sum the degrees of
the two actors in the dyad (i,j).

Additionally, there is an optional scaling method, which can be chosen using
the 'scaling' method. When the 'prop' scaling method is applied, the degree
count is divided by two times the total number of past events. This scaling
converts the statistic into a fraction, representing the proportion of past
events in which at least one actor in the dyad was involved. For the first
timepoint, where no events have previously occurred, it is assumed that each
actor is equally likely to be involved in an event. In this case, the
statistic is set to 1 divided by the total number of actors (N).

The totaldegreeDyad effect is defined for the tie-oriented model and is
applicable to both directed and undirected events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ totaldegreeDyad()
remstats(reh = reh_tie, tie_effects = effects)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{totaldegreeSender}
\alias{totaldegreeSender}
\alias{totaldegree}
\title{totaldegreeSender}
\usage{
totaldegreeSender(scaling = c("none", "prop", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the degree statistic. Default is
to not scale the statistic (scaling = "none"). Alternatively,
scaling of the raw degree counts by two times the number of past events at
time t can be requested with 'prop' or standardization of the raw degree
counts per time point can be requested with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for an `totaldegreeSender` effect in the
tie-oriented model or the sender activity rate step of the actor-oriented
model.
}
\details{
A total degree of the sender effect refers to the tendency for actors to
send events if they have send and received more past events. The statistic
at timepoint \emph{t} for dyad \emph{(i,j)} (tie-oriented model) or sender
\emph{i} (actor-oriented model) is equal to the number of events
send and received by actor \emph{i} before timepoint \emph{t}. Note that the
'totaldegreeSender' effect is only defined for directed events.

Optionally, a scaling method can be set with \code{scaling}. By scaling the
degree count by the total number of past events times two, the statistic
refers to the fraction of past events times two that involved actor i. At
the first time point, when no events did previously occur, it is assumed
that every actor is equally likely to send a message and the statistic is
set equal to 1/n, where n refers to the number of actors.
}
\examples{
effects <- ~ totaldegreeSender()
reh_tie <- remify::remify(history, model = "tie")
remstats(reh = reh_tie, tie_effects = effects)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, sender_effects = effects)

}
\seealso{
\code{\link{indegreeSender}}, \code{\link{indegreeReceiver}},
\code{\link{outdegreeSender}}, \code{\link{outdegreeReceiver}}, or
\code{\link{totaldegreeReceiver}} for other types of degree effects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{recencyContinue}
\alias{recencyContinue}
\alias{recency}
\title{recencyContinue}
\usage{
recencyContinue(consider_type = "ignore")
}
\arguments{
\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a recency continue effect in the
\code{effects} argument of \code{\link{tomstats}} or the
\code{receiver_effects} argument of \code{\link{aomstats}}.
}
\details{
The recencyContinue effect refers to a recency statistic similar to what is
described in Vu et al. (2017) and Mulder and Leenders (2019). For each
timepoint t, for directed dyad (i,j) the statistic is equal to 1/(the time
that has past since the dyad was last active + 1).
}
\examples{
effects <- ~ recencyContinue()
reh_tie <- remify::remify(history, model = "tie")
remstats(tie_effects = effects, reh = reh_tie)

reh_actor <- remify::remify(history, model = "actor")
remstats(receiver_effects = effects, reh = reh_actor)

}
\seealso{
\code{\link{rrankSend}}, \code{\link{rrankReceive}},
\code{\link{recencySendSender}}, \code{\link{recencyReceiveSender}},
\code{\link{recencyReceiveSender}} and \code{\link{recencyReceiveReceiver}}
for other type of recency effects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack_stats.R
\name{stack_stats}
\alias{stack_stats}
\title{Stack remstats for model fitting}
\usage{
stack_stats(stats, reh)
}
\arguments{
\item{stats}{A \code{tomstats} object (output of \code{remstats()} or
\code{tomstats()}).}

\item{reh}{A \code{remify} object (output of \code{remify::remify2()}).}
}
\value{
A list with elements:
\describe{
  \item{remstats_stack}{Data frame in long format with columns: \code{event}
    (event index), \code{dyad} (dyad index 1..D), all statistic columns,
    \code{obs} (1 = observed event, 0 = non-event), and
    \code{log_interevent} (log inter-event time; only for interval timing).}
  \item{subset}{Integer vector of length 2: first and last event index.}
  \item{D}{Number of dyads in the risk set.}
  \item{E}{Number of events (time points).}
  \item{ordinal}{Logical: whether the ordinal likelihood applies.}
}
}
\description{
Stack a \code{tomstats} object into long format suitable for fitting with
standard R model functions. Each row corresponds to one dyad in the risk
set at one event time point.
}
\details{
For interval timing (\code{ordinal = FALSE}), the stacked data can be
fitted with a Poisson GLM using \code{log_interevent} as an offset:
\code{glm(obs ~ -1 + . - event - dyad - log_interevent,
          offset = log_interevent, family = poisson, data = stacked$remstats_stack)}

For ordinal timing (\code{ordinal = TRUE}), use conditional logistic
regression stratified by event:
\code{survival::clogit(obs ~ -1 + . - event - dyad +
                       strata(event), data = stacked$remstats_stack)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{both_male_long}
\alias{both_male_long}
\title{Exogenous Dyad Attribute in Long Format: both_male_long}
\format{
A data frame with the following columns:
\describe{
 \item{actor1}{Numeric id of the first actor in the dyad.}
 \item{actor2}{Numeric id of the second actor in the dyad.}
 \item{both_male}{Binary indicator (1 for male-male dyads, 0 otherwise).}
}
}
\source{
Simulated exogenous information on actors in a social network.
}
\usage{
data(both_male_long)
}
\description{
A data frame representing exogenous attributes of dyads in a social network in long format. Each row indicates whether a dyad consists of two male actors (sex=0) in the original matrix `info_both_male_wide`.
}
\examples{
data(both_male_long)
head(both_male_long)

}
\seealso{
\code{\link{tie}} for the function using this data, \code{\link{both_male_wide}} for the data in wide format, and \code{\link{info}} for an overview of the actor exogenous attributes.
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{psABB}
\alias{psABB}
\title{psABB}
\usage{
psABB()
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for a participation shift AB-B in the sender step of the actor-oriented model.
}
\details{
The AB-B participation shift refers to the tendency for immediate reciprocation (the next sender is the previous receiver). For each timepoint t, the psABBA statistic is equal to one for the actor (i.e, the previous event receiver) that will create the participation shift if it would occur as sender in the edgelist at time t and equal to zero for the actors that will not create this participation shift. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed.
}
\examples{
reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, sender_effects = ~ psABB())

}
\seealso{
\code{\link{psABA}} or \code{\link{psABX}} for exploring alternative participation shifts in the sender step of the actor-oriented model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.tomstats}
\alias{plot.tomstats}
\title{Plotting Relational Event Network Statistics Trajectories}
\usage{
\method{plot}{tomstats}(x, effect, subset = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{\link{tomstats}} containing relational
event network statistics.}

\item{effect}{A character string specifying the name of the effect in 'x' or
an integer indicating the index of the effect to be plotted.}

\item{subset}{An optional vector specifying a subset of dyads to be used for
plotting. By default, a maximum of 5 unique dyads are used for plotting.}

\item{...}{Additional arguments passed to plot().}
}
\description{
Generate line plots to visualize the trajectories of a specified effect in a
\code{\link{tomstats}} object.
}
\details{
This function creates line plots to illustrate the temporal trajectories of a
specified effect in a relational event network, as captured by a
\code{\link{tomstats}} object. The 'effect' parameter allows users to choose a
specific effect for visualization, either by providing the effect's name or
its index within the 'tomstats' object. An optional 'subset' parameter enables
users to focus on specific dyads. If 'subset' is not specified, a default
maximum of 5 unique dyads is plotted. These dyads are randomly selected to
represent trajectories across the range of different endpoints for the effect
(excluding zero).
}
\examples{
library(remstats)
# Load data
data(history)
# Prepare data
reh <- remify::remify(edgelist = history[,1:3], model = "tie")
# Compute effects
stats <- remstats(reh, tie_effects = ~ inertia())
# Plot the 'inertia' trajectories for 5 dyads
plot(stats, effect = "inertia")
# Plot the 'inertia' trajectory for a specific dyad
plot(stats, effect = "inertia", subset = 60)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{rrankSend}
\alias{rrankSend}
\alias{recencyRank}
\alias{rrank}
\title{rrankSend}
\usage{
rrankSend(consider_type = "ignore")
}
\arguments{
\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for a recency rank send effect in the
\code{effects} argument of \code{\link{tomstats}} or the
\code{receiver_effects} argument of \code{\link{aomstats}}.
}
\details{
The rrankSend effect refers to a rank-based recency effect, as described in
section 2.2.5 of Butts (2008). For each timepoint t, for directed dyad (i,j)
the statistic is equal to the inverse of the rank of receiver j among the
actors to which sender i has most recently send past events. Note that the
'rrankSend' effect is only defined for directed events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ rrankSend()
remstats(tie_effects = effects, reh = reh_tie)

reh_actor <- remify::remify(history, model = "actor")
remstats(receiver_effects = effects, reh = reh_actor)

}
\seealso{
\code{\link{rrankReceive}}, \code{\link{recencySendSender}},
\code{\link{recencySendReceiver}}, \code{\link{recencyReceiveSender}},
\code{\link{recencyReceiveReceiver}} and \code{\link{recencyContinue}} for
other type of recency effects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aomstats.R
\name{aomstats}
\alias{aomstats}
\title{aomstats}
\usage{
aomstats(
  reh,
  sender_effects = NULL,
  receiver_effects = NULL,
  attr_actors = NULL,
  attr_dyads = NULL,
  memory = c("full", "window", "decay", "interval"),
  memory_value = NA,
  start = 2,
  stop = Inf,
  display_progress = FALSE
)
}
\arguments{
\item{reh}{an object of class \code{"\link[remify]{remify}"} characterizing 
the relational event history.}

\item{sender_effects}{an object of class \code{"\link[stats]{formula}"} (or
one that can be coerced to that class): a symbolic description of the
effects in the sender activity rate step of the actor-oriented model for
which statistics are computed, see `Details'}

\item{receiver_effects}{an object of class \code{"\link[stats]{formula}"}
(or one that can be coerced to that class): a symbolic description of the
effects in the receiver choice step of model for which statistics are
computed, see `Details'}

\item{attr_actors}{optionally, an object of class
\code{"\link[base]{data.frame}"} that contains exogenous attributes for 
actors (see Details).}

\item{attr_dyads}{optionally, an object of class \code{data.frame} or 
\code{matrix} containing attribute information for dyads (see Details).}

\item{memory}{The memory to be used. See `Details'.}

\item{memory_value}{Numeric value indicating the memory parameter. Default 
is \code{NA}, which is only valid for \code{memory = "full"} (no memory
parameter required). See `Details'.}

\item{start}{an optional integer value, specifying the index of the first
time or event in the relational event history for which statistics must be 
computed (see 'Details'). Default is \code{2}: the first event has no history and is used only to initialize statistics, not to fit the model.}

\item{stop}{an optional integer value, specifying the index of the last
time or event in the relational event history for which statistics must be 
computed (see 'Details')}

\item{display_progress}{should a progress bar for the computation of the
endogenous statistics be shown (TRUE) or not (FALSE)?}
}
\value{
An object of class 'aomstats'. List with in the first element the
statistics for the sender activity rate step and in the second element the
statistics for the receiver choice step. Statistics are computed once per 
unique time point (per-timepoint "pt" method), so the number of rows in the 
output equals \code{reh$M} (the number of  unique time points), which may be 
less than the total number of observed events when simultaneous events are 
present. The 'aomstats' object has the following attributes: 
  \describe{
    \item{\code{model}}{Type of model that is estimated.}
    \item{\code{formula}}{Model formula(s), obtained from the formula(s) 
    inputted to 'sender_effects' and/or 'receiver_effects'.}
    \item{\code{actors}}{The set of actors used to construct the statistics, 
    obtained from the remify object inputted to 'reh'.}
  }
}
\description{
Computes statistics for the sender activity rate step and receiver choice
step in actor-oriented relational event models (e.g., see Stadtfeld & Block,
2017).
}
\section{Effects}{

The statistics to be computed are defined symbolically and should be
supplied to the \code{sender_effects} and/or \code{receiver_effects}
arguments in the form \code{~ effects}. The terms are separated by +
operators. For example: \code{receiver_effects = ~ inertia() + otp()}.
Interactions between two effects can be included with * or :
operators. For example: \code{receivereffects = ~ inertia():otp()}.  A list
of available effects can be obtained with \code{\link{actor_effects}()}.

The majority of the statistics can be scaled in some way, see
the documentation of the \code{scaling} argument in the separate effect
functions for more information on this.
}

\section{attr_actors}{

For the computation of the \emph{exogenous} statistics an attributes object
with the exogenous covariate information has to be supplied to the
\code{attr_actors} argument in either \code{remstats()} or in the separate
effect functions supplied to the \code{..._effects} arguments (e.g., see
\code{\link{send}}). This \code{attr_actors} object should be constructed as
follows: A dataframe with rows referring to the attribute value of actor
\emph{i} at timepoint \emph{t}. A `name` column is required that contains the
actor name (corresponding to the actor names in the relational event
history). A `time` column is required that contains the time when attributes
change (set to zero if none of the attributes vary over time). Subsequent
columns contain the attributes that are called in the specifications of
exogenous statistics (column name corresponding to the string supplied to
the \code{variable} argument in the effect function). Note that the
procedure for the exogenous effects `tie' and `event' deviates from this,
here the exogenous covariate information has to be specified in a different
way, see \code{\link{tie}} and \code{\link{event}}.
}

\section{attr_dyads}{
  
For the computation of the \emph{dyad exogenous} statistics with \code{tie()}, an attributes object with the exogenous covariates information per dyad has to be supplied. This is a \code{data.frame} or \code{matrix} containing attribute information for dyads. If \code{attr_dyads} is a \code{data.frame}, the first two columns should represent "actor1" and "actor2" (for directed events, "actor1" corresponds to the sender, and "actor2" corresponds to the receiver). Additional columns can represent dyads' exogenous attributes. If attributes vary over time, include a column named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond to "actor1", columns to "actor2", and cells contain dyads' exogenous attributes.
}

\section{Memory}{

The default `memory` setting is `"full"`, which implies that at each time
point $t$ the entire event history before $t$ is included in the computation
of the statistics. Alternatively, when `memory` is set to `"window"`, only
the past event history within a given time window is considered (see
Mulders & Leenders, 2019). This length of this time window is set by the
`memory_value` parameter. For example, when `memory_value = 100` and `memory
= "window"`, at time point $t$ only the past events that happened at most
100 time units ago are included in the computation of the statistics.
A third option is to set `memory` to `"interval"`. In this case, the past
event history within a given time interval is considered. For example, when
`"memory_value" = c(50, 100)` and `memory = "window"`, at time point $t$
only the past events that happened between 50 and 100 time units ago are
included in the computation of the statistics. Finally, the fourth option is
to set `memory` to `"decay"`. In this case, the weight of the past event in
the computation of the statistics depend on the elapsed time between $t$ and
the past event. This weight is determined based on an exponential decay
function with half-life parameter `memory_value` (see Brandes et al., 2009).
}

\section{Event weights}{

Note that if the relational event history contains a column that is named
``weight'', it is assumed that these affect the endogenous statistics. These
affect the computation of all endogenous statistics with a few exceptions
that follow logically from their definition (e.g., the recenyContinue
statistic does depend on time since the event and not on event weights).
}

\section{Subset of the relational event history}{

Optionally, statistics can be computed for a slice of the relational event
sequence - but based on the entire history. This is achieved by setting the
start and stop values equal to the index of the first and last event for
which statistics are requested. For example, start = 5 and stop = 5 computes
the statistics for only the 5th event in the relational event sequence,
based on the history that consists of events 1-4.
}

\examples{
library(remstats)

# Load the data
data(history)
data(info)

# Prepare the data
reh <- remify::remify(edgelist = history, model = "actor")

# Define the sender effects
seff <- ~ send("extraversion")

# Define the receiver_effects
reff <- ~ receive("agreeableness") + inertia() + otp()

# Compute the statistics
aomstats(
  reh = reh, sender_effects = seff, receiver_effects = reff,
  attr_actors = info
)

}
\references{
Stadtfeld, C., & Block, P. (2017). Interactions, actors, and
time: Dynamic network actor models for relational events. Sociological
Science, 4, 318–352. \doi{10.15195/v4.a14}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{inertia}
\alias{inertia}
\title{inertia}
\usage{
inertia(scaling = c("none", "prop", "std"), consider_type = "ignore")
}
\arguments{
\item{scaling}{the method for scaling the inertia statistic. Default is to
not scale the statistic (scaling = "none"). Alternatively, the
statistics can be scaled by specifying 'prop', in which raw counts are
divided by the outdegree of the sender at time t (see 'details') or
standardization of the raw counts per time point can be requested with 'std'.}

\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to 
compute the statistic.
}
\description{
Specifies the statistic for an inertia effect in the tie-oriented model or
the receiver choice step of the actor-oriented model.
}
\details{
An inertia effect refers to the tendency for dyads to repeatedly interact
with each other (tie-oriented model) or for actors to repeatedly choose the
same actor as receiver of their events (actor-oriented model). The statistic
at timepoint \emph{t} for dyad \emph{(i,j)} resp. receiver \emph{j} is
equal to the number of \emph{(i,j)} events before timepoint \emph{t}.

Optionally, a scaling method can be set with \code{scaling}. By scaling the
inertia count by the outdegree of the sender ("prop"), the statistic refers
to the fraction of messages send by actor i that were send to actor j. If
actor i hasn't send any messages yet it can be assumed that every actor is
equally likely to receive a message from i and the statistic is set equal to
1/(n-1), where n refers to the number of actors. The resulting statistic is
similar to the "FrPSndSnd" statistic in the R package 'relevent', or the
persistence statistic in Section 2.2.2 of Butts (2008). Note that this
scaling method is only defined for directed events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ inertia()
remstats(reh = reh_tie, tie_effects = effects)

reh_actor <- remify::remify(history, model = "actor")
remstats(reh = reh_actor, receiver_effects = effects)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{psABBY}
\alias{psABBY}
\title{psABBY}
\usage{
psABBY(consider_type = "ignore")
}
\arguments{
\item{consider_type}{character. Controls how event types are handled:
\code{"ignore"} (default): aggregate over all event types (one statistic);
\code{"separate"}: compute C type-specific statistics, where the type-c
statistic for a dyad reflects past type-c events on that actor pair
;
\code{"interact"}: compute C^2 statistics capturing past-event-type x
dyad-type interactions (only meaningful with \code{extend_riskset_by_type=TRUE} in remify object).
Also accepts \code{FALSE} (-> \code{"ignore"}) and \code{TRUE}
(-> \code{"separate"}) for backward compatibility.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for a participation shift AB-BY.
}
\details{
The AB-BY participation shift refers to one of Gibson's (2003) dyadic participation shifts. The AB-BY pshift refers to a tendency for turn receiving (here, the next sender is the previous receiver and the next receiver is not in the current previous). For each timepoint t, the psABBY statistic is equal to one for the dyads that will create the participation shift if they would occur in the edgelist at time t and equal to zero for the dyads that will not create this participation shift. If consider_type is set to TRUE, the type of the AB event and the type of the BY events have to be equal. If it is set to FALSE, the participation shift is set to one for every BY event, regardless of the event type. If multiple events in the edgelist occur at the same time point, the order of these events determines whether the p-shift is observed. Note that the AB-BY pshift is only defined for directed events.
}
\examples{
reh_tie <- remify::remify(history, model = "tie")
effects <- ~ psABBY()
remstats(reh = reh_tie, tie_effects = effects)

}
\seealso{
\code{\link{psABBA}}, \code{\link{psABXA}}, \code{\link{psABXB}},
\code{\link{psABXY}} or \code{\link{psABAY}} for other dyadic participation
shifts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effects.R
\name{tie}
\alias{tie}
\alias{dyad}
\title{tie}
\usage{
tie(variable, attr_dyads = NULL, scaling = c("none", "std"), x, variableName)
}
\arguments{
\item{variable}{A string specifying the attribute to compute the statistic. If \code{attr_dyads} is a \code{data.frame}, this refers to the column name in \code{attr_actors}. If \code{attr_dyads} is a \code{matrix}, this corresponds to the name of the exogenous attribute, used to label the statistic in the resulting \code{remstats} object.}

\item{attr_dyads}{A \code{data.frame} or \code{matrix} containing attribute information for dyads. If \code{attr_dyads} is a \code{data.frame}, the first two columns should represent "actor1" and "actor2" (for directed events, "actor1" corresponds to the sender, and "actor2" corresponds to the receiver). Additional columns can represent dyads' exogenous attributes. If attributes vary over time, include a column named "time". If \code{attr_dyads} is a \code{matrix}, the rows correspond to "actor1", columns to "actor2", and cells contain dyads' exogenous attributes.}

\item{scaling}{The method for scaling the statistic. The default is no scaling. Alternatively, standardization of the statistic per time point can be requested with "std".}

\item{x}{Deprecated argument. Please use 'attr_dyads' instead.}

\item{variableName}{Deprecated argument. Please use 'variable' instead.}
}
\value{
List with all information required by `remstats::remstats()` to compute the statistic.
}
\description{
Specifies the statistic for a "tie" (or, "dyad") effect.
}
\details{
The "tie" effect or "dyad" effect refers to an exogenous dyad attribute that influences dyad \emph{(i,j)}'s interaction rate (in tie-oriented models) or the probability of actor \emph{j} being chosen as a receiver for the event sent by the active sender \emph{i} (in actor-oriented models). The statistic represents the value of the exogenous attribute for dyad \emph{(i,j)} in the \code{attr_dyads} data.
}
\examples{
data(history)
data(both_male_long)
effect <- ~ tie(variable = "both_male", attr_dyads = both_male_long)
reh <- remify::remify(history, model = "tie")
remstats(reh = reh, tie_effects = effect)

data(both_male_wide)
effect <- ~ tie(variable = "both_male", attr_dyads = both_male_wide)
reh <- remify::remify(history, model = "tie")
remstats(reh = reh, tie_effects = effect)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tie_effects.R
\name{tie_effects}
\alias{tie_effects}
\title{tie_effects}
\usage{
tie_effects(directed = NULL, endogenous = NULL)
}
\arguments{
\item{directed}{logical value. The function outputs all statistics in the 
tie-oriented model for directed events if true, or all statistics in the 
tie-oriented model for undirected events if false.}

\item{endogenous}{logical value. The function outputs all endogenous 
statistics in the tie-oriented model if true, or all exogenous statistics if 
false}
}
\value{
Returns a list of available effects and their corresponding statistics.
}
\description{
Overview of statistics in the tie-oriented model, see Details.
}
\details{
Overview of statistics in the tie-oriented model.

Baseline:
\itemize{
 \item \code{\link{baseline}}
}

Exogenous statistics:
\itemize{
 \item \code{\link{send}()}
 \item \code{\link{receive}()}
 \item \code{\link{tie}()}
 \item \code{\link{same}()}
 \item \code{\link{difference}()}
 \item \code{\link{average}()}
 \item \code{\link{minimum}()}
 \item \code{\link{maximum}()}
 \item \code{\link{event}()}
 \item \code{\link{userStat}()}
}

Endogenous statistics:
\itemize{
 \item \code{\link{indegreeSender}()}
 \item \code{\link{indegreeReceiver}()}
 \item \code{\link{outdegreeSender}()}
 \item \code{\link{outdegreeReceiver}()}
 \item \code{\link{totaldegreeSender}()}
 \item \code{\link{totaldegreeReceiver}()}
 \item \code{\link{totaldegreeDyad}()}
 \item \code{\link{degreeMin}()}
 \item \code{\link{degreeMax}()}
 \item \code{\link{degreeDiff}()}
 \item \code{\link{inertia}()}
 \item \code{\link{reciprocity}()}
 \item \code{\link{otp}()}
 \item \code{\link{itp}()}
 \item \code{\link{osp}()}
 \item \code{\link{isp}()}
 \item \code{\link{sp}()}
 \item \code{\link{psABBA}()}
 \item \code{\link{psABBY}()}
 \item \code{\link{psABXA}()}
 \item \code{\link{psABXB}()}
 \item \code{\link{psABXY}()}
 \item \code{\link{psABAY}()}
 \item \code{\link{psABAB}()}
 \item \code{\link{rrankSend}()}
 \item \code{\link{rrankReceive}()}
 \item \code{\link{recencySendSender}()}
 \item \code{\link{recencySendReceiver}()}
 \item \code{\link{recencyReceiveSender}()}
 \item \code{\link{recencyReceiveReceiver}()}
 \item \code{\link{recencyContinue}()}
 \item \code{\link{FEtype}()}
}

\strong{Directed vs. undirected availability}
The set of available effects depends on whether events are directed.
Use \code{\link{tie_effects}(directed = TRUE)} or
\code{\link{tie_effects}(directed = FALSE)} to obtain the authoritative list
for the current package version.

For undirected events (\code{directed = FALSE}), the available tie-model effects are:
\itemize{
  \item \code{\link{baseline}}, \code{\link{FEtype}}
  \item \code{\link{tie}}, \code{\link{same}}, \code{\link{difference}},
        \code{\link{average}}, \code{\link{minimum}}, \code{\link{maximum}}
  \item \code{\link{event}}, \code{\link{userStat}}
  \item \code{\link{inertia}}, \code{\link{totaldegreeDyad}},
        \code{\link{degreeMin}}, \code{\link{degreeMax}}, \code{\link{degreeDiff}}
  \item \code{\link{sp}}, \code{\link{psABAY}}, \code{\link{psABAB}},
        \code{\link{recencyContinue}}
}
}
\examples{
# List of available effects 
tie_effects()

# List of available effects for undirected networks
tie_effects(directed = FALSE)

# List of available endogenous effects for undirected networks
tie_effects(directed = FALSE, endogenous = TRUE)

}
